using SimpleJSON;
using System.Collections;
using System.Diagnostics.CodeAnalysis;
using System.Text.RegularExpressions;
using System;
using UnityEngine;
using static EntryPointUtils;

sealed class AutoUpdate : MVRScript
{
    const string NATURALIS_VERSION = "1.1.8";
    string _packagePrefix;

    public override void Init() => StartCoroutine(InitCo());

    IEnumerator InitCo()
    {
        yield return new WaitForEndOfFrame();
        while(SuperController.singleton.isLoading)
        {
            yield return null;
        }

        string packageId = GetPackageId(manager, GetPluginId(name));
        _packagePrefix = string.IsNullOrEmpty(packageId) ? "" : $"{packageId}:/";

        var naturalisBM = this.FirstOwnAtomStorableByRegex(NaturalisBMRegex);
        var naturalisTM = this.FirstOwnAtomStorableByRegex(NaturalisTMRegex);
        if(naturalisBM || naturalisTM)
        {
            yield return MigrateFromNaturalis(naturalisBM, naturalisTM);
            yield return new WaitForSeconds(1.5f);
            manager.RemovePluginWithUID(GetPluginId(name));
        }
        else
        {
            pluginLabelJSON.val = $"{Constant.NATURALIS} not found on this atom.";
        }
    }

    IEnumerator MigrateFromNaturalis(JSONStorable naturalisBM, JSONStorable naturalisTM)
    {
        var sourceStorable = naturalisBM ? naturalisBM : naturalisTM;
        string pluginId = GetPluginId(sourceStorable.name);
        string versionStr = sourceStorable.GetStringParamValue(Constant.NATURALIS_VERSION);
        if(string.IsNullOrEmpty(versionStr))
        {
            pluginLabelJSON.val = $"Unidentified {Constant.NATURALIS} version. ???";
            yield break;
        }

        var bootyMagicJSON = naturalisBM ? naturalisBM.GetJSON(true, true, true) : null;
        var tittyMagicJSON = naturalisTM ? naturalisTM.GetJSON(true, true, true) : null;

        manager.RemovePluginWithUID(pluginId);
        pluginLabelJSON.val = $"Adding {Constant.NATURALIS}...";
        AddNaturalis(manager, _packagePrefix);
        yield return new WaitForSeconds(1);
        yield return RestorePlugin(bootyMagicJSON, versionStr, true);
        yield return RestorePlugin(tittyMagicJSON, versionStr, false);
    }

    IEnumerator RestorePlugin(JSONClass sourceJSON, string versionStr, bool isBootyMagic)
    {
        var regex = isBootyMagic ? NaturalisBMRegex : NaturalisTMRegex;
        var targetStorable = this.FirstOwnAtomStorableByRegex(regex);
        while(!targetStorable)
        {
            yield return new WaitForSeconds(0.1f);
            targetStorable = this.FirstOwnAtomStorableByRegex(regex);
        }

        while(targetStorable.GetBoolParamValue(Constant.INTERNAL_IS_CALIBRATING))
        {
            yield return null;
        }

        yield return new WaitForSeconds(0.05f);
        if(sourceJSON != null)
        {
            var jc = isBootyMagic
                ? ConvertBootyMagicJSON(sourceJSON, versionStr, NATURALIS_VERSION, pluginLabelJSON)
                : ConvertTittyMagicJSON(sourceJSON, versionStr, NATURALIS_VERSION, pluginLabelJSON);
            targetStorable.RestoreFromJSON(jc);
        }
    }
}

// ReSharper disable MemberCanBePrivate.Global
// ReSharper disable UnusedMember.Global

static class EntryPointUtils
{
    public static Regex NaturalisBMRegex { get; } = new Regex($@"^plugin#\d+_{Constant.NATURALIS}.{Constant.BOOTY_MAGIC}", RegexOptions.Compiled);
    public static Regex NaturalisTMRegex { get; } = new Regex($@"^plugin#\d+_{Constant.NATURALIS}.{Constant.TITTY_MAGIC}", RegexOptions.Compiled);

    public static JSONClass ConvertBootyMagicJSON(JSONClass jc, string sourceVersionStr, string targetVersionStr, JSONStorableString pluginLabelJSON)
    {
        /* Assume any empty versionStr == pre v1.1 */
        if(!string.IsNullOrEmpty(sourceVersionStr) && !new SemVer(sourceVersionStr, true).IsMajorMinorLte(new SemVer(targetVersionStr, true)))
        {
            pluginLabelJSON.val = "Loading with defaults...";
            return new JSONClass()
            {
                ["id"] = $"plugin#0_{Constant.NATURALIS}.{Constant.BOOTY_MAGIC}",
                ["enabled"] = { AsBool = true },
            };
        }

        pluginLabelJSON.val = "Restoring plugin params...";
        /* any pre v1.1 -> v1.1 */
        jc["Rotation Spring Offset"].AsFloat = GetFloatValue(jc, "Spring Offset", 0.00f);
        jc["Rotation Damper Offset"].AsFloat = GetFloatValue(jc, "Damper Offset", 0.00f);
        jc["Rotation Spring Back Gravity"].AsFloat = GetFloatValue(jc, "Spring Back Gravity", 1.00f);
        jc["Rotation Spring Forward Gravity"].AsFloat = GetFloatValue(jc, "Spring Forward Gravity", 1.00f);
        return jc;
    }

    public static JSONClass ConvertTittyMagicJSON(JSONClass jc, string sourceVersionStr, string targetVersionStr, JSONStorableString pluginLabelJSON)
    {
        /* Assume any empty versionStr == pre v1.1 */
        if(!string.IsNullOrEmpty(sourceVersionStr) && !new SemVer(sourceVersionStr, true).IsMajorMinorLte(new SemVer(targetVersionStr, true)))
        {
            pluginLabelJSON.val = "Loading with defaults...";
            return new JSONClass()
            {
                ["id"] = $"plugin#0_{Constant.NATURALIS}.{Constant.TITTY_MAGIC}",
                ["enabled"] = { AsBool = true },
            };
        }

        pluginLabelJSON.val = "Restoring plugin params...";
        /* any pre v1.1 -> v1.1 */
        jc["Force Physics Depth Out Multiplier"].AsFloat = GetFloatValue(jc, "Force Physics Depth Out", 1.00f);
        jc["Force Physics Depth In Multiplier"].AsFloat = GetFloatValue(jc, "Force Physics Depth In", 1.00f);
        jc["Rotation Spring Offset"].AsFloat = GetFloatValue(jc, "Spring Offset", 0.00f);
        jc["Rotation Damper Offset"].AsFloat = GetFloatValue(jc, "Damper Offset", 0.00f);
        jc["Rotation Spring Back Gravity"].AsFloat = GetFloatValue(jc, "Spring Back Gravity", 1.00f);
        jc["Rotation Spring Forward Gravity"].AsFloat = GetFloatValue(jc, "Spring Forward Gravity", 1.00f);
        jc["Rotation Damper Back Gravity"].AsFloat = GetFloatValue(jc, "Damper Back Gravity", 1.00f);
        jc["Rotation Damper Forward Gravity"].AsFloat = GetFloatValue(jc, "Damper Forward Gravity", 1.00f);
        jc["Rotation Damper Sideways In Gravity"].AsFloat = GetFloatValue(jc, "Damper Sideways In Gravity", 1.00f);
        return jc;
    }

    public static float GetFloatValue(JSONClass jc, string key, float defaultValue)
    {
        if(!jc.HasKey(key))
        {
            return defaultValue;
        }

        float value = jc[key].AsFloat;
        jc.Remove(key);
        return value;
    }

    public static bool GetBoolValue(JSONClass jc, string key, bool defaultValue)
    {
        if(!jc.HasKey(key))
        {
            return defaultValue;
        }

        bool value = jc[key].AsBool;
        jc.Remove(key);
        return value;
    }

    public static string GetPluginId(string pluginStorableName) =>
        pluginStorableName.Substring(0, pluginStorableName.IndexOf('_'));

    public static string GetPackageId(MVRPluginManager manager, string pluginUid)
    {
        string filename = manager.GetJSON()["plugins"][pluginUid].Value;
        int idx = filename.IndexOf(":/", StringComparison.Ordinal);
        return idx >= 0 ? filename.Substring(0, idx) : null;
    }

    public static void AddNaturalis(MVRPluginManager manager, string packagePrefix)
    {
        var plugin = manager.CreatePlugin();
        plugin.pluginURLJSON.val = $"{packagePrefix}Custom/Scripts/{Constant.EVERLASTER}/{Constant.NATURALIS}/{Constant.NATURALIS}.cslist";
    }
}

static class JSONStorableExtensions
{
    public static JSONStorable FirstOwnAtomStorableByRegex(this JSONStorable self, Regex regex)
    {
        var storableIds = self.containingAtom.GetStorableIDs();
        foreach(string id in storableIds)
        {
            if(regex.IsMatch(id))
            {
                var storable = self.containingAtom.GetStorableByID(id);
                if(storable && storable.containingAtom && storable != self)
                {
                    return storable;
                }
            }
        }

        return null;
    }
}

sealed class SemVer
{
    readonly Version _version;

    public SemVer(string versionStr, bool removePrereleaseTag = false)
    {
        try
        {
            /* Without removePrereleaseTag set to true, a versionStr containing a prerelease tag
             * will result in a null version, and all comparisons will return false.
             */
            _version = removePrereleaseTag
                ? new Version(versionStr.Split('-')[0])
                : new Version(versionStr);
        }
        catch(Exception)
        {
            // Debug.Log("Invalid version string: " + versionStr);
        }
    }

    public bool IsGte(string other) => IsGte(new SemVer(other));
    public bool IsMajorGte(string other) => IsMajorGte(new SemVer(other));
    public bool IsMajorMinorGte(string other) => IsMajorMinorGte(new SemVer(other));

    public bool IsLte(string other) => IsLte(new SemVer(other));
    public bool IsMajorLte(string other) => IsMajorLte(new SemVer(other));
    public bool IsMajorMinorLte(string other) => IsMajorMinorLte(new SemVer(other));

    public bool IsMajorMinorEqual(string other) => IsMajorMinorEqual(new SemVer(other));

    public bool IsGte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        return _version >= other._version;
    }

    public bool IsMajorGte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        return _version.Major >= other._version.Major;
    }

    public bool IsMajorMinorGte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        var majorMinorA = new Version(_version.Major, _version.Minor);
        var majorMinorB = new Version(other._version.Major, other._version.Minor);

        return majorMinorA >= majorMinorB;
    }

    public bool IsLte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        return _version <= other._version;
    }

    public bool IsMajorLte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        return _version.Major <= other._version.Major;
    }

    public bool IsMajorMinorLte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        var majorMinorA = new Version(_version.Major, _version.Minor);
        var majorMinorB = new Version(other._version.Major, other._version.Minor);

        return majorMinorA <= majorMinorB;
    }

    public bool IsMajorMinorEqual(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        var majorMinorA = new Version(_version.Major, _version.Minor);
        var majorMinorB = new Version(other._version.Major, other._version.Minor);

        return majorMinorA == majorMinorB;
    }
}

[SuppressMessage("ReSharper", "InconsistentNaming")]
static class Constant
{
    const string INTERNAL = "internal_";
    public const string EVERLASTER = "everlaster";
    public const string NATURALIS = "Naturalis";
    public const string BOOTY_MAGIC = "BootyMagic";
    public const string TITTY_MAGIC = "TittyMagic";
    [SuppressMessage("ReSharper", "InconsistentNaming")]
    public readonly static string NATURALIS_VERSION = $"{NATURALIS} Version";
    public readonly static string INTERNAL_IS_CALIBRATING = $"{INTERNAL}isCalibrating";
    public readonly static string INTERNAL_IS_MIGRATION_SCRIPT = $"{INTERNAL}isMigrationScript";
}
