using SimpleJSON;
using System.Collections;
using System.Diagnostics.CodeAnalysis;
using System.Text.RegularExpressions;
using System;
using UnityEngine;
using static EntryPointUtils;

sealed class AutoUpgradeFromTittyMagic5 : MVRScript
{
    readonly Regex _tittyMagicRegex = new Regex($@"^plugin#\d+_{Constant.TITTY_MAGIC}.Script", RegexOptions.Compiled);

    string _packagePrefix;

    public override void Init() => StartCoroutine(InitCo());

    IEnumerator InitCo()
    {
        yield return new WaitForEndOfFrame();
        while(SuperController.singleton.isLoading)
        {
            yield return null;
        }

        string packageId = GetPackageId(manager, GetPluginId(name));
        _packagePrefix = string.IsNullOrEmpty(packageId) ? "" : $"{packageId}:/";

        var tittyMagic = this.FirstOwnAtomStorableByRegex(_tittyMagicRegex);
        if(tittyMagic)
        {
            yield return MigrateFromTittyMagic(tittyMagic);
            SuperController.LogMessage(
                "Successfully upgraded from TittyMagic 5 to Naturalis." +
                " Any triggers in the scene targeting TittyMagic 5 parameters will need to be cleaned up or fixed manually."
            );
            yield return new WaitForSeconds(1.5f);
            manager.RemovePluginWithUID(GetPluginId(name));
        }
        else
        {
            pluginLabelJSON.val = $"{Constant.TITTY_MAGIC} not found on this atom.";
        }
    }

    IEnumerator MigrateFromTittyMagic(JSONStorable tittyMagic)
    {
        string versionStr = tittyMagic.IsStringJSONParam("version") ? tittyMagic.GetStringParamValue("version") : "";
        if(string.IsNullOrEmpty(versionStr))
        {
            pluginLabelJSON.val = $"{Constant.TITTY_MAGIC} version not identified / too old";
            yield break;
        }

        if(versionStr.StartsWith("v"))
        {
            versionStr = versionStr.Substring(1);
        }

        if(!versionStr.StartsWith("5."))
        {
            pluginLabelJSON.val = $"Only upgrade from {Constant.TITTY_MAGIC} 5 is supported";
            yield break;
        }

        if(versionStr.Split('-').Length > 1)
        {
            pluginLabelJSON.val = $"Upgrading pre-release v{versionStr} is unsupported";
            yield break;
        }

        var tittyMagicJSON = tittyMagic.GetJSON(true, true, true);
        manager.RemovePluginWithUID(GetPluginId(tittyMagic.name));
        pluginLabelJSON.val = "Adding Naturalis...";
        AddNaturalis(manager, _packagePrefix);

        yield return new WaitForSeconds(1);
        var naturalisTMNew = this.FirstOwnAtomStorableByRegex(NaturalisTMRegex);
        while(!naturalisTMNew || naturalisTMNew.GetBoolParamValue(Constant.INTERNAL_IS_CALIBRATING))
        {
            yield return null;
        }

        yield return new WaitForSeconds(0.1f);
        pluginLabelJSON.val = "Restoring plugin params...";
        var jc = ConvertToNaturalisTittyMagicJSON(tittyMagicJSON, versionStr);
        naturalisTMNew.RestoreFromJSON(jc);
    }

    /* Converts TittyMagic 5 JSON to Naturalis TittyMagic JSON.
     * - LDP is disabled (even if you have it enabled by default in Naturalis User Preferences).
     * - Auto-Adjust Gravity Counterforce is enabled (default in Naturalis).
     * - Mass Offset is not applied, set to 0 instead since the mass calculation is different in Naturalis.
     * - All physics parameter offsets are also not applied, except for Angle Targets.
     *   The default physics values differ too much with Naturalis that the result could be quite different with the TittyMagic 5 offset values.
     *   With soft physics parameters, the gradient also complicates things.
     * - Quickness is converted from the 0 to 100 range to the -1 to 1 range, with 70 in TittyMagic corresponding to 0 in Naturalis.
     * - The overall gravity physics multipliers are converted to Naturalis' parameter specific multipliers.
     * - Everything else should be migrated as is - softness, soft physics on/off, self-collision, adaptive friction, nipple erection,
     *   gravity offset morphing, and hard collider properties.
     */
    static JSONClass ConvertToNaturalisTittyMagicJSON(JSONClass jc, string versionStr)
    {
        float baseGravity = GetFloatValue(jc, "gravityPhysicsBase", 1f);
        float forwardGravity = GetFloatValue(jc, "gravityPhysicsForward", 1f) * baseGravity;
        float backGravity = GetFloatValue(jc, "gravityPhysicsBack", 1f) * baseGravity;
        float leftRightGravity = GetFloatValue(jc, "gravityPhysicsLeftRight", 1f) * baseGravity;

        float legacyQuickness = GetFloatValue(jc, "breastQuickness", 70f);
        float quickness = legacyQuickness <= 70f
            ? Mathf.Lerp(-1f, 0f, legacyQuickness / 70f)
            : Mathf.Lerp(0f, 1f, (legacyQuickness - 70f) / 30f);

        var result = new JSONClass
        {
            ["id"] = "plugin#0_Naturalis.TittyMagic",
            ["enabled"] = { AsBool = true },
            ["Breast Soft Physics Enabled"] = { AsBool = GetBoolValue(jc, "on", false) },
            ["Local Distortion Physics Enabled"] = { AsBool = false },
            ["Breast Soft Physics Self Collide"] = { AsBool = GetBoolValue(jc, "allowSelfCollision", false) },
            ["Use Adaptive Friction"] = { AsBool = GetBoolValue(jc, "enableAdaptiveFriction", false) },
            // ["Breast Joint Mass Offset"] = { AsFloat = GetFloatValue(jc, "massOffset", 0f) },
            // ["Center Of Gravity Offset"] = { AsFloat = GetFloatValue(jc, "centerOfGravityPercentOffset", 0f) },
            // ["Spring Offset"] = { AsFloat = GetFloatValue(jc, "springOffset", 0f) },
            // ["Damper Offset"] = { AsFloat = GetFloatValue(jc, "damperOffset", 0f) },
            // ["In/Out Spring Offset"] = { AsFloat = GetFloatValue(jc, "positionSpringZOffset", 0f) },
            // ["In/Out Damper Offset"] = { AsFloat = GetFloatValue(jc, "positionDamperZOffset", 0f) },
            ["Up/Down Angle Target Offset"] = { AsFloat = GetFloatValue(jc, "targetRotationXOffset", 0f) },
            ["Right/Left Angle Target Offset"] = { AsFloat = GetFloatValue(jc, "targetRotationYOffset", 0f) },
            // ["Fat Spring Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesSpringOffset", 0f) },
            // ["Fat Spring Gradient Middle Shift"] = { AsFloat = -0.4f },
            // ["Fat Spring Gradient Outer Multiplier"] = { AsFloat = GetFloatValue(jc, "softVerticesSpringOffset", 0f) },
            // ["Fat Spring Gradient Center Multiplier"] = { AsFloat = 1.5f },
            // ["Fat Spring Gradient Nipple Multiplier"] = { AsFloat = 2.25f },
            // ["Fat Damper Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesDamperOffset", 0f) },
            // ["Fat Damper Gradient Middle Shift"] = { AsFloat = 0f },
            // ["Fat Damper Gradient Outer Multiplier"] = { AsFloat = 1.8f },
            // ["Fat Damper Gradient Center Multiplier"] = { AsFloat = 1.5f },
            // ["Fat Damper Gradient Nipple Multiplier"] = { AsFloat = 1.8f },
            // ["Fat Mass Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesMassOffset", 0f) },
            // ["Fat Mass Gradient Middle Shift"] = { AsFloat = 0f },
            // ["Fat Mass Gradient Outer Multiplier"] = { AsFloat = 1f },
            // ["Fat Mass Gradient Center Multiplier"] = { AsFloat = 1f },
            // ["Fat Mass Gradient Nipple Multiplier"] = { AsFloat = 1f },
            // ["Fat Collider Radius Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesColliderRadiusOffset", 0f) },
            // ["Fat Collider Radius Gradient Middle Shift"] = { AsFloat = 0f },
            // ["Fat Collider Radius Gradient Outer Multiplier"] = { AsFloat = 1.1f },
            // ["Fat Collider Radius Gradient Center Multiplier"] = { AsFloat = 1.25f },
            // ["Fat Collider Radius Gradient Nipple Multiplier"] = { AsFloat = 0.33f },
            // ["Fat Collider Depth Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesColliderAdditionalNormalOffsetOffset", 0f) },
            // ["Fat Collider Depth Gradient Middle Shift"] = { AsFloat = 0f },
            // ["Fat Collider Depth Gradient Outer Multiplier"] = { AsFloat = 1f },
            // ["Fat Collider Depth Gradient Center Multiplier"] = { AsFloat = 1f },
            // ["Fat Collider Depth Gradient Nipple Multiplier"] = { AsFloat = 1f },
            // ["Fat Distance Limit Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesDistanceLimitOffset", 0f) },
            // ["Fat Distance Limit Gradient Middle Shift"] = { AsFloat = -0.8f },
            // ["Fat Distance Limit Gradient Outer Multiplier"] = { AsFloat = 1f },
            // ["Fat Distance Limit Gradient Center Multiplier"] = { AsFloat = 0.8f },
            // ["Fat Distance Limit Gradient Nipple Multiplier"] = { AsFloat = 0.86f },
            // ["Fat Back Force Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesBackForceOffset", 0f) },
            // ["Fat Back Force Max Force Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesBackForceMaxForceOffset", 0f) },
            // ["Fat Back Force Threshold Offset"] = { AsFloat = GetFloatValue(jc, "softVerticesBackForceThresholdDistanceOffset", 0f) },
            ["Twist Angle Target"] = { AsFloat = GetFloatValue(jc, "targetRotationZ", 1) },
            ["Gravity Counterforce (%)"] = { AsFloat = 10f },
            ["Spring Back Gravity"] = { AsFloat = backGravity },
            ["Spring Forward Gravity"] = { AsFloat = forwardGravity },
            ["Damper Back Gravity"] = { AsFloat = backGravity },
            ["Damper Forward Gravity"] = { AsFloat = forwardGravity },
            ["Damper Sideways In Gravity"] = { AsFloat = leftRightGravity },
            ["In/Out Spring Back Gravity"] = { AsFloat = backGravity },
            ["In/Out Spring Forward Gravity"] = { AsFloat = forwardGravity },
            // ["Force Physics Up Multiplier"] =
            ["Force Physics Base Multiplier"] = { AsFloat = GetFloatValue(jc, "forcePhysicsBase", 1) },
            ["Force Physics Depth Out Multiplier"] = { AsFloat = GetFloatValue(jc, "forcePhysicsForward", 1) },
            ["Force Physics Depth In Multiplier"] = { AsFloat = GetFloatValue(jc, "forcePhysicsBack", 1) },
            // ["Force Physics Sideways In Multiplier"] =
            // ["Force Physics Sideways Out Multiplier"] =
            ["Friction Offset"] = { AsFloat = GetFloatValue(jc, "frictionOffset", 0f) },
            ["Dry Skin Friction"] = { AsFloat = 1f },
            ["Pectoral1 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral1ColliderRadius", 0f) },
            ["Pectoral1 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral1ColliderCenterX", 0f) },
            ["Pectoral1 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral1ColliderCenterY", 0f) },
            ["Pectoral1 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral1ColliderCenterZ", 0f) },
            ["Pectoral2 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral2ColliderRadius", 0f) },
            ["Pectoral2 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral2ColliderCenterX", 0f) },
            ["Pectoral2 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral2ColliderCenterY", 0f) },
            ["Pectoral2 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral2ColliderCenterZ", 0f) },
            ["Pectoral3 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral3ColliderRadius", 0f) },
            ["Pectoral3 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral3ColliderCenterX", 0f) },
            ["Pectoral3 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral3ColliderCenterY", 0f) },
            ["Pectoral3 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral3ColliderCenterZ", 0f) },
            ["Pectoral4 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral4ColliderRadius", 0f) },
            ["Pectoral4 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral4ColliderCenterX", 0f) },
            ["Pectoral4 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral4ColliderCenterY", 0f) },
            ["Pectoral4 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral4ColliderCenterZ", 0f) },
            ["Pectoral5 Radius Offset"] = { AsFloat = GetFloatValue(jc, "pectoral5ColliderRadius", 0f) },
            ["Pectoral5 X Offset"] = { AsFloat = GetFloatValue(jc, "pectoral5ColliderCenterX", 0f) },
            ["Pectoral5 Y Offset"] = { AsFloat = GetFloatValue(jc, "pectoral5ColliderCenterY", 0f) },
            ["Pectoral5 Z Offset"] = { AsFloat = GetFloatValue(jc, "pectoral5ColliderCenterZ", 0f) },
            ["Gravity Offset Morphing"] = { AsFloat = GetFloatValue(jc, "gravityOffsetMorphing", 1f) },
            ["Force Morphing Base Multiplier"] = { AsFloat = GetFloatValue(jc, "forceMorphingBase", 1f) },
            ["Force Morphing Up"] = { AsFloat = GetFloatValue(jc, "forceMorphingUp", 1f) },
            ["Force Morphing Depth In"] = { AsFloat = GetFloatValue(jc, "forceMorphingBack", 1f) },
            ["Force Morphing Depth Out"] = { AsFloat = GetFloatValue(jc, "forceMorphingForward", 1f) },
            ["Nipple Erection"] = { AsFloat = GetFloatValue(jc, "nippleErection", 0f) },
            ["Breast Softness"] = { AsFloat = GetFloatValue(jc, "breastSoftness", 70f) },
            ["Breast Quickness Offset"] = { AsFloat = quickness },
        };

        if(versionStr == "5.0.0")
        {
            result["Pectoral1 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral1ColliderForce", 0.50f);
            result["Pectoral2 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral2ColliderForce", 0.50f);
            result["Pectoral3 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral3ColliderForce", 0.50f);
            result["Pectoral4 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral4ColliderForce", 0.50f);
            result["Pectoral5 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral5ColliderForce", 0.50f);
            result["Hard Colliders Base Mass Multiplier"].AsFloat = GetFloatValue(jc, "combinedColliderForce", 1.00f);
        }
        else
        {
            result["Pectoral1 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral1CollisionForce", 0.50f);
            result["Pectoral2 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral2CollisionForce", 0.50f);
            result["Pectoral3 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral3CollisionForce", 0.50f);
            result["Pectoral4 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral4CollisionForce", 0.50f);
            result["Pectoral5 Mass Multiplier"].AsFloat = GetFloatValue(jc, "pectoral5CollisionForce", 0.50f);
            result["Hard Colliders Base Mass Multiplier"].AsFloat = GetFloatValue(jc, "baseCollisionForce", 1.00f);
        }

        if(versionStr.StartsWith("5.0.") || versionStr.StartsWith("5.1."))
        {
            result["Force Morphing Sideways In"].AsFloat = GetFloatValue(jc, "forceMorphingLeftRight", 1.00f);
            result["Force Morphing Sideways Out"].AsFloat = GetFloatValue(jc, "forceMorphingLeftRight", 1.00f);
        }
        else
        {
            result["Force Morphing Sideways In"].AsFloat = GetFloatValue(jc, "forceMorphingSidewaysIn", 1f);
            result["Force Morphing Sideways Out"].AsFloat = GetFloatValue(jc, "forceMorphingSidewaysOut", 1f);
        }

        result["pluginLabel"] = jc["pluginLabel"].Value;
        return result;
    }
}

// ReSharper disable MemberCanBePrivate.Global
// ReSharper disable UnusedMember.Global

static class EntryPointUtils
{
    public static Regex NaturalisBMRegex { get; } = new Regex($@"^plugin#\d+_{Constant.NATURALIS}.{Constant.BOOTY_MAGIC}", RegexOptions.Compiled);
    public static Regex NaturalisTMRegex { get; } = new Regex($@"^plugin#\d+_{Constant.NATURALIS}.{Constant.TITTY_MAGIC}", RegexOptions.Compiled);

    public static JSONClass ConvertBootyMagicJSON(JSONClass jc, string sourceVersionStr, string targetVersionStr, JSONStorableString pluginLabelJSON)
    {
        /* Assume any empty versionStr == pre v1.1 */
        if(!string.IsNullOrEmpty(sourceVersionStr) && !new SemVer(sourceVersionStr, true).IsMajorMinorLte(new SemVer(targetVersionStr, true)))
        {
            pluginLabelJSON.val = "Loading with defaults...";
            return new JSONClass()
            {
                ["id"] = $"plugin#0_{Constant.NATURALIS}.{Constant.BOOTY_MAGIC}",
                ["enabled"] = { AsBool = true },
            };
        }

        pluginLabelJSON.val = "Restoring plugin params...";
        /* any pre v1.1 -> v1.1 */
        jc["Rotation Spring Offset"].AsFloat = GetFloatValue(jc, "Spring Offset", 0.00f);
        jc["Rotation Damper Offset"].AsFloat = GetFloatValue(jc, "Damper Offset", 0.00f);
        jc["Rotation Spring Back Gravity"].AsFloat = GetFloatValue(jc, "Spring Back Gravity", 1.00f);
        jc["Rotation Spring Forward Gravity"].AsFloat = GetFloatValue(jc, "Spring Forward Gravity", 1.00f);
        return jc;
    }

    public static JSONClass ConvertTittyMagicJSON(JSONClass jc, string sourceVersionStr, string targetVersionStr, JSONStorableString pluginLabelJSON)
    {
        /* Assume any empty versionStr == pre v1.1 */
        if(!string.IsNullOrEmpty(sourceVersionStr) && !new SemVer(sourceVersionStr, true).IsMajorMinorLte(new SemVer(targetVersionStr, true)))
        {
            pluginLabelJSON.val = "Loading with defaults...";
            return new JSONClass()
            {
                ["id"] = $"plugin#0_{Constant.NATURALIS}.{Constant.TITTY_MAGIC}",
                ["enabled"] = { AsBool = true },
            };
        }

        pluginLabelJSON.val = "Restoring plugin params...";
        /* any pre v1.1 -> v1.1 */
        jc["Force Physics Depth Out Multiplier"].AsFloat = GetFloatValue(jc, "Force Physics Depth Out", 1.00f);
        jc["Force Physics Depth In Multiplier"].AsFloat = GetFloatValue(jc, "Force Physics Depth In", 1.00f);
        jc["Rotation Spring Offset"].AsFloat = GetFloatValue(jc, "Spring Offset", 0.00f);
        jc["Rotation Damper Offset"].AsFloat = GetFloatValue(jc, "Damper Offset", 0.00f);
        jc["Rotation Spring Back Gravity"].AsFloat = GetFloatValue(jc, "Spring Back Gravity", 1.00f);
        jc["Rotation Spring Forward Gravity"].AsFloat = GetFloatValue(jc, "Spring Forward Gravity", 1.00f);
        jc["Rotation Damper Back Gravity"].AsFloat = GetFloatValue(jc, "Damper Back Gravity", 1.00f);
        jc["Rotation Damper Forward Gravity"].AsFloat = GetFloatValue(jc, "Damper Forward Gravity", 1.00f);
        jc["Rotation Damper Sideways In Gravity"].AsFloat = GetFloatValue(jc, "Damper Sideways In Gravity", 1.00f);
        return jc;
    }

    public static float GetFloatValue(JSONClass jc, string key, float defaultValue)
    {
        if(!jc.HasKey(key))
        {
            return defaultValue;
        }

        float value = jc[key].AsFloat;
        jc.Remove(key);
        return value;
    }

    public static bool GetBoolValue(JSONClass jc, string key, bool defaultValue)
    {
        if(!jc.HasKey(key))
        {
            return defaultValue;
        }

        bool value = jc[key].AsBool;
        jc.Remove(key);
        return value;
    }

    public static string GetPluginId(string pluginStorableName) =>
        pluginStorableName.Substring(0, pluginStorableName.IndexOf('_'));

    public static string GetPackageId(MVRPluginManager manager, string pluginUid)
    {
        string filename = manager.GetJSON()["plugins"][pluginUid].Value;
        int idx = filename.IndexOf(":/", StringComparison.Ordinal);
        return idx >= 0 ? filename.Substring(0, idx) : null;
    }

    public static void AddNaturalis(MVRPluginManager manager, string packagePrefix)
    {
        var plugin = manager.CreatePlugin();
        plugin.pluginURLJSON.val = $"{packagePrefix}Custom/Scripts/{Constant.EVERLASTER}/{Constant.NATURALIS}/{Constant.NATURALIS}.cslist";
    }
}

static class JSONStorableExtensions
{
    public static JSONStorable FirstOwnAtomStorableByRegex(this JSONStorable self, Regex regex)
    {
        var storableIds = self.containingAtom.GetStorableIDs();
        foreach(string id in storableIds)
        {
            if(regex.IsMatch(id))
            {
                var storable = self.containingAtom.GetStorableByID(id);
                if(storable && storable.containingAtom && storable != self)
                {
                    return storable;
                }
            }
        }

        return null;
    }
}

sealed class SemVer
{
    readonly Version _version;

    public SemVer(string versionStr, bool removePrereleaseTag = false)
    {
        try
        {
            /* Without removePrereleaseTag set to true, a versionStr containing a prerelease tag
             * will result in a null version, and all comparisons will return false.
             */
            _version = removePrereleaseTag
                ? new Version(versionStr.Split('-')[0])
                : new Version(versionStr);
        }
        catch(Exception)
        {
            // Debug.Log("Invalid version string: " + versionStr);
        }
    }

    public bool IsGte(string other) => IsGte(new SemVer(other));
    public bool IsMajorGte(string other) => IsMajorGte(new SemVer(other));
    public bool IsMajorMinorGte(string other) => IsMajorMinorGte(new SemVer(other));

    public bool IsLte(string other) => IsLte(new SemVer(other));
    public bool IsMajorLte(string other) => IsMajorLte(new SemVer(other));
    public bool IsMajorMinorLte(string other) => IsMajorMinorLte(new SemVer(other));

    public bool IsMajorMinorEqual(string other) => IsMajorMinorEqual(new SemVer(other));

    public bool IsGte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        return _version >= other._version;
    }

    public bool IsMajorGte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        return _version.Major >= other._version.Major;
    }

    public bool IsMajorMinorGte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        var majorMinorA = new Version(_version.Major, _version.Minor);
        var majorMinorB = new Version(other._version.Major, other._version.Minor);

        return majorMinorA >= majorMinorB;
    }

    public bool IsLte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        return _version <= other._version;
    }

    public bool IsMajorLte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        return _version.Major <= other._version.Major;
    }

    public bool IsMajorMinorLte(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        var majorMinorA = new Version(_version.Major, _version.Minor);
        var majorMinorB = new Version(other._version.Major, other._version.Minor);

        return majorMinorA <= majorMinorB;
    }

    public bool IsMajorMinorEqual(SemVer other)
    {
        if(_version == null || other._version == null)
        {
            return false;
        }

        var majorMinorA = new Version(_version.Major, _version.Minor);
        var majorMinorB = new Version(other._version.Major, other._version.Minor);

        return majorMinorA == majorMinorB;
    }
}

[SuppressMessage("ReSharper", "InconsistentNaming")]
static class Constant
{
    const string INTERNAL = "internal_";
    public const string EVERLASTER = "everlaster";
    public const string NATURALIS = "Naturalis";
    public const string BOOTY_MAGIC = "BootyMagic";
    public const string TITTY_MAGIC = "TittyMagic";
    [SuppressMessage("ReSharper", "InconsistentNaming")]
    public readonly static string NATURALIS_VERSION = $"{NATURALIS} Version";
    public readonly static string INTERNAL_IS_CALIBRATING = $"{INTERNAL}isCalibrating";
    public readonly static string INTERNAL_IS_MIGRATION_SCRIPT = $"{INTERNAL}isMigrationScript";
}
