using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using NAT.Common;
using NAT.Common.Components;
using NAT.Common.Models;
using System;
using UnityEngine;

namespace NAT
{
    sealed class Person
    {
        internal Atom Atom { get; }
        readonly string _packageId;
        internal DAZCharacterSelector Geometry { get; private set; }
        internal DAZSkinV2 Skin { get; private set; }

        GenerateDAZMorphsControlUI _morphsControlUI;
        internal JSONStorable SoftBodyPhysicsEnabler { get; private set; }
        internal JSONStorableBool CollisionEnabledJsb { get; private set; }
        internal JSONStorableFloat ScaleJsf { get; private set; }

        internal AdjustJoints GluteControl { get; private set; }
        internal AdjustJoints BreastControl { get; private set; }
        internal DAZPhysicsMesh GlutePhysicsMesh { get; private set; }
        internal DAZPhysicsMesh BreastPhysicsMesh { get; private set; }
        internal List<AutoColliderGroup> AutoColliderGroups { get; private set; }
        internal AutoColliderGroup LeftPectoralAutoColliderGroup { get; private set; }
        internal AutoColliderGroup RightPectoralAutoColliderGroup { get; private set; }

        internal Rigidbody HipRb { get; private set; }
        internal Rigidbody ChestRb { get; private set; }

        internal CalibratedJoint LeftBreast { get; private set; }
        internal CalibratedJoint RightBreast { get; private set; }
        internal CalibratedJoint LeftGlute { get; private set; }
        internal CalibratedJoint RightGlute { get; private set; }

        string _morphsPath;
        string _packagePrefix;
        PresetManagerControl _posePmc;
        PresetManager _posePm;
        PresetManager _morphPm;

        internal Transform AtomUIContent { get; private set; }

        internal bool CharacterReady { get; private set; }
        internal bool SkinReady { get; private set; }
        internal bool IsFemale { get; private set; }
        internal CalibrationState CalibrationState { get; }
        internal bool IsFreezeGrabbing { get; private set; }
        internal bool SetupDone { get; private set; }
        internal bool FutaSetupDone { get; private set; }

        internal Person(Atom atom, string packageId)
        {
            Atom = atom;
            _packageId = packageId;
            CalibrationState = new CalibrationState(this);
        }

        internal JSONStorable GetStorable(string id) => Atom.GetStorableByID(id);

        internal bool PectoralRbCollisionsEnabled(bool enabled) => enabled
            ? LeftBreast.Rigidbody.detectCollisions || RightBreast.Rigidbody.detectCollisions
            : !LeftBreast.Rigidbody.detectCollisions || !RightBreast.Rigidbody.detectCollisions;

        internal bool HasDummySkin() => Geometry.selectedCharacter.name == "femaledummy";

        internal IEnumerator WaitForGeometryAndSkinReady(float limit)
        {
            float timeout = Time.unscaledTime + limit;
            while(!(CharacterReady && SkinReady) && Time.unscaledTime < timeout)
            {
                Geometry = (DAZCharacterSelector) Atom.GetStorableByID("geometry");
                CharacterReady = Geometry.selectedCharacter.ready;
                SkinReady = HasDummySkin() || Atom.GetStorableByID("skin");
                yield return new WaitForSecondsRealtime(0.5f);
            }
        }

        internal IEnumerator RefreshSkinCo()
        {
            CharacterReady = false;
            SkinReady = false;
            Skin = null;

            const float limit = 15f;
            yield return WaitForGeometryAndSkinReady(limit);

            if(!CharacterReady)
            {
                Loggr.Error($"Selected character {Geometry.selectedCharacter.name} was not ready after {limit.ToString()} seconds of waiting. Try reloading");
                yield break;
            }

            if(!SkinReady)
            {
                Loggr.Error($"Person skin materials not found after {limit.ToString()} seconds of waiting. Try reloading");
                yield break;
            }

            Skin = Atom.GetComponentInChildren<DAZCharacter>().skin;
        }

        internal bool IsSelectedCharacterReady() => Geometry.selectedCharacter.ready;

        internal void Setup()
        {
            try
            {
                IsFemale = !Geometry.selectedCharacter.isMale;
                Skin = Atom.GetComponentInChildren<DAZCharacter>().skin;
                _morphsControlUI = IsFemale ? Geometry.morphsControlUI : Geometry.morphsControlUIOtherGender;
                SoftBodyPhysicsEnabler = Atom.GetStorableByID("SoftBodyPhysicsEnabler");
                CollisionEnabledJsb = Atom.GetStorableByID("AtomControl").GetBoolJSONParam("collisionEnabled");
                ScaleJsf = Atom.GetStorableByID("rescaleObject").GetFloatJSONParam("scale");

                if(IsFemale)
                {
                    GluteControl = (AdjustJoints) Atom.GetStorableByID("GluteControl");
                    GlutePhysicsMesh = (DAZPhysicsMesh) Atom.GetStorableByID("LowerPhysicsMesh");
                    HipRb = Utils.FindRigidbody(Atom, "hip");
                    LeftGlute = new Glute(GluteControl.joint1, this);
                    RightGlute = new Glute(GluteControl.joint2, this);
                }

                BreastControl = (AdjustJoints) Atom.GetStorableByID(IsFemale ? "BreastControl" : "PectoralControl");
                BreastPhysicsMesh = (DAZPhysicsMesh) Atom.GetStorableByID("BreastPhysicsMesh");

                ChestRb = Utils.FindRigidbody(Atom, "chest");
                LeftBreast = new Breast(BreastControl.joint2, Atom.rigidbodies.FirstOrDefault(rb => rb.name == "lNipple"), this);
                RightBreast = new Breast(BreastControl.joint1, Atom.rigidbodies.FirstOrDefault(rb => rb.name == "rNipple"), this);

                AutoColliderGroups = Atom.autoColliderBatchUpdaters.First().GetComponentsInChildren<AutoColliderGroup>(true).ToList();
                LeftPectoralAutoColliderGroup = AutoColliderGroups.First(x => x.name == "FemaleAutoColliderslPectoral");
                RightPectoralAutoColliderGroup = AutoColliderGroups.First(x => x.name == "FemaleAutoCollidersrPectoral");

                SaveOriginalUseColliders();
                EnableUseColliders();

                string packagePath = string.IsNullOrEmpty(_packageId) ? "" : $"{_packageId}:/";
                _morphsPath = $"{packagePath}Custom/Atom/Person/Morphs/female/{Constant.EVERLASTER}";
                _packagePrefix = string.IsNullOrEmpty(_packageId) ? "" : $"{_packageId}:";

                _posePmc = Atom.presetManagerControls.Find(pmc => pmc.name == "PosePresets");
                _posePm = Atom.GetStorableByID("PosePresets").GetComponentInChildren<PresetManager>();
                _posePm.postLoadEvent.AddListener(OnPosePresetLoaded);

                _morphPm = Atom.GetStorableByID("MorphPresets").GetComponentInChildren<PresetManager>();
                _morphPm.postLoadEvent.AddListener(ZeroAllDuplicateMorphs);

                AtomUIContent = Atom.transform.Find("UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content");

                Atom.mainController.onGrabStartHandlers -= OnStartGrab;
                Atom.mainController.onGrabEndHandlers -= OnEndGrab;
                Atom.mainController.onGrabStartHandlers += OnStartGrab;
                Atom.mainController.onGrabEndHandlers += OnEndGrab;

                SetupDone = true;
            }
            catch(Exception e)
            {
                Loggr.Error($"Error in Person setup: {e}");
            }
        }

        internal void SetupFuta()
        {
            TransplantBreastDazMesh();
            ParentBreastPhysicsMesh("MaleMorphers");
            EnablePectoralAutoColliderGroups();
            IgnoreMaleChestCollidersOnFemalePectorals();
            FutaSetupDone = true;
        }

        void TransplantBreastDazMesh()
        {
            var skin = Atom.GetComponentInChildren<DAZMergedSkinV2>();
            var breastDazMesh = BreastPhysicsMesh.GetComponent<DAZPhysicsMesh>();
            IgnoreMaleChestCollidersOnPhysicsMesh(breastDazMesh);
            breastDazMesh.InitColliders();
            breastDazMesh.skin = skin;
            RegisterMesh(breastDazMesh);
            breastDazMesh.Init();
            breastDazMesh.enabled = true;
        }

        void IgnoreMaleChestCollidersOnPhysicsMesh(DAZPhysicsMesh breastDazMesh)
        {
            try
            {
                var maleChest = AutoColliderGroups.First(x => x.name == "MaleAutoColliderschest");
                foreach(var vertGroup in breastDazMesh.softVerticesGroups)
                {
                    var ignoreAutoColliderGroups = vertGroup.ignoreAutoColliderGroups.ToList();
                    if(!ignoreAutoColliderGroups.Contains(maleChest))
                    {
                        ignoreAutoColliderGroups.Add(maleChest);
                    }

                    vertGroup.ignoreAutoColliderGroups = ignoreAutoColliderGroups.ToArray();
                }
            }
            catch(Exception e)
            {
                Loggr.Error($"IgnoreMaleChestCollidersOnPhysicsMesh: {e}");
            }
        }

        void RegisterMesh(DAZPhysicsMesh mesh)
        {
            var characterRun = Atom.GetComponentInChildren<DAZCharacterRun>();
            if(!characterRun)
            {
                Loggr.Error("RegisterMesh: DAZCharacterRun not found");
                return;
            }

            var newMeshes = characterRun.physicsMeshes.ToList();
            if(!newMeshes.Contains(mesh))
            {
                newMeshes.Add(mesh);
                characterRun.physicsMeshes = newMeshes.ToArray();
            }
        }

        internal void ParentBreastPhysicsMesh(string parentName)
        {
            var parent = Geometry.transform.Find(parentName);
            if(!parent)
            {
                Loggr.Error($"Transform {parentName} not found in geometry");
                return;
            }

            BreastPhysicsMesh.transform.parent = parent;
        }

        void EnablePectoralAutoColliderGroups()
        {
            EnableAutoColliderGroup(LeftPectoralAutoColliderGroup);
            EnableAutoColliderGroup(RightPectoralAutoColliderGroup);
        }

        void EnableAutoColliderGroup(AutoColliderGroup group)
        {
            try
            {
                group.enabled = true;
                group.gameObject.SetActive(true);
                foreach(var ac in group.GetAutoColliders())
                {
                    ac.enabled = true;
                    ac.gameObject.SetActive(true);
                    ac.jointRB.detectCollisions = CollisionEnabledJsb.val;
                }
            }
            catch(Exception e)
            {
                Loggr.Error($"Error enabling AutoColliderGroup {(group == null ? "null" : group.name)}: {e}");
            }
        }

        void IgnoreMaleChestCollidersOnFemalePectorals()
        {
            try
            {
                var maleChest = AutoColliderGroups.First(x => x.name == "MaleAutoColliderschest");
                IgnoreColliderGroup(LeftPectoralAutoColliderGroup, maleChest);
                IgnoreColliderGroup(RightPectoralAutoColliderGroup, maleChest);
            }
            catch(Exception e)
            {
                Loggr.Error($"IgnoreMaleChestCollidersOnFemalePectorals: {e}");
            }
        }

        static void IgnoreColliderGroup(AutoColliderGroup group, AutoColliderGroup ignoreGroup)
        {
            var ignoreColliderGroups = group.ignoreColliderGroups.ToList();
            if(!ignoreColliderGroups.Contains(ignoreGroup))
            {
                ignoreColliderGroups.Add(ignoreGroup);
            }

            group.ignoreColliderGroups = ignoreColliderGroups.ToArray();
            group.InitColliders();
        }

        bool _originalUseAdvancedColliders;
        bool _originalUseAuxBreastColliders;

        void SaveOriginalUseColliders()
        {
            _originalUseAdvancedColliders = Geometry.useAdvancedColliders;
            _originalUseAuxBreastColliders = Geometry.useAuxBreastColliders;
        }

        internal void EnableUseColliders()
        {
            /* Advanced colliders must be enabled for collider visualizer, force morphing and hard collider handler */
            Geometry.useAdvancedColliders = true;
            Geometry.useAuxBreastColliders = true;
        }

        void OnStartGrab(FreeControllerV3 controller)
        {
            if(controller.GetBoolParamValue("freezeAtomPhysicsWhenGrabbed"))
            {
                IsFreezeGrabbing = true;
            }
        }

        void OnPosePresetLoaded()
        {
            if(_posePmc.GetStringParamValue("presetName").StartsWith($"{Constant.EVERLASTER}/{Constant.NATURALIS}/"))
            {
                return;
            }

            ZeroAllDuplicateMorphs();
        }

        void OnEndGrab(FreeControllerV3 controller) => IsFreezeGrabbing = false;

        internal void SetGlutesUseGravity(bool value)
        {
            LeftGlute.Rigidbody.useGravity = value;
            RightGlute.Rigidbody.useGravity = value;
        }

        internal void SetBreastsUseGravity(bool value)
        {
            LeftBreast.Rigidbody.useGravity = value;
            RightBreast.Rigidbody.useGravity = value;
        }

        internal void SetPectoralRbCollisions(bool value)
        {
            LeftBreast.Rigidbody.detectCollisions = value;
            RightBreast.Rigidbody.detectCollisions = value;
        }

        internal DAZMorph GetBreastMorph(string fileName) => GetOwnMorph($"{Constant.TITTY_MAGIC}/{fileName}.vmi");
        internal DAZMorph GetGluteMorph(string fileName) => GetOwnMorph($"{Constant.BOOTY_MAGIC}/{fileName}.vmi");
        DAZMorph GetOwnMorph(string subPath) => GetMorph($"{_morphsPath}/{subPath}");

        DAZMorph GetMorph(string uid)
        {
            var dazMorph = _morphsControlUI.GetMorphByUid(uid);
            if(dazMorph == null)
            {
                Loggr.Error($"Morph with uid '{uid}' not found");
            }

            return dazMorph;
        }

        internal PresetParams GetPosePresetParams() => PresetParams.GetParamsFromControl(_posePmc);
        internal void SetPosePresetParams(PresetParams presetParams) => presetParams.SetParamsToControl(_posePmc);

        internal void SavePosePreset(
            string name,
            bool storePoseMorphsInPose = true,
            bool storeBonesInPose = true,
            bool storeRootNodeInPose = false
        )
        {
            new PresetParams(
                $"{Constant.EVERLASTER}/{Constant.NATURALIS}/{name}",
                storeRootNodeInPose,
                storeBonesInPose,
                storePoseMorphsInPose
            ).SetParamsToControl(_posePmc);
            _posePmc.CallAction("StorePreset");
        }

        internal void LoadPosePreset(string name, bool fromPackage)
        {
            string prefix = fromPackage ? _packagePrefix : "";
            _posePmc.SetStringParamValue("presetName", $"{prefix}{Constant.EVERLASTER}/{Constant.NATURALIS}/{name}");
            _posePmc.CallAction("LoadPreset");
        }

        internal void DeletePosePreset(string name)
        {
            _posePmc.SetStringParamValue("presetName", $"{Constant.EVERLASTER}/{Constant.NATURALIS}/{name}");
            _posePm.DeletePreset();
        }

        readonly Dictionary<string, List<string>> _morphNamePrefixes = new Dictionary<string, List<string>>
        {
            { "BM_", UidPrefixesSet("BM_") },
            { "TM_", UidPrefixesSet("TM_") },
            { "TM3_", UidPrefixesSet("TM3_") },
            { "TM5_", UidPrefixesSet("TM5_") },
        };

        static List<string> UidPrefixesSet(string prefix) => new List<string>
        {
            $"{prefix}{MorphPrefix.DN}",
            $"{prefix}{MorphPrefix.UP}",
            $"{prefix}{MorphPrefix.FW}",
            $"{prefix}{MorphPrefix.BK}",
            $"{prefix}{MorphPrefix.LT}",
            $"{prefix}{MorphPrefix.RT}",
        };

        void ZeroAllDuplicateMorphs()
        {
            ZeroDuplicateBootyMagicMorphs();
            ZeroDuplicateTittyMagicMorphs();
        }

        internal void ZeroDuplicateBootyMagicMorphs() =>
            ZeroDuplicateMorphs("BM_", new[] { $"{Constant.EVERLASTER}.{Constant.NATURALIS}." });

        internal void ZeroDuplicateTittyMagicMorphs()
        {
            string[] packageUidPrefixes = { $"{Constant.EVERLASTER}.{Constant.TITTY_MAGIC}.", $"{Constant.EVERLASTER}.{Constant.BOOTY_MAGIC}." };
            ZeroDuplicateMorphs("TM_", packageUidPrefixes);
            ZeroDuplicateMorphs("TM3_", packageUidPrefixes);
            ZeroDuplicateMorphs("TM5_", packageUidPrefixes);
        }

        void ZeroDuplicateMorphs(string prefix, string[] packageUidPrefixes)
        {
            foreach(var morph in Geometry.morphBank1.morphs)
            {
                if(IsDuplicateLocalMorph(prefix, morph) || IsDuplicatePackageMorph(prefix, packageUidPrefixes, morph))
                {
                    morph.morphValue = 0;
                }
            }
        }

        bool IsDuplicateLocalMorph(string prefix, DAZMorph morph) =>
            !morph.isInPackage &&
            morph.morphValue != 0 &&
            MorphNameStartsWithAny(_morphNamePrefixes[prefix], morph);

        bool IsDuplicatePackageMorph(string prefix, IEnumerable<string> packageUidPrefixes, DAZMorph morph) =>
            morph.isInPackage &&
            morph.morphValue != 0 &&
            MorphNameStartsWithAny(_morphNamePrefixes[prefix], morph) &&
            morph.packageUid != _packageId && // packageId is null when running plugin locally
            MorphPackageUidStartsWithAny(packageUidPrefixes, morph);

        static bool MorphNameStartsWithAny(IEnumerable<string> tests, DAZMorph morph)
        {
            foreach(string test in tests)
            {
                if(morph.morphName.StartsWith(test))
                {
                    return true;
                }
            }

            return false;
        }

        static bool MorphPackageUidStartsWithAny(IEnumerable<string> tests, DAZMorph morph)
        {
            foreach(string test in tests)
            {
                if(morph.packageUid.StartsWith(test))
                {
                    return true;
                }
            }

            return false;
        }

        internal Vector3 SkinVertsAveragePosition(int[] indices)
        {
            var sum = Vector3.zero;
            foreach(int t in indices)
            {
                sum += Skin.rawSkinnedVerts[t];
            }

            return sum / indices.Length;
        }

        internal void Destroy()
        {
            _posePm.postLoadEvent.RemoveListener(OnPosePresetLoaded);
            _morphPm.postLoadEvent.RemoveListener(ZeroAllDuplicateMorphs);

            if(IsFemale)
            {
                Geometry.useAdvancedColliders = _originalUseAdvancedColliders;
                Geometry.useAuxBreastColliders = _originalUseAuxBreastColliders;
            }

            if(Atom && Atom.mainController)
            {
                Atom.mainController.onGrabStartHandlers -= OnStartGrab;
                Atom.mainController.onGrabEndHandlers -= OnEndGrab;
            }
        }
    }
}
