using NAT.Common;
using System.Collections;
using UnityEngine;

namespace NAT
{
    sealed class PersonMonitor : MonoBehaviour
    {
        Person _person;
        JSONStorable _bootyMagic;
        JSONStorable _tittyMagic;
        JSONStorableBool _breastInOutJsb;
        DAZCharacter _selectedCharacter;

        float _fixedDeltaTime;

        internal bool IsBreastSoftPhysicsOn() => _globalSoftPhysicsOn && _atomSoftPhysicsOn && _breastSoftPhysicsOn;

        internal bool IsGluteSoftPhysicsOn() => _globalSoftPhysicsOn && _atomSoftPhysicsOn && _gluteSoftPhysicsOn;

        bool _globalSoftPhysicsOn;
        bool _atomSoftPhysicsOn;
        bool _breastSoftPhysicsOn;
        bool _gluteSoftPhysicsOn;

        internal bool Initialized { get; private set; }

        internal void Init(Person person, JSONStorable bootyMagic, JSONStorable tittyMagic)
        {
            enabled = false; // will be enabled after first-time calibration
            _person = person;
            _bootyMagic = bootyMagic;
            _tittyMagic = tittyMagic;
            _selectedCharacter = _person.Geometry.selectedCharacter;
            _fixedDeltaTime = Time.fixedDeltaTime;
            _atomSoftPhysicsOn = _person.SoftBodyPhysicsEnabler.GetBoolParamValue("enabled");
            _globalSoftPhysicsOn = UserPreferences.singleton.softPhysics;

            if(_person.IsFemale)
            {
                _breastInOutJsb = _person.GetStorable("BreastInOut").GetBoolJSONParam("enabled");
                if(_tittyMagic.IsEnabledNullSafe() && _person.IsFemale)
                {
                    /* prevent breasts being flattened due to breastInOut morphs on scene load with plugin already present */
                    _breastInOutJsb.val = true;
                    _breastInOutJsb.val = false;
                }

                _breastInOutJsb.setJSONCallbackFunction += OnBreastInOutChanged;
            }

            StartCoroutine(FinishInit());
        }

        IEnumerator FinishInit()
        {
            if(_bootyMagic.IsEnabledNullSafe())
            {
                while(!_bootyMagic || !_bootyMagic.IsBoolJSONParam(Constant.GLUTE_SOFT_PHYSICS_ENABLED))
                {
                    yield return null;
                }

                _gluteSoftPhysicsOn = _bootyMagic.GetBoolParamValue(Constant.GLUTE_SOFT_PHYSICS_ENABLED);
            }

            if(_tittyMagic.IsEnabledNullSafe())
            {
                while(!_tittyMagic || !_tittyMagic.IsBoolJSONParam(Constant.BREAST_SOFT_PHYSICS_ENABLED))
                {
                    yield return null;
                }

                _breastSoftPhysicsOn = _tittyMagic.GetBoolParamValue(Constant.BREAST_SOFT_PHYSICS_ENABLED);
            }

            Initialized = true;
        }

        void OnBreastInOutChanged(JSONStorableBool jsb)
        {
            if(!_tittyMagic.IsEnabledNullSafe() || !_person.IsFemale)
            {
                return;
            }

            if(jsb.val)
            {
                jsb.val = false;
                Loggr.Message("Auto Breast In/Out Morphs disabled - directional force morphing works better without them.");
            }
        }

        internal void Check(bool runCallbacks = false)
        {
            if(!Initialized || !enabled)
            {
                return;
            }

            if(_tittyMagic.IsEnabledNullSafe())
            {
                CheckBreasts();
            }

            if(_bootyMagic.IsEnabledNullSafe())
            {
                CheckGlutes();
            }

            if(_tittyMagic.IsEnabledNullSafe() || _bootyMagic.IsEnabledNullSafe())
            {
                CheckCommon();

                if(_selectedCharacter != _person.Geometry.selectedCharacter)
                {
                    StartCoroutine(RefreshCharacterCo());
                }
            }

            CheckSoftPhysicsToggled();
            CheckPhysicsRateChanged();

            if(runCallbacks)
            {
                OnBreastInOutChanged(_breastInOutJsb);
            }
        }

        void CheckBreasts()
        {
            _person.SetBreastsUseGravity(false);
            if(_person.PectoralRbCollisionsEnabled(true))
            {
                _person.SetPectoralRbCollisions(false);
            }

            if(!_person.Geometry.useAuxBreastColliders)
            {
                _person.Geometry.useAuxBreastColliders = true;
                Loggr.Message("Breast Hard Colliders enabled.");
            }
        }

        void CheckGlutes()
        {
            _person.SetGlutesUseGravity(false);
        }

        void CheckCommon()
        {
            if(_person.IsFemale)
            {
                if(!_person.Geometry.useAdvancedColliders)
                {
                    _person.Geometry.useAdvancedColliders = true;
                    Loggr.Message("Advanced Colliders enabled - they are necessary for directional force morphing and hard colliders to work.");
                }
            }
        }

        void CheckSoftPhysicsToggled()
        {
            bool atomSoftPhysicsOn = _person.SoftBodyPhysicsEnabler.GetBoolParamValue("enabled");
            bool globalSoftPhysicsOn = UserPreferences.singleton.softPhysics;

            if(_bootyMagic.IsEnabledNullSafe())
            {
                bool gluteSoftPhysicsOn = _person.GlutePhysicsMesh.on;
                string location = LocationWhereStillDisabled(gluteSoftPhysicsOn, _gluteSoftPhysicsOn, atomSoftPhysicsOn, globalSoftPhysicsOn);
                if(!string.IsNullOrEmpty(location))
                {
                    Loggr.Message($"Soft physics is still disabled in {location}");
                }

                bool enabledNow = globalSoftPhysicsOn && atomSoftPhysicsOn && gluteSoftPhysicsOn;
                if(enabledNow != IsGluteSoftPhysicsOn())
                {
                    _bootyMagic.SetBoolParamValue(Constant.INTERNAL_SYNC_SOFT_PHYSICS, enabledNow);
                }

                _gluteSoftPhysicsOn = gluteSoftPhysicsOn;
            }

            if(_tittyMagic.IsEnabledNullSafe())
            {
                bool breastSoftPhysicsOn = _person.BreastPhysicsMesh.on;
                string location = LocationWhereStillDisabled(breastSoftPhysicsOn, _breastSoftPhysicsOn, atomSoftPhysicsOn, globalSoftPhysicsOn);
                if(!string.IsNullOrEmpty(location))
                {
                    Loggr.Message($"Soft physics is still disabled in {location}");
                }

                bool enabledNow = globalSoftPhysicsOn && atomSoftPhysicsOn && breastSoftPhysicsOn;
                if(enabledNow != IsBreastSoftPhysicsOn())
                {
                    _tittyMagic.SetBoolParamValue(Constant.INTERNAL_SYNC_SOFT_PHYSICS, enabledNow);
                }

                _breastSoftPhysicsOn = breastSoftPhysicsOn;
            }

            _atomSoftPhysicsOn = atomSoftPhysicsOn;
            _globalSoftPhysicsOn = globalSoftPhysicsOn;
        }

        void CheckPhysicsRateChanged()
        {
            float value = Time.fixedDeltaTime;
            if(Mathf.Abs(value - _fixedDeltaTime) > 0.001f)
            {
                _tittyMagic.CallActionNullSafe(Constant.INTERNAL_ON_PHYSICS_RATE_CHANGED);
                _bootyMagic.CallActionNullSafe(Constant.INTERNAL_ON_PHYSICS_RATE_CHANGED);
            }

            _fixedDeltaTime = value;
        }

        bool _inRefreshCharacterCo;

        IEnumerator RefreshCharacterCo()
        {
            if(_inRefreshCharacterCo)
            {
                yield break;
            }

            _inRefreshCharacterCo = true;
            while(!_person.IsSelectedCharacterReady())
            {
                yield return null;
            }

            if(_selectedCharacter.isMale == _person.Geometry.selectedCharacter.isMale)
            {
                yield return _person.RefreshSkinCo();
                _tittyMagic.CallActionNullSafe(Constant.INTERNAL_ON_CHARACTER_CHANGED);
                _bootyMagic.CallActionNullSafe(Constant.INTERNAL_ON_CHARACTER_CHANGED);
                _selectedCharacter = _person.Geometry.selectedCharacter;
            }
            else
            {
                Loggr.Message("Changing gender while the plugin is active is not supported. Disable the plugin, change the gender and then reload the plugin.");
                _person.Geometry.selectedCharacter = _selectedCharacter;
            }

            _inRefreshCharacterCo = false;
        }

        string LocationWhereStillDisabled(bool currEnabled, bool prevEnabled, bool atomSoftPhysicsOn, bool globalSoftPhysicsOn)
        {
            if(currEnabled && !prevEnabled)
            {
                if(!atomSoftPhysicsOn && !globalSoftPhysicsOn)
                {
                    return "Control & Physics 1 and in User Preferences";
                }

                if(!atomSoftPhysicsOn)
                {
                    return "Control & Physics 1";
                }

                if(!globalSoftPhysicsOn)
                {
                    return "User Preferences";
                }
            }
            else if(atomSoftPhysicsOn && !_atomSoftPhysicsOn)
            {
                if(!currEnabled && !globalSoftPhysicsOn)
                {
                    return "the plugin UI and in User Preferences";
                }

                if(!currEnabled)
                {
                    return "the plugin UI";
                }

                if(!globalSoftPhysicsOn)
                {
                    return "User Preferences";
                }
            }
            else if(globalSoftPhysicsOn && !_globalSoftPhysicsOn)
            {
                if(!currEnabled && !atomSoftPhysicsOn)
                {
                    return "the plugin UI and in Control & Physics 1";
                }

                if(!currEnabled)
                {
                    return "the plugin UI";
                }

                if(!atomSoftPhysicsOn)
                {
                    return "Control & Physics 1";
                }
            }

            return null;
        }

        void Update()
        {
            if(Time.frameCount % 60 == 0)
            {
                Check();
            }
        }

        void OnDestroy()
        {
            if(_breastInOutJsb != null)
            {
                _breastInOutJsb.setJSONCallbackFunction -= OnBreastInOutChanged;
            }
        }
    }
}
