using NAT.Common;
using NAT.Common.Models;
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.Store
{
    sealed class ClothingStore
    {
        const string UID_PREFIX = "clothing:";
        readonly DAZCharacterSelector _geometry;
        readonly bool _isFemale;
        internal Dictionary<string, ClothingItem> Items { get; } = new Dictionary<string, ClothingItem>();
        internal Dictionary<object, ClothingItem> ItemsByRef { get; } = new Dictionary<object, ClothingItem>();
        internal List<ClothingItem> EnabledItems { get; } = new List<ClothingItem>();
        readonly List<ClothingItem> _retryItems = new List<ClothingItem>();

        internal Action OnItemsReloadedHandlers { get; set; }
        internal Action<ClothingItem> OnItemToggledHandlers { get; set; }
        internal Action<ClothingItem, bool> OnItemAllowDetachToggledHandlers { get; set; }

        internal ClothingStore(Person person)
        {
            _geometry = person.Geometry;
            _isFemale = person.IsFemale;
            SyncItems();
            EnabledItems.AddRange(Items.Values.Where(item => item.EnabledJsb?.val ?? false));
        }

        internal void ReloadItems()
        {
            SyncItems();
            foreach(var pair in Items)
            {
                var item = pair.Value;
                if(item.EnabledJsb.val)
                {
                    if(!EnabledItems.Contains(item))
                    {
                        EnabledItems.Add(item);
                    }
                }
                else
                {
                    EnabledItems.Remove(item);
                }
            }

            OnItemsReloadedHandlers?.Invoke();
        }

        void SyncItems()
        {
            var duplicates = new Dictionary<string, int>();
            foreach(string name in _geometry.GetBoolParamNames())
            {
                if(!name.StartsWith(UID_PREFIX))
                {
                    continue;
                }

                string uid = name.Replace(UID_PREFIX, "");
                if(Items.ContainsKey(uid) || uid == "Clothing Creator")
                {
                    continue;
                }

                var dazItem = _geometry.GetClothingItem(uid);
                if(
                    (_isFemale && dazItem.gender != DAZDynamicItem.Gender.Female) ||
                    (!_isFemale && dazItem.gender != DAZDynamicItem.Gender.Male)
                )
                {
                    continue;
                }

                var enabledJsb = _geometry.GetBoolJSONParam(name);
                string label = BuildLabel(duplicates, dazItem);
                var item = new ClothingItem(dazItem, label, enabledJsb);

                enabledJsb.setJSONCallbackFunction += OnToggleItem;
                item.RefreshControls();
                AddOnItemAllowDetachHandler(item);
                Items[uid] = item;
            }

            _retryItems.Clear();
            var needRetry = Items
                .Where(kvp => kvp.Value.NeedsRetry)
                .Select(kvp => kvp.Value);
            _retryItems.AddRange(needRetry);
        }

        static string BuildLabel(Dictionary<string, int> duplicates, DAZClothingItem dazItem)
        {
            string label = $"{dazItem.displayName}";
            if(dazItem.IsInPackage)
            {
                label += $" [{dazItem.packageUid}]";
                return label;
            }

            label += $" [{dazItem.creatorName}]";
            int count;
            if(duplicates.ContainsKey(label))
            {
                duplicates[label]++;
                count = duplicates[label];
            }
            else
            {
                duplicates[label] = 1;
                count = 1;
            }

            return count > 1 ? $"{label} ({count})" : label;
        }

        void OnToggleItem(JSONStorableBool jsb)
        {
            try
            {
                string uid = jsb.name.Replace(UID_PREFIX, "");
                var item = Items[uid];
                if(item.EnabledJsb.val)
                {
                    EnabledItems.Add(item);
                    item.RefreshControls();
                    if(item.NeedsRetry)
                    {
                        _retryItems.Add(item);
                    }
                    else
                    {
                        AddOnItemAllowDetachHandler(item);
                    }
                }
                else
                {
                    EnabledItems.Remove(item);
                }

                OnItemToggledHandlers?.Invoke(item);
            }
            catch(Exception e)
            {
                Loggr.Error($"OnToggleItem: {e}");
            }
        }

        void AddOnItemAllowDetachHandler(ClothingItem item)
        {
            if(!item.SimControl)
            {
                return;
            }

            var jsb = item.SimControl.GetBoolJSONParam("allowDetach");
            ItemsByRef[jsb] = item;
            jsb.setJSONCallbackFunction -= OnToggleItemAllowDetach;
            jsb.setJSONCallbackFunction += OnToggleItemAllowDetach;
        }

        void OnToggleItemAllowDetach(JSONStorableBool jsb)
        {
            try
            {
                OnItemAllowDetachToggledHandlers?.Invoke(ItemsByRef[jsb], jsb.val);
            }
            catch(Exception e)
            {
                Loggr.Error($"OnToggleItemAllowDetach: {e}");
            }
        }

        internal void Execute()
        {
            if(Time.frameCount % 60 != 0)
            {
                return;
            }

            for(int i = _retryItems.Count - 1; i >= 0; i--)
            {
                var item = _retryItems[i];
                item.RefreshControls();
                if(!item.NeedsRetry)
                {
                    _retryItems.RemoveAt(i);
                    AddOnItemAllowDetachHandler(item);
                }
            }
        }

        internal void Destroy()
        {
            foreach(var pair in Items)
            {
                var item = pair.Value;
                if(item.EnabledJsb != null)
                {
                    item.EnabledJsb.setJSONCallbackFunction -= OnToggleItem;
                }

                if(item.SimControl != null)
                {
                    item.SimControl.GetBoolJSONParam("allowDetach").setJSONCallbackFunction -= OnToggleItemAllowDetach;
                }
            }
        }
    }
}
