using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using NAT;
using NAT.BM;
using NAT.BM.Handlers;
using NAT.BM.UI;
using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Storables;
using NAT.Common.UI;
using SimpleJSON;
using UnityEngine;

namespace Naturalis
{
    sealed class BootyMagic : ScriptBase
    {
        public override bool ShouldIgnore() => false;

        const string CALIBRATION_DATA_VERSION = "1.1.0";

        protected override void OnUIEnabled()
        {
            if(!Person.IsFemale)
            {
                return;
            }

            AddGravityControls();

            if(Tabs.IsMainTabActive())
            {
                if(!Tabs.IsNestedWindowActive())
                {
                    Tabs.MainWindow.Refresh();
                }
                else if(Tabs.GetActiveNestedWindow() is DevWindow)
                {
                    _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(true);
                }
            }
            else if(Tabs.IsHardCollidersTabActive())
            {
                if(!enabled)
                {
                    Tabs.GoToMainWindow();
                }

                if(_hardColliderHandler.ColliderVisualizer)
                {
                    _hardColliderHandler.ColliderVisualizer.ShowPreviewsJSON.val = true;
                }
            }
            else if(Tabs.IsPhysicsTabActive())
            {
                if(!Tabs.IsNestedWindowActive())
                {
                    if(_softPhysicsHandler.SoftPhysicsSynced)
                    {
                        Tabs.PhysicsWindow.Refresh();
                    }
                }
                else if(Tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(true);
                    var softParameterWindow = (SoftParameterWindow) Tabs.GetActiveNestedWindow();
                    softParameterWindow.Refresh();
                    softParameterWindow.SetShowColliderPreviews();
                }
            }
            else if(Tabs.IsClothingTabActive())
            {
                // TODO if soft physics section is active
                _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(true);
            }
        }

        protected override void OnUIDisabled()
        {
            if(!Person.IsFemale)
            {
                return;
            }

            _gravityPhysicsHandler.UnsetGravityDirection();

            if(Bootstrap.CalibrateOnUIClosedJsb.val && CalibrationStateHandler.Pending)
            {
                StartCalibration();
            }

            if(_hardColliderHandler.ColliderVisualizer)
            {
                _hardColliderHandler.ColliderVisualizer.ShowPreviewsJSON.val = false;
            }

            if(_softPhysicsHandler.ColliderVisualizer)
            {
                _softPhysicsHandler.ColliderVisualizer.ShowPreviewsJSON.val = false;
            }

            Tabs.ClosePopups();
            _softPhysicsHandler.SoftJointDiagnostics?.DrawServiceSetEnabled(false);

            if(Tabs.IsMainTabActive())
            {
                if(Tabs.GetActiveNestedWindow() is SettingsWindow)
                {
                    Tabs.MainWindow.OnReturn();
                }
            }
        }

        AngleCalculator _angleCalculator;
        GluteJointPhysicsHandler _jointPhysicsHandler;
        GluteSoftPhysicsHandler _softPhysicsHandler;
        GluteGravityPhysicsHandler _gravityPhysicsHandler;
        GluteForcePhysicsHandler _forcePhysicsHandler;
        FrictionHandler _frictionHandler;
        HardColliderHandler _hardColliderHandler;
        GluteForceMorphHandler _forceMorphHandler;
        HipClothingHandler _clothingHandler;
        MorphListener _gluteMorphListener;

        public override void Init()
        {
            base.Init(Constant.BOOTY_MAGIC, Colors.peachPink);
            StartCoroutine(SetupRoutine());
        }

        IEnumerator SetupRoutine()
        {
            yield return InitBootstrapCo();
            if(Bootstrap.Done == false)
            {
                Initialized = false;
                yield break;
            }

            Person = Bootstrap.Person;

            if(!Person.IsFemale)
            {
                var errorJss = new JSONStorableString("Info", $"<b>{Constant.BOOTY_MAGIC} is only supported on a female character.</b>");
                var textField = CreateTextField(errorJss);
                textField.backgroundColor = Color.clear;
                Initialized = false;
                yield break;
            }

            Provider = new Provider(this);

            /* Used to store version in save JSON and communicate version to other plugin instances */
            var versionJss = new StorableString(Constant.NATURALIS_VERSION, "")
            {
                val = Bootstrap.NATURALIS_VERSION,
            };
            PreRegisterInternalParam(versionJss, false);

            _angleCalculator = Provider.GetInstance<AngleCalculator>();
            _angleCalculator.Init(
                Person.LeftGlute.Rigidbody,
                Person.RightGlute.Rigidbody,
                Person.HipRb
            );
            SetupHandlers();
            SetupGluteMorphListener();
            SetupStorables();
            SetupCallbacks();
            SetupCalibrationHelper();

            Bindings = Bindings ?? new Bindings(this); // Might already be setup in OnBindingsListRequested.
            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);

            SetupUI();
            RegisterParams();

            if(!RestoringFromJSON)
            {
                PresetHandler.LoadDefaultPreset(Person.IsFemale, true);
                if(!InitializedFromDefaultPreset)
                {
                    _clothingHandler.AutoLoadDefaultProfiles();
                }

                StartCalibration(true);
            }

            Initialized = true;
        }

        IEnumerator InitBootstrapCo()
        {
            Bootstrap = gameObject.AddComponent<Bootstrap>();
            Bootstrap.Init(this);
            while(!Bootstrap || Bootstrap.Done == null)
            {
                yield return null;
            }
        }

        void SetupHandlers()
        {
            PresetHandler = new PluginPresetHandler(this, $"{DomainName}Presets");
            _jointPhysicsHandler = Provider.GetInstance<GluteJointPhysicsHandler>();
            _softPhysicsHandler = Provider.GetInstance<GluteSoftPhysicsHandler>();
            _gravityPhysicsHandler = Provider.GetInstance<GluteGravityPhysicsHandler>();
            gravityPhysicsHandler = _gravityPhysicsHandler;
            _forcePhysicsHandler = Provider.GetInstance<GluteForcePhysicsHandler>();
            _frictionHandler = Provider.GetInstance<FrictionHandler>();
            _hardColliderHandler = Provider.GetInstance<HardColliderHandler>();
            _forceMorphHandler = Provider.GetInstance<GluteForceMorphHandler>();
            _clothingHandler = Provider.GetInstance<HipClothingHandler>();

            foreach(Handler handler in Provider.GetHandlers())
            {
                handler.Init<BootyMagic>(Provider);
            }

            Bootstrap.ClothingStore.OnItemsReloadedHandlers += _clothingHandler.OnItemsReloaded;
            Bootstrap.ClothingStore.OnItemToggledHandlers += _clothingHandler.OnItemToggled;
            Bootstrap.ClothingStore.OnItemAllowDetachToggledHandlers += _clothingHandler.OnItemAllowDetachToggled;
        }

        void SetupGluteMorphListener()
        {
            var indices = _jointPhysicsHandler.leftVolumeCalculationVerts.Concat(_jointPhysicsHandler.rightVolumeCalculationVerts);
            _gluteMorphListener = new MorphListener(
                new HashSet<int>(indices),
                new HashSet<string>
                {
                    "FBMFitnessDetails",
                    "FBMHeight",
                    "NyssaH",
                    "Norma",
                    "Alessa",
                    "Eisa",
                    "Izarra",
                    "bellybulge",
                    "Athletic",
                    "Body Tone",
                    "Bodybuilder Details",
                    "Ribs Width",
                    "Distort 1",
                    "Distort 2",
                    "FBMBodyTone",
                    "FBMBodybuilderDetails",
                    "Center Gap Smooth",
                    "PBMSacralDimples",
                    "Glute Line Sharpness",
                    "G3F_ThighsInnerSlim",
                    "Inner Thigh Gap2",
                    "Inner Thigh Gap",
                    "!Bend Helper",
                    "Pubic Area Width",
                    "Centre Crease",
                    "WaistBellyIn",
                    "BackLessCurve",
                    "TeethFixer",
                    "Thigh inner contour",
                    "G3F_Hip Leg Space",
                    "Doggy Style Adjust",
                    "Glute crease",
                    "deepbulge_bot",
                }
            );

            if(Bootstrap.CheckForSizeChangesJsb.val)
            {
                ProcessMorphsForListener();
            }
        }

        void SetupStorables()
        {
            SoftnessJsf = new StorableFloat("Glute Softness", 70f, 0f, 100f)
            {
                setCallbackFunction = _ => SetCalibrationPending(SoftnessJsf),
            };
            QuicknessJsf = new StorableFloat("Glute Quickness Offset", 0.00f, -1.00f, 1.00f)
            {
                setCallbackFunction = _ => SetCalibrationPending(SoftnessJsf),
            };

            LeftRightAngleOffsetJsf = new StorableFloat("Left/Right Angle Offset", 0.0f, -30.0f, 30.0f)
            {
                setCallbackFunction = _ => SyncMorphsAngleOffsets(),
            };
            UpDownAngleOffsetJsf = new StorableFloat("Up/Down Angle Offset", 0.0f, -15.0f, 15.0f)
            {
                setCallbackFunction = _ => SyncMorphsAngleOffsets(),
            };
            DepthOffsetJsf = new StorableFloat("Depth Offset", 0.0f, -15.0f, 15.0f)
            {
                setCallbackFunction = _ => SyncMorphsAngleOffsets(),
            };
            SyncMorphsAngleOffsets();

            CalibrationAction = new StorableAction(Constant.CALIBRATE, () => StartCalibration(true));
            var calibrateOnNextPendingAction = new StorableAction(
                Constant.CALIBRATE_ON_NEXT_PENDING,
                () => CalibrationStateHandler.CalibrateOnNextPending = true
            );

            bool softPhysicsOn =
                UserPreferences.singleton.softPhysics &&
                containingAtom.GetStorableByID("SoftBodyPhysicsEnabler").GetBoolParamValue("enabled") &&
                _softPhysicsHandler.SoftPhysicsOnJsb.val;
            var syncSoftPhysicsJsb = new StorableBool(Constant.INTERNAL_SYNC_SOFT_PHYSICS, softPhysicsOn)
            {
                setCallbackFunction = OnSoftPhysicsToggled,
            };

            var onPhysicsRateChangedAction = new StorableAction(Constant.INTERNAL_ON_PHYSICS_RATE_CHANGED, () =>
            {
                if(!CalibrationStateHandler.Pending && !CalibrationStateHandler.CalibrateOnNextPending)
                {
                    Loggr.LogMessage("Physics rate changed - calibration set to pending.");
                }

                CalibrationStateHandler.SetCalibrationPending(Constant.PHYSICS_RATE_CHANGED);
            });

            var onCharacterChangedAction = new StorableAction(Constant.INTERNAL_ON_CHARACTER_CHANGED, () =>
            {
                _frictionHandler.Refresh();
                UIModManager.SetFrictionUIVisibility();
            });

            var unsetPendingJss = new StorableString(Constant.INTERNAL_UNSET_CALIBRATION_PENDING, "")
            {
                setJSONCallbackFunction = jss =>
                {
                    CalibrationStateHandler.SetCalibrationPending(jss.val, false);
                    jss.valNoCallback = "";
                },
            };

            var processMorphsAction = new StorableAction(Constant.INTERNAL_PROCESS_MORPHS, () =>
            {
                if(!_gluteMorphListener.ProcessingDone)
                {
                    StartCoroutine(ProcessMorphsForListenerCo());
                }
            });

            PreRegisterGeneralParam(SoftnessJsf);
            PreRegisterGeneralParam(QuicknessJsf);
            PreRegisterGeneralParam(CalibrationAction);
            PreRegisterGeneralParam(calibrateOnNextPendingAction);
            PreRegisterInternalParam(syncSoftPhysicsJsb);
            PreRegisterInternalParam(onPhysicsRateChangedAction);
            PreRegisterInternalParam(onCharacterChangedAction);
            PreRegisterInternalParam(unsetPendingJss);
            PreRegisterInternalParam(processMorphsAction);
        }

        IEnumerator ProcessMorphsForListenerCo()
        {
            Loggr.LogMessage("Processing all morphs to be able to detect changes to glute morphs. This could take a while if you have lots of morphs.");
            yield return null;
            ProcessMorphsForListener();
        }

        void ProcessMorphsForListener() => _gluteMorphListener.ProcessMorphs(Person.Geometry.morphBank1);

        internal override void SyncMorphsAngleOffsets(float? timer = null)
        {
            float x = LeftRightAngleOffsetJsf.val;
            float y = -UpDownAngleOffsetJsf.val + 5.0f;
            float z = -0.001f * DepthOffsetJsf.val;

            var leftMorphsOffsets = _clothingHandler.LeftMorphsOffsets + _clothingHandler.LeftMorphsOffsetsCompensation;
            var rightMorphsOffsets = _clothingHandler.RightMorphsOffsets + _clothingHandler.RightMorphsOffsetsCompensation;

            var offsetLeft = new Vector3(-x + leftMorphsOffsets.x, y + leftMorphsOffsets.y, z + leftMorphsOffsets.z);
            var offsetRight = new Vector3(x + rightMorphsOffsets.x, y + rightMorphsOffsets.y, z + rightMorphsOffsets.z);

            if(timer != null && timer.Value >= Time.deltaTime)
            {
                var transition = CoroutineUtils.ExponentialVector3ArrayTransition(
                    timer.Value,
                    new Vector3TransitionParams(leftMeasurementOffset, offsetLeft, value => leftMeasurementOffset = value),
                    new Vector3TransitionParams(rightMeasurementOffset, offsetRight, value => rightMeasurementOffset = value)
                );
                StartCoroutine(transition);
            }
            else
            {
                leftMeasurementOffset = offsetLeft;
                rightMeasurementOffset = offsetRight;
            }
        }

        void SetupCallbacks()
        {
            Person.ScaleJsf.setCallbackFunction += OnAtomScaleChanged;
            SuperController.singleton.onBeforeSceneSaveHandlers += OnBeforeSceneSave;
            SuperController.singleton.onSceneSavedHandlers += OnSceneSaved;
        }

        void SetupCalibrationHelper()
        {
            CalibrationStateHandler = gameObject.AddComponent<CalibrationStateHandler>();
            CalibrationStateHandler.Init(this, () => StartCalibration());
            PreRegisterInternalParam(CalibrationStateHandler.CalibratingJsb);
        }

        // https://github.com/vam-community/vam-plugins-interop-specs/blob/main/keybindings.md
        public void OnBindingsListRequested(List<object> bindingsList)
        {
            Bindings = Bindings ?? new Bindings(this); // Might already be setup in Init.
            bindingsList.Add(Bindings.NamespaceDict);
            bindingsList.AddRange(Bindings.GetActionsList());
        }

        void SetupUI()
        {
            var uiModManager = gameObject.AddComponent<UIModManager>();
            uiModManager.Init(Provider);
            UIModManager = uiModManager;

            Tabs = new Tabs(this)
            {
                MainWindow = new MainWindow(this),
                PhysicsWindow = new PhysicsWindow(this),
                MorphsWindow = new MorphsWindow(this),
                HardCollidersWindow = new HardCollidersWindow(this),
                ClothingWindow = new ClothingWindow(this),
            };
        }

        void Update()
        {
            if(Initialized != true)
            {
                return;
            }

            try
            {
                if(!isUpdatePaused)
                {
                    if(Bootstrap.CheckForSizeChangesJsb.val)
                    {
                        if(_gluteMorphListener.ChangeDetected())
                        {
                            CalibrationStateHandler.SetCalibrationPending(Constant.SIZE_CHANGED, _gluteMorphListener.IsChanged);
                        }
                    }

                    if(_softPhysicsHandler.SoftJointDiagnostics?.IsDrawingClothingCoverage ?? false)
                    {
                        _clothingHandler.HandleToggleCoverageJoint();
                    }

                    if(pluginUIEventsListener.IsEnabled)
                    {
                        Tabs.GetActiveWindow()?.UpdateDiagnostics();
                    }
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"Update error: {e}");
            }
        }

        void FixedUpdate()
        {
            if(Initialized != true ||
                RestoringFromJSON ||
                !Person.Skin ||
                Person.IsFreezeGrabbing ||
                Bootstrap.IsGrabbingSubscene ||
                !SuperController.singleton.autoSimulation
            )
            {
                return;
            }

            if(isUpdatePaused)
            {
                if(inCalibrationCo && _gravityPhysicsHandler.IsSimulatingGravity)
                {
                    SimulateGravity();
                }

                return;
            }

            try
            {
                Person.LeftGlute.Measure(leftMeasurementOffset);
                Person.RightGlute.Measure(rightMeasurementOffset);

                if(_gravityPhysicsHandler.IsSimulatingGravity)
                {
                    SimulateGravity();
                }
                else
                {
                    _angleCalculator.CalculateAngles();
                    _gravityPhysicsHandler.Execute(_softPhysicsHandler.SoftPhysicsOn);

                    if(!Bootstrap.ForceUpdatesPaused())
                    {
                        _forcePhysicsHandler.Execute();
                        _forceMorphHandler.Execute();
                    }
                }

                _hardColliderHandler.UpdateDistanceDiffs();
                _hardColliderHandler.UpdateFriction(_frictionHandler.MaxHardColliderFriction);
                _clothingHandler.Execute();
                _jointPhysicsHandler.Execute();
                _softPhysicsHandler.Execute();
            }
            catch(Exception e)
            {
                Loggr.LogError($"FixedUpdate error: {e}");
                isUpdatePaused = true;
            }
        }

        bool _updatingMass;

        void SimulateGravity()
        {
            string direction = _gravityPhysicsHandler.GravitySimulationDirection;
            _angleCalculator.OverrideAngles(direction);
            _gravityPhysicsHandler.AddGravitySimulationForce(!inCalibrationCo);

            if(!_updatingMass)
            {
                if(inCalibrationCo && direction != Direction.DOWN)
                {
                    Person.LeftGlute.Measure(Vector3.zero);
                    Person.RightGlute.Measure(Vector3.zero);
                }

                _gravityPhysicsHandler.Execute(_softPhysicsHandler.SoftPhysicsOn);
                _forcePhysicsHandler.Execute();
                _forceMorphHandler.Execute();
            }
        }

        protected override IEnumerator CalibrationCo(JSONClass calibrationData = null)
        {
            inCalibrationCo = true;
            isUpdatePaused = true;
            _gravityPhysicsHandler.UnsetGravityDirection();
            /* Ensure everything is ready for calibration to proceed in case of e.g. appearance preset load */
            {
                /* Ensures glute soft physics enabled value is correct on first calibration when soft physics enabled is synced to SoftPhysicsHandler */
                while(!Bootstrap.PersonMonitor.Initialized)
                {
                    yield return null;
                }

                /* Can occur when loading a look and VaM pauses to load assets */
                while(!Person.IsSelectedCharacterReady())
                {
                    yield return null;
                }

                /* Must be unfrozen before checking for hard collider rigidbodies */
                yield return CalibrationStateHandler.WaitForAtomPhysicsUnfrozen();

                Bootstrap.PersonMonitor.Check(true);
                Bootstrap.PersonMonitor.enabled = false;

                ((GluteColliderHandler) _hardColliderHandler).ConfigureAutoColliders();

                /* Sanity check. Might occur (?) even after hard colliders are enabled by settings monitor */
                float timeout = Time.unscaledTime + 5;
                while(!_hardColliderHandler.RigidbodiesFound() && Time.unscaledTime < timeout)
                {
                    yield return null;
                }

                if(!_hardColliderHandler.RigidbodiesFound())
                {
                    Loggr.LogError("Calibration error: collider rigidbodies not found after 5 seconds of waiting");
                    yield break;
                }
            }

            bool doFullCalibration =
                calibrationData == null ||
                !calibrationData.HasKey("mass") ||
                !calibrationData.HasKey("left_glute") ||
                !calibrationData.HasKey("right_glute") ||
                !calibrationData.HasKey("soft_joints") ||
                !calibrationData.HasKey("hard_colliders");

            if(!doFullCalibration)
            {
                try
                {
                    CalibrateFromJSON(calibrationData);
                }
                catch(Exception e)
                {
                    doFullCalibration = true;
                    Loggr.LogError($"Error loading calibration data from JSON, continuing with full calibration. Error: {e}");
                }
            }

            if(doFullCalibration)
            {
                yield return FullCalibrationCo();
            }

            SaveCalibratedValues();
            CalibrationStateHandler.UpdateMassWhenCalibrating = false;
            isUpdatePaused = false;
            inCalibrationCo = false;
            FirstCalibrationDone = true;

            /* Update UI */
            if(Tabs.IsMainTabActive())
            {
                if(!Tabs.IsNestedWindowActive())
                {
                    Tabs.MainWindow.Refresh();
                }
            }
            else if(Tabs.IsPhysicsTabActive())
            {
                if(!Tabs.IsNestedWindowActive())
                {
                    Tabs.PhysicsWindow.Refresh();
                }
                else if(Tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    ((SoftParameterWindow) Tabs.GetActiveNestedWindow()).Refresh();
                }
            }
            else if(Tabs.IsClothingTabActive())
            {
                Tabs.ClothingWindow.Refresh();
            }
        }

        void CalibrateFromJSON(JSONClass jc)
        {
            _softPhysicsHandler.SyncSoftPhysicsOn();
            _jointPhysicsHandler.UpdateMassValueAndAmounts(jc["mass"].AsFloat);

            Softness = CalculateSoftness();
            Quickness = QuicknessJsf.val;
            SetMultipliersAndUpdatePhysics();

            Person.LeftGlute.CalibrateFromJSON(jc["left_glute"].AsObject);
            Person.RightGlute.CalibrateFromJSON(jc["right_glute"].AsObject);
            _softPhysicsHandler.CalibrateFromJSON(jc["soft_joints"].AsObject);

            _hardColliderHandler.SyncCollidersMass();
            _hardColliderHandler.CalibrateFromJSON(jc["hard_colliders"].AsObject);
            _hardColliderHandler.SyncAllOffsets();

            _clothingHandler.SetAllAdjustmentValues();
            _clothingHandler.SyncAdjustments();
            SyncMorphsAngleOffsets();

            CalibrationStateHandler.EnableSettingsMonitor();
        }

        IEnumerator FullCalibrationCo()
        {
            /* Must be called before calibration pose is applied because the current pose must be stored without duplicate morphs having >0 values. */
            ZeroDynamicAdjustments();
            Person.ZeroDuplicateBootyMagicMorphs();

            yield return CalibrationStateHandler.Begin();
            if(CalibrationStateHandler.Cancelling)
            {
                CalibrationStateHandler.Cancelling = false;
                yield break;
            }

            _softPhysicsHandler.SyncSoftPhysicsOn();

            {
                _gravityPhysicsHandler.EnableGravitySimulation(Direction.DOWN);
                yield return UpdateMassAndPhysicsCo();
                _jointPhysicsHandler.SetGravitySimulationDamper();
                yield return new WaitForSeconds(0.25f);

                float minTime = Time.time + 0.40f;
                float maxTime = Time.time + 1.00f;
                const float tolerance = 0.003f;
                while(
                    Time.time < minTime ||
                    (
                        (Person.LeftGlute.Moved(tolerance) || Person.RightGlute.Moved(tolerance)) &&
                        Time.time < maxTime
                    )
                )
                {
                    yield return null;
                }

                Person.LeftGlute.Calibrate();
                Person.RightGlute.Calibrate();

                if(_softPhysicsHandler.IsLdpUpdateOn())
                {
                    _softPhysicsHandler.CalibrateSoftJointsDistortion(Direction.DOWN);
                }

                _hardColliderHandler.SyncCollidersMass();
                _hardColliderHandler.CalibrateColliders();
                _hardColliderHandler.SyncAllOffsets();
            }

            if(_softPhysicsHandler.IsLdpUpdateOn())
            {
                foreach(string direction in new[] { Direction.FORWARD, Direction.UP, Direction.LEFT, Direction.RIGHT, Direction.BACK })
                {
                    _gravityPhysicsHandler.EnableGravitySimulation(direction);
                    yield return new WaitForSeconds(0.25f);

                    float timeout = Time.time + 1.50f;
                    while(_softPhysicsHandler.AnySoftJointMoved() && Time.time < timeout)
                    {
                        yield return null;
                    }

                    _softPhysicsHandler.CalibrateSoftJointsDistortion(direction);
                }
            }

            _clothingHandler.SetAllAdjustmentValues();
            _clothingHandler.SyncAdjustments();
            SyncMorphsAngleOffsets();

            _gravityPhysicsHandler.StopGravitySimulation();
            _jointPhysicsHandler.UnsetGravitySimulationDamper();
            yield return CalibrationStateHandler.Finish();
        }

        void ZeroDynamicAdjustments()
        {
            Person.LeftGlute.ResetMeasurements();
            Person.RightGlute.ResetMeasurements();
            _softPhysicsHandler.ResetSoftJointsDistortion();
            _softPhysicsHandler.ResetSoftJointsClothingAdjustments();
            _hardColliderHandler.ResetDistanceDiffs();
            _gravityPhysicsHandler.ResetAll();
            _forcePhysicsHandler.ResetAll();
            _forceMorphHandler.ResetAll();
            _clothingHandler.ResetAll();
        }

        float CalculateSoftness() => Calc.SoftnessBaseCurve(SoftnessJsf.val / 100f);

        IEnumerator UpdateMassAndPhysicsCo()
        {
            _updatingMass = true;
            if(CalibrationStateHandler.UpdateMassWhenCalibrating)
            {
                Softness = 0;
                Quickness = 0;
                _jointPhysicsHandler.UpdateBaseValues(false);
                _softPhysicsHandler.Execute(_jointPhysicsHandler.NormalizedRealMass, _jointPhysicsHandler.NormalizedMass, false);

                yield return new WaitForSeconds(0.25f);
                /* Mass is calculated multiple times because each new mass value changes the exact glute
                 * shape and therefore the estimated volume.
                 */
                _jointPhysicsHandler.CalculateMass();
                yield return Utils.WaitAndRepeat(
                    () => _jointPhysicsHandler.CalculateMass(),
                    (int) Mathf.Lerp(4, 12, _jointPhysicsHandler.NormalizedRealMass),
                    0.13f
                );
            }

            Softness = CalculateSoftness();
            Quickness = QuicknessJsf.val;
            SetMultipliersAndUpdatePhysics();

            _updatingMass = false;
        }

        void SetMultipliersAndUpdatePhysics()
        {
            float normalizedRealMass = _jointPhysicsHandler.NormalizedRealMass;
            _gravityPhysicsHandler.SetMultipliers(normalizedRealMass, Softness);
            _forceMorphHandler.SetMultipliers(normalizedRealMass, Softness);
            _forcePhysicsHandler.SetMultipliers(normalizedRealMass, Softness);
            _hardColliderHandler.UpdateFrictionSizeMultipliers();
            _clothingHandler.SetMultipliers(normalizedRealMass);

            _jointPhysicsHandler.UpdateBaseValues();
            _softPhysicsHandler.Execute(normalizedRealMass, _jointPhysicsHandler.NormalizedMass);
            _frictionHandler.CalculateFriction();
        }

        void SaveCalibratedValues()
        {
            SoftnessJsf.SaveCalibrated();
            QuicknessJsf.SaveCalibrated();
            _gluteMorphListener.SaveCalibrated();
            _forceMorphHandler.SaveCalibrated();
            _jointPhysicsHandler.SaveCalibrated();
            _gravityPhysicsHandler.SaveCalibrated();
            _softPhysicsHandler.SaveCalibrated();
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            // Is set to "" for some reason, even though initialization sets it to the correct value
            SetStringParamValue(Constant.NATURALIS_VERSION, Bootstrap.NATURALIS_VERSION);
            var jc = base.GetJSON(includePhysical, includeAppearance, forceStore);

            if(Person.IsFemale)
            {
                if(_isSavingScene && FirstCalibrationDone && !CalibrationStateHandler.Pending)
                {
                    jc["stored_in_scene"].AsBool = true; // Unused, might be useful later
                    var calibrationData = new JSONClass
                    {
                        ["version"] = CALIBRATION_DATA_VERSION,
                        ["mass"] = { AsFloat = _jointPhysicsHandler.CalibratedMass },
                        ["left_glute"] = Person.LeftGlute.SerializeCalibrationData(),
                        ["right_glute"] = Person.RightGlute.SerializeCalibrationData(),
                        ["soft_joints"] = _softPhysicsHandler.SerializeCalibrationData(),
                        ["hard_colliders"] = _hardColliderHandler.SerializeCalibrationData(),
                    };
                    jc["calibration_data"] = calibrationData;
                }

                var activeClothingProfiles = new JSONArray();
                foreach(var profile in _clothingHandler.ActiveProfiles)
                {
                    if(!string.IsNullOrEmpty(profile.LoadedFullPath))
                    {
                        activeClothingProfiles.Add(new JSONClass
                        {
                            // Hacky way to ensure that path is updated to SELF:/ by package manager when packaging the scene with the active preset files in the same var
                            ["presetName"] = profile.LoadedFullPath,
                        });
                    }
                }

                jc["Active Clothing Profiles"] = activeClothingProfiles;
            }

            needsStore = true;
            return jc;
        }

        public override void RestoreFromJSON(
            JSONClass jc,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            JSONArray presetAtoms = null,
            bool setMissingToDefault = true
        )
        {
            RestoringFromJSON = true;

            /* Disable early to allow correct enabled value to be used during Init */
            if(jc.HasKey("enabled") && !jc["enabled"].AsBool)
            {
                enabled = false;
            }

            /* Prevent overriding versionJss.val from JSON. Version stored in JSON just for information,
             * but could be intercepted here and used to save a "loadedFromVersion" value.
             */
            if(jc.HasKey(Constant.NATURALIS_VERSION))
            {
                jc[Constant.NATURALIS_VERSION] = Bootstrap.NATURALIS_VERSION;
            }

            /* Naturalis v1.0.0 -> v1.0.1 fixes TODO remove? v1.1 .cs migration should handle */
            JSONUtils.RenameKey(jc, "In/Out Damper Back Gravity", "In/Out Spring Back Gravity");
            JSONUtils.RenameKey(jc, "In/Out Damper Forward Gravity", "In/Out Spring Forward Gravity");

            StartCoroutine(
                RestoreFromJSONCo(
                    jc,
                    restorePhysical,
                    restoreAppearance,
                    presetAtoms,
                    setMissingToDefault
                )
            );
        }

        IEnumerator RestoreFromJSONCo(
            JSONClass jc,
            bool restorePhysical,
            bool restoreAppearance,
            JSONArray presetAtoms,
            bool setMissingToDefault
        )
        {
            while(Initialized == null)
            {
                yield return null;
            }

            if(Initialized == false)
            {
                yield break;
            }

            ConvertColliderParamNames(jc);
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            RestoringFromJSON = false;

            if(enabled)
            {
                JSONClass calibrationData = null;
                string scenePackageId = null;
                if(jc.HasKey("calibration_data"))
                {
                    #if VAM_GT_1_21
                    {
                        scenePackageId = FileUtils.ParsePackageIdFromPath(SuperController.singleton.LoadedSceneName);
                    }
                    #endif
                    var tmpData = jc["calibration_data"].AsObject;
                    if(
                        tmpData.HasKey("version") &&
                        new SemVer(tmpData["version"].Value).IsMajorMinorEqual(CALIBRATION_DATA_VERSION)
                    )
                    {
                        calibrationData = tmpData;
                    }
                }

                // ReSharper disable once ExpressionIsAlwaysNull
                _clothingHandler.AutoLoadProfiles(JSONUtils.GetJSONArrayEnumerable(jc, "Active Clothing Profiles"), scenePackageId);
                StartCalibration(true, calibrationData);
            }
        }

        internal override void RestoreFromPresetJSON(JSONClass jc, string fullPath)
        {
            if(Initialized != true || RestoringFromJSON)
            {
                Loggr.LogMessage("Preset file must be loaded after initialization.");
                return;
            }

            jc["enabled"].AsBool = enabled;
            ConvertColliderParamNames(jc);
            base.RestoreFromJSON(jc);
            string packageId = FileUtils.ParsePackageIdFromPath(fullPath);
            _clothingHandler.AutoLoadProfiles(JSONUtils.GetJSONArrayEnumerable(jc, "Active Clothing Profiles"), packageId);
            SetLoadedPresetName(fullPath);

            if(Bootstrap.CalibrateOnPresetLoadedJsb.val)
            {
                StartCalibration(true);
            }
        }

        internal override void InitializeFromPresetJSON(JSONClass jc, string fullPath)
        {
            jc["enabled"].AsBool = enabled;
            base.RestoreFromJSON(jc);
            string packageId = FileUtils.ParsePackageIdFromPath(fullPath);
            _clothingHandler.AutoLoadProfiles(JSONUtils.GetJSONArrayEnumerable(jc, "Active Clothing Profiles"), packageId);
            SetLoadedPresetName(fullPath);
            InitializedFromDefaultPreset = true;
        }

        /* Naturalis v1.0 beta 4 and earlier -> Naturalis v1.0 */
        static void ConvertColliderParamNames(JSONClass jc)
        {
            string[] colliderIds =
            {
                "Glute1",
                "Glute1 (1)",
                "Glute1 (2)",
                "Glute1 (3)",
                "Glute1 (4)",
                "Glute1 (5)",
                "Glute1 (6)",
                "Glute1 (7)",
            };

            string[] paramNames =
            {
                Constant.MASS_MULTIPLIER,
                Constant.RADIUS_OFFSET,
                Constant.X_OFFSET,
                Constant.Y_OFFSET,
                Constant.Z_OFFSET,
            };

            foreach(string colliderId in colliderIds)
            {
                foreach(string paramName in paramNames)
                {
                    string key = colliderId.ToLower() + paramName;
                    if(jc.HasKey(key))
                    {
                        float value = jc[key].AsFloat;
                        jc.Remove(key);
                        jc[$"{colliderId} {paramName}"].AsFloat = value;
                    }
                }
            }
        }

        void OnAtomScaleChanged(float _)
        {
            if(Initialized != true || !enabled)
            {
                return;
            }

            if(!CalibrationStateHandler.Pending && !CalibrationStateHandler.CalibrateOnNextPending)
            {
                Loggr.LogMessage("Person atom scale changed - calibration set to pending.");
            }

            CalibrationStateHandler.SetCalibrationPending(Constant.SIZE_CHANGED);
        }

        bool _isSavingScene;

        void OnBeforeSceneSave()
        {
            if(Initialized != true || !enabled)
            {
                return;
            }

            try
            {
                _isSavingScene = true;
                isUpdatePaused = true;
                _forceMorphHandler.ResetAll();
                _clothingHandler.ResetMorphs();
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnBeforeSceneSave error: {e}");
            }
        }

        void OnSceneSaved()
        {
            if(Initialized != true || !enabled)
            {
                return;
            }

            try
            {
                _isSavingScene = false;
                if(!CalibrationStateHandler.CalibratingJsb.val)
                {
                    isUpdatePaused = false;
                    _clothingHandler.SyncMorphsOffsetsCompensation();
                    SyncMorphsAngleOffsets();
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnSceneSaved error: {e}");
            }
        }

        void OnSoftPhysicsToggled(bool value)
        {
            _softPhysicsHandler.ReverseSyncSoftPhysicsOn();
            if(!CalibrationStateHandler.CalibrateOnNextPending)
            {
                Loggr.LogMessage($"Soft physics {(value ? "enabled" : "disabled")} - updates paused, calibration set to pending.");
            }

            isUpdatePaused = true;
            CalibrationStateHandler.SetCalibrationPending(Constant.SOFT_PHYSICS_CHANGED);
            _softPhysicsHandler.SoftPhysicsSynced = false;

            if(Tabs.IsPhysicsTabActive())
            {
                if(Tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    if(!value)
                    {
                        Tabs.PhysicsWindow.OnReturn();
                    }
                }
                else if(!value)
                {
                    Tabs.PhysicsWindow.Refresh();
                }
            }

            if(!value)
            {
                _softPhysicsHandler.SoftJointDiagnostics.VisualizeLdpJsb.val = false;
            }
        }

        void OnEnable()
        {
            if(Initialized != true)
            {
                return;
            }

            try
            {
                Bootstrap.PersonMonitor.Check(true);
                Person.EnableUseColliders();
                _hardColliderHandler.EnableMultiplyFriction();
                if(_hardColliderHandler.ColliderVisualizer)
                {
                    _hardColliderHandler.ColliderVisualizer.enabled = true;
                }

                _softPhysicsHandler.EnableMultiplyFriction();
                if(_softPhysicsHandler.ColliderVisualizer)
                {
                    _softPhysicsHandler.ColliderVisualizer.enabled = true;
                }

                _softPhysicsHandler.SoftJointDiagnostics?.DrawServiceSetEnabled(true);
                StartCalibration(true);
                UIModManager.enabled = true;
                Person.SetGlutesUseGravity(false);
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnEnable error: {e}");
            }
        }

        void OnDisable()
        {
            if(Initialized != true)
            {
                return;
            }

            try
            {
                if(calibrationCo != null)
                {
                    StopCoroutine(calibrationCo);
                }

                _hardColliderHandler.RestoreOriginalPhysics();
                if(_hardColliderHandler.ColliderVisualizer)
                {
                    _hardColliderHandler.ColliderVisualizer.enabled = false;
                }

                _jointPhysicsHandler.RestoreOriginalPhysics();
                _softPhysicsHandler.RestoreOriginalPhysics();
                if(_softPhysicsHandler.ColliderVisualizer)
                {
                    _softPhysicsHandler.ColliderVisualizer.enabled = false;
                }

                _softPhysicsHandler.SoftJointDiagnostics?.DrawServiceSetEnabled(false);
                UIModManager.enabled = false;

                CalibrationStateHandler.Finish(Bootstrap.TittyMagicStorable.IsEnabledNullSafe());
                _forceMorphHandler.ResetAll();
                _clothingHandler.ResetMorphs();
                Person.SetGlutesUseGravity(true);
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnDisable error: {e}");
            }
        }

        new void OnDestroy()
        {
            try
            {
                base.OnDestroy();
                if(Person != null)
                {
                    Person.ScaleJsf.setCallbackFunction -= OnAtomScaleChanged;
                }

                _frictionHandler?.Destroy();
                _hardColliderHandler?.Destroy();
                _softPhysicsHandler?.Destroy();

                Destroy(Bootstrap);
                Bootstrap = null;

                SuperController.singleton.onSceneSavedHandlers -= OnSceneSaved;
                SuperController.singleton.onBeforeSceneSaveHandlers -= OnBeforeSceneSave;
                SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
            }
            catch(Exception e)
            {
                if(Initialized == true)
                {
                    SuperController.LogError($"OnDestroy error: {e}");
                }
                else
                {
                    Debug.LogError($"OnDestroy error: {e}");
                }
            }
        }
    }
}
