using ColliderEditor;
using NAT.Common;
using NAT.Common.Components;
using NAT.Common.Handlers;
using NAT.Common.Models;
using NAT.Common.Storables;
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.BM.Handlers
{
    sealed class GluteColliderHandler : HardColliderHandler
    {
        GluteJointPhysicsHandler _jointPhysicsHandler;
        List<AutoColliderPair> _autoColliderPairs;

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            relativeToRigidbody = Person.HipRb;
            _jointPhysicsHandler = provider.GetInstance<GluteJointPhysicsHandler>();

            ConfigureAutoColliders();
            InitColliderVisualizer(); // depends on alt colliders being enabled first
            LoadSettings();
            EnableMultiplyFriction();

            var options = HardColliderGroups.Select(c => c.VisualizerId).ToList();
            var displayOptions = HardColliderGroups.Select(c => c.Id).ToList();
            ColliderGroupsJssc = new JSONStorableStringChooser(
                "colliderGroup",
                options,
                displayOptions,
                "",
                "Collider"
            )
            {
                setCallbackFunction = value => ColliderVisualizer.SelectablesJSON.val = value,
                val = options[0],
            };

            ColliderVisualizer.GroupsJSON.setJSONCallbackFunction = jssc =>
            {
                ColliderVisualizer.SelectedPreviewOpacityJSON.val = jssc.val == "Off" ? 0 : 1;
                if(jssc.val != "Off")
                {
                    ColliderVisualizer.SelectablesJSON.val = ColliderGroupsJssc.val;
                }
            };
            BaseMassMultiplierJsf = new StorableFloat($"Hard Colliders Base {Constant.MASS_MULTIPLIER}", 1.00f, 0.01f, 2.00f)
            {
                setCallbackFunction = _ => SyncCollidersMass(),
            };
            HighlightAllJsb = new JSONStorableBool(
                "Highlight All Colliders",
                false,
                value => ColliderVisualizer.PreviewOpacityJSON.val = value ? 1.00f : 0.67f
            );
        }

        void InitColliderVisualizer()
        {
            ColliderVisualizer = Script.gameObject.AddComponent<ColliderVisualizer>();
            var groups = new List<Group>
            {
                new Group("Off", @"$off"), //match nothing
                new Group("Both glutes", @"^((AutoCollider)?FemaleAutoColliders)?[LR]Glute1"),
                new Group("Left glute", @"^((AutoCollider)?FemaleAutoColliders)?LGlute1"),
            };
            ColliderVisualizer.Init(Script, groups);
            ColliderVisualizer.PreviewOpacityJSON.val = 0.67f;
            ColliderVisualizer.PreviewOpacityJSON.defaultVal = 0.67f;
            ColliderVisualizer.SelectedPreviewOpacityJSON.val = 1;
            ColliderVisualizer.SelectedPreviewOpacityJSON.defaultVal = 1;
            ColliderVisualizer.GroupsJSON.val = "Left glute";
            ColliderVisualizer.GroupsJSON.defaultVal = "Left glute";
            ColliderVisualizer.HighlightMirrorJSON.val = true;

            foreach(string option in new[] { Constant.SELECT, "Other", "All" })
            {
                ColliderVisualizer.GroupsJSON.choices.Remove(option);
            }
        }

        protected override void LoadSettings()
        {
            HardColliderGroups = _autoColliderPairs.Select(pair => NewHardColliderGroup(pair.Left, pair.Right)).ToList();
        }

        internal void ConfigureAutoColliders()
        {
            const string nameStart = "FemaleAutoColliders";
            var lGlute = Person.AutoColliderGroups.First(group => group.name == $"{nameStart}LGlute");
            var rGlute = Person.AutoColliderGroups.First(group => group.name == $"{nameStart}RGlute");
            var lGluteAlt = Person.AutoColliderGroups.First(group => group.name == $"{nameStart}LGluteAlt");
            var rGluteAlt = Person.AutoColliderGroups.First(group => group.name == $"{nameStart}RGluteAlt");

            lGlute.enabled = false;
            rGlute.enabled = false;
            lGluteAlt.enabled = true;
            rGluteAlt.enabled = true;

            var leftAutoColliders = lGluteAlt.GetAutoColliders();
            var rightAutoColliders = rGluteAlt.GetAutoColliders();

            IgnoreSoftColliders(leftAutoColliders);
            IgnoreSoftColliders(rightAutoColliders);

            if(_autoColliderPairs == null)
            {
                _autoColliderPairs = new List<AutoColliderPair>();
                for(int i = 0; i < leftAutoColliders.Length; i++)
                {
                    _autoColliderPairs.Add(new AutoColliderPair { Left = leftAutoColliders[i], Right = rightAutoColliders[i] });
                }
            }
        }

        void IgnoreSoftColliders(AutoCollider[] autoColliders)
        {
            foreach(var softVerticesGroup in Person.GlutePhysicsMesh.softVerticesGroups)
            {
                foreach(var softVerticesSet in softVerticesGroup.softVerticesSets)
                {
                    foreach(var ac in autoColliders)
                    {
                        Physics.IgnoreCollision(softVerticesSet.jointCollider, ac.jointCollider);
                        if(softVerticesSet.jointCollider2)
                        {
                            Physics.IgnoreCollision(softVerticesSet.jointCollider2, ac.jointCollider);
                        }
                    }
                }
            }
        }

        static Dictionary<string, Scaler> GetScalingConfig(string id)
        {
            switch(id)
            {
                case "Glute1":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = 0.04f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = -0.36f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.05f,
                                Max = 40,
                            }
                        },
                    };
                case "Glute1 (1)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                    };
                case "Glute1 (2)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = 0.57f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = -1.06f,
                                Max = 40,
                                MassCurve = x => 0.12f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.47f,
                                Max = 40,
                                MassCurve = x => -0.14f * x,
                            }
                        },
                    };
                case "Glute1 (3)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = -0.67f,
                                Max = 40,
                                MassCurve = x => 0.43f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.57f,
                                Max = 40,
                                MassCurve = x => -0.60f * x,
                            }
                        },
                    };
                case "Glute1 (4)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                                MassCurve = x => -0.21f * x,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = -0.18f,
                                Max = 40,
                                MassCurve = x => -0.13f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.06f,
                                Max = 40,
                                MassCurve = x => -0.37f * x,
                            }
                        },
                    };
                case "Glute1 (5)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                                MassCurve = x => -0.17f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                                MassCurve = x => -0.12f * x,
                            }
                        },
                    };
                case "Glute1 (6)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                                MassCurve = x => -0.16f * x,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                                MassCurve = x => 0.04f * x,
                            }
                        },
                    };
                case "Glute1 (7)":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = 0.00f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = -0.04f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = -0.82f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.23f,
                                Max = 40,
                                MassCurve = x => -0.15f * x,
                            }
                        },
                    };
                default: throw new ArgumentException($"Invalid id {id}");
            }
        }

        HardColliderGroup NewHardColliderGroup(AutoCollider left, AutoCollider right)
        {
            string id = left.name.Replace("AutoColliderFemaleAutoCollidersL", ""); // -> Glute1, Glute1 (1), Glute1 (2) etc.
            var leftCollider = new HardCollider(left);
            var rightCollider = new HardCollider(right);
            leftCollider.Collider.GetComponent<SphereCollider>().enabled = true;
            rightCollider.Collider.GetComponent<SphereCollider>().enabled = true;
            var scalingConfig = GetScalingConfig(id);
            string visualizerId = ColliderVisualizer.SelectablesJSON.choices.Find(option => option.EndsWith($"L{id}"));
            var group = new HardColliderGroup(
                id,
                visualizerId,
                leftCollider,
                rightCollider,
                scalingConfig[Constant.RADIUS_OFFSET],
                scalingConfig[Constant.X_OFFSET],
                scalingConfig[Constant.Y_OFFSET],
                scalingConfig[Constant.Z_OFFSET],
                1.0f // TODO
            )
            {
                MassMultiplierJsf = new StorableFloat($"{id} {Constant.MASS_MULTIPLIER}", 1.00f, 0.01f, 2.00f),
                RadiusJsf = new StorableFloat($"{id} {Constant.RADIUS_OFFSET}", 0, -1f, 1f),
                RightJsf = new StorableFloat($"{id} {Constant.X_OFFSET}", 0, -1f, 1f),
                UpJsf = new StorableFloat($"{id} {Constant.Y_OFFSET}", 0, -1f, 1f),
                LookJsf = new StorableFloat($"{id} {Constant.Z_OFFSET}", 0, -1f, 1f),
            };

            group.MassMultiplierJsf.setCallbackFunction = _ => SyncColliderMass(group, AllCollidersMass / HardColliderGroups.Count);
            group.RadiusJsf.setCallbackFunction = _ => SyncRadius(group);
            group.RightJsf.setCallbackFunction = _ => SyncPosition(group);
            group.UpJsf.setCallbackFunction = _ => SyncPosition(group);
            group.LookJsf.setCallbackFunction = _ => SyncPosition(group);

            return group;
        }

        void SyncRadius(HardColliderGroup group)
        {
            group.UpdateDimensions(_jointPhysicsHandler.NormalizedMass, Script.Softness);
            SyncSizeAuto();
        }

        void SyncPosition(HardColliderGroup group)
        {
            group.UpdatePosition(_jointPhysicsHandler.NormalizedMass, Script.Softness);
            SyncSizeAuto();
        }

        internal override void SyncAllOffsets()
        {
            float mass = _jointPhysicsHandler.NormalizedRealMass;
            float softness = Script.Softness;
            foreach(var group in HardColliderGroups)
            {
                group.UpdateDimensions(mass, softness);
                group.UpdatePosition(mass, softness);
            }

            SyncSizeAuto();
        }

        internal override void UpdateFrictionSizeMultipliers()
        {
            frictionSizeMultiplierLeft = 1f;
            frictionSizeMultiplierRight = 1f;
        }

        internal override void UpdateClothingAdjustmentAmounts(float massMultiplierLeft, float massMultiplierRight)
        {
            foreach(var group in HardColliderGroups)
            {
                group.SetClothingMassMultipliers(massMultiplierLeft, massMultiplierRight);
            }
        }

        float AllCollidersMass => BaseMassMultiplierJsf.val * _jointPhysicsHandler.CollidersMassAmount;

        internal override void SyncCollidersMass()
        {
            float massPerCollider = AllCollidersMass / HardColliderGroups.Count;
            foreach(var group in HardColliderGroups)
            {
                SyncColliderMass(group, massPerCollider);
            }
        }

        static void SyncColliderMass(HardColliderGroup group, float mass) => group.UpdateRigidbodyMass(group.MassMultiplierJsf.val * mass);

        internal override void PreRegisterParams()
        {
            Script.PreRegisterHardCollidersParam(BaseMassMultiplierJsf);
            foreach(var group in HardColliderGroups)
            {
                Script.PreRegisterHardCollidersParam(group.MassMultiplierJsf);
                Script.PreRegisterHardCollidersParam(group.RadiusJsf);
                Script.PreRegisterHardCollidersParam(group.RightJsf);
                Script.PreRegisterHardCollidersParam(group.UpJsf);
                Script.PreRegisterHardCollidersParam(group.LookJsf);
            }
        }

        internal override void CalibrateColliders()
        {
            var gluteCenterLeft = CollidersCenter(_jointPhysicsHandler.leftVolumeCalculationVerts);
            var gluteCenterRight = CollidersCenter(_jointPhysicsHandler.rightVolumeCalculationVerts);
            foreach(var group in HardColliderGroups)
            {
                group.Calibrate(gluteCenterLeft, gluteCenterRight, Person.HipRb);
            }
        }

        internal override void UpdateDistanceDiffs()
        {
            var gluteCenterLeft = CollidersCenter(_jointPhysicsHandler.leftVolumeCalculationVerts);
            var gluteCenterRight = CollidersCenter(_jointPhysicsHandler.rightVolumeCalculationVerts);
            foreach(var group in HardColliderGroups)
            {
                group.UpdateDistanceDiffs(gluteCenterLeft, gluteCenterRight, Person.HipRb);
            }
        }
    }
}
