using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;
using System.Collections.Generic;
using UnityEngine;
using static NAT.Direction;

namespace NAT.BM.Handlers
{
    sealed class GluteForcePhysicsHandler : ForceHandler
    {
        GluteJointPhysicsHandler _jointPhysicsHandler;
        GluteSoftPhysicsHandler _softPhysicsHandler;

        List<JointPhysicsParam> _leftParams, _rightParams;

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _jointPhysicsHandler = provider.GetInstance<GluteJointPhysicsHandler>();
            _softPhysicsHandler = provider.GetInstance<GluteSoftPhysicsHandler>();
            BaseJsf = NewMultiplierJsf("Force Physics Base Multiplier");
            UpJsf = NewMultiplierJsf("Force Physics Up Multiplier");
            DownJsf = NewMultiplierJsf("Force Physics Down Multiplier");
            DepthOutJsf = NewMultiplierJsf("Force Physics Depth Out Multiplier");
            DepthInJsf = NewMultiplierJsf("Force Physics Depth In Multiplier");
            SidewaysInJsf = NewMultiplierJsf("Force Physics Sideways In Multiplier");
            SidewaysOutJsf = NewMultiplierJsf("Force Physics Sideways Out Multiplier");
            LoadSettings();
        }

        protected override void OnMultiplierValueChanged(JSONStorableFloat jsf)
        {
            var storable = (StorableFloat) jsf;
            if(_softPhysicsHandler.IsLdpUpdateOn())
            {
                Script.SetCalibrationPending(storable);
            }
        }

        protected override void LoadSettings()
        {
            var centerOfGravity = _jointPhysicsHandler.CenterOfGravity;
            var inOutDamper = _jointPhysicsHandler.InOutDamper;
            var upDownAngleTarget = _jointPhysicsHandler.UpDownAngleTarget;
            var rightLeftAngleTarget = _jointPhysicsHandler.RightLeftAngleTarget;

            centerOfGravity.SetForceCalculators(CenterOfGravityCalculators());
            inOutDamper.SetForceCalculators(InOutDamperCalculators());
            upDownAngleTarget.SetForceCalculators(UpDownAngleTargetCalculators());
            rightLeftAngleTarget.SetForceCalculators(RightLeftAngleTargetCalculators());

            _leftParams = new List<JointPhysicsParam>
            {
                centerOfGravity.Left,
                inOutDamper.Left,
                // upDownAngleTarget.Left,
                // rightLeftAngleTarget.Left,
            };

            _rightParams = new List<JointPhysicsParam>
            {
                centerOfGravity.Right,
                inOutDamper.Right,
                // upDownAngleTarget.Right,
                // rightLeftAngleTarget.Right,
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> CenterOfGravityCalculators()
        {
            var backCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = -0.0975f,
                    MultiplierB = -0.065f,
                    CurveA = x => 1 - x,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var forwardCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = 0.0975f,
                    MultiplierB = 0.065f,
                    CurveA = x => 1 - x,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var left = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, backCalculator },
                { FORWARD, forwardCalculator },
            };
            var right = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, backCalculator },
                { FORWARD, forwardCalculator },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> InOutDamperCalculators()
        {
            var depthInCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = 1.5f,
                    MultiplierB = -4f,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var depthOutCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = 1.5f,
                    MultiplierB = -4f,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var left = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, depthInCalculator },
                { FORWARD, depthOutCalculator },
            };
            var right = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, depthInCalculator },
                { FORWARD, depthOutCalculator },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> UpDownAngleTargetCalculators()
        {
            var upCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = 5.33f,
                    MultiplierB = 4.0f,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var downCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = -1.0f,
                    MultiplierB = -0.75f,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var left = new Dictionary<string, PhysicsCalculator>
            {
                { UP, upCalculator },
                { DOWN, downCalculator },
            };
            var right = new Dictionary<string, PhysicsCalculator>
            {
                { UP, upCalculator },
                { DOWN, downCalculator },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> RightLeftAngleTargetCalculators() =>
            new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        {
                            SIDEWAYS_IN, new PhysicsCalculator
                            {
                                PrimaryConfig = new MultiplierConfig
                                {
                                    MultiplierA = 5.33f,
                                    MultiplierB = 4.0f,
                                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                                },
                            }
                        },
                        {
                            SIDEWAYS_OUT, new PhysicsCalculator
                            {
                                PrimaryConfig = new MultiplierConfig
                                {
                                    MultiplierA = -5.33f,
                                    MultiplierB = -4.0f,
                                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                                },
                            }
                        },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        {
                            SIDEWAYS_IN, new PhysicsCalculator
                            {
                                PrimaryConfig = new MultiplierConfig
                                {
                                    MultiplierA = -5.33f,
                                    MultiplierB = -4.0f,
                                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                                },
                            }
                        },
                        {
                            SIDEWAYS_OUT, new PhysicsCalculator
                            {
                                PrimaryConfig = new MultiplierConfig
                                {
                                    MultiplierA = 5.33f,
                                    MultiplierB = 4.0f,
                                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                                },
                            }
                        },
                    }
                },
            };

        internal override void PreRegisterParams()
        {
            Script.PreRegisterForcePhysicsParam(BaseJsf);
            // Script.PreRegisterForcePhysicsParam(UpJsf);
            // Script.PreRegisterForcePhysicsParam(DownJsf);
            Script.PreRegisterForcePhysicsParam(DepthOutJsf);
            Script.PreRegisterForcePhysicsParam(DepthInJsf);
            // Script.PreRegisterForcePhysicsParam(SidewaysInJsf);
            // Script.PreRegisterForcePhysicsParam(SidewaysOutJsf);
        }

        /* TODO duplicates ForceMorphHandler */

        float _mass;
        float _softness;

        float _upDownLimit;
        float _depthInLimit;
        float _depthOutLimit;
        float _sidewaysLimit;

        float _upMultiplier;
        float _downMultiplier;
        float _depthInMultiplier;
        float _depthOutMultiplier;
        float _sidewaysMultiplier;

        static float SoftnessMultiplier(float x) => Calc.InverseSmoothStep(x, 0.80f, 0.03f);

        internal void SetMultipliers(float mass, float softness)
        {
            _mass = mass;
            _softness = SoftnessMultiplier(softness);
            SetUpDownMultipliers();
            SetDepthInMultiplier();
            SetDepthOutMultiplier();
            SetSidewaysMultiplier();
        }

        void SetUpDownMultipliers()
        {
            _upDownLimit = 50 * Mathf.Lerp(1, 1.20f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.50f, 1.25f, _softness);
            _upMultiplier = softnessMultiplier * 1.50f * Mathf.Lerp(1, 1.80f, _mass);
            _downMultiplier = softnessMultiplier * 1.80f * Mathf.Lerp(1, 1.80f, _mass);
        }

        void SetDepthInMultiplier()
        {
            _depthInLimit = 0.045f * Mathf.Lerp(1, 3.00f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.50f, 1.00f, _softness);
            _depthInMultiplier = softnessMultiplier * 4.40f * Mathf.Lerp(1, 0.50f, _mass);
        }

        void SetDepthOutMultiplier()
        {
            _depthOutLimit = 0.045f * Mathf.Lerp(1, 3.00f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.65f, 1.125f, _softness);
            _depthOutMultiplier = softnessMultiplier * 2.20f * Mathf.Lerp(1, 1.50f, _mass);
        }

        void SetSidewaysMultiplier()
        {
            _sidewaysLimit = 50 * Mathf.Lerp(1, 1.20f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.50f, 1.25f, _softness);
            _sidewaysMultiplier = softnessMultiplier * 1.10f * Mathf.Lerp(1, 1.80f, _mass);
        }

        /* TODO END */

        internal void Execute()
        {
            AdjustUpPhysics();
            AdjustDownPhysics();
            AdjustDepthOutPhysics();
            AdjustDepthInPhysics();
            AdjustSidewaysPhysics();
        }

        static float UpDownForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f);

        float UpEffect(float angle) =>
            _upMultiplier *
            UserUpMultiplier() *
            UpDownForceEffectCurve(Mathf.Abs(angle) / _upDownLimit);

        void AdjustUpPhysics()
        {
            if(Person.LeftGlute.AngleY <= 0)
            {
                UpdatePhysicsLeft(UP, UpEffect(Person.LeftGlute.AngleY));
            }
            else
            {
                ResetPhysicsLeft(UP);
            }

            if(Person.RightGlute.AngleY <= 0)
            {
                UpdatePhysicsRight(UP, UpEffect(Person.RightGlute.AngleY));
            }
            else
            {
                ResetPhysicsRight(UP);
            }
        }

        float DownEffect(float angle) => _downMultiplier *
            UserDownMultiplier() *
            UpDownForceEffectCurve(Mathf.Abs(angle) / _upDownLimit);

        void AdjustDownPhysics()
        {
            if(Person.LeftGlute.AngleY <= 0)
            {
                ResetPhysicsLeft(DOWN);
            }
            else
            {
                UpdatePhysicsLeft(DOWN, DownEffect(Person.LeftGlute.AngleY));
            }

            if(Person.RightGlute.AngleY <= 0)
            {
                ResetPhysicsRight(DOWN);
            }
            else
            {
                UpdatePhysicsRight(DOWN, DownEffect(Person.RightGlute.AngleY));
            }
        }

        static float DepthForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f);

        float DepthInEffect(float depthDiff) =>
            _depthInMultiplier *
            UserDepthInMultiplier() *
            DepthForceEffectCurve(Mathf.Abs(depthDiff) / _depthInLimit);

        void AdjustDepthInPhysics()
        {
            if(Person.LeftGlute.DepthDiff >= 0)
            {
                ResetPhysicsLeft(BACK);
            }
            else
            {
                UpdatePhysicsLeft(BACK, DepthInEffect(Person.LeftGlute.DepthDiff));
            }

            if(Person.RightGlute.DepthDiff >= 0)
            {
                ResetPhysicsRight(BACK);
            }
            else
            {
                UpdatePhysicsRight(BACK, DepthInEffect(Person.RightGlute.DepthDiff));
            }
        }

        float DepthOutEffect(float depthDiff) => _depthOutMultiplier *
            UserDepthOutMultiplier() *
            DepthForceEffectCurve(Mathf.Abs(depthDiff) / _depthOutLimit);

        void AdjustDepthOutPhysics()
        {
            if(Person.LeftGlute.DepthDiff >= 0)
            {
                UpdatePhysicsLeft(FORWARD, DepthOutEffect(Person.LeftGlute.DepthDiff));
            }
            else
            {
                ResetPhysicsLeft(FORWARD);
            }

            if(Person.RightGlute.DepthDiff >= 0)
            {
                UpdatePhysicsRight(FORWARD, DepthOutEffect(Person.RightGlute.DepthDiff));
            }
            else
            {
                ResetPhysicsRight(FORWARD);
            }
        }

        static float SidewaysForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f);

        float SidewaysInEffect(float angle) =>
            _sidewaysMultiplier *
            UserSidewaysInMultiplier() *
            SidewaysForceEffectCurve(Mathf.Abs(angle) / _sidewaysLimit);

        float SidewaysOutEffect(float angle) =>
            _sidewaysMultiplier *
            UserSidewaysOutMultiplier() *
            SidewaysForceEffectCurve(Mathf.Abs(angle) / _sidewaysLimit);

        void AdjustSidewaysPhysics()
        {
            if(Person.LeftGlute.AngleX <= 0)
            {
                // right force on left breast
                ResetPhysicsLeft(SIDEWAYS_OUT);
                UpdatePhysicsLeft(SIDEWAYS_IN, SidewaysInEffect(Person.LeftGlute.AngleX));
            }
            else
            {
                // left force on left breast
                ResetPhysicsLeft(SIDEWAYS_IN);
                UpdatePhysicsLeft(SIDEWAYS_OUT, SidewaysOutEffect(Person.LeftGlute.AngleX));
            }

            if(Person.RightGlute.AngleX <= 0)
            {
                // right force on right breast
                ResetPhysicsRight(SIDEWAYS_IN);
                UpdatePhysicsRight(SIDEWAYS_OUT, SidewaysOutEffect(Person.RightGlute.AngleX));
            }
            else
            {
                // left force on right breast
                ResetPhysicsRight(SIDEWAYS_OUT);
                UpdatePhysicsRight(SIDEWAYS_IN, SidewaysInEffect(Person.RightGlute.AngleX));
            }
        }

        void UpdatePhysicsLeft(string direction, float effect)
        {
            foreach(var param in _leftParams)
            {
                param.UpdateForceValue(direction, effect, _mass, _softness);
            }
        }

        void UpdatePhysicsRight(string direction, float effect)
        {
            foreach(var param in _rightParams)
            {
                param.UpdateForceValue(direction, effect, _mass, _softness);
            }
        }

        internal override void ResetAll()
        {
            foreach(string direction in new[]
            {
                LEFT,
                RIGHT,
                UP,
                DOWN,
                FORWARD,
                BACK,
            })
            {
                ResetPhysicsLeft(direction);
                ResetPhysicsRight(direction);
            }
        }

        void ResetPhysicsLeft(string direction)
        {
            foreach(var param in _leftParams)
            {
                param.ResetForceValue(direction);
            }
        }

        void ResetPhysicsRight(string direction)
        {
            foreach(var param in _rightParams)
            {
                param.ResetForceValue(direction);
            }
        }
    }
}
