using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using System.Collections.Generic;
using UnityEngine;
using static NAT.Direction;

namespace NAT.BM.Handlers
{
    sealed class GluteGravityPhysicsHandler : GravityPhysicsHandler
    {
        GluteJointPhysicsHandler _jointPhysicsHandler;

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _jointPhysicsHandler = provider.GetInstance<GluteJointPhysicsHandler>();
            JointMass = _jointPhysicsHandler.JointMass;
            LoadSettings();
            Person.SetGlutesUseGravity(false);
        }

        protected override void LoadSettings()
        {
            {
                var paramGroup = _jointPhysicsHandler.RotationSpring;
                var leftCalculators = SpringCalculators(Side.LEFT);
                var rightCalculators = SpringCalculators(Side.RIGHT);
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText = "Rotation spring decreases when leaning forward or back.";
            }

            {
                var paramGroup = _jointPhysicsHandler.InOutSpring;
                var leftCalculators = InOutSpringCalculators(Side.LEFT);
                var rightCalculators = InOutSpringCalculators(Side.RIGHT);
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText = "In/Out Spring decreases when leaning forward or back.";
            }

            {
                var paramGroup = _jointPhysicsHandler.UpDownAngleTarget;
                var leftCalculators = UpDownAngleTargetCalculators(Side.LEFT);
                var rightCalculators = UpDownAngleTargetCalculators(Side.RIGHT);
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText = "Up/Down Angle Target decreases when upright and increases when upside down.";
            }

            {
                var paramGroup = _jointPhysicsHandler.RightLeftAngleTarget;
                var leftCalculators = RightLeftAngleTargetCalculatorsLeft();
                var rightCalculators = RightLeftAngleTargetCalculatorsRight();
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText =
                    "Right/Left Angle Target is positive when leaning to the right and negative when leaning to the left (the other way around for the right glute).";
            }
        }

        static Dictionary<string, PhysicsCalculator> SpringCalculators(string side) => new Dictionary<string, PhysicsCalculator>
        {
            { BACK, BackSpringCalculator(side, BACK) },
            { FORWARD, ForwardSpringCalculator(side) },
        };

        static PhysicsCalculator BackSpringCalculator(string side, string direction) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_SPRING.NAME, side), direction),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0,
                MultiplierA = -0.75f,
                MultiplierB = -3.75f,
                CurveA = x => 1 - x,
            },
        };

        static PhysicsCalculator ForwardSpringCalculator(string side) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_SPRING.NAME, side), FORWARD),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0,
                MultiplierA = 7.5f,
                MultiplierB = 2.5f,
                CurveA = x => 1 - x,
            },
        };

        static Dictionary<string, PhysicsCalculator> InOutSpringCalculators(string side) => new Dictionary<string, PhysicsCalculator>
        {
            { BACK, InOutSpringCalculator(side, BACK) },
            { FORWARD, InOutSpringCalculator(side, FORWARD) },
        };

        static PhysicsCalculator InOutSpringCalculator(string side, string direction) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(IN_OUT_SPRING.NAME, side), direction),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = -25f,
                MultiplierA = -100f,
                MultiplierB = -125f,
                CurveA = x => Calc.Exponential1(1 - x, 0.3f, 10f, 0.84f, 1.01f),
                CurveB = Calc.InOutSpringSoftnessCurve,
            },
        };

        static PhysicsCalculator UpAngleTargetCalculator(string side) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(BREAST_UP_DOWN_ANGLE_TARGET.NAME, side), UP),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 2.50f,
            },
        };

        static PhysicsCalculator DownAngleTargetCalculator(string side) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(BREAST_UP_DOWN_ANGLE_TARGET.NAME, side), DOWN),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = -2.50f,
            },
        };

        static Dictionary<string, PhysicsCalculator> UpDownAngleTargetCalculators(string side) => new Dictionary<string, PhysicsCalculator>
        {
            { UP, UpAngleTargetCalculator(side) },
            { DOWN, DownAngleTargetCalculator(side) },
        };

        static Dictionary<string, PhysicsCalculator> RightLeftAngleTargetCalculatorsLeft() => new Dictionary<string, PhysicsCalculator>
        {
            {
                SIDEWAYS_IN, new PhysicsCalculator
                {
                    MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_IN),
                    PrimaryConfig = new MultiplierConfig
                    {
                        BaseValue = 1.00f,
                        MultiplierB = 2.00f,
                    },
                }
            },
            {
                SIDEWAYS_OUT, new PhysicsCalculator
                {
                    MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_OUT),
                    PrimaryConfig = new MultiplierConfig
                    {
                        BaseValue = -1.50f,
                        MultiplierB = -3.00f,
                    },
                }
            },
        };

        static Dictionary<string, PhysicsCalculator> RightLeftAngleTargetCalculatorsRight() => new Dictionary<string, PhysicsCalculator>
        {
            {
                SIDEWAYS_IN, new PhysicsCalculator
                {
                    MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_IN),
                    PrimaryConfig = new MultiplierConfig
                    {
                        BaseValue = -1.00f,
                        MultiplierB = -2.00f,
                    },
                }
            },
            {
                SIDEWAYS_OUT, new PhysicsCalculator
                {
                    MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_OUT),
                    PrimaryConfig = new MultiplierConfig
                    {
                        BaseValue = 1.50f,
                        MultiplierB = 3.00f,
                    },
                }
            },
        };

        internal void SetMultipliers(float mass, float softness) // TODO
        {
            upGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            downGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            forwardGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            backGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            sidewaysInGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            sidewaysOutGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            SetGravityValues();
        }

        bool _softPhysicsOn;
        float _mass;
        float _softness;
        float _leftRoll;
        float _rightRoll;
        float _leftPitch;
        float _rightPitch;

        internal void Execute(bool softPhysicsOn)
        {
            _softPhysicsOn = softPhysicsOn;
            _mass = _jointPhysicsHandler.NormalizedMass;
            _softness = Script.Softness;
            _leftRoll = AngleCalculator.SmoothLeftRoll;
            _rightRoll = AngleCalculator.SmoothRightRoll;
            _leftPitch = AngleCalculator.Smooth2XLeftPitch;
            _rightPitch = AngleCalculator.Smooth2XRightPitch;

            if(!IsSimulatingGravity)
            {
                ApplyRealGravity();
            }

            AdjustSidewaysPhysics();
            AdjustVerticalPhysics();
            AdjustHorizontalPhysics();
        }

        void AdjustSidewaysPhysics()
        {
            float leftEffect = Mathf.Abs(_leftRoll);
            if(_leftRoll > 0)
            {
                ResetLeftGlute(SIDEWAYS_IN);
                UpdateLeftGlute(SIDEWAYS_OUT, leftEffect);
            }
            else
            {
                ResetLeftGlute(SIDEWAYS_OUT);
                UpdateLeftGlute(SIDEWAYS_IN, leftEffect);
            }

            float rightEffect = Mathf.Abs(_rightRoll);
            if(_rightRoll > 0)
            {
                ResetRightGlute(SIDEWAYS_OUT);
                UpdateRightGlute(SIDEWAYS_IN, rightEffect);
            }
            else
            {
                ResetRightGlute(SIDEWAYS_IN);
                UpdateRightGlute(SIDEWAYS_OUT, rightEffect);
            }
        }

        void AdjustVerticalPhysics()
        {
            float leftEffect = AngleCalculator.UpDownAdjustByAngle(_leftPitch) * AngleCalculator.RollMultiplier(_leftRoll);
            if(_leftPitch > 1 || _leftPitch < -1)
            {
                // upside down, leaning forward or back
                ResetLeftGlute(DOWN);
                UpdateLeftGlute(UP, leftEffect);
            }
            else
            {
                ResetLeftGlute(UP);
                UpdateLeftGlute(DOWN, leftEffect);
            }

            float rightEffect = AngleCalculator.UpDownAdjustByAngle(_rightPitch) * AngleCalculator.RollMultiplier(_rightRoll);
            if(_rightPitch > 1 || _rightPitch < -1)
            {
                // upside down, leaning forward or back
                ResetRightGlute(DOWN);
                UpdateRightGlute(UP, rightEffect);
            }
            else
            {
                ResetRightGlute(UP);
                UpdateRightGlute(DOWN, rightEffect);
            }
        }

        void AdjustHorizontalPhysics()
        {
            float leftEffect = AngleCalculator.DepthAdjustByAngle(_leftPitch) * AngleCalculator.RollMultiplier(_leftRoll);

            if(_leftPitch > 0)
            {
                // leaning forward, upright or upside down
                ResetLeftGlute(BACK);
                UpdateLeftGlute(FORWARD, leftEffect);
            }
            else
            {
                // leaning back
                ResetLeftGlute(FORWARD);
                UpdateLeftGlute(BACK, leftEffect);
            }

            float rightEffect = AngleCalculator.DepthAdjustByAngle(_rightPitch) * AngleCalculator.RollMultiplier(_rightRoll);
            if(_rightPitch > 0)
            {
                // leaning forward, upright or upside down
                ResetRightGlute(BACK);
                UpdateRightGlute(FORWARD, rightEffect);
            }
            else
            {
                // leaning back
                UpdateRightGlute(BACK, rightEffect);
                ResetRightGlute(FORWARD);
            }
        }

        void UpdateLeftGlute(string direction, float effect)
        {
            _jointPhysicsHandler.RotationSpring.Left.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.InOutSpring.Left.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.UpDownAngleTarget.Left.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.RightLeftAngleTarget.Left.UpdateGravityValue(direction, effect, _mass, _softness);
        }

        void UpdateRightGlute(string direction, float effect)
        {
            _jointPhysicsHandler.RotationSpring.Right.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.InOutSpring.Right.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.UpDownAngleTarget.Right.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.RightLeftAngleTarget.Right.UpdateGravityValue(direction, effect, _mass, _softness);
        }

        internal void ResetAll()
        {
            foreach(string direction in new[]
            {
                SIDEWAYS_IN,
                SIDEWAYS_OUT,
                UP,
                DOWN,
                FORWARD,
                BACK,
            })
            {
                ResetLeftGlute(direction);
                ResetRightGlute(direction);
            }
        }

        void ResetLeftGlute(string direction)
        {
            _jointPhysicsHandler.RotationSpring.Left.ResetGravityValue(direction);
            _jointPhysicsHandler.InOutSpring.Left.ResetGravityValue(direction);
            _jointPhysicsHandler.UpDownAngleTarget.Left.ResetGravityValue(direction);
            _jointPhysicsHandler.RightLeftAngleTarget.Left.ResetGravityValue(direction);
        }

        void ResetRightGlute(string direction)
        {
            _jointPhysicsHandler.RotationSpring.Right.ResetGravityValue(direction);
            _jointPhysicsHandler.InOutSpring.Right.ResetGravityValue(direction);
            _jointPhysicsHandler.UpDownAngleTarget.Right.ResetGravityValue(direction);
            _jointPhysicsHandler.RightLeftAngleTarget.Right.ResetGravityValue(direction);
        }
    }
}
