using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using System.Collections.Generic;
using System.Text;
using UnityEngine;
using static NAT.Side;

namespace NAT.BM.Handlers
{
    sealed class GluteJointPhysicsHandler : JointPhysicsHandler
    {
        GluteGravityPhysicsHandler _gravityPhysicsHandler;
        internal float CollidersMassAmount { get; private set; }

        /* Only the verts that move on a person with soft physics off */
        internal readonly int[] leftVolumeCalculationVerts =
        {
            7877, 283, 7821, 7833, 2391, 288, 2392, 7878, 7875, 7876, 7864, 7825, 7826,
            7845, 274, 2390, 8249, 296, 281, 7849, 7824, 8212, 265, 2388, 4036, 287, 4039,
        };
        internal readonly int[] rightVolumeCalculationVerts =
        {
            18557, 11211, 18501, 18513, 13221, 11216, 13222, 18558, 18555, 18556, 18544, 18505, 18506, 18525,
            11202, 13220, 18920, 11224, 11209, 18529, 18504, 18883, 11193, 13218, 14812, 11215, 14815,
        };

        // JointPhysicsParamGroup RightLeftSpring { get; set; }
        // JointPhysicsParamGroup RightLeftDamper { get; set; }

        // JSONStorableBool XMotionEnabledJsb { get; set; }
        // JSONStorableBool YMotionEnabledJsb { get; set; }
        // JSONStorableBool ZMotionEnabledJsb { get; set; }

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _gravityPhysicsHandler = provider.GetInstance<GluteGravityPhysicsHandler>();
            // XMotionEnabledJsb = new JSONStorableBool("Right/Left Motion Enabled", true, SetXMotionEnabled);
            // YMotionEnabledJsb = new JSONStorableBool("Up/Down Motion Enabled", false, SetYMotionEnabled);
            // ZMotionEnabledJsb = new JSONStorableBool("In/Out Motion Enabled", true, SetZMotionEnabled);
            // SetXMotionEnabled(XMotionEnabledJsb.val);
            // SetYMotionEnabled(YMotionEnabledJsb.val);
            // SetZMotionEnabled(ZMotionEnabledJsb.val);
            LoadSettings();
        }

        // void SetXMotionEnabled(bool value)
        // {
        //     var motion = value ? ConfigurableJointMotion.Limited : ConfigurableJointMotion.Locked;
        //     Person.LeftGlute.Joint.xMotion = motion;
        //     Person.RightGlute.Joint.xMotion = motion;
        // }

        // void SetYMotionEnabled(bool value)
        // {
        //     var motion = value ? ConfigurableJointMotion.Limited : ConfigurableJointMotion.Locked;
        //     Person.LeftGlute.Joint.yMotion = motion;
        //     Person.RightGlute.Joint.yMotion = motion;
        // }

        // void SetZMotionEnabled(bool value)
        // {
        //     var motion = value ? ConfigurableJointMotion.Limited : ConfigurableJointMotion.Locked;
        //     Person.LeftGlute.Joint.zMotion = motion;
        //     Person.RightGlute.Joint.zMotion = motion;
        // }

        internal void CalculateMass()
        {
            float leftVolume = CalculateVolume(leftVolumeCalculationVerts);
            float rightVolume = CalculateVolume(rightVolumeCalculationVerts);
            float volume = (leftVolume + rightVolume) / 2;
            /* Interpolates volume range [0.77, 2.70] to mass range [1.00, 4.00] */
            UpdateMassValueAndAmounts(1.55f * volume - 0.19f);
        }

        internal void UpdateMassValueAndAmounts(float mass)
        {
            CalibratedMass = mass;
            const float min = GLUTE_JOINT_MASS.MIN;
            const float max = GLUTE_JOINT_MASS.MAX;
            float jointToCollidersRatio = Mathf.Lerp(0.92f, 0.78f, Mathf.InverseLerp(min, max, mass));
            JointMass.UpdateMassValue(jointToCollidersRatio * mass);
            JointMass.UpdateOffsetMinMax();
            CollidersMassAmount = (1 - jointToCollidersRatio) * mass;
            NormalizedMass = Mathf.InverseLerp(min, max, JointMass.Left.Value);
            NormalizedRealMass = Mathf.InverseLerp(min, max, JointMass.Left.BaseValueJsf.val);
        }

        float CalculateVolume(int[] vertexIndices)
        {
            var rawSkinnedVerts = Person.Skin.rawSkinnedVerts;
            var positions = new Vector3[vertexIndices.Length];
            for(int i = 0; i < vertexIndices.Length; i++)
            {
                positions[i] = Calc.RelativePosition(Person.HipRb, rawSkinnedVerts[vertexIndices[i]]);
            }

            return 1000 * 0.5f * VolumeCalculator.SphereVolume(positions, Person.ScaleJsf.val);
        }

        protected override void LoadSettings()
        {
            JointMass = new JointPhysicsParamGroup(Script)
            {
                Name = GLUTE_JOINT_MASS.NAME,
                ValueFormat = "F3",
                Left = GluteMassParam(LEFT, Person.LeftGlute),
                Right = GluteMassParam(RIGHT, Person.RightGlute),
                RequiresCalibration = true,
                InfoText = new StringBuilder()
                    .AppendBold(GLUTE_JOINT_MASS.NAME)
                    .AppendLine(" is the mass of the joint.")
                    .AppendLine()
                    .Append("Since mass represents glute size, other physics parameters are adjusted based on its value.")
                    .ToString(),
            };

            CenterOfGravity = new JointPhysicsParamGroup(Script)
            {
                Name = CENTER_OF_GRAVITY.NAME,
                ValueFormat = "F2",
                Left = CenterOfGravityParam(LEFT, Person.LeftGlute),
                Right = CenterOfGravityParam(RIGHT, Person.RightGlute),
                RequiresCalibration = true,
                InfoText = new StringBuilder()
                    .AppendBold(CENTER_OF_GRAVITY.NAME)
                    .AppendLine(" is the position of the joint's center of mass.")
                    .AppendLine()
                    .Append("At 0, the center of mass is inside the body at the joint. At 1, it is at the skin surface.")
                    .ToString(),
            };

            RotationSpring = new JointPhysicsParamGroup(Script)
            {
                Name = ROTATION_SPRING.NAME,
                ValueFormat = "F0",
                Left = RotationSpringParam(LEFT, Person.LeftGlute),
                Right = RotationSpringParam(RIGHT, Person.RightGlute),
                RequiresCalibration = true,
                InfoText = new StringBuilder()
                    .AppendBold(ROTATION_SPRING.NAME)
                    .AppendLine(" pushes the joint towards its angle target.")
                    .AppendLine()
                    .Append("The angle target is defined by the Up/Down and Left/Right Angle Target parameters.")
                    .ToString(),
            };

            RotationDamper = new JointPhysicsParamGroup(Script)
            {
                Name = ROTATION_DAMPER.NAME,
                ValueFormat = "F2",
                Left = RotationDamperParam(LEFT, Person.LeftGlute),
                Right = RotationDamperParam(RIGHT, Person.RightGlute),
                DependsOnPhysicsRate = true,
                InfoText = new StringBuilder()
                    .AppendBold(ROTATION_DAMPER.NAME)
                    .AppendLine(" reduces oscillation around the joint's angle target.")
                    .AppendLine()
                    .Append("The higher the damper, the quicker glutes will stop swinging.")
                    .ToString(),
            };

            // RightLeftSpring = new JointPhysicsParamGroup(Script)
            // {
            //     Name = RIGHT_LEFT_SPRING.NAME,
            //     ValueFormat = "0",
            //     Left = RightLeftSpringParam(LEFT, Person.LeftGlute),
            //     Right = RightLeftSpringParam(RIGHT, Person.RightGlute),
            //     RequiresCalibration = true,
            //     InfoText = new StringBuilder()
            //         .AppendBold(RIGHT_LEFT_SPRING.NAME)
            //         .Append(" pushes the joint towards its position target along the X axis.")
            //         .ToString(),
            // };

            // RightLeftDamper = new JointPhysicsParamGroup(Script)
            // {
            //     Name = RIGHT_LEFT_DAMPER.NAME,
            //     valueFormat = "F2",
            //     Left = RightLeftDamperParam(LEFT, Person.LeftGlute),
            //     Right = RightLeftDamperParam(RIGHT, Person.RightGlute),
            //     DependsOnPhysicsRate = true,
            //     InfoText = new StringBuilder()
            //         .AppendBold(RIGHT_LEFT_DAMPER.NAME)
            //         .Append(" reduces oscillation around the joint's position target along the X axis.")
            //         .ToString(),
            // };

            InOutSpring = new JointPhysicsParamGroup(Script)
            {
                Name = IN_OUT_SPRING.NAME,
                ValueFormat = "0",
                Left = InOutSpringParam(LEFT, Person.LeftGlute),
                Right = InOutSpringParam(RIGHT, Person.RightGlute),
                RequiresCalibration = true,
                InfoText = new StringBuilder()
                    .AppendBold(IN_OUT_SPRING.NAME)
                    .AppendLine(" pushes the joint towards its position target along the Z axis.")
                    .AppendLine()
                    .Append("Directional force morphing along the forward-back axis depends on In/Out Spring being suitably low for the given glute mass.")
                    .ToString(),
            };

            InOutDamper = new JointPhysicsParamGroup(Script)
            {
                Name = IN_OUT_DAMPER.NAME,
                ValueFormat = "0",
                Left = InOutDamperParam(LEFT, Person.LeftGlute),
                Right = InOutDamperParam(RIGHT, Person.RightGlute),
                DependsOnPhysicsRate = true,
                InfoText = new StringBuilder()
                    .AppendBold(IN_OUT_DAMPER.NAME)
                    .Append(" reduces oscillation around the joint position target along the Z axis.")
                    .ToString(),
            };

            var leftUpDownAngleTarget = UpDownAngleTargetParam(LEFT);
            var leftRightLeftAngleTarget = RightLeftAngleTargetParam(LEFT);

            var rightUpDownAngleTarget = UpDownAngleTargetParam(RIGHT);
            var rightRightLeftAngleTarget = RightLeftAngleTargetParam(RIGHT);

            leftUpDownAngleTarget.Sync = value => SyncAngleTarget(Person.LeftGlute, value, leftRightLeftAngleTarget.Value);
            rightUpDownAngleTarget.Sync = value => SyncAngleTarget(Person.RightGlute, value, rightRightLeftAngleTarget.Value);

            UpDownAngleTarget = new JointPhysicsParamGroup(Script)
            {
                Name = GLUTE_UP_DOWN_ANGLE_TARGET.NAME,
                ValueFormat = "F2",
                Left = leftUpDownAngleTarget,
                Right = rightUpDownAngleTarget,
                RequiresCalibration = true,
                InfoText = new StringBuilder()
                    .AppendBold(GLUTE_UP_DOWN_ANGLE_TARGET.NAME)
                    .AppendLine(" is the joint's vertical target angle. Negative values pull glutes down, positive values push them up.")
                    .AppendLine()
                    .Append("Generally it's better to leave the offset at 0, and adjust the Glute Weight gravity multipliers instead.")
                    .ToString(),
            };

            RightLeftAngleTarget = new JointPhysicsParamGroup(Script)
            {
                Name = RIGHT_LEFT_ANGLE_TARGET.NAME,
                ValueFormat = "F2",
                Left = leftRightLeftAngleTarget,
                Right = rightRightLeftAngleTarget,
                RequiresCalibration = true,
                RightInverted = true,
                InfoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_ANGLE_TARGET.NAME)
                    .Append(" is the joint's horizontal target angle. A negative offset pulls glutes apart, while a positive offset pushes them together.")
                    .ToString(),
            };

            ParamGroups = new List<JointPhysicsParamGroup>
            {
                CenterOfGravity,
                RotationSpring,
                RotationDamper,
                // RightLeftSpring,
                // RightLeftDamper,
                InOutSpring,
                InOutDamper,
                UpDownAngleTarget,
                RightLeftAngleTarget,
            };

            JointMass.SetCallbackFunctions();
            foreach(var group in ParamGroups)
            {
                group.SetCallbackFunctions();
            }
        }

        JointPhysicsParam GluteMassParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            GLUTE_JOINT_MASS.NAME,
            GLUTE_JOINT_MASS.MIN,
            GLUTE_JOINT_MASS.MAX
        )
        {
            Sync = value => SyncGluteMass(target.Rigidbody, value),
        };

        JointPhysicsParam CenterOfGravityParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            CENTER_OF_GRAVITY.NAME,
            CENTER_OF_GRAVITY.MIN,
            CENTER_OF_GRAVITY.MAX
        )
        {
            Sync = value => SyncCenterOfGravity(target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0.33f,
                MultiplierA = 0.75f,
                MultiplierB = 1.33f,
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = -0.06f,
                MultiplierA = 0.60f,
                MultiplierB = 0.00f,
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = 0.06f,
                MultiplierA = 0.60f,
                MultiplierB = 0.00f,
            },
        };

        JointPhysicsParam RotationSpringParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            ROTATION_SPRING.NAME,
            ROTATION_SPRING.MIN,
            ROTATION_SPRING.MAX
        )
        {
            Sync = value => SyncRotationSpring(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 66f,
                MultiplierA = 0.80f,
                MultiplierB = -0.46f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
                CurveB = x => Calc.InverseSmoothStep(x, 0.10f, 0.13f),
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = 6f,
                MultiplierA = 0.67f,
                MultiplierB = 0.00f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = -9f,
                MultiplierA = 0.67f,
                MultiplierB = 0.00f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
            },
        };

        JointPhysicsParam RotationDamperParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            ROTATION_DAMPER.NAME,
            ROTATION_DAMPER.MIN,
            ROTATION_DAMPER.MAX
        )
        {
            Sync = value => SyncDamper(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0.58f,
                MultiplierA = 1.00f,
                MultiplierB = -0.55f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
                CurveB = x => Calc.InverseSmoothStep(x, 0.24f, 0.61f),
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = -0.03f,
                MultiplierA = 8.375f,
                MultiplierB = 0.00f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = 0.015f,
                MultiplierA = 8.375f,
                MultiplierB = 0.00f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
            },
        };

        // JointPhysicsParam RightLeftSpringParam(string side, CalibratedJoint target) => new JointPhysicsParam(
        //     Script,
        //     side,
        //     RIGHT_LEFT_SPRING.NAME,
        //     RIGHT_LEFT_SPRING.MIN,
        //     RIGHT_LEFT_SPRING.MAX
        // )
        // {
        //     Sync = value => SyncRightLeftSpring(target.Joint, target.Rigidbody, value),
        //     PrimaryConfig = new MultiplierConfig
        //     {
        //         BaseValue = 380f,
        //     },
        // };

        // JointPhysicsParam RightLeftDamperParam(string side, CalibratedJoint target) => new JointPhysicsParam(
        //     Script,
        //     side,
        //     RIGHT_LEFT_DAMPER.NAME,
        //     RIGHT_LEFT_DAMPER.MIN,
        //     RIGHT_LEFT_DAMPER.MAX
        // )
        // {
        //     Sync = value => SyncRightLeftDamper(target.Joint, target.Rigidbody, value),
        //     PrimaryConfig = new MultiplierConfig
        //     {
        //         BaseValue = 10f,
        //     },
        // };

        JointPhysicsParam InOutSpringParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            IN_OUT_SPRING.NAME,
            IN_OUT_SPRING.MIN,
            IN_OUT_SPRING.MAX
        )
        {
            Sync = value => SyncInOutSpring(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 900f,
                MultiplierA = -0.15f,
                MultiplierB = -0.30f,
                CurveA = x => 1 - x, // TODO smoothstep
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = 35f,
                MultiplierA = -0.28f,
                MultiplierB = 0.00f,
                CurveA = x => 1 - x,
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = -52f,
                MultiplierA = -0.28f,
                MultiplierB = 0.00f,
                CurveA = x => 1 - x,
            },
        };

        JointPhysicsParam InOutDamperParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            IN_OUT_DAMPER.NAME,
            IN_OUT_DAMPER.MIN,
            IN_OUT_DAMPER.MAX
        )
        {
            Sync = value => SyncInOutDamper(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 9.50f,
                MultiplierA = 0.67f,
                MultiplierB = -0.40f,
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = -2.00f,
                MultiplierA = 1.33f,
                MultiplierB = 0.00f,
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = 3.00f,
                MultiplierA = 1.33f,
                MultiplierB = 0.00f,
            },
        };

        static JointPhysicsParam UpDownAngleTargetParam(string side) => new JointPhysicsParam(
            side,
            GLUTE_UP_DOWN_ANGLE_TARGET.NAME,
            GLUTE_UP_DOWN_ANGLE_TARGET.MIN,
            GLUTE_UP_DOWN_ANGLE_TARGET.MAX
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 10.00f,
            },
        };

        static JointPhysicsParam RightLeftAngleTargetParam(string side) => new JointPhysicsParam(
            side,
            RIGHT_LEFT_ANGLE_TARGET.NAME,
            RIGHT_LEFT_ANGLE_TARGET.MIN,
            RIGHT_LEFT_ANGLE_TARGET.MAX
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        // Reimplements AdjustJoints.cs method SyncMass
        void SyncGluteMass(Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            if(Mathf.Abs(rb.mass - value) <= 0.001f)
            {
                return;
            }

            rb.mass = value;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncCenterOfGravity
        void SyncCenterOfGravity(Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var newCenterOfMass = Vector3.Lerp(
                Person.GluteControl.lowCenterOfGravity,
                Person.GluteControl.highCenterOfGravity,
                value
            );
            if(Calc.IsVectorEqualWithin(1 / 100f, rb.centerOfMass, newCenterOfMass))
            {
                return;
            }

            rb.centerOfMass = newCenterOfMass;
            rb.WakeUp();
        }

        void SyncRotationSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            float scalePow = Mathf.Pow(1.7f, Person.GluteControl.scale - 1f);

            float scaledSpring = value * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionSpring - scaledSpring) <= 1)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionSpring = scaledSpring;
            slerpDrive.maximumForce = DEFAULT_SLERP_MAX_FORCE * scalePow;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionSpring = 2 * scaledSpring;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionSpring = 2 * scaledSpring;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.spring = 2 * scaledSpring * Person.GluteControl.limitSpringMultiplier; // TODO check
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.spring = 2 * scaledSpring * Person.GluteControl.limitSpringMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        void SyncDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            float scalePow = Mathf.Pow(1.7f, Person.GluteControl.scale - 1f);

            float scaledDamper = value * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionDamper - scaledDamper) <= 0.01f)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionDamper = scaledDamper;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionDamper = 2 * scaledDamper;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionDamper = 2 * scaledDamper;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.damper = 2 * scaledDamper * Person.GluteControl.limitDamperMultiplier; // TODO check
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.damper = 2 * scaledDamper * Person.GluteControl.limitDamperMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        // void SyncRightLeftSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        // {
        //     if(!Script.enabled)
        //     {
        //         return;
        //     }
        //
        //     var drive = joint.xDrive;
        //     if(Mathf.Abs(drive.positionSpring - value) <= 1)
        //     {
        //         return;
        //     }
        //
        //     drive.positionSpring = value;
        //     joint.xDrive = drive;
        //     rb.WakeUp();
        // }

        // void SyncRightLeftDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        // {
        //     if(!Script.enabled)
        //     {
        //         return;
        //     }
        //
        //     var drive = joint.xDrive;
        //     if(Mathf.Abs(drive.positionDamper - value) <= 0.01f)
        //     {
        //         return;
        //     }
        //
        //     drive.positionDamper = value;
        //     joint.xDrive = drive;
        //     rb.WakeUp();
        // }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        void SyncInOutSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.zDrive;
            if(Mathf.Abs(drive.positionSpring - value) <= 1)
            {
                return;
            }

            drive.positionSpring = value;
            joint.zDrive = drive;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        void SyncInOutDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.zDrive;
            if(Mathf.Abs(drive.positionDamper - value) <= 0.01f)
            {
                return;
            }

            drive.positionDamper = value;
            joint.zDrive = drive;
            rb.WakeUp();
        }

        void SyncAngleTarget(CalibratedJoint target, float x, float y)
        {
            if(!Script.enabled)
            {
                return;
            }

            target.SetRotation(x, y, 0);
        }

        internal override void PreRegisterParams()
        {
            Script.PreRegisterJointPhysicsParam(JointMass.Left.OffsetJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.LinkDirectionsJsb);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.UpGravityJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.DownGravityJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.BackGravityJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.ForwardGravityJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.SidewaysGravityJsf);

            foreach(var paramGroup in ParamGroups)
            {
                var leftParam = paramGroup.Left;
                Script.PreRegisterJointPhysicsParam(leftParam.OffsetJsf);
                foreach(var pair in leftParam.GravityCalculators)
                {
                    Script.PreRegisterJointPhysicsParam(pair.Value.MultiplierJsf);
                }
            }
        }

        internal void SaveCalibrated()
        {
            JointMass.SaveCalibrated();
            foreach(var group in ParamGroups)
            {
                group.SaveCalibrated();
            }
        }

        internal void UpdateBaseValues(bool updateOffsetMinMax = true)
        {
            foreach(var group in ParamGroups)
            {
                group.UpdateBaseValue(
                    NormalizedMass,
                    Script.Softness,
                    Script.Quickness
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }
        }

        internal void SetGravitySimulationDamper()
        {
            float interpolant = Calc.InverseSmoothStep(NormalizedMass, 0.50f, 0.10f);
            RotationDamper.OverrideValue(Mathf.Lerp(2.5f, 5.0f, interpolant));
        }

        internal void UnsetGravitySimulationDamper() =>
            RotationDamper.UpdateBaseValue(
                NormalizedMass,
                Script.Softness,
                Script.Quickness
            );

        internal void RestoreOriginalPhysics()
        {
            if(Script.Initialized == true)
            {
                return;
            }

            foreach(string name in new[]
            {
                "mass",
                "centerOfGravityPercent",
                "spring",
                "damper",
                "positionSpringZ",
                "positionDamperZ",
                "targetRotationX",
                "targetRotationY",
            })
            {
                /* Set a value that is different from the original, then restore the original
                 * in order to trigger VaM's internal sync
                 */
                var paramJsf = Person.GluteControl.GetFloatJSONParam(name);
                float original = paramJsf.val;
                paramJsf.valNoCallback = Mathf.Abs(paramJsf.val - paramJsf.min) > 0.01f
                    ? paramJsf.min
                    : paramJsf.max;
                paramJsf.val = original;
            }
        }
    }
}
