using NAT.Common;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using UnityEngine;

namespace NAT.BM.Models
{
    sealed class SoftJointGradientGroup
    {
        // TODO when center is shared between multiple gradient groups
        internal SoftJoint Center { get; set; }
        internal SoftJoint Main1 { get; set; }
        internal SoftJoint Main2 { get; set; }
        internal SoftJoint Main3 { get; set; }
        internal SoftJoint Main4 { get; set; }
        internal SoftJoint Outer { get; set; }

        // ratios of distances to total distance
        float _main1Ratio;
        float _main2Ratio;
        float _main3Ratio;
        float _main4Ratio;

        internal void CalculateDistances()
        {
            float total;
            float mainToOuter;
            float centerToMain1 = Vector3.Distance(Center.JointRb.position, Main1.JointRb.position);
            if(Main2 != null)
            {
                float main1ToMain2 = Vector3.Distance(Main1.JointRb.position, Main2.JointRb.position);
                if(Main3 != null)
                {
                    float main2ToMain3 = Vector3.Distance(Main2.JointRb.position, Main3.JointRb.position);
                    if(Main4 != null)
                    {
                        float main3ToMain4 = Vector3.Distance(Main3.JointRb.position, Main4.JointRb.position);
                        mainToOuter = Vector3.Distance(Main4.JointRb.position, Outer.JointRb.position);
                        total = centerToMain1 + main1ToMain2 + main2ToMain3 + main3ToMain4 + mainToOuter;
                        _main4Ratio = (centerToMain1 + main1ToMain2 + main2ToMain3 + main3ToMain4) / total;
                    }
                    else
                    {
                        mainToOuter = Vector3.Distance(Main3.JointRb.position, Outer.JointRb.position);
                        total = centerToMain1 + main1ToMain2 + main2ToMain3 + mainToOuter;
                    }

                    _main3Ratio = (centerToMain1 + main1ToMain2 + main2ToMain3) / total;
                }
                else
                {
                    mainToOuter = Vector3.Distance(Main2.JointRb.position, Outer.JointRb.position);
                    total = centerToMain1 + main1ToMain2 + mainToOuter;
                }

                _main2Ratio = (centerToMain1 + main1ToMain2) / total;
            }
            else
            {
                mainToOuter = Vector3.Distance(Main1.JointRb.position, Outer.JointRb.position);
                total = centerToMain1 + mainToOuter;
            }

            _main1Ratio = centerToMain1 / total;
        }

        /* Value is parameter's overall base value based on softness, quickness and breast mass.
         * Individual joints get adjusted values based on the gradient.
         */
        internal void SetGradientBaseValues(string paramName, GradientConfig gradientConfig, float value)
        {
            Center.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 0));
            Main1.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main1Ratio));
            Main2?.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main2Ratio));
            Main3?.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main3Ratio));
            Main4?.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main4Ratio));
            Outer.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 1));
        }

        // TODO called unnecessarily multiple times with same inputs, could cache values
        // Gradient starts from outer (ratio 1) and ends in areola (ratio 0)
        static float CalculateJointBaseValue(GradientConfig config, float value, float ratio)
        {
            float middle = 1 -
                Mathf.Lerp(
                    0.01f,
                    0.99f,
                    Mathf.InverseLerp(config.MiddleShiftJsf.min, config.MiddleShiftJsf.max, config.MiddleShiftJsf.val)
                );

            float result;

            if(ratio > middle)
            {
                float smoothRatio = Calc.InverseSmoothStepWithMax(
                    ratio - middle,
                    1 - middle,
                    config.OuterToMiddleCurvature,
                    config.OuterToMiddleMidPoint
                );
                result = Mathf.Lerp(value, value * config.OuterMultiplierJsf.val, smoothRatio);
            }
            else
            {
                result = config.CenterMultiplierJsf.val * value;
            }

            return result;
        }

        internal void SetAdditionalOffsets(string paramName, float value)
        {
            Center.SetAdditionalOffset(paramName, value);
            Main1.SetAdditionalOffset(paramName, value);
            Main2?.SetAdditionalOffset(paramName, value);
            Main3?.SetAdditionalOffset(paramName, value);
            Main4?.SetAdditionalOffset(paramName, value);
            Outer.SetAdditionalOffset(paramName, value);
        }
    }
}
