// ReSharper disable UnusedMember.Local
using NAT.BM.Handlers;
using NAT.Common;
using NAT.Common.Diagnostics;
using NAT.Common.Models;
using NAT.Common.UI;
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.BM.UI
{
    sealed class DevWindow : Window
    {
        readonly GluteJointPhysicsHandler _jointPhysicsHandler;
        readonly GluteSoftPhysicsHandler _softPhysicsHandler;
        readonly GluteForceMorphHandler _forceMorphHandler;

        readonly JSONStorableString _leftDebugArea;
        readonly JSONStorableString _jointDiagnosticsJss;
        readonly JSONStorableStringChooser _softParamJssc;
        SoftPhysicsParamGroup _selectedParamGroup;

        readonly CalibratedJoint _leftJoint, _rightJoint;

        internal DevWindow(ScriptBase script) : base(script)
        {
            _jointPhysicsHandler = script.Provider.GetInstance<GluteJointPhysicsHandler>();
            _softPhysicsHandler = script.Provider.GetInstance<GluteSoftPhysicsHandler>();
            _forceMorphHandler = script.Provider.GetInstance<GluteForceMorphHandler>();
            _leftDebugArea = new JSONStorableString("leftDebugArea", "");
            _jointDiagnosticsJss = new JSONStorableString("jointDiagnostics", "");
            var options = _softPhysicsHandler.SoftParamGroups.Select(group => group.Name).ToList();
            _softParamJssc = new JSONStorableStringChooser("softPhysicsParams", options, options.First(), "Params")
            {
                setCallbackFunction = OnParamOptionSelected,
            };
            OnParamOptionSelected(_softParamJssc.val);

            _leftJoint = script.Person.LeftGlute;
            _rightJoint = script.Person.RightGlute;
        }

        internal Vector3 Deg2Rad(Vector3 eulerAngles)
        {
            float x = Mathf.Deg2Rad * eulerAngles.x;
            float y = Mathf.Deg2Rad * eulerAngles.y;
            float z = Mathf.Deg2Rad * eulerAngles.z;
            return new Vector3(x, y, z);
        }

        static IEnumerable<LineV3> PositionMarker(Vector3 pos, Color color, float size = 0.003f) => new List<LineV3>
        {
            new LineV3
            {
                From = pos + new Vector3(-size, 0, 0),
                To = pos + new Vector3(size, 0, 0),
                Color = color,
            },
            new LineV3
            {
                From = pos + new Vector3(0, -size, 0),
                To = pos + new Vector3(0, size, 0),
                Color = color,
            },
            new LineV3
            {
                From = pos + new Vector3(0, 0, -size),
                To = pos + new Vector3(0, 0, size),
                Color = color,
            },
        };

        void OnParamOptionSelected(string value)
        {
            _selectedParamGroup = _softPhysicsHandler.SoftParamGroups.First(group => group.Name == value);
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddElement(() =>
            {
                var textField = script.CreateTextField(_leftDebugArea, rightSide);
                textField.UItext.fontSize = 28;
                textField.height = Constant.UI_MAX_HEIGHT - 195;
                return textField;
            });

        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(50, rightSide);
            AddElement(() =>
            {
                var popup = script.CreateFilterablePopup(_softPhysicsHandler.SoftJointDiagnostics.SoftJointRbDiagnosticJssc, rightSide);
                script.ConfigureCustomPopup(popup, 480f, false);
                script.AddPrevNextButtons(popup, true);
                return popup;
            });

            AddElement(() =>
            {
                var popup = script.CreateFilterablePopup(_softParamJssc, rightSide);
                script.ConfigureCustomPopup(popup, 0f, false);
                script.AddPrevNextButtons(popup, true);
                return popup;
            });

            AddElement(() =>
            {
                var textField = script.CreateTextField(_jointDiagnosticsJss, true);
                textField.UItext.fontSize = 28;
                textField.height = 250;
                return textField;
            });
        }

        protected override void PostBuild() =>
            _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(true);

        protected override void OnClose() =>
            _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(false);

        internal override void UpdateDiagnostics()
        {
            if(Time.frameCount % 6 != 0)
            {
                return;
            }

            // DebugMorphAngleLimits();
        }

        void DebugMorphAngleLimits()
        {
            diagnosticsSb
                .Clear()
                .AppendFormat("NRealM: {0:0.000}", _jointPhysicsHandler.NormalizedRealMass)
                .AppendLine()
                .AppendLine()
                .AppendFormat("UpLimit: {0:0.000}", _forceMorphHandler.UpLimit)
                .AppendLine()
                .AppendFormat("DepthOutLimit: {0:0.000}", _forceMorphHandler.DepthOutLimit)
                .AppendLine()
                .AppendFormat("SidewaysOutLimit: {0:0.000}", _forceMorphHandler.SidewaysOutLimit)
                .AppendLine()
                .AppendFormat("SidewaysInLimit: {0:0.000}", _forceMorphHandler.SidewaysInLimit)
                .AppendLine()
                .AppendLine()
                .AppendFormat("UpDownNormalizer: {0:0.000}", _forceMorphHandler.UpDownNormalizer)
                .AppendLine()
                .AppendFormat("DepthInNormalizer: {0:0.000}", _forceMorphHandler.DepthInNormalizer)
                .AppendLine()
                .AppendFormat("DepthOutNormalizer: {0:0.000}", _forceMorphHandler.DepthOutNormalizer)
                .AppendLine()
                .AppendFormat("SidewaysNormalizer: {0:0.000}", _forceMorphHandler.SidewaysNormalizer)
                .AppendLine()
                .AppendLine()
                .AppendFormat("AngleY L: {0:0.000}", _leftJoint.AngleY)
                .AppendLine()
                .AppendFormat("AngleY R: {0:0.000}", _rightJoint.AngleY)
                .AppendLine()
                .AppendFormat("AngleX L: {0:0.000}", _leftJoint.AngleX)
                .AppendLine()
                .AppendFormat("AngleX R: {0:0.000}", _rightJoint.AngleX)
                .AppendLine()
                .AppendFormat("DepthDiff L: {0:0.000}", _leftJoint.DepthDiff)
                .AppendLine()
                .AppendFormat("DepthDiff R: {0:0.000}", _rightJoint.DepthDiff)
                .AppendLine();
            _leftDebugArea.val = diagnosticsSb.ToString();
        }

        void DebugSkinVerts()
        {
            try
            {
                string str = "";
                var rawSkinnedVerts = script.Person.Skin.rawSkinnedVerts;
                foreach(int i in _jointPhysicsHandler.leftVolumeCalculationVerts)
                {
                    var vertex = rawSkinnedVerts[i];
                    str += $"\n\n{i}: {vertex.ToPrettyString()}";
                }

                _leftDebugArea.val = str;
            }
            catch(Exception e)
            {
                Debug.Log($"DebugSkinVerts: {e}");
            }
        }

        void DebugSoftJoints()
        {
            try
            {
                if(_selectedParamGroup != null)
                {
                    if(Time.frameCount % 6 == 0)
                    {
                        _jointDiagnosticsJss.val = _softPhysicsHandler.SoftJointDiagnostics.GetSelectedJointValues(_selectedParamGroup);
                    }
                }

                _softPhysicsHandler.SoftJointDiagnostics.ClearLines();
                _softPhysicsHandler.SoftJointDiagnostics.DrawSelectedJoint();
                if(_selectedParamGroup != null)
                {
                    _softPhysicsHandler.SoftJointDiagnostics.DrawDiagnostics(_selectedParamGroup);
                }
            }
            catch(Exception e)
            {
                Debug.Log($"DebugSoftJoints: {e}");
            }
        }
    }
}
