using ColliderEditor;
using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.UI;
using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace NAT.BM.UI
{
    sealed class HardCollidersWindow : Window
    {
        readonly Dictionary<string, UIDynamic> _sectionElements = new Dictionary<string, UIDynamic>();
        readonly HardColliderHandler _hardColliderHandler;
        readonly ColliderVisualizer _visualizer;
        readonly JSONStorableString _allCollidersMassDisplayValueJss;

        internal HardCollidersWindow(ScriptBase script) : base(script)
        {
            _hardColliderHandler = script.Provider.GetInstance<HardColliderHandler>();
            _visualizer = _hardColliderHandler.ColliderVisualizer;
            _allCollidersMassDisplayValueJss = new JSONStorableString("allCollidersMassDisplayValue", "");
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            AddShowPreviewsPopupChangeHandler();
            AddColliderPopupChangeHandler();
            _visualizer.enabled = true;
            _visualizer.ShowPreviewsJSON.val = true;
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddHeaderTextField("Configure Hard Colliders", rightSide);

            AddInfoTextField(
                "Hard colliders apply a force to the back to the glute joint, making the glute move in response to collision.",
                rightSide
            );

            AddElement(_hardColliderHandler.BaseMassMultiplierJsf.name, () =>
            {
                var slider = script.CreateSlider(_hardColliderHandler.BaseMassMultiplierJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Base Mass Multiplier";
                return slider;
            });

            AddElement(() => CreateValueTextField(_allCollidersMassDisplayValueJss, 28, 45, rightSide));

            AddInfoTextField(
                new StringBuilder()
                    .AppendLine("One third of Glute Mass is applied to hard colliders, the rest is applied to the glute joint.")
                    .AppendLine()
                    .AppendBold("Collider mass")
                    .AppendLine(" affects the glute joint's inertia as well as the amount of force applied back to the glute joint.")
                    .AppendLine()
                    .AppendBold("Radius")
                    .AppendLine(" adjusts the size of the selected collider.")
                    .AppendLine()
                    .AppendBold("X, Y and Z offsets")
                    .AppendLine(" can be used to modify the position of the selected collider.")
                    .AppendLine()
                    .Append("The closer the collider matches the volume of the glute, the more accurately it responds to touch.")
                    .ToString(),
                rightSide,
                525f
            );

            AddElement(_visualizer.GroupsJSON.name, () =>
            {
                var popup = script.CreatePopup(_visualizer.GroupsJSON, rightSide);
                popup.label = "Show Previews";
                popup.popup.labelText.color = Color.black;
                popup.popup.popupPanel.offsetMin += new Vector2(5, popup.popup.popupPanelHeight + 55);
                popup.popup.popupPanel.offsetMax += new Vector2(5, popup.popup.popupPanelHeight + 55);
                return popup;
            });

            AddElement(_visualizer.XRayPreviewsJSON.name, () =>
            {
                var toggle = script.CreateToggle(_visualizer.XRayPreviewsJSON, rightSide);
                toggle.label = "Use XRay Previews";
                return toggle;
            });

            AddElement(_hardColliderHandler.HighlightAllJsb.name, () =>
                script.CreateToggle(_hardColliderHandler.HighlightAllJsb, rightSide)
            );
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(40, rightSide);
            AddElement(_hardColliderHandler.ColliderGroupsJssc.name, () =>
            {
                var popup = script.CreateFilterablePopup(_hardColliderHandler.ColliderGroupsJssc, rightSide);
                script.ConfigureCustomPopup(popup, 520f, false);
                script.AddPrevNextButtons(popup, true);
                return popup;
            });
        }

        protected override void OnClose()
        {
            _visualizer.ShowPreviewsJSON.val = false;
            _visualizer.enabled = false;
        }

        void RebuildSection(string colliderId, bool rightSide = true)
        {
            var hardColliderGroup = _hardColliderHandler.HardColliderGroups
                .Find(group => group.VisualizerId == colliderId);

            AddColliderSectionElement(hardColliderGroup.MassMultiplierJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.MassMultiplierJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Mass Multiplier";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.Left.MassDisplayValueJss.name, () =>
                CreateValueTextField(hardColliderGroup.Left.MassDisplayValueJss, 28, 45, rightSide)
            );

            AddColliderSectionElement(hardColliderGroup.RadiusJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.RadiusJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Radius Offset";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.RightJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.RightJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "X Offset";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.UpJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.UpJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Y Offset";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.LookJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.LookJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Z Offset";
                return slider;
            });
        }

        void AddColliderSectionElement(string key, Func<UIDynamic> createElement)
        {
            _sectionElements[key] = createElement();
        }

        internal override void Clear()
        {
            base.Clear();
            ClearSection();
            RemoveShowPreviewsPopupChangeHandler();
            RemoveColliderPopupChangeHandler();
        }

        void ClearSection()
        {
            foreach(var pair in _sectionElements)
            {
                script.RemoveElement(pair.Value);
            }

            _sectionElements.Clear();
        }

        void AddShowPreviewsPopupChangeHandler()
        {
            var uiDynamicPopup = GetUIDynamicAs<UIDynamicPopup>(_visualizer.GroupsJSON.name);
            if(uiDynamicPopup)
            {
                uiDynamicPopup.popup.onValueChangeHandlers += OnShowPreviewsPopupValueChanged;
                GetUIDynamic(_visualizer.XRayPreviewsJSON.name).SetActiveStyle(uiDynamicPopup.popup.currentValue != "Off");
            }
        }

        void RemoveShowPreviewsPopupChangeHandler()
        {
            var popup = GetUIDynamicAs<UIDynamicPopup>(_visualizer.GroupsJSON.name);
            if(popup)
            {
                popup.popup.visible = false;
                popup.popup.onValueChangeHandlers -= OnShowPreviewsPopupValueChanged;
            }
        }

        void OnShowPreviewsPopupValueChanged(string value)
        {
            GetUIDynamic(_visualizer.XRayPreviewsJSON.name).SetActiveStyle(value != "Off");
            GetUIDynamic(_hardColliderHandler.HighlightAllJsb.name).SetActiveStyle(value != "Off");
        }

        void AddColliderPopupChangeHandler()
        {
            var popup = GetUIDynamicAs<UIDynamicPopup>(_hardColliderHandler.ColliderGroupsJssc.name);
            if(popup)
            {
                popup.popup.onValueChangeHandlers += OnColliderPopupValueChanged;
                RebuildSection(_hardColliderHandler.ColliderGroupsJssc.val);
            }
        }

        void RemoveColliderPopupChangeHandler()
        {
            var popup = GetUIDynamicAs<UIDynamicPopup>(_hardColliderHandler.ColliderGroupsJssc.name);
            if(popup)
            {
                popup.popup.visible = false;
                popup.popup.onValueChangeHandlers -= OnColliderPopupValueChanged;
            }
        }

        void OnColliderPopupValueChanged(string value)
        {
            ClearSection();
            RebuildSection(value);
        }

        internal override void UpdateDiagnostics()
        {
            if(!isBuilt || Time.frameCount % 6 != 0)
            {
                return;
            }

            float mass = 0f;
            foreach(var group in _hardColliderHandler.HardColliderGroups)
            {
                mass += group.Left.GetRigidbodyMass();
            }

            _allCollidersMassDisplayValueJss.val = diagnosticsSb
                .Clear()
                .AppendFormat("All Colliders Mass: <b>{0:0.000} kg</b>", mass)
                .ToString();
        }
    }
}
