using NAT.BM.Handlers;
using NAT.Common;
using NAT.Common.UI;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.BM.UI
{
    sealed class MorphsWindow : Window
    {
        readonly GluteForceMorphHandler _forceMorphHandler;

        internal MorphsWindow(ScriptBase script) : base(script)
        {
            _forceMorphHandler = script.Provider.GetInstance<GluteForceMorphHandler>();
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            GetUIDynamic(_forceMorphHandler.BaseJsf.name).AddListener(UpdateAllSliderColors);
            UpdateAllSliderColors(0);
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddHeaderTextField("Directional Force Morphing", rightSide);

            AddInfoTextField(
                "Adjust the amount of morphing and the joint's neutral position and rotation.",
                rightSide,
                85f,
                28
            );

            AddElement(() =>
            {
                var slider = script.CreateSlider(script.UpDownAngleOffsetJsf, rightSide);
                slider.valueFormat = "F1";
                slider.label = script.UpDownAngleOffsetJsf.name + " (°)";
                return slider;
            });

            AddInfoTextField(
                "Up/Down morphing is based on the joint rotating away from the calibrated vertical angle (plus offset).",
                rightSide,
                140f,
                28
            );

            AddElement(() =>
            {
                var slider = script.CreateSlider(script.DepthOffsetJsf, rightSide);
                slider.valueFormat = "F1";
                slider.label = script.DepthOffsetJsf.name + " (mm)";
                return slider;
            });

            AddInfoTextField(
                "Depth morphing is based on the joint's movement away from the calibrated Z axis position (plus offset).",
                rightSide,
                140f,
                28
            );

            AddElement(() =>
            {
                var slider = script.CreateSlider(script.LeftRightAngleOffsetJsf, rightSide);
                slider.valueFormat = "F1";
                slider.label = script.LeftRightAngleOffsetJsf.name + " (°)";
                return slider;
            });

            AddInfoTextField(
                "Sideways morphing is based on the joint rotating away from the calibrated horizontal angle (plus offset).",
                rightSide,
                140f,
                28
            );
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(5, rightSide);
            AddElement(_forceMorphHandler.BaseJsf.name, () =>
            {
                var slider = script.CreateSlider(_forceMorphHandler.BaseJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Base Multiplier";
                return slider;
            });
            AddSpacer(5, rightSide);
            AddMultiplierSlider(_forceMorphHandler.UpJsf, "Up", rightSide);
            AddMultiplierSlider(_forceMorphHandler.DownJsf, "Down", rightSide);
            AddSpacer(5, rightSide);
            AddMultiplierSlider(_forceMorphHandler.DepthOutJsf, "Depth Out", rightSide);
            AddMultiplierSlider(_forceMorphHandler.DepthInJsf, "Depth In", rightSide);
            AddSpacer(5, rightSide);
            AddMultiplierSlider(_forceMorphHandler.SidewaysInJsf, "Sideways In", rightSide);
            AddMultiplierSlider(_forceMorphHandler.SidewaysOutJsf, "Sideways Out", rightSide);
        }

        void AddMultiplierSlider(JSONStorableFloat storable, string label, bool rightSide) => AddElement(storable.name, () =>
        {
            var slider = script.CreateSlider(storable, rightSide);
            slider.valueFormat = "F2";
            slider.label = label;
            slider.AddListener((float _) => UpdateSliderColor(storable));
            return slider;
        });

        void UpdateAllSliderColors(float _)
        {
            UpdateSliderColor(_forceMorphHandler.UpJsf);
            UpdateSliderColor(_forceMorphHandler.DownJsf);
            UpdateSliderColor(_forceMorphHandler.DepthOutJsf);
            UpdateSliderColor(_forceMorphHandler.DepthInJsf);
            UpdateSliderColor(_forceMorphHandler.SidewaysInJsf);
            UpdateSliderColor(_forceMorphHandler.SidewaysOutJsf);
        }

        void UpdateSliderColor(JSONStorableFloat storable)
        {
            var slider = GetUIDynamicAs<UIDynamicSlider>(storable.name);
            if(slider)
            {
                var images = slider.slider.gameObject.transform.GetComponentsInChildren<Image>();
                var fillImage = images.First(image => image.name == "Fill");
                var handleImage = images.First(image => image.name == "Handle");
                var color = MultiplierSliderColor(_forceMorphHandler.BaseJsf.val * storable.val);
                fillImage.color = color;
                handleImage.color = color;
            }
        }

        static Color MultiplierSliderColor(float value) => value <= 1
            ? Color.Lerp(new Color(1, 1, 1, 0.25f), Color.white, value)
            : Color.Lerp(Color.white, new Color(1.0f, 0.2f, 0.2f), (value - 1) / 3);
    }
}
