using NAT.BM.Handlers;
using NAT.Common;
using NAT.Common.Models;
using NAT.Common.UI;
using System.Text;
using UnityEngine;

namespace NAT.BM.UI
{
    sealed class ParameterWindow : Window
    {
        readonly GluteJointPhysicsHandler _jointPhysicsHandler;
        readonly JointPhysicsParamGroup _paramGroup;
        readonly JointPhysicsParam _param;
        readonly JSONStorableString _valueJss = new JSONStorableString("Value", "");
        readonly JSONStorableString _gravityTargetsJss = new JSONStorableString("Gravity Targets", "");
        readonly StringBuilder _valueSb = new StringBuilder();
        readonly StringBuilder _gravityTargetsSb = new StringBuilder();

        internal ParameterWindow(ScriptBase script, JointPhysicsParamGroup paramGroup) : base(script)
        {
            var provider = script.Provider;
            _jointPhysicsHandler = provider.GetInstance<GluteJointPhysicsHandler>();
            _paramGroup = paramGroup;
            _param = _paramGroup.Left;
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddInfoTextField(
                "\n".Size(8) + _paramGroup.InfoText,
                rightSide,
                280f,
                28
            );

            if(_param.HasAdjustableGravityPhysics())
            {
                AddHeaderTextField("Gravity Physics", rightSide);

                AddInfoTextField(
                    "\n".Size(8) +
                    _paramGroup.GravityInfoText,
                    rightSide,
                    150f,
                    28
                );

                AddElement(() =>
                {
                    var textField = script.CreateTextField(_gravityTargetsJss, rightSide);
                    textField.UItext.fontSize = 28;
                    textField.height = 250;
                    textField.backgroundColor = Color.clear;
                    return textField;
                });
            }
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddElement(() =>
            {
                var toggle = script.CreateToggle(_paramGroup.ApplyChangesOnlyToLeftParamJsb, rightSide);
                toggle.label = "Apply Changes Only To Left Glute";
                return toggle;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(_param.OffsetJsf, rightSide);
                slider.valueFormat = _paramGroup.ValueFormat;
                slider.label = "Offset";
                return slider;
            });

            AddElement(() => CreateValueTextField(_valueJss, 28, 45, rightSide));

            if(_param.HasAdjustableGravityPhysics())
            {
                AddSpacer(170, rightSide);
                foreach(var pair in _param.GravityCalculators)
                {
                    AddElement(() =>
                    {
                        var slider = script.CreateSlider(pair.Value.MultiplierJsf, rightSide);
                        slider.label = $"{pair.Key} Gravity";
                        return slider;
                    });
                }
            }
        }

        protected override void PostBuild() => Refresh();

        internal override void UpdateDiagnostics()
        {
            if(Time.frameCount % 6 != 0)
            {
                return;
            }

            _valueJss.val = _valueSb
                .Clear()
                .AppendFormat("Value: <b>{0}</b>", _param.Value.ToString(_paramGroup.ValueFormat))
                .ToString();

            if(_param.HasAdjustableGravityPhysics())
            {
                _gravityTargetsSb
                    .Clear()
                    .AppendSize("\n", 8)
                    .AppendLine("Left glute gravity target values:");

                foreach(var pair in _param.GravityCalculators)
                {
                    var config = pair.Value;
                    float targetValue =
                        _param.OffsetJsf.val +
                        _param.BaseValueJsf.val +
                        config.CalculateDynamic(1f, _jointPhysicsHandler.NormalizedMass, script.Softness, true);
                    targetValue = Mathf.Clamp(targetValue, _param.BaseValueJsf.min, _param.BaseValueJsf.max);
                    _gravityTargetsSb
                        .AppendLine()
                        .AppendFormat("{0}: ", pair.Key)
                        .AppendBold(targetValue.ToString(_paramGroup.ValueFormat));
                }

                _gravityTargetsJss.val = _gravityTargetsSb.ToString();
            }
        }
    }
}
