using NAT.BM.Handlers;
using NAT.Common;
using NAT.Common.Models;
using NAT.Common.UI;
using System.Collections.Generic;

namespace NAT.BM.UI
{
    sealed class PhysicsWindow : Window
    {
        readonly GluteJointPhysicsHandler _jointPhysicsHandler;
        readonly GluteSoftPhysicsHandler _softPhysicsHandler;
        readonly GluteGravityPhysicsHandler _gravityPhysicsHandler;

        readonly Dictionary<string, Window> _parameterWindows = new Dictionary<string, Window>();

        internal PhysicsWindow(ScriptBase script) : base(script)
        {
            var provider = script.Provider;
            _jointPhysicsHandler = provider.GetInstance<GluteJointPhysicsHandler>();
            _softPhysicsHandler = provider.GetInstance<GluteSoftPhysicsHandler>();
            _gravityPhysicsHandler = provider.GetInstance<GluteGravityPhysicsHandler>();
            AddParameterWindow(Constant.GLUTE_WEIGHT, new WeightWindow(script, _jointPhysicsHandler.JointMass));

            foreach(var paramGroup in _jointPhysicsHandler.ParamGroups)
            {
                AddParameterWindow(paramGroup.Name, new ParameterWindow(script, paramGroup));
            }

            foreach(var paramGroup in _softPhysicsHandler.SoftParamGroups)
            {
                AddParameterWindow(paramGroup.Name, new SoftParameterWindow(script, paramGroup));
            }

            foreach(var paramGroup in _softPhysicsHandler.ParamGroups)
            {
                AddParameterWindow(paramGroup.Name, new ParameterWindow(script, paramGroup));
            }
        }

        void AddParameterWindow(string key, Window window)
        {
            AddNestedWindow(window);
            _parameterWindows[key] = window;
        }

        /* Buttons are active if soft physics is currently enabled and in sync, and settingsMonitor agrees */

        bool IsSoftPhysicsOnAndSynced() =>
            _softPhysicsHandler.SoftPhysicsOn &&
            _softPhysicsHandler.SoftPhysicsSynced &&
            script.Bootstrap.PersonMonitor.IsGluteSoftPhysicsOn();

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();

            UpdateMainParamButtons();
            UpdateSoftParamButtons(IsSoftPhysicsOnAndSynced());
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddHeaderTextField("Joint Physics Parameters", rightSide);
            AddSpacer(5, rightSide);
            AddPhysicsParamButton(Constant.GLUTE_WEIGHT, rightSide);
            foreach(var paramGroup in _jointPhysicsHandler.ParamGroups)
            {
                AddPhysicsParamButton(paramGroup.Name, rightSide);
            }
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddHeaderTextField("Soft Physics Parameters", rightSide);
            AddSpacer(5, rightSide);
            foreach(var paramGroup in _softPhysicsHandler.SoftParamGroups)
            {
                AddSoftPhysicsParamButton(paramGroup, rightSide);
            }

            foreach(var paramGroup in _softPhysicsHandler.ParamGroups)
            {
                AddSoftPhysicsParamButton(paramGroup, rightSide);
            }
        }

        void AddPhysicsParamButton(string groupName, bool rightSide) => AddElement($"{groupName}Button", () =>
        {
            var button = script.CreateButton(groupName, rightSide);
            button.height = 80.7f;
            button.buttonText.fontSize = 30;
            button.AddListener(() =>
            {
                ClearSelf();
                ActiveNestedWindow = _parameterWindows[groupName];
                ActiveNestedWindow.Rebuild();
            });

            return button;
        });

        void AddSoftPhysicsParamButton(SoftPhysicsParamGroup paramGroup, bool rightSide) => AddSoftPhysicsParamButton(paramGroup.Name, rightSide);
        void AddSoftPhysicsParamButton(JointPhysicsParamGroup paramGroup, bool rightSide) => AddSoftPhysicsParamButton(paramGroup.Name, rightSide);

        void AddSoftPhysicsParamButton(string paramName, bool rightSide) => AddElement($"{paramName}Button", () =>
        {
            var button = script.CreateButton(paramName, rightSide);
            button.height = 80.7f;
            button.buttonText.fontSize = 30;
            button.AddListener(() =>
            {
                ClearSelf();
                ActiveNestedWindow = _parameterWindows[paramName];
                ActiveNestedWindow.Rebuild();
            });

            return button;
        });

        void UpdateMainParamButtons()
        {
            SyncParamButton(Constant.GLUTE_WEIGHT, _jointPhysicsHandler.JointMass.IsModified || _gravityPhysicsHandler.IsGravityModified);

            foreach(var paramGroup in _jointPhysicsHandler.ParamGroups)
            {
                SyncParamButton(paramGroup.Name, paramGroup.IsModified);
            }
        }

        void UpdateSoftParamButtons(bool active)
        {
            foreach(var paramGroup in _softPhysicsHandler.SoftParamGroups)
            {
                SyncParamButton(paramGroup.Name, paramGroup.IsModified, active);
            }

            foreach(var paramGroup in _softPhysicsHandler.ParamGroups)
            {
                SyncParamButton(paramGroup.Name, paramGroup.IsModified, active);
            }
        }

        internal override void Refresh()
        {
            UpdateSoftParamButtons(_softPhysicsHandler.SoftPhysicsOn);
        }

        internal override void OnMainParamAdjusted(PhysicsParamGroup paramGroup)
        {
            if(!script.Tabs.IsActive(this))
            {
                return;
            }

            if(paramGroup == _jointPhysicsHandler.JointMass)
            {
                SyncParamButton(Constant.GLUTE_WEIGHT, paramGroup.IsModified || _gravityPhysicsHandler.IsGravityModified);
                return;
            }

            if(
                paramGroup == _softPhysicsHandler.FatBackForce ||
                paramGroup == _softPhysicsHandler.FatBackForceThreshold ||
                paramGroup == _softPhysicsHandler.FatBackForceMaxForce
            )
            {
                bool active = IsSoftPhysicsOnAndSynced();
                SyncParamButton(paramGroup.Name, paramGroup.IsModified, active);
                return;
            }

            SyncParamButton(paramGroup.Name, paramGroup.IsModified);
        }

        internal override void OnSoftParamAdjusted(PhysicsParamGroup paramGroup)
        {
            if(!script.Tabs.IsActive(this))
            {
                return;
            }

            SyncParamButton(paramGroup.Name, paramGroup.IsModified, IsSoftPhysicsOnAndSynced());
        }

        void SyncParamButton(string name, bool modified, bool active = true)
        {
            var button = GetUIDynamicAs<UIDynamicButton>($"{name}Button");
            string label = $"  {name}";
            if(modified && active)
            {
                label += " *".Bold();
            }

            button.label = label;
            button.SetActiveStyle(active, true);
        }
    }
}
