using NAT.Common;
using NAT.Common.Models;
using NAT.Common.UI;
using NAT.BM.Handlers;
using NAT.Common.Handlers;
using System.Text;
using UnityEngine;

namespace NAT.BM.UI
{
    sealed class WeightWindow : Window
    {
        readonly GravityPhysicsHandler _gravityPhysicsHandler;
        readonly JointPhysicsParamGroup _paramGroup;
        readonly JointPhysicsParam _param;
        readonly JSONStorableString _valueJss = new JSONStorableString("Value", "");
        readonly JSONStorableString _gravityValuesJss = new JSONStorableString("Gravity Values", "");
        readonly StringBuilder _valueSb = new StringBuilder();
        readonly StringBuilder _gravityValuesSb = new StringBuilder();

        internal WeightWindow(ScriptBase script, JointPhysicsParamGroup paramGroup) : base(script)
        {
            var provider = script.Provider;
            _gravityPhysicsHandler = provider.GetInstance<GluteGravityPhysicsHandler>();
            _paramGroup = paramGroup;
            _param = _paramGroup.Left;
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddInfoTextField(
                "\n".Size(8) + _paramGroup.InfoText,
                rightSide,
                220f,
                28
            );

            AddHeaderTextField("Glute Gravity", rightSide);

            AddInfoTextField(
                new StringBuilder()
                    .AppendSize("\n", 8)
                    .AppendLine("Adjust the force of gravity on the glute joint relative to different rotations of the joint.")
                    .AppendLine()
                    .Append("\"Down Gravity\" makes glutes hang down when standing up,")
                    .AppendLine(" while \"Back Gravity\" pulls glutes out when leaning back.")
                    .AppendLine()
                    .Append("The effective gravity is a result of scene gravity, background multipliers, and the multipliers on the right.")
                    .ToString(),
                rightSide,
                395f,
                28
            );

            AddElement(() =>
            {
                var textField = script.CreateTextField(_gravityValuesJss, rightSide);
                textField.UItext.fontSize = 28;
                textField.height = 350f;
                textField.backgroundColor = Color.clear;
                return textField;
            });
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddElement(() =>
            {
                var toggle = script.CreateToggle(_paramGroup.ApplyChangesOnlyToLeftParamJsb, rightSide);
                toggle.label = "Apply Changes Only To Left Glute";
                return toggle;
            });

            AddElement(() =>
            {
                var storable = _param.OffsetJsf;
                var slider = script.CreateSlider(storable, rightSide);
                slider.valueFormat = _paramGroup.ValueFormat;
                slider.label = "Offset";
                return slider;
            });

            AddElement(() => CreateValueTextField(_valueJss, 28, 45f, rightSide));

            AddSpacer(110, rightSide);
            AddElement(() => script.CreateToggle(_gravityPhysicsHandler.LinkDirectionsJsb, rightSide));

            foreach(var multiplier in _gravityPhysicsHandler.UserMultipliers)
            {
                AddElement(multiplier.name, () =>
                {
                    var slider = script.CreateSlider(multiplier, rightSide);
                    slider.valueFormat = "F2";
                    return slider;
                });
            }
        }

        internal override void UpdateDiagnostics()
        {
            if(Time.frameCount % 6 != 0)
            {
                return;
            }

            _valueJss.val = _valueSb
                .Clear()
                .AppendFormat("Value: <b>{0}</b>", _param.Value.ToString(_paramGroup.ValueFormat))
                .ToString();

            var gravity = _gravityPhysicsHandler.CustomGravityLeft;
            _gravityValuesJss.val = _gravityValuesSb
                .Clear()
                .AppendSize("\n", 8)
                .AppendItalic("Scene gravity (m/s^2):")
                .AppendLine()
                .AppendSize("\n", 8)
                .AppendFormat(" x: {0:0.00}", Physics.gravity.x)
                .AppendLine()
                .AppendFormat(" y: {0:0.00}", Physics.gravity.y)
                .AppendLine()
                .AppendFormat(" z: {0:0.00}", Physics.gravity.z)
                .AppendLine()
                .AppendLine()
                .AppendItalic("Glute joint gravity (m/s^2):")
                .AppendLine()
                .AppendSize("\n", 8)
                .AppendFormat(" x: {0:0.00}", gravity.x)
                .AppendLine()
                .AppendFormat(" y: {0:0.00}", gravity.y)
                .AppendLine()
                .AppendFormat(" z: {0:0.00}", gravity.z)
                .AppendLine()
                .ToString();
        }
    }
}
