using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Components
{
    sealed class UIMod : MonoBehaviour
    {
        readonly List<RectTransformReposition> _repositionedRects = new List<RectTransformReposition>();
        readonly List<GameObject> _inactivatedUIGameObjects = new List<GameObject>();
        readonly List<GameObject> _customUIGameObjects = new List<GameObject>();

        internal string Id { get; private set; }
        Transform _container;
        string _targetName;
        UnityEventsListener _containerUIEventsListener;
        internal UnityEventsListener TargetUIEventsListener { get; set; }

        internal bool IsVisible() => TargetUIEventsListener && TargetUIEventsListener.IsEnabled;

        Func<UIMod, IEnumerator> _applyChanges;
        internal Func<bool> CheckPluginEnabled { get; set; }
        internal bool UseDisabledStyle { get; set; }

        internal void Init(
            string uiModId,
            Transform container,
            string targetTransformName,
            Func<UIMod, IEnumerator> changesFunc
        )
        {
            Id = uiModId;
            _container = container;
            _targetName = targetTransformName;
            _applyChanges = changesFunc;
        }

        internal void DeferApply()
        {
            if(!isActiveAndEnabled)
            {
                Debug.Log($"UIMod {Id} in is not active and enabled");
                return;
            }

            if(!this)
            {
                Debug.Log($"UIMod {Id} doesn't exist");
                return;
            }

            if(_container.gameObject.activeInHierarchy)
            {
                StartCoroutine(ApplyCo());
            }
            /* Add listener to wait for the transform to be active in hierarchy */
            else
            {
                _containerUIEventsListener = _container.gameObject.AddComponent<UnityEventsListener>();
                _containerUIEventsListener.EnabledHandlers += () => StartCoroutine(ApplyCo());
            }
        }

        bool _modified;
        internal Transform Target { get; private set; }

        IEnumerator ApplyCo()
        {
            if(_modified)
            {
                yield break;
            }

            DestroyImmediate(_containerUIEventsListener);
            _containerUIEventsListener = null;

            /* Set modified immediately to prevent double apply in case eventListener
             * gets enabled multiple times in a short period of time
             */
            _modified = true;

            if(_targetName != null)
            {
                float timeout = Time.unscaledTime + 1;
                while(Time.unscaledTime < timeout)
                {
                    yield return new WaitForSeconds(0.1f);
                    Target = _container.Find(_targetName);
                }

                if(!Target)
                {
                    Loggr.Error($"Failed to modify {Id} UI on target {_targetName} - could not find transform");
                    yield break;
                }
            }

            yield return _applyChanges(this);
            if(!CheckPluginEnabled())
            {
                Disable();
            }
        }

        internal void InactivateChildren()
        {
            foreach(Transform child in Target)
            {
                if(child.gameObject.activeSelf)
                {
                    _inactivatedUIGameObjects.Add(child.gameObject);
                    child.gameObject.SetActive(false);
                }
            }
        }

        internal void AddCustomObject(GameObject go)
        {
            _customUIGameObjects.Add(go);
        }

        internal void RepositionTransform(Transform t, Vector2 originalPosition, Vector2 newPosition)
        {
            _repositionedRects.Add(new RectTransformReposition(t.GetComponent<RectTransform>(), originalPosition, newPosition));
        }

        internal void Enable()
        {
            if(!_modified)
            {
                DeferApply();
            }
            else
            {
                foreach(var go in _customUIGameObjects)
                {
                    if(UseDisabledStyle)
                    {
                        var elements = go.GetComponentsInChildren<UIDynamic>();
                        foreach(var element in elements)
                        {
                            element.SetActiveStyle(true, false, true);
                        }
                    }
                    else
                    {
                        go.SetActive(true);
                    }
                }

                foreach(var go in _inactivatedUIGameObjects)
                {
                    go.SetActive(false);
                }

                foreach(var repositionedRect in _repositionedRects)
                {
                    repositionedRect.Apply();
                }
            }
        }

        internal void Disable()
        {
            foreach(var go in _customUIGameObjects)
            {
                if(UseDisabledStyle)
                {
                    var elements = go.GetComponentsInChildren<UIDynamic>();
                    foreach(var element in elements)
                    {
                        element.SetActiveStyle(false, true);
                    }
                }
                else
                {
                    go.SetActive(false);
                }
            }

            foreach(var go in _inactivatedUIGameObjects)
            {
                go.SetActive(true);
            }

            foreach(var repositionedRect in _repositionedRects)
            {
                repositionedRect.RestoreOriginal();
            }
        }

        void OnDestroy()
        {
            DestroyImmediate(_containerUIEventsListener);
            _containerUIEventsListener = null;
            DestroyImmediate(TargetUIEventsListener);
            TargetUIEventsListener = null;
            foreach(var go in _customUIGameObjects)
            {
                Destroy(go);
            }
        }
    }
}
