using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

namespace NAT.Common.Diagnostics
{
    sealed class SoftJointDiagnostics
    {
        readonly Color _gridColor = new Color(1f, 1f, 1f, 0.050f);
        readonly Color _gradientColor = new Color(1f, 1f, 1f, 0.125f);
        readonly float _markerSize = Utils.IsVR() ? 0.001f : 0.003f;

        readonly SoftJoint[] _leftJoints, _rightJoints;
        DrawService _drawService;
        readonly Camera _camera;

        internal JSONStorableBool VisualizeGradientJsb { get; }
        internal JSONStorableBool VisualizeLdpJsb { get; }
        internal JSONStorableStringChooser SoftJointRbDiagnosticJssc { get; }

        SoftJoint _selected;

        internal bool IsDrawingClothingCoverage { get; set; }

        internal SoftJointDiagnostics(SoftJoint[] leftJoints, SoftJoint[] rightJoints, DrawService drawService, Camera camera)
        {
            _leftJoints = leftJoints;
            _rightJoints = rightJoints;
            _drawService = drawService;
            _camera = camera;

            // TODO deprecate
            var options = new List<string> { Constant.SELECT };
            foreach(var softJoint in _leftJoints)
            {
                options.Add(softJoint.Id);
            }

            SoftJointRbDiagnosticJssc = new JSONStorableStringChooser(
                "softJointRb",
                options,
                options,
                Constant.SELECT,
                "Soft Joint",
                value => _selected = _leftJoints.FirstOrDefault(softJoint => softJoint.Id == value)
            );

            _drawService.Init(Color.gray, true);

            VisualizeGradientJsb = new JSONStorableBool("Visualize Gradient", true);
            VisualizeLdpJsb = new JSONStorableBool("Visualize Local Distortion", false);
        }

        internal void DrawServiceSetEnabled(bool value)
        {
            if(!_drawService)
            {
                return;
            }

            ClearLines();
            IsDrawingClothingCoverage = false;
            _drawService.enabled = value;
        }

        internal void ClearLines() => _drawService.Lines.Clear();

        internal void DrawDiagnostics(SoftPhysicsParamGroup paramGroup)
        {
            if(!_drawService.enabled || !(VisualizeGradientJsb.val || VisualizeLdpJsb.val))
            {
                return;
            }

            string paramName = paramGroup.Name;
            var param = paramGroup.Left;
            float visualizationMin = -Mathf.Max(paramGroup.VisualizationLimit, param.OffsetJsf.min - param.OffsetJsf.val);
            float visualizationMax = Mathf.Min(paramGroup.VisualizationLimit, param.OffsetJsf.max - param.OffsetJsf.val);

            if(VisualizeGradientJsb.val)
            {
                float baseWithOffset = param.BaseValueJsf.val + param.OffsetJsf.val;
                foreach(var softJoint in _leftJoints)
                {
                    var color = JointRelativeColor(
                        _gradientColor,
                        softJoint.GetBaseValue(paramName) - baseWithOffset,
                        baseWithOffset,
                        visualizationMin,
                        visualizationMax
                    );
                    AddPositionMarker(softJoint, color);
                }
            }

            if(VisualizeLdpJsb.val)
            {
                if(paramGroup.VisualizeRelativeChange)
                {
                    foreach(var softJoint in _leftJoints)
                    {
                        float jointBaseValue = softJoint.GetBaseValue(paramName);
                        var color = JointRelativeColor(
                            _gridColor,
                            softJoint.GetDistortionValue(paramName),
                            jointBaseValue,
                            visualizationMin,
                            visualizationMax
                        );
                        AddJointConnections(softJoint, color);
                    }
                }
                else
                {
                    foreach(var softJoint in _leftJoints)
                    {
                        var color = JointColor(
                            _gridColor,
                            softJoint.GetDistortionValue(paramName),
                            visualizationMin,
                            visualizationMax
                        );
                        AddJointConnections(softJoint, color);
                    }
                }
            }
        }

        internal void DrawClothingCoverage(ClothingAdjustmentProfile profile)
        {
            if(!_drawService.enabled)
            {
                return;
            }

            IsDrawingClothingCoverage = true;
            foreach(var pair in profile.CoveredSoftJoints)
            {
                foreach(var softJoint in pair.Value)
                {
                    AddPositionMarker(softJoint, Colors.transparentGreen);
                }
            }

            foreach(var pair in profile.ExposedSoftJoints)
            {
                foreach(var softJoint in pair.Value)
                {
                    AddPositionMarker(softJoint, Colors.transparentLightGray);
                }
            }

            if(profile.HoveredOverJoint != null)
            {
                AddRectangle(profile.HoveredOverJoint, Colors.desaturatedRed);
            }
        }

        internal void DrawClothingAdjustments(SoftPhysicsParamGroup paramGroup)
        {
            if(!_drawService.enabled || paramGroup == null)
            {
                return;
            }

            DrawClothingAdjustmentsSide(paramGroup, paramGroup.Left, _leftJoints);
            DrawClothingAdjustmentsSide(paramGroup, paramGroup.Right, _rightJoints);
        }

        void DrawClothingAdjustmentsSide(SoftPhysicsParamGroup paramGroup, SoftPhysicsParam param, IEnumerable<SoftJoint> softJoints)
        {
            string paramName = paramGroup.Name;
            float visualizationMin = -Mathf.Max(paramGroup.ClothingAdjustmentVisualizationLimit, param.OffsetJsf.min - param.OffsetJsf.val);
            float visualizationMax = Mathf.Min(paramGroup.ClothingAdjustmentVisualizationLimit, param.OffsetJsf.max - param.OffsetJsf.val);

            if(paramGroup.VisualizeRelativeChange)
            {
                foreach(var softJoint in softJoints)
                {
                    float jointBaseValue = softJoint.GetBaseValue(paramName);
                    var color = JointRelativeColor(
                        _gradientColor,
                        softJoint.GetClothingAdjustmentValue(paramName),
                        jointBaseValue,
                        visualizationMin,
                        visualizationMax
                    );
                    AddPositionMarker(softJoint, color);
                }
            }
            else
            {
                foreach(var softJoint in softJoints)
                {
                    var color = JointColor(
                        _gradientColor,
                        softJoint.GetClothingAdjustmentValue(paramName),
                        visualizationMin,
                        visualizationMax
                    );
                    AddPositionMarker(softJoint, color);
                }
            }
        }

        internal void DrawSelectedJoint()
        {
            if(_selected != null)
            {
                AddPositionMarker(_selected, Color.green);
                AddJointConnections(_selected, Color.white);
            }
        }

        internal string GetSelectedJointValues(SoftPhysicsParamGroup paramGroup)
        {
            if(_selected == null)
            {
                return "";
            }

            string paramName = paramGroup.Name;
            string format = paramGroup.ValueFormat;

            var sb = new StringBuilder();
            float paramBase = _selected.GetBaseValue(paramName);
            float paramOffset = paramGroup.Left.OffsetJsf.val;
            sb.AppendLine();
            sb.AppendLine($"Base+Offset value: {(paramBase + paramOffset).ToString(format).Bold()}");
            sb.AppendLine($"Base value: {paramBase.ToString(format).Bold()}");
            sb.AppendLine($"Offset value: {paramOffset.ToString(format).Bold()}");
            sb.AppendLine($"Distortion: {(10 * _selected.Distortion).ToString(format)} mm");
            sb.AppendLine($"Distortion value: {_selected.GetDistortionValue(paramName).ToString(format).Bold()}");
            sb.AppendLine($"Real value: {_selected.GetRealValue(paramName).ToString(format).Bold()}");
            sb.Append("\n".Size(6));
            return sb.ToString();
        }

        void AddPositionMarker(SoftJoint softJoint, Color color) =>
            _drawService.Lines.AddRange(PositionMarker(softJoint.JointRb.position, color, _markerSize));

        void AddRectangle(SoftJoint softJoint, Color color) =>
            _drawService.Lines.AddRange(Rectangle(softJoint.JointRb.position, color, _markerSize + 0.001f));

        void AddJointConnections(SoftJoint softJoint, Color color) =>
            _drawService.Lines.AddRange(JointConnections(softJoint, color));

        IEnumerable<LineV3> PositionMarker(Vector3 pos, Color color, float size)
        {
            var viewportPos = _camera.WorldToViewportPoint(pos);
            size /= viewportPos.z;
            float sizeX = _camera.aspect < 1 ? size : size / _camera.aspect;
            float sizeY = _camera.aspect < 1 ? size * _camera.aspect : size;

            yield return new LineV3
            {
                From = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x - sizeX, viewportPos.y, viewportPos.z)),
                To = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x + sizeX, viewportPos.y, viewportPos.z)),
                Color = color,
            };
            yield return new LineV3
            {
                From = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x, viewportPos.y - sizeY, viewportPos.z)),
                To = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x, viewportPos.y + sizeY, viewportPos.z)),
                Color = color,
            };
        }

        IEnumerable<LineV3> Rectangle(Vector3 pos, Color color, float size = 0.003f)
        {
            var viewportPos = _camera.WorldToViewportPoint(pos);
            size /= viewportPos.z;
            float sizeX = _camera.aspect < 1 ? size : size / _camera.aspect;
            float sizeY = _camera.aspect < 1 ? size * _camera.aspect : size;

            yield return new LineV3
            {
                From = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x - sizeX, viewportPos.y - sizeY, viewportPos.z)),
                To = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x + sizeX, viewportPos.y - sizeY, viewportPos.z)),
                Color = color,
            };
            yield return new LineV3
            {
                From = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x - sizeX, viewportPos.y + sizeY, viewportPos.z)),
                To = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x + sizeX, viewportPos.y + sizeY, viewportPos.z)),
                Color = color,
            };
            yield return new LineV3
            {
                From = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x - sizeX, viewportPos.y - sizeY, viewportPos.z)),
                To = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x - sizeX, viewportPos.y + sizeY, viewportPos.z)),
                Color = color,
            };
            yield return new LineV3
            {
                From = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x + sizeX, viewportPos.y - sizeY, viewportPos.z)),
                To = _camera.ViewportToWorldPoint(new Vector3(viewportPos.x + sizeX, viewportPos.y + sizeY, viewportPos.z)),
                Color = color,
            };
        }

        static IEnumerable<LineV3> JointConnections(SoftJoint joint, Color color)
        {
            var origin = joint.JointRb.position;
            foreach(var target in joint.NearbyJoints)
            {
                yield return new LineV3 { From = origin, To = (origin + target.JointRb.position) / 2, Color = color };
            }
        }

        static Color JointColor(Color baseColor, float diff, float min, float max) => diff < 0
            ? Color.Lerp(Color.cyan, baseColor, Mathf.InverseLerp(min, 0, diff))
            : Color.Lerp(baseColor, Color.red, Mathf.InverseLerp(0, max, diff));

        static Color JointRelativeColor(Color baseColor, float diff, float offset, float min, float max)
        {
            if(diff < 0)
            {
                min -= offset;
                return min >= 0
                    ? Color.cyan
                    : Color.Lerp(Color.cyan, baseColor, Mathf.InverseLerp(min, 0, diff));

            }

            max += offset;
            return max <= 0
                ? Color.red
                : Color.Lerp(baseColor, Color.red, Mathf.InverseLerp(0, max, diff));
        }

        internal void Destroy()
        {
            Object.Destroy(_drawService);
            _drawService = null;
        }
    }
}
