using NAT.Common.Components;
using NAT.Common.Storables;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using UnityEngine;

namespace NAT.Common.Handlers
{
    sealed class CalibrationStateHandler : MonoBehaviour
    {
        ScriptBase _script;
        Bootstrap _bootstrap;
        Person _person;
        string _domainName;
        internal StorableBool CalibratingJsb { get; private set; }
        MotionAnimationMaster _motionAnimationMaster;
        Action _calibrationAction;
        internal bool CalibrateOnNextPending { get; set; }

        readonly HashSet<string> _pendingReasons = new HashSet<string>();
        internal bool Pending { get; private set; }
        internal bool Cancelling { get; set; }
        internal bool UpdateMassWhenCalibrating { get; set; }

        internal void Init(ScriptBase script, Action calibrationAction)
        {
            _script = script;
            _domainName = _script.DomainName;
            _bootstrap = _script.Bootstrap;
            _person = _script.Person;
            CalibratingJsb = new StorableBool(Constant.INTERNAL_IS_CALIBRATING, false)
            {
                setCallbackFunction = _ => UpdateButtonText(),
            };
            _motionAnimationMaster = SuperController.singleton.motionAnimationMaster;
            _calibrationAction = calibrationAction;
        }

        UIDynamicButton _calibrationButton;

        internal UIDynamicButton CalibrationButton
        {
            set
            {
                _calibrationButton = value;
                UpdateButtonText();
            }
        }

        internal void SetCalibrationPending(string reason, bool value = true)
        {
            if(value)
            {
                _pendingReasons.Add(reason);
            }
            else
            {
                _pendingReasons.Remove(reason);
            }

            Pending = _pendingReasons.Count > 0;
            if(Pending && CalibrateOnNextPending)
            {
                _calibrationAction();
                CalibrateOnNextPending = false;
            }
            else
            {
                UpdateButtonText();
            }

            UpdateMassWhenCalibrating = false;
            foreach(string x in _pendingReasons)
            {
                if(x == Constant.PHYSICS_RATE_CHANGED ||
                    x == Constant.SIZE_CHANGED ||
                    x == Constant.SOFT_PHYSICS_CHANGED ||
                    x == BREAST_JOINT_MASS.NAME ||
                    x == GLUTE_JOINT_MASS.NAME)
                {
                    UpdateMassWhenCalibrating = true;
                    break;
                }
            }
        }

        void ResetCalibrationPending()
        {
            CalibrateOnNextPending = false;
            Pending = false;
            _pendingReasons.Clear();
            UpdateButtonText();
        }

        bool _showCalibrationDone;

        void UpdateButtonText()
        {
            if(!_calibrationButton)
            {
                return;
            }

            if(CalibratingJsb.val)
            {
                _calibrationButton.label = $"Calibrating {_domainName}...";
                _calibrationButton.SetActiveStyle(false, true);
            }
            else if(Pending)
            {
                _calibrationButton.label = $"{_domainName} Calibration Pending".Color(new Color(0.75f, 0.00f, 0.00f));
                _calibrationButton.SetActiveStyle(true);
            }
            else if(_showCalibrationDone)
            {
                _calibrationButton.label = "Calibration Done";
                _calibrationButton.SetActiveStyle(false, true);
            }
            else
            {
                _calibrationButton.label = $"  Calibrate {_domainName}";
                _calibrationButton.SetActiveStyle(gameObject.GetComponent<MVRScript>().IsEnabledNullSafe(), true);
            }
        }

        internal IEnumerator WaitForAtomPhysicsUnfrozen()
        {
            if(_person.Atom.isPhysicsFrozen)
            {
                Loggr.Message($"Calibration will start when physics is unfrozen on atom '{_person.Atom.uid}'.");
            }

            while(_person.Atom.isPhysicsFrozen)
            {
                yield return null;
            }
        }

        bool _queued;
        bool _deferOnGlobalActions;
        bool? _globalFrozen;
        bool? _playing;
        readonly Dictionary<Atom, Dictionary<Rigidbody, RigidbodyConstraints>> _saveRigidbodyConstraints = new Dictionary<Atom, Dictionary<Rigidbody, RigidbodyConstraints>>();
        readonly List<FreeControllerState> _freeControllerStates = new List<FreeControllerState>();

        bool _deferOnLocalActions;

        internal IEnumerator Begin()
        {
            ResetCalibrationPending();

            if(CalibratingJsb.val)
            {
                if(!_queued)
                {
                    _queued = true;
                }
                else
                {
                    Cancelling = true;
                    yield break;
                }
            }

            while(_queued)
            {
                yield return null;
            }

            CalibratingJsb.val = true;

            /* The instance which started calibrating first has control */
            if(IsLocalOtherCalibrating() || IsGlobalOtherCalibrating())
            {
                // Debug.Log($"{_person.Atom.uid}.{_domainName}: deferring to other (global)");
                _deferOnGlobalActions = true;
            }
            else
            {
                // Debug.Log($"{_person.Atom.uid}.{_domainName}: not deferring (global)");
                SetAtomStates();

                if(_bootstrap.FreezeMotionSoundJsb.val && _globalFrozen == null)
                {
                    _globalFrozen = Utils.GlobalAnimationFrozen();
                    if(!_globalFrozen.Value)
                    {
                        SuperController.singleton.SetFreezeAnimation(true);
                    }
                }
                else if(_bootstrap.PauseSceneAnimationJsb.val && _playing == null)
                {
                    _playing = _motionAnimationMaster.activeWhilePlaying.activeSelf;
                    if(_playing.Value)
                    {
                        _motionAnimationMaster.StopPlayback();
                    }
                }
            }

            if(IsLocalOtherCalibrating())
            {
                // Debug.Log($"{_person.Atom.uid}.{_domainName}: deferring to other (local)");
                _deferOnLocalActions = true;
            }
            else
            {
                // Debug.Log($"{_person.Atom.uid}.{_domainName}: not deferring (local)");
                _person.CalibrationState.Apply();
            }
        }

        bool IsLocalOtherCalibrating() => _script.GetOtherPlugin().GetBoolParamValueNullSafe(Constant.INTERNAL_IS_CALIBRATING);

        bool IsGlobalOtherCalibrating()
        {
            var otherPersonAtoms = Utils.GetAtomsByType("Person").Where(atom => atom.uid != _script.containingAtom.uid);
            foreach(var atom in otherPersonAtoms)
            {
                var tittyMagic = FindPluginStorableOnAtom(atom, Constant.TITTY_MAGIC);
                var bootyMagic = FindPluginStorableOnAtom(atom, Constant.BOOTY_MAGIC);
                if(tittyMagic.GetBoolParamValueNullSafe(Constant.INTERNAL_IS_CALIBRATING) || bootyMagic.GetBoolParamValueNullSafe(Constant.INTERNAL_IS_CALIBRATING))
                {
                    return true;
                }
            }

            return false;
        }

        JSONStorable FindPluginStorableOnAtom(Atom atom, string domainName)
        {
            var regex = new Regex($@"^plugin#\d+_{Constant.NATURALIS}.{domainName}$", RegexOptions.Compiled);
            return atom.FindStorablesByRegexMatch(regex).Find(storable => MinVersionStorableValue(storable, "1.0"));
        }

        bool MinVersionStorableValue(JSONStorable storable, string ver)
        {
            if(storable.IsStringJSONParam(Constant.NATURALIS_VERSION))
            {
                string version = storable.GetStringParamValue(Constant.NATURALIS_VERSION);
                bool match = new SemVer(version, true).IsGte(ver);
                return _script.IsDevelopmentEnv ? match || version == "0.0.0" : match;
            }

            return false;
        }

        void SetAtomStates()
        {
            var atoms = SuperController.singleton.GetAtoms();
            var personAtoms = atoms.Where(atom => atom.type == "Person").ToList();
            var allPersonsLinkableRigidbodies = new HashSet<Rigidbody>();

            _saveRigidbodyConstraints.Clear();
            _freeControllerStates.Clear();

            foreach(var atom in personAtoms)
            {
                FreezeAtomRigidbodies(atom);
                allPersonsLinkableRigidbodies.UnionWith(atom.linkableRigidbodies);
            }

            foreach(var atom in atoms)
            {
                UnlinkAtomControllers(atom, allPersonsLinkableRigidbodies);
            }
        }

        void FreezeAtomRigidbodies(Atom atom)
        {
            try
            {
                var freezableRigidbodies = atom.rigidbodies.Where(rigidbody =>
                    !rigidbody.name.StartsWith("AutoCollider") &&
                    rigidbody.name != "LGlute" &&
                    rigidbody.name != "RGlute" &&
                    rigidbody.name != "lPectoral" &&
                    rigidbody.name != "rPectoral" &&
                    rigidbody.name != "lNipple" &&
                    rigidbody.name != "rNipple"
                );

                var atomRigidbodyConstraints = new Dictionary<Rigidbody, RigidbodyConstraints>();

                foreach(var rigidbody in freezableRigidbodies)
                {
                    atomRigidbodyConstraints[rigidbody] = rigidbody.constraints;
                    rigidbody.constraints = RigidbodyConstraints.FreezeAll;
                }

                _saveRigidbodyConstraints[atom] = atomRigidbodyConstraints;
            }
            catch(Exception e)
            {
                Loggr.Error($"Failed to freeze rigidbodies: {e}");
            }
        }

        void UnlinkAtomControllers(Atom atom, HashSet<Rigidbody> allPersonsLinkableRigidbodies)
        {
            try
            {
                foreach(var controller in atom.freeControllers)
                {
                    if(allPersonsLinkableRigidbodies.Contains(controller.linkToRB) && FreeControllerState.IsControllerLinked(controller))
                    {
                        var state = new FreeControllerState(controller);
                        state.SetStateOn();
                        _freeControllerStates.Add(state);
                    }
                }
            }
            catch(Exception e)
            {
                Loggr.Error($"Failed to unlink controllers: {e}");
            }
        }

        internal IEnumerator Finish()
        {
            if(!CalibratingJsb.val)
            {
                yield break;
            }

            if(!_deferOnLocalActions && !_queued)
            {
                while(IsLocalOtherCalibrating())
                {
                    yield return new WaitForSecondsRealtime(0.1f);
                }

                _person.CalibrationState.Restore();
                // Debug.Log($"{_person.Atom.uid}.{_domainName}: finishing up (local)");
            }

            if(!_deferOnGlobalActions && !_queued)
            {
                while(IsLocalOtherCalibrating() || IsGlobalOtherCalibrating())
                {
                    yield return new WaitForSecondsRealtime(0.1f);
                }

                RestoreAtomControllerStates();
                RestorePersonsRigidbodyConstraints();
                _saveRigidbodyConstraints.Clear();

                if(_globalFrozen.HasValue)
                {
                    if(!_globalFrozen.Value)
                    {
                        SuperController.singleton.SetFreezeAnimation(false);
                    }

                    _globalFrozen = null;
                }
                else if(_playing.HasValue)
                {
                    if(_playing.Value)
                    {
                        _motionAnimationMaster.StartPlayback();
                    }

                    _playing = null;
                }

                // Debug.Log($"{_person.Atom.uid}.{_domainName}: finishing up (global)");
            }

            if(!_queued)
            {
                CalibratingJsb.val = false;
                EnableSettingsMonitor();
                StartCoroutine(UpdateButtonTextCo());
            }

            _deferOnLocalActions = false;
            _deferOnGlobalActions = false;
            _queued = false;
        }

        IEnumerator UpdateButtonTextCo()
        {
            _showCalibrationDone = true;
            UpdateButtonText();
            yield return new WaitForSeconds(1);
            _showCalibrationDone = false;
            UpdateButtonText();
        }

        internal void Finish(bool isOtherEnabled)
        {
            if(!CalibratingJsb.val)
            {
                ResetCalibrationPending();
                return;
            }

            if(!isOtherEnabled)
            {
                _person.CalibrationState.Restore();
                RestoreAtomControllerStates();
                RestorePersonsRigidbodyConstraints();
                _saveRigidbodyConstraints.Clear();

                if(_globalFrozen.HasValue)
                {
                    if(!_globalFrozen.Value)
                    {
                        SuperController.singleton.SetFreezeAnimation(false);
                    }

                    _globalFrozen = null;
                }
                else if(_playing.HasValue)
                {
                    if(_playing.Value)
                    {
                        _motionAnimationMaster.StartPlayback();
                    }

                    _playing = null;
                }
            }

            CalibratingJsb.val = false;
            _deferOnLocalActions = false;
            _deferOnGlobalActions = false;
            _queued = false;
        }

        void RestoreAtomControllerStates()
        {
            foreach(var state in _freeControllerStates)
            {
                state.Restore();
            }

            _freeControllerStates.Clear();
        }

        void RestorePersonsRigidbodyConstraints()
        {
            try
            {
                foreach(var atom in SuperController.singleton.GetAtoms())
                {
                    if(atom.type != "Person")
                    {
                        continue;
                    }

                    Dictionary<Rigidbody, RigidbodyConstraints> atomRigidbodyConstraints;
                    if(_saveRigidbodyConstraints.TryGetValue(atom, out atomRigidbodyConstraints))
                    {
                        foreach(var pair in atomRigidbodyConstraints)
                        {
                            var rigidbody = pair.Key;
                            if(rigidbody)
                            {
                                rigidbody.constraints = pair.Value;
                            }
                        }
                    }

                    foreach(var rbAttrs in atom.GetComponentsInChildren<RigidbodyAttributes>())
                    {
                        rbAttrs.SyncTensor();
                    }
                }
            }
            catch(Exception e)
            {
                Loggr.Error($"Failed to reset rigidbody constraints: {e}");
            }
        }

        internal void EnableSettingsMonitor() => _bootstrap.PersonMonitor.enabled = !IsLocalOtherCalibrating();
    }
}
