using MVR.FileManagementSecure;
using NAT.Common.Models;
using SimpleJSON;
using System;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Handlers
{
    sealed class ClothingPresetHandler : PresetHandler
    {
        readonly string _subDir;
        readonly ClothingHandler _clothingHandler;
        internal bool IsSettingTarget { get; set; }

        internal ClothingPresetHandler(ScriptBase script, ClothingHandler clothingHandler, string subDir) : base(script, subDir)
        {
            _subDir = subDir;
            _clothingHandler = clothingHandler;
            var loadPresetWithPathAction = NewLoadWithPathAction("LoadClothingProfileWithPath");
            loadPresetWithPathAction.actionCallback = path => LoadByTriggerAction(path, false);
            script.PreRegisterClothingParam(loadPresetWithPathAction);
            var loadTargetProfileWithPathAction = NewLoadWithPathAction("LoadClothingTargetProfileWithPath");
            loadTargetProfileWithPathAction.actionCallback = path => LoadByTriggerAction(path, true);
            script.PreRegisterClothingParam(loadTargetProfileWithPathAction);
        }

        internal void UpdatePresetDir(string itemSubDir)
        {
            presetDir = $"{FileUtils.savesDir}/{_subDir}/{itemSubDir}";
            lastDir = presetDir;
            SavePresetJsu.suggestedPath = lastDir;
            LoadPresetJsu.suggestedPath = lastDir;
        }

        protected override List<ShortCut> GetShortCutsForDirectory()
        {
            var shortCuts = base.GetShortCutsForDirectory();
            if(string.IsNullOrEmpty(script.Bootstrap.PackageId) || string.IsNullOrEmpty(script.Bootstrap.PackageFullPath))
            {
                return shortCuts;
            }

            string displayName = $"{FileUtils.savesDir}/{_clothingHandler.ClothingProfilesDir}/MeshedVR";
            string path = $"{script.Bootstrap.PackageFullPath}:/{displayName}";
            shortCuts.Add(new ShortCut
            {
                package = script.Bootstrap.PackageId,
                displayName = displayName,
                path = path,
            });

            return shortCuts;
        }

        protected override void LoadAction(string path)
        {
            // Debug.Log($"{script.DomainName}: PresetHandler: LoadAction");
            var json = script.LoadJSON(path);
            if(json != null)
            {
                _clothingHandler.LoadSelectedItemProfile(path, json.AsObject, IsSettingTarget);
                IsSettingTarget = false;
            }
        }

        void LoadByTriggerAction(string path, bool setTarget)
        {
            if(string.IsNullOrEmpty(path))
            {
                return;
            }

            OnPathSelected(ref path);
            LoadEnabledItemProfileFromPath(path, setTarget);
            LoadPresetJsu.val = ""; // ensure empty value when next browse cancelled
        }

        internal void LoadEnabledItemProfileFromPath(string path, bool setTarget = false)
        {
            // Debug.Log($"{script.DomainName}: PresetHandler: LoadEnabledProfileFromPath");
            var json = script.LoadJSON(path);
            if(json != null)
            {
                _clothingHandler.LoadEnabledItemProfileByPath(path, json.AsObject, setTarget);
            }
        }

        protected override void SaveAction(string path)
        {
            var jc = _clothingHandler.GetSelectedProfileJSON();
            script.SaveJSON(jc, path);
            SuperController.singleton.DoSaveScreenshot(path);
            _clothingHandler.SetSelectedProfileLoadedInfo(path);
        }

        internal void SaveDefaultPreset()
        {
            FileUtils.EnsureDirExists(presetDir);
            string path = $"{presetDir}/Default.json";
            var jc = _clothingHandler.GetSelectedProfileJSON();
            script.SaveJSON(jc, path);
            _clothingHandler.SetSelectedProfileLoadedInfo(path);
        }

        internal void LoadDefaultPreset(ClothingItem item, bool autoLoading = false)
        {
            // Debug.Log($"{script.DomainName}: PresetHandler: LoadDefaultPreset");
            string uid = item.DazItem.uid;
            string path = $"{presetDir}/Default.json";

            JSONNode json = null;
            if(FileUtils.FileExists(path))
            {
                json = script.LoadJSON(path);
            }
            else if(item.IsBuiltIn)
            {
                // path = $"{script.Bootstrap.PackageId}:/{path}"; // Used just for displaying the path in the UI
                string builtInPresetDir = $"{script.Bootstrap.PackageId}:/{FileUtils.savesDir}/{_clothingHandler.ClothingProfilesDir}/{item.DazItem.creatorName}";
                path = $"{builtInPresetDir}/{item.DazItem.displayName}.json";
                if(FileUtils.FileExists(path))
                {
                    json = script.LoadJSON(path);
                }
            }
            else
            {
                int packageUidSeparatorIdx = uid.IndexOf(":/", StringComparison.Ordinal);
                if(packageUidSeparatorIdx != -1)
                {
                    string packageUid = uid.Substring(0, packageUidSeparatorIdx);
                    path = $"{packageUid}:/{path}";
                }

                if(FileUtils.FileExists(path))
                {
                    json = script.LoadJSON(path);
                }
            }

            if(autoLoading)
            {
                if(json != null)
                {
                    // Debug.Log($"Auto loading default preset for {uid} from {path}");
                    _clothingHandler.LoadItemByUidProfile(uid, path, json.AsObject);
                }
            }
            else
            {
                if(json != null)
                {
                    _clothingHandler.LoadSelectedItemProfile(path, json.AsObject);
                }
                else
                {
                    script.Loggr.LogMessage("Default preset not saved yet.");
                }
            }
        }
    }
}
