using NAT.Common.Models;
using NAT.Common.Storables;
using Naturalis.Common.Handlers.Configs;
using SimpleJSON;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Handlers.Configs
{
    class ClothingAdjustmentParams
    {
        internal Dictionary<string, AdjustmentFloatParam> AllParams { get; } = new Dictionary<string, AdjustmentFloatParam>();

        internal AdjustmentFloatParam YForce { get; }
        internal AdjustmentFloatParam XForce { get; }
        internal AdjustmentFloatParam ZForce { get; }
        internal AdjustmentFloatParam YXHoldPct { get; }
        internal AdjustmentFloatParam ZHoldPct { get; }
        internal AdjustmentFloatParam YXDampeningPct { get; }
        internal AdjustmentFloatParam ZDampeningPct { get; }
        internal AdjustmentFloatParam YMorphsOffset { get; }
        internal AdjustmentFloatParam XMorphsOffset { get; }
        internal AdjustmentFloatParam ZMorphsOffset { get; }

        internal Dictionary<string, AdjustmentFloatParam> SoftPhysicsParams { get; } = new Dictionary<string, AdjustmentFloatParam>();
        internal Dictionary<string, AdjustmentFloatParam> MorphParams { get; } = new Dictionary<string, AdjustmentFloatParam>();

        internal ClothingAdjustmentParams(
            JSONStorableFloat yForceJsf,
            JSONStorableFloat xForceJsf,
            JSONStorableFloat zForceJsf,
            JSONStorableFloat yxHoldPctJsf,
            JSONStorableFloat zHoldPctJsf,
            JSONStorableFloat yxDampeningPctJsf,
            JSONStorableFloat zDampeningPctJsf,
            JSONStorableFloat yMorphsOffsetJsf,
            JSONStorableFloat xMorphsOffsetJsf,
            JSONStorableFloat zMorphsOffsetJsf,
            Dictionary<string, JSONStorableFloat> softPhysicsJsfs,
            Dictionary<string, ClothingMorphConfig> morphConfigs,
            string prefix,
            string suffix
        )
        {
            YForce = new AdjustmentFloatParam(StorableFloat.Copy(yForceJsf, prefix, suffix, true));
            XForce = new AdjustmentFloatParam(StorableFloat.Copy(xForceJsf, prefix, suffix, true));
            ZForce = new AdjustmentFloatParam(StorableFloat.Copy(zForceJsf, prefix, suffix, true));
            YXHoldPct = new AdjustmentFloatParam(StorableFloat.Copy(yxHoldPctJsf, prefix, suffix, true));
            ZHoldPct = new AdjustmentFloatParam(StorableFloat.Copy(zHoldPctJsf, prefix, suffix, true));
            YXDampeningPct = new AdjustmentFloatParam(StorableFloat.Copy(yxDampeningPctJsf, prefix, suffix, true));
            ZDampeningPct = new AdjustmentFloatParam(StorableFloat.Copy(zDampeningPctJsf, prefix, suffix, true));
            YMorphsOffset = new AdjustmentFloatParam(StorableFloat.Copy(yMorphsOffsetJsf, prefix, suffix, true));
            XMorphsOffset = new AdjustmentFloatParam(StorableFloat.Copy(xMorphsOffsetJsf, prefix, suffix, true));
            ZMorphsOffset = new AdjustmentFloatParam(StorableFloat.Copy(zMorphsOffsetJsf, prefix, suffix, true));

            AllParams.Add(yForceJsf.name, YForce);
            AllParams.Add(xForceJsf.name, XForce);
            AllParams.Add(zForceJsf.name, ZForce);
            AllParams.Add(yxHoldPctJsf.name, YXHoldPct);
            AllParams.Add(zHoldPctJsf.name, ZHoldPct);
            AllParams.Add(yxDampeningPctJsf.name, YXDampeningPct);
            AllParams.Add(zDampeningPctJsf.name, ZDampeningPct);
            AllParams.Add(yMorphsOffsetJsf.name, YMorphsOffset);
            AllParams.Add(xMorphsOffsetJsf.name, XMorphsOffset);
            AllParams.Add(zMorphsOffsetJsf.name, ZMorphsOffset);

            if(softPhysicsJsfs != null)
            {
                foreach(var pair in softPhysicsJsfs)
                {
                    var adjustmentFloatParam = new AdjustmentFloatParam(StorableFloat.Copy(pair.Value, prefix, suffix, true));
                    SoftPhysicsParams[pair.Key] = adjustmentFloatParam;
                    AllParams.Add(pair.Value.name, adjustmentFloatParam);
                }
            }

            foreach(var pair in morphConfigs)
            {
                MorphParams[pair.Key] = new AdjustmentFloatParam(StorableFloat.Copy(pair.Value.StorableFloat, prefix, suffix, true));
            }
        }

        internal void SetAllAdjustmentValues(
            float upDownMassMultiplier,
            float sidewaysMassMultiplier,
            float inOutMassMultiplier,
            bool isBootyMagic
        )
        {
            SetYForce(YForce.StorableParam.val, upDownMassMultiplier);
            SetXForce(XForce.StorableParam.val, sidewaysMassMultiplier);
            SetZForce(ZForce.StorableParam.val, inOutMassMultiplier, !isBootyMagic);
            SetPercentageValue(YXHoldPct);
            SetPercentageValue(ZHoldPct);
            SetPercentageValue(YXDampeningPct);
            SetPercentageValue(ZDampeningPct);
            SetYMorphsOffset(YMorphsOffset.StorableParam.val, isBootyMagic);
            SetXMorphsOffset(XMorphsOffset.StorableParam.val, isBootyMagic);
            SetZMorphsOffset(ZMorphsOffset.StorableParam.val, isBootyMagic);
            foreach(var pair in SoftPhysicsParams)
            {
                SetPercentageValue(pair.Value);
            }
        }

        internal void SetYForce(float value, float multiplier) => YForce.AdjustedValue = CalculateForceValue(value, multiplier);
        internal void SetXForce(float value, float multiplier) => XForce.AdjustedValue = CalculateForceValue(value, multiplier);
        internal void SetZForce(float value, float multiplier, bool invert) => ZForce.AdjustedValue = CalculateForceValue(value, multiplier, invert);

        static float CalculateForceValue(float value, float multiplier, bool invert = false)
        {
            float force = value * Constant.G * multiplier;
            if(invert)
            {
                force = -force;
            }

            return force;
        }

        internal static void SetPercentageValue(AdjustmentFloatParam param) => param.AdjustedValue = Calc.PercentageValue(param.StorableParam.val);
        internal void SetYMorphsOffset(float value, bool invert) => YMorphsOffset.AdjustedValue = CalculateMorphsOffsetValue(value, invert);
        internal void SetXMorphsOffset(float value, bool invert) => XMorphsOffset.AdjustedValue = CalculateMorphsOffsetValue(value, invert);
        internal void SetZMorphsOffset(float value, bool invert) => ZMorphsOffset.AdjustedValue = CalculateMorphsOffsetValue(value, invert, 0.001f);
        static float CalculateMorphsOffsetValue(float value, bool invert, float multiplier = 1) => multiplier * (invert ? -value : value);

        internal void CopyFrom(ClothingAdjustmentParams profile)
        {
            foreach(var pair in AllParams)
            {
                pair.Value.CopyFrom(profile.AllParams[pair.Key]);
            }

            foreach(var pair in profile.MorphParams)
            {
                MorphParams[pair.Key].SetVal(pair.Value.Value);
            }
        }

        internal void RestoreFromJSON(JSONClass jc)
        {
            foreach(var pair in AllParams)
            {
                pair.Value.SetValFromJSON(jc);
            }

            foreach(var pair in MorphParams)
            {
                var param = pair.Value;
                param.SetValNoCallback(jc[param.Name].AsFloat);
            }
        }

        internal void Reset()
        {
            foreach(var pair in AllParams)
            {
                pair.Value.SetDefaultValNoCallback();
            }

            foreach(var pair in MorphParams)
            {
                pair.Value.SetDefaultValNoCallback();
            }
        }

        internal bool IsSaved()
        {
            foreach(var pair in AllParams)
            {
                if(!pair.Value.IsSaved)
                {
                    return false;
                }
            }

            foreach(var pair in MorphParams)
            {
                if(!pair.Value.IsSaved)
                {
                    return false;
                }
            }

            return true;
        }

        internal void SaveValues()
        {
            foreach(var pair in AllParams)
            {
                pair.Value.SaveValue();
            }

            foreach(var pair in MorphParams)
            {
                pair.Value.SaveValue();
            }
        }
    }
}
