using NAT.Common.Storables;
using NAT.Common.Models;
using Naturalis.Common.Handlers.Configs;
using SimpleJSON;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.Common.Handlers.Configs
{
    sealed class ClothingAdjustmentProfile
    {
        readonly ClothingItem _item;
        internal string ParamNamePrefix { get; private set; }
        internal string LoadedName { get; set; } = "None";
        internal string TargetName { get; set; } = "";
        internal string LoadedFullPath { get; set; }
        internal AdjustmentStringParam TargetFullPath { get; }
        internal AdjustmentBoolParam Symmetric { get; }
        internal AdjustmentBoolParam OffWhenUndressing { get; }

        internal Dictionary<string, AdjustmentBoolParam> ActiveStorableBools { get; } = new Dictionary<string, AdjustmentBoolParam>();
        internal AdjustmentBoolParam LeftActive => ActiveStorableBools[Side.LEFT];
        internal AdjustmentBoolParam RightActive => ActiveStorableBools[Side.RIGHT];

        internal Dictionary<string, ClothingAdjustmentParams> Params { get; } = new Dictionary<string, ClothingAdjustmentParams>();
        ClothingAdjustmentParams LeftParams => Params[Side.LEFT];
        ClothingAdjustmentParams RightParams => Params[Side.RIGHT];
        internal ClothingAdjustmentParams RightCachedParams { get; set; }

        internal Dictionary<string, ClothingAdjustmentParams> TargetParams { get; } = new Dictionary<string, ClothingAdjustmentParams>
        {
            { Side.LEFT, null },
            { Side.RIGHT, null },
        };

        ClothingAdjustmentParams LeftTargetParams => TargetParams[Side.LEFT];
        ClothingAdjustmentParams RightTargetParams => TargetParams[Side.RIGHT];
        internal AdjustmentFloatParam Transition { get; }

        internal Dictionary<string, List<SoftJoint>> CoveredSoftJoints { get; } = new Dictionary<string, List<SoftJoint>>
        {
            { Side.LEFT, new List<SoftJoint>() },
            { Side.RIGHT, new List<SoftJoint>() },
        };

        internal Dictionary<string, List<SoftJoint>> ExposedSoftJoints { get; } = new Dictionary<string, List<SoftJoint>>
        {
            { Side.LEFT, new List<SoftJoint>() },
            { Side.RIGHT, new List<SoftJoint>() },
        };

        readonly Dictionary<string, List<SoftJoint>> _savedCoveredSoftJoints = new Dictionary<string, List<SoftJoint>>
        {
            { Side.LEFT, new List<SoftJoint>() },
            { Side.RIGHT, new List<SoftJoint>() },
        };

        internal SoftJoint HoveredOverJoint { get; set; }

        internal bool IsRightSelected { get; set; }
        internal bool IsActive => LeftActive.Value || RightActive.Value;
        internal bool IsCoverageMapped { get; set; }
        internal bool HasTarget => TargetParams[Side.LEFT] != null || TargetParams[Side.RIGHT] != null;
        internal bool IsSymmetricUnset { get; set; }

        internal ClothingAdjustmentProfile(
            ClothingItem item,
            string paramNamePrefix,
            StorableBool symmetricJsb,
            StorableBool leftActiveJsb,
            StorableBool rightActiveJsb,
            StorableFloat transition,
            StorableBool profileOffWhenUndressingJsb
        )
        {
            _item = item;
            ParamNamePrefix = paramNamePrefix;
            Symmetric = new AdjustmentBoolParam(symmetricJsb);
            ActiveStorableBools[Side.LEFT] = new AdjustmentBoolParam(leftActiveJsb);
            ActiveStorableBools[Side.RIGHT] = new AdjustmentBoolParam(rightActiveJsb);
            Transition = new AdjustmentFloatParam(transition);
            OffWhenUndressing = new AdjustmentBoolParam(profileOffWhenUndressingJsb);
            TargetFullPath = new AdjustmentStringParam(new StorableString("Target Full Path", ""));
        }

        internal string GetItemName() => _item.DazItem.displayName;

        internal float GetLeftXForce() => GetParamValue(LeftParams.XForce, LeftTargetParams?.XForce);
        internal float GetRightXForce() => GetParamValue(RightParams.XForce, RightTargetParams?.XForce);
        internal float GetLeftYForce() => GetParamValue(LeftParams.YForce, LeftTargetParams?.YForce);
        internal float GetRightYForce() => GetParamValue(RightParams.YForce, RightTargetParams?.YForce);
        internal float GetLeftZForce() => GetParamValue(LeftParams.ZForce, LeftTargetParams?.ZForce);
        internal float GetRightZForce() => GetParamValue(RightParams.ZForce, RightTargetParams?.ZForce);
        internal float GetLeftYXHoldPct() => GetParamValue(LeftParams.YXHoldPct, LeftTargetParams?.YXHoldPct);
        internal float GetRightYXHoldPct() => GetParamValue(RightParams.YXHoldPct, RightTargetParams?.YXHoldPct);
        internal float GetLeftZHoldPct() => GetParamValue(LeftParams.ZHoldPct, LeftTargetParams?.ZHoldPct);
        internal float GetRightZHoldPct() => GetParamValue(RightParams.ZHoldPct, RightTargetParams?.ZHoldPct);
        internal float GetLeftYXDampeningPct() => GetParamValue(LeftParams.YXDampeningPct, LeftTargetParams?.YXDampeningPct);
        internal float GetRightYXDampeningPct() => GetParamValue(RightParams.YXDampeningPct, RightTargetParams?.YXDampeningPct);
        internal float GetLeftZDampeningPct() => GetParamValue(LeftParams.ZDampeningPct, LeftTargetParams?.ZDampeningPct);
        internal float GetRightZDampeningPct() => GetParamValue(RightParams.ZDampeningPct, RightTargetParams?.ZDampeningPct);
        internal float GetLeftYMorphsOffset() => GetParamValue(LeftParams.YMorphsOffset, LeftTargetParams?.YMorphsOffset);
        internal float GetRightYMorphsOffset() => GetParamValue(RightParams.YMorphsOffset, RightTargetParams?.YMorphsOffset);
        internal float GetLeftXMorphsOffset() => GetParamValue(LeftParams.XMorphsOffset, LeftTargetParams?.XMorphsOffset);
        internal float GetRightXMorphsOffset() => GetParamValue(RightParams.XMorphsOffset, RightTargetParams?.XMorphsOffset);
        internal float GetLeftZMorphsOffset() => GetParamValue(LeftParams.ZMorphsOffset, LeftTargetParams?.ZMorphsOffset);
        internal float GetRightZMorphsOffset() => GetParamValue(RightParams.ZMorphsOffset, RightTargetParams?.ZMorphsOffset);

        float GetParamValue(AdjustmentFloatParam param, AdjustmentFloatParam transitionParam) => transitionParam == null
            ? param.AdjustedValue
            : Mathf.Lerp(param.AdjustedValue, transitionParam.AdjustedValue, Transition.AdjustedValue);

        internal float GetLeftMorphParamValue(string id) => LeftTargetParams != null
            ? Mathf.Lerp(LeftParams.MorphParams[id].Value, LeftTargetParams.MorphParams[id].Value, Transition.AdjustedValue)
            : LeftParams.MorphParams[id].Value;

        internal float GetRightMorphParamValue(string id) => RightTargetParams != null
            ? Mathf.Lerp(RightParams.MorphParams[id].Value, RightTargetParams.MorphParams[id].Value, Transition.AdjustedValue)
            : RightParams.MorphParams[id].Value;

        internal float GetSoftPhysicsParamValue(string side, string paramName)
        {
            float value = Params[side].SoftPhysicsParams[paramName].AdjustedValue;
            return TargetParams[side] != null
                ? Mathf.Lerp(value, TargetParams[side].SoftPhysicsParams[paramName].AdjustedValue, Transition.AdjustedValue)
                : value;
        }

        internal void SetAllAdjustmentValues(
            float upDownMassMultiplier,
            float sidewaysMassMultiplier,
            float inOutMassMultiplier,
            bool isBootyMagic
        )
        {
            SetTransitionValue();
            LeftParams.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
            if(!Symmetric.StorableParam.val)
            {
                RightParams.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
            }

            SetTargetAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
        }

        internal void SetTargetAdjustmentValues(
            float upDownMassMultiplier,
            float sidewaysMassMultiplier,
            float inOutMassMultiplier,
            bool isBootyMagic
        )
        {
            LeftTargetParams?.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
            if(!Symmetric.StorableParam.val)
            {
                RightTargetParams?.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
            }
        }

        internal void SetTransitionValue() => Transition.AdjustedValue = 0.01f * Transition.StorableParam.val;

        internal void SyncSymmetric(bool value)
        {
            if(value)
            {
                RightCachedParams = Params[Side.RIGHT];
                Params[Side.RIGHT] = Params[Side.LEFT];
            }
            else
            {
                Params[Side.RIGHT] = RightCachedParams;
                if(!IsSymmetricUnset)
                {
                    Params[Side.RIGHT].CopyFrom(Params[Side.LEFT]);
                    IsSymmetricUnset = true;
                }
            }
        }

        internal bool HasSoftJoints()
        {
            foreach(var pair in CoveredSoftJoints)
            {
                if(pair.Value.Count > 0)
                {
                    return true;
                }
            }

            return false;
        }

        internal void RestoreFromJSON(JSONClass jc)
        {
            Symmetric.SetValNoCallback(jc[Symmetric.Name].AsBool);
            LeftActive.SetValNoCallback(jc[LeftActive.Name].AsBool);
            Transition.SetValNoCallback(jc[Transition.StorableParam.name].AsFloat);
            LeftParams.RestoreFromJSON(jc);
            if(Symmetric.Value)
            {
                Params[Side.RIGHT] = LeftParams;
                RightActive.SetValNoCallback(LeftActive.Value);
                IsSymmetricUnset = false;
            }
            else
            {
                Params[Side.RIGHT] = RightCachedParams;
                RightParams.RestoreFromJSON(jc);
                RightActive.SetValNoCallback(jc[RightActive.Name].AsBool);
                IsSymmetricUnset = true;
            }
        }

        internal void SaveValues()
        {
            Symmetric.SaveValue();
            LeftActive.SaveValue();
            RightActive.SaveValue();
            Transition.SaveValue();
            _savedCoveredSoftJoints[Side.LEFT] = CoveredSoftJoints[Side.LEFT].ToList();
            _savedCoveredSoftJoints[Side.RIGHT] = CoveredSoftJoints[Side.RIGHT].ToList();
            LeftParams.SaveValues();
            TargetFullPath.SaveValue();
            if(!Symmetric.Value)
            {
                RightParams.SaveValues();
            }
        }

        internal void Reset(bool isFemale)
        {
            Symmetric.SetDefaultValNoCallback();
            LeftActive.SetDefaultValNoCallback();
            RightActive.SetDefaultValNoCallback();
            Transition.SetDefaultValNoCallback();
            LeftParams.Reset();
            if(Symmetric.Value)
            {
                Params[Side.RIGHT] = LeftParams;
                RightActive.SetValNoCallback(LeftActive.Value);
                IsSymmetricUnset = false;
            }
            else
            {
                RightParams.Reset();
                IsSymmetricUnset = true;
            }

            CoveredSoftJoints[Side.LEFT].Clear();
            CoveredSoftJoints[Side.RIGHT].Clear();
            ExposedSoftJoints[Side.LEFT].Clear();
            ExposedSoftJoints[Side.RIGHT].Clear();
            if(isFemale)
            {
                IsCoverageMapped = false;
            }

            LoadedName = "None";
            ResetTarget();
            SaveValues();
        }

        internal void ResetTarget()
        {
            TargetParams[Side.LEFT] = null;
            TargetParams[Side.RIGHT] = null;
            TargetName = "";
            TargetFullPath.SetDefaultValNoCallback();
        }

        internal void SetProfileOffAsTarget(ClothingAdjustmentParams offParams)
        {
            TargetParams[Side.LEFT] = offParams;
            TargetParams[Side.RIGHT] = offParams;
            TargetName = "Profile Off";
            TargetFullPath.SetVal("Profile Off");
        }

        internal void SetValuesFromCurrentTransitionState()
        {
            if(LeftTargetParams != RightTargetParams)
            {
                Symmetric.SetVal(false);
            }

            SetParamsOnSideFromCurrentTransitionState(Side.LEFT);
            if(!Symmetric.Value)
            {
                SetParamsOnSideFromCurrentTransitionState(Side.RIGHT);
            }

            Transition.StorableParam.valNoCallback = 0f;
        }

        void SetParamsOnSideFromCurrentTransitionState(string side)
        {
            var targetParams = TargetParams[side];
            foreach(var pair in Params[side].AllParams)
            {
                var param = pair.Value;
                var targetParam = targetParams.AllParams[pair.Key];
                param.SetValNoCallback(Mathf.Lerp(param.Value, targetParam.Value, Transition.AdjustedValue));
            }

            foreach(var pair in Params[side].MorphParams)
            {
                var param = pair.Value;
                var targetParam = targetParams.MorphParams[pair.Key];
                param.SetValNoCallback(Mathf.Lerp(param.Value, targetParam.Value, Transition.AdjustedValue));
            }
        }

        internal bool IsSaved =>
            Symmetric.IsSaved &&
            LeftActive.IsSaved &&
            RightActive.IsSaved &&
            CoveredJointsIsSaved(Side.LEFT) &&
            CoveredJointsIsSaved(Side.RIGHT) &&
            Transition.IsSaved &&
            TargetFullPath.IsSaved &&
            (Symmetric.Value ? LeftParams.IsSaved() : LeftParams.IsSaved() && RightParams.IsSaved());

        bool CoveredJointsIsSaved(string side) => CoveredSoftJoints[side]
            .OrderBy(joint => joint.SoftVerticesSet.targetVertex)
            .SequenceEqual(_savedCoveredSoftJoints[side].OrderBy(joint => joint.SoftVerticesSet.targetVertex));
    }
}
