using NAT.Common.Storables;
using UnityEngine;

namespace NAT.Common.Handlers.Configs
{
    sealed class LocalDistortionPhysicsConfig
    {
        float _compressionMax;
        float _compressionMultiplier;
        float _compressionSlope;

        internal StorableFloat CompressionEffectJsf { get; set; }
        internal StorableFloat CompressionSensitivityJsf { get; set; }

        float _stretchingMax;
        float _stretchingMultiplier;
        float _stretchingSlope;

        internal StorableFloat StretchingEffectJsf { get; set; }
        internal StorableFloat StretchingSensitivityJsf { get; set; }

        internal void SetCompressionMaxAndMultiplier(float value)
        {
            _compressionMax = CalculateMax(value);
            _compressionMultiplier = CalculateMultiplier(value);
        }

        internal void SetStretchingMaxAndMultiplier(float value)
        {
            _stretchingMax = CalculateMax(value);
            _stretchingMultiplier = CalculateMultiplier(value);
        }

        internal void SetCompressionSlope(float value) =>
            _compressionSlope = CalculateSlope(value);

        internal void SetStretchingSlope(float value) =>
            _stretchingSlope = CalculateSlope(value);

        static float CalculateMax(float value) => value > 0
            ? Calc.RoundToDecimals(value / 2, 100)
            : Calc.RoundToDecimals(-value / 2, 100);

        static float CalculateMultiplier(float value) => value > 0
            ? Calc.RoundToDecimals(Calc.Exponential1(value, 0.7f, 4.9f, 1.5f), 100)
            : -Calc.RoundToDecimals(Calc.Exponential1(-value, 0.7f, 4.9f, 1.5f), 100);

        static float CalculateSlope(float value) => Mathf.Lerp(-0.33f, 0.75f, 1 - value);

        internal float CalculateDistortionOffset(
            float limit,
            float distortionBase,
            float distortion
        )
        {
            if(distortion < 0) // compression
            {
                if(_compressionMultiplier == 0)
                {
                    return 0;
                }

                float value = Calc.InverseSmoothStep(
                    -distortion / (_compressionMax * distortionBase),
                    Calc.ClampSmoothStepCurvature(_compressionSlope),
                    0.5f
                );

                return _compressionMultiplier < 0
                    ? -Mathf.Lerp(0, limit, -_compressionMultiplier * value)
                    : Mathf.Lerp(0, limit, _compressionMultiplier * value);
            }
            else // stretching
            {
                if(_stretchingMultiplier == 0)
                {
                    return 0;
                }

                float value = Calc.InverseSmoothStep(
                    distortion / (_stretchingMax * distortionBase),
                    Calc.ClampSmoothStepCurvature(_stretchingSlope),
                    0.5f
                );

                return _stretchingMultiplier < 0
                    ? -Mathf.Lerp(0, limit, -_stretchingMultiplier * value)
                    : Mathf.Lerp(0, limit, _stretchingMultiplier * value);
            }
        }

        internal bool IsModified =>
            CompressionEffectJsf.IsModified ||
            CompressionSensitivityJsf.IsModified ||
            StretchingEffectJsf.IsModified ||
            StretchingSensitivityJsf.IsModified;
    }
}
