using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;

namespace NAT.Common.Handlers
{
    abstract class FrictionHandler : Handler
    {
        internal StorableFloat FrictionOffsetJsf { get; private set; }
        internal JSONStorableFloat FrictionValueJsf { get; private set; }
        internal StorableBool AdaptiveFrictionJsb { get; private set; }
        internal StorableFloat DrySkinFrictionJsf { get; private set; }
        StorableAction _calculateFrictionAction;
        internal float MaxHardColliderFriction { get; set; }

        JSONStorable _skinMaterialsStorable;
        protected JSONStorableFloat glossJsf;
        protected JSONStorableFloat specularBumpinessJsf;

        internal bool IsEnabled { get; private set; }

        protected void SetupStorables()
        {
            AdaptiveFrictionJsb = new StorableBool("Use Adaptive Friction", false)
            {
                setCallbackFunction = value =>
                {
                    Script.UIModManager.DrySkinFrictionSlider.SetActiveStyle(value, true);
                    CalculateFriction();
                },
            };
            FrictionOffsetJsf = new StorableFloat("Friction Offset", 0, -0.5f, 0.5f)
            {
                setCallbackFunction = _ => CalculateFriction(),
            };
            FrictionValueJsf = new JSONStorableFloat("Friction Value", 1.00f, 0.00f, 1.00f, false)
            {
                setCallbackFunction = UpdateSoftColliders,
            };
            DrySkinFrictionJsf = new StorableFloat("Dry Skin Friction", 1.000f, 0.000f, 1.000f)
            {
                setCallbackFunction = _ => CalculateFriction(),
            };
            _calculateFrictionAction = new StorableAction(Constant.INTERNAL_CALCULATE_FRICTION, CalculateFriction);

            _skinMaterialsStorable = Script.containingAtom.GetStorableByID("skin");
            if(_skinMaterialsStorable)
            {
                SetGlossAndSpecularStorables();
                IsEnabled = true;
            }
            else
            {
                IsEnabled = false;
            }
        }

        internal void Refresh()
        {
            _skinMaterialsStorable = Script.containingAtom.GetStorableByID("skin");
            IsEnabled = _skinMaterialsStorable;
            if(IsEnabled)
            {
                SetGlossAndSpecularStorables();
                LoadSettings();
                CalculateFriction();
            }
        }

        protected static PhysicsCalculator FatColliderRadiusCalculator() => new PhysicsCalculator
        {
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0,
                MultiplierA = 0.0090f,
                MultiplierB = 0,
                CurveA = x => 1.75f * Calc.InverseSmoothStep(x, 0.4f, 0.7f),
            },
        };

        protected abstract void UpdateSoftColliders(float friction);

        void SetGlossAndSpecularStorables()
        {
            glossJsf = _skinMaterialsStorable.GetFloatJSONParam("Gloss");
            glossJsf.setJSONCallbackFunction = _ =>
            {
                Script.GetOtherPlugin().CallActionNullSafe(Constant.INTERNAL_CALCULATE_FRICTION);
                CalculateFriction();
            };

            specularBumpinessJsf = _skinMaterialsStorable.GetFloatJSONParam("Specular Bumpiness");
            specularBumpinessJsf.setJSONCallbackFunction = _ =>
            {
                Script.GetOtherPlugin().CallActionNullSafe(Constant.INTERNAL_CALCULATE_FRICTION);
                CalculateFriction();
            };
        }

        internal override void PreRegisterParams()
        {
            Script.PreRegisterFrictionParam(AdaptiveFrictionJsb);
            Script.PreRegisterFrictionParam(FrictionOffsetJsf);
            Script.PreRegisterFrictionParam(DrySkinFrictionJsf);
            Script.PreRegisterInternalParam(_calculateFrictionAction);
        }

        internal abstract void CalculateFriction();

        internal void Destroy()
        {
            if(glossJsf != null)
            {
                glossJsf.setJSONCallbackFunction = null;
            }

            if(specularBumpinessJsf != null)
            {
                specularBumpinessJsf.setJSONCallbackFunction = null;
            }
        }
    }
}
