using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;
using NAT.Store;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.Common.Handlers
{
    abstract class GravityPhysicsHandler : Handler
    {
        protected AngleCalculator AngleCalculator { get; private set; }
        protected PhysicsParamGroup JointMass { get; set; }
        ClothingStore _clothingStore;
        HairStore _hairStore;

        Rigidbody _leftRb, _rightRb;
        Rigidbody[] _gravityRigidbodies;

        internal StorableBool LinkDirectionsJsb { get; private set; }
        internal StorableFloat UpGravityJsf { get; private set; }
        internal StorableFloat DownGravityJsf { get; private set; }
        internal StorableFloat BackGravityJsf { get; private set; }
        internal StorableFloat ForwardGravityJsf { get; private set; }
        internal StorableFloat SidewaysGravityJsf { get; private set; }
        internal StorableFloat[] UserMultipliers { get; private set; }

        float _transition;
        Vector3 _leftPrevSimulationForce, _rightPrevSimulationForce = Vector3.zero;
        Vector3 _prevDefaultGravityForce = Vector3.zero;

        protected float upGravityMultiplier,
            downGravityMultiplier,
            forwardGravityMultiplier,
            backGravityMultiplier,
            sidewaysOutGravityMultiplier,
            sidewaysInGravityMultiplier;

        float _upGravityLeft,
            _upGravityRight,
            _downGravityLeft,
            _downGravityRight,
            _backGravityLeft,
            _backGravityRight,
            _forwardGravityLeft,
            _forwardGravityRight,
            _sidewaysInGravityLeft,
            _sidewaysInGravityRight,
            _sidewaysOutGravityLeft,
            _sidewaysOutGravityRight;

        internal float ExtraDownGravityLeft { get; private set; }
        internal float ExtraDownGravityRight { get; private set; }

        internal Vector3 RefDownGravityLeft { get; private set; }
        internal Vector3 RefDownGravityRight { get; private set; }
        internal Vector3 RefUpGravityLeft { get; private set; }
        internal Vector3 RefUpGravityRight { get; private set; }
        internal Vector3 RefLeftGravityLeft { get; private set; }
        internal Vector3 RefLeftGravityRight { get; private set; }
        internal Vector3 RefRightGravityLeft { get; private set; }
        internal Vector3 RefRightGravityRight { get; private set; }
        internal Vector3 RefBackGravityLeft { get; private set; }
        internal Vector3 RefBackGravityRight { get; private set; }
        internal Vector3 RefForwardGravityLeft { get; private set; }
        internal Vector3 RefForwardGravityRight { get; private set; }

        internal Vector3 CustomGravityLeft { get; private set; }
        internal Vector3 CustomGravityRight { get; private set; }

        internal Dictionary<string, StorableAction> SimulateGravityActions { get; private set; }
        internal StorableAction StopGravitySimulationAction { get; private set; }
        internal bool IsSimulatingGravity { get; private set; }
        internal string GravitySimulationDirection { get; private set; }

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            AngleCalculator = provider.GetInstance<AngleCalculator>();
            _clothingStore = Script.Bootstrap.ClothingStore;
            _hairStore = Script.Bootstrap.HairStore;

            if(Script.IsBootyMagic)
            {
                _leftRb = Person.LeftGlute.Rigidbody;
                _rightRb = Person.RightGlute.Rigidbody;

            }
            else if(Script.IsTittyMagic)
            {
                _leftRb = Person.LeftBreast.Rigidbody;
                _rightRb = Person.RightBreast.Rigidbody;
            }

            _gravityRigidbodies = Person.Atom.realRigidbodies
                .Where(rigidbody => rigidbody.useGravity)
                .Where(rigidbody => !rigidbody.name.EndsWith("Link"))
                .Where(rigidbody => !rigidbody.isKinematic)
                .Where(rigidbody => rigidbody.gameObject.activeSelf)
                .Where(rigidbody => rigidbody != Person.LeftGlute?.Rigidbody && rigidbody != Person.RightGlute?.Rigidbody)
                .Where(rigidbody => rigidbody != Person.LeftBreast.Rigidbody && rigidbody != Person.RightBreast.Rigidbody)
                .ToArray();

            float max = Script.IsBootyMagic ? 3.00f : 2.00f;
            LinkDirectionsJsb = new StorableBool("Link Directions", true)
            {
                setCallbackFunction = OnToggleLinkMultipliers,
            };
            UpGravityJsf = new StorableFloat("Up Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };
            DownGravityJsf = new StorableFloat("Down Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };
            BackGravityJsf = new StorableFloat("Back Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };
            ForwardGravityJsf = new StorableFloat("Forward Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };
            SidewaysGravityJsf = new StorableFloat("Sideways Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };

            UserMultipliers = new[] { UpGravityJsf, DownGravityJsf, BackGravityJsf, ForwardGravityJsf, SidewaysGravityJsf };

            string[] gravityDirections =
            {
                Direction.UP,
                Direction.DOWN,
                Direction.BACK,
                Direction.FORWARD,
                Direction.LEFT,
                Direction.RIGHT,
            };

            SimulateGravityActions = new Dictionary<string, StorableAction>();
            foreach(string direction in gravityDirections)
            {
                SimulateGravityActions[direction] = new StorableAction($"Simulate {direction} Gravity", () => SetGravityDirection(direction));
            }

            StopGravitySimulationAction = new StorableAction("Stop Gravity Simulation", UnsetGravityDirection);
        }

        void OnToggleLinkMultipliers(bool value)
        {
            if(Script.RestoringFromJSON)
            {
                return;
            }

            if(value)
            {
                SyncMultipliersToJsf(UpGravityJsf);
            }
        }

        void OnSetGravityMultiplier(JSONStorableFloat jsf)
        {
            if(LinkDirectionsJsb.val)
            {
                SyncMultipliersToJsf(jsf);
            }

            FinishSetGravityMultiplier(jsf);
            SetGravityValues();
        }

        void SyncMultipliersToJsf(JSONStorableFloat reference)
        {
            float multiplier = reference.val;
            foreach(var jsf in UserMultipliers)
            {
                if(jsf == reference)
                {
                    continue;
                }

                jsf.valNoCallback = multiplier;
                FinishSetGravityMultiplier(jsf);
            }
        }

        void FinishSetGravityMultiplier(JSONStorableFloat jsf)
        {
            var storable = (StorableFloat) jsf;
            Script.SetCalibrationPending(storable);
        }

        internal override void PreRegisterParams()
        {
            foreach(var pair in SimulateGravityActions)
            {
                Script.PreRegisterGeneralParam(pair.Value);
            }

            Script.PreRegisterGeneralParam(StopGravitySimulationAction);
        }

        internal void AddGravitySimulationForce(bool fullBody = true)
        {
            var simulatedGravityLeft = Utils.DirectionVector(_leftRb, GravitySimulationDirection) * Physics.gravity.magnitude;
            var simulatedGravityRight = Utils.DirectionVector(_rightRb, GravitySimulationDirection) * Physics.gravity.magnitude;
            var leftForce = CalculateCustomGravityLeft(simulatedGravityLeft);
            var rightForce = CalculateCustomGravityRight(simulatedGravityRight);
            var defaultGravityForce = (leftForce + rightForce).normalized * Physics.gravity.magnitude;

            if(JointMass.ApplyChangesOnlyToLeftParamJsb.val)
            {
                CustomGravityLeft = leftForce;
                CustomGravityRight = rightForce;
                if(_transition < 1f)
                {
                    _transition += 0.020f / Utils.PhysicsRateMultiplier();
                    CustomGravityLeft = Vector3.Lerp(_leftPrevSimulationForce, CustomGravityLeft, _transition);
                    CustomGravityRight = Vector3.Lerp(_rightPrevSimulationForce, CustomGravityRight, _transition);
                    defaultGravityForce = Vector3.Lerp(_prevDefaultGravityForce, defaultGravityForce, _transition);
                    _leftPrevSimulationForce = CustomGravityLeft;
                    _rightPrevSimulationForce = CustomGravityRight;
                    _prevDefaultGravityForce = defaultGravityForce;
                }
            }
            else
            {
                CustomGravityLeft = 0.5f * (leftForce + rightForce);
                CustomGravityRight = CustomGravityLeft;
                if(_transition < 1f)
                {
                    _transition += 0.020f / Utils.PhysicsRateMultiplier();
                    CustomGravityLeft = Vector3.Lerp(_leftPrevSimulationForce, CustomGravityLeft, _transition);
                    CustomGravityRight = CustomGravityLeft;
                    defaultGravityForce = Vector3.Lerp(_prevDefaultGravityForce, defaultGravityForce, _transition);
                    _leftPrevSimulationForce = CustomGravityLeft;
                    _rightPrevSimulationForce = CustomGravityRight;
                    _prevDefaultGravityForce = defaultGravityForce;
                }
            }

            _leftRb.AddForce(CustomGravityLeft, ForceMode.Acceleration);
            _rightRb.AddForce(CustomGravityRight, ForceMode.Acceleration);

            if(fullBody)
            {
                foreach(var rigidbody in _gravityRigidbodies)
                {
                    rigidbody.AddForce(defaultGravityForce, ForceMode.Acceleration);
                }

                if(_transition <= 1f)
                {
                    _clothingStore.EnabledItems.ForEach(item => item.SetSimForce(defaultGravityForce));
                    _hairStore.EnabledItems.ForEach(item => item.SetSimForce(defaultGravityForce));
                }
            }
        }

        void SetGravityDirection(string direction)
        {
            if(GravitySimulationDirection == direction || Script.CalibrationStateHandler.CalibratingJsb.val)
            {
                return;
            }

            EnableGravitySimulation(direction);
            Script.SetButtonColors(direction);
        }

        internal void EnableGravitySimulation(string direction)
        {
            GravitySimulationDirection = direction;
            IsSimulatingGravity = true;
            _transition = 0f;
            foreach(var rigidbody in _gravityRigidbodies)
            {
                rigidbody.useGravity = false;
            }
        }

        internal void UnsetGravityDirection()
        {
            if(Script.CalibrationStateHandler.CalibratingJsb.val)
            {
                return;
            }

            StopGravitySimulation();
            Script.ResetButtonColors();
        }

        // TODO store useGravity values, restore original values on stop
        // TODO unnecessary foreaches post-calibration
        internal void StopGravitySimulation()
        {
            foreach(var rigidbody in _gravityRigidbodies)
            {
                rigidbody.useGravity = rigidbody != _leftRb && rigidbody != _rightRb;
            }

            _clothingStore.EnabledItems.ForEach(item => item.UnsetSimForce());
            _hairStore.EnabledItems.ForEach(item => item.UnsetSimForce());

            IsSimulatingGravity = false;
            GravitySimulationDirection = null;
            _leftPrevSimulationForce = Vector3.zero;
        }

        readonly Quaternion _upright = Quaternion.Euler(Vector3.zero);
        readonly Quaternion _upsideDown = Quaternion.Euler(new Vector3(180, 0, 0));
        readonly Quaternion _left = Quaternion.Euler(new Vector3(0, -90, 0));
        readonly Quaternion _right = Quaternion.Euler(new Vector3(0, 90, 0));
        readonly Quaternion _forward = Quaternion.Euler(new Vector3(90, 0, 0));
        readonly Quaternion _backward = Quaternion.Euler(new Vector3(-90, 0, 0));

        internal void SetGravityValues()
        {
            _upGravityLeft = UpGravityJsf.val * upGravityMultiplier;
            _downGravityLeft = DownGravityJsf.val * downGravityMultiplier;
            _forwardGravityLeft = ForwardGravityJsf.val * forwardGravityMultiplier;
            _backGravityLeft = BackGravityJsf.val * backGravityMultiplier;
            _sidewaysInGravityLeft = SidewaysGravityJsf.val * sidewaysInGravityMultiplier;
            _sidewaysOutGravityLeft = SidewaysGravityJsf.val * sidewaysOutGravityMultiplier;

            if(JointMass.ApplyChangesOnlyToLeftParamJsb.val)
            {
                _upGravityRight = upGravityMultiplier;
                _downGravityRight = downGravityMultiplier;
                _forwardGravityRight = forwardGravityMultiplier;
                _backGravityRight = backGravityMultiplier;
                _sidewaysInGravityRight = sidewaysInGravityMultiplier;
                _sidewaysOutGravityRight = sidewaysOutGravityMultiplier;
            }
            else
            {
                _upGravityRight = _upGravityLeft;
                _downGravityRight = _downGravityLeft;
                _forwardGravityRight = _forwardGravityLeft;
                _backGravityRight = _backGravityLeft;
                _sidewaysInGravityRight = _sidewaysInGravityLeft;
                _sidewaysOutGravityRight = _sidewaysOutGravityLeft;
            }

            ExtraDownGravityLeft = Mathf.Clamp(_downGravityLeft - 1f, 0, _downGravityLeft - 1f);
            ExtraDownGravityRight = Mathf.Clamp(_downGravityRight - 1f, 0, _downGravityRight - 1f);
            // TODO callbacks for actual gravity being manipulated at runtime

            RefDownGravityLeft = CalculateReferenceGravity(_upright, false);
            RefDownGravityRight = CalculateReferenceGravity(_upright, true);
            RefUpGravityLeft = CalculateReferenceGravity(_upsideDown, false);
            RefUpGravityRight = CalculateReferenceGravity(_upsideDown, true);
            RefLeftGravityLeft = CalculateReferenceGravity(_left, false);
            RefLeftGravityRight = CalculateReferenceGravity(_left, true);
            RefRightGravityLeft = CalculateReferenceGravity(_right, false);
            RefRightGravityRight = CalculateReferenceGravity(_right, true);
            RefBackGravityLeft = CalculateReferenceGravity(_forward, false);
            RefBackGravityRight = CalculateReferenceGravity(_forward, true);
            RefForwardGravityLeft = CalculateReferenceGravity(_backward, false);
            RefForwardGravityRight = CalculateReferenceGravity(_backward, true);
        }

        protected void ApplyRealGravity()
        {
            var leftForce = CalculateCustomGravityLeft(Physics.gravity);
            var rightForce = CalculateCustomGravityRight(Physics.gravity);
            if(JointMass.ApplyChangesOnlyToLeftParamJsb.val)
            {
                CustomGravityLeft = leftForce;
                CustomGravityRight = rightForce;
            }
            else
            {
                CustomGravityLeft = 0.5f * (leftForce + rightForce);
                CustomGravityRight = CustomGravityLeft;
            }

            _leftRb.AddForce(CustomGravityLeft, ForceMode.Acceleration);
            _rightRb.AddForce(CustomGravityRight, ForceMode.Acceleration);
        }

        Vector3 CalculateCustomGravityLeft(Vector3 gravity)
        {
            var localGravity = _leftRb.transform.InverseTransformDirection(gravity);
            var customGravity = GravityVectorLeft(localGravity);
            return _leftRb.transform.TransformDirection(customGravity);
        }

        Vector3 CalculateCustomGravityRight(Vector3 gravity)
        {
            var localGravity = _rightRb.transform.InverseTransformDirection(gravity);
            var customGravity = GravityVectorRight(localGravity);
            return _rightRb.transform.TransformDirection(customGravity);
        }

        Vector3 CalculateReferenceGravity(Quaternion relativeTo, bool right)
        {
            var localGravity = Quaternion.Inverse(relativeTo) * Physics.gravity;
            var customGravity = right ? GravityVectorRight(localGravity) : GravityVectorLeft(localGravity);
            return relativeTo * customGravity;
        }

        Vector3 GravityVectorLeft(Vector3 localGravity) =>
            new Vector3(
                localGravity.x * (localGravity.x > 0 ? _sidewaysInGravityLeft : _sidewaysOutGravityLeft),
                localGravity.y * (localGravity.y > 0 ? _upGravityLeft : _downGravityLeft),
                localGravity.z * (localGravity.z > 0 ? _forwardGravityLeft : _backGravityLeft)
            );

        Vector3 GravityVectorRight(Vector3 localGravity) =>
            new Vector3(
                localGravity.x * (localGravity.x > 0 ? _sidewaysOutGravityRight : _sidewaysInGravityRight),
                localGravity.y * (localGravity.y > 0 ? _upGravityRight : _downGravityRight),
                localGravity.z * (localGravity.z > 0 ? _forwardGravityRight : _backGravityRight)
            );

        protected static PhysicsCalculator DefaultPhysicsCalculator() => new PhysicsCalculator
        {
            PrimaryConfig = new MultiplierConfig(),
            SecondaryConfig = new MultiplierConfig(),
        };

        protected static StorableFloat NewGravityMultiplierJsf(string paramName, string direction) =>
            new StorableFloat($"{paramName} {direction} Gravity", 1, 0, 3);

        internal void SaveCalibrated()
        {
            foreach(var jsf in UserMultipliers)
            {
                jsf.SaveCalibrated();
            }
        }

        internal bool IsGravityModified =>
            UpGravityJsf.IsModified ||
            DownGravityJsf.IsModified ||
            ForwardGravityJsf.IsModified ||
            BackGravityJsf.IsModified ||
            SidewaysGravityJsf.IsModified;
    }
}
