using ColliderEditor;
using NAT.Common.Models;
using NAT.Common.Storables;
using SimpleJSON;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Handlers
{
    abstract class HardColliderHandler : Handler
    {
        protected Rigidbody relativeToRigidbody;

        internal ColliderVisualizer ColliderVisualizer { get; set; }
        internal JSONStorableStringChooser ColliderGroupsJssc { get; set; }
        internal List<HardColliderGroup> HardColliderGroups { get; set; }
        internal StorableFloat BaseMassMultiplierJsf { get; set; }
        internal JSONStorableBool HighlightAllJsb { get; set; }

        protected float frictionSizeMultiplierLeft;
        protected float frictionSizeMultiplierRight;

        internal void ResetDistanceDiffs()
        {
            foreach(var group in HardColliderGroups)
            {
                group.ResetDistanceDiffs();
            }
        }

        internal void EnableMultiplyFriction()
        {
            foreach(var group in HardColliderGroups)
            {
                group.EnableMultiplyFriction();
            }
        }

        internal abstract void SyncAllOffsets();

        protected void SyncSizeAuto()
        {
            foreach(var group in HardColliderGroups)
            {
                group.AutoColliderSizeSet();
            }

            float averageRadius = 0f;
            foreach(var group in HardColliderGroups)
            {
                averageRadius += group.Left.AutoCollider.colliderRadius;
            }

            averageRadius /= HardColliderGroups.Count;

            foreach(var group in HardColliderGroups)
            {
                group.UpdateMaxFrictionalDistance(
                    frictionSizeMultiplierLeft,
                    frictionSizeMultiplierRight,
                    averageRadius
                );
            }

            ColliderVisualizer.SyncPreviews();
        }

        internal abstract void UpdateFrictionSizeMultipliers();

        internal void UpdateFriction(float maxHardColliderFriction)
        {
            foreach(var group in HardColliderGroups)
            {
                group.UpdateFriction(maxHardColliderFriction);
            }
        }

        internal abstract void UpdateClothingAdjustmentAmounts(float massMultiplierLeft, float massMultiplierRight);

        internal abstract void SyncCollidersMass();
        internal abstract void CalibrateColliders();
        internal abstract void UpdateDistanceDiffs();

        protected Vector3 CollidersCenter(int[] vertexIndices) =>
            Calc.RelativePosition(relativeToRigidbody, Person.SkinVertsAveragePosition(vertexIndices));

        internal JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass();

            foreach(var group in HardColliderGroups)
            {
                var groupJSON = new JSONClass
                {
                    ["left_dist"] = { AsFloat = group.Left.SqrNeutralDistance },
                    ["right_dist"] = { AsFloat = group.Right.SqrNeutralDistance },
                };
                jc[group.GetJSONId()] = groupJSON;
            }

            return jc;
        }

        internal void CalibrateFromJSON(JSONClass jc)
        {
            foreach(var group in HardColliderGroups)
            {
                var groupJSON = jc[group.GetJSONId()].AsObject;
                group.Left.SqrNeutralDistance = groupJSON["left_dist"].AsFloat;
                group.Right.SqrNeutralDistance = groupJSON["right_dist"].AsFloat;
            }
        }

        internal void RestoreOriginalPhysics()
        {
            /* Required for restoring default collider mass. Enabled in case disabled
             * programmatically, and not yet restored back on by PersonMonitor.
             */
            if(Script.IsTittyMagic)
            {
                Person.EnableUseColliders();
            }

            foreach(var group in HardColliderGroups)
            {
                group.RestoreDefaults();
            }
        }

        internal bool RigidbodiesFound()
        {
            foreach(var group in HardColliderGroups)
            {
                if(!group.HasRigidbodies())
                {
                    return false;
                }
            }

            return true;
        }

        internal void Destroy()
        {
            Object.Destroy(ColliderVisualizer);
            ColliderVisualizer = null;
        }
    }
}
