using ColliderEditor;
using NAT.Common.Diagnostics;
using NAT.Common.Models;
using NAT.Common.Storables;
using SimpleJSON;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Handlers
{
    abstract class SoftPhysicsHandler : Handler
    {
        protected AngleCalculator angleCalculator;
        internal ColliderVisualizer ColliderVisualizer { get; set; }
        internal SoftJointDiagnostics SoftJointDiagnostics { get; set; }

        internal List<SoftJoint> AllSoftJointsList { get; } = new List<SoftJoint>();
        internal Dictionary<string, SoftJoint[]> AllSoftJoints { get; set; } = new Dictionary<string, SoftJoint[]>();
        internal Dictionary<string, SoftJoint[]> OuterSoftJoints { get; set; }
        internal Dictionary<string, SoftJoint[]> MainSoftJoints { get; set; }
        internal Dictionary<string, SoftJoint[]> CenterSoftJoints { get; set; }
        internal Dictionary<string, Rigidbody[]> CenterSoftJointsRigidbodies { get; } = new Dictionary<string, Rigidbody[]>();

        internal SoftPhysicsParamGroup[] SoftParamGroups { get; set; }
        internal JointPhysicsParamGroup[] ParamGroups { get; set; }

        internal SoftPhysicsParamGroup FatSpring { get; set; }
        internal SoftPhysicsParamGroup FatDamper { get; set; }
        internal SoftPhysicsParamGroup FatMass { get; set; }
        internal SoftPhysicsParamGroup FatColliderRadius { get; set; }
        internal SoftPhysicsParamGroup FatColliderDepth { get; set; }
        internal SoftPhysicsParamGroup FatColliderFriction { get; set; }
        internal SoftPhysicsParamGroup FatDistanceLimit { get; set; }
        internal JointPhysicsParamGroup FatBackForce { get; set; }
        internal JointPhysicsParamGroup FatBackForceMaxForce { get; set; }
        internal JointPhysicsParamGroup FatBackForceThreshold { get; set; }

        protected CustomFixedUpdate softJointsFixedUpdate;

        internal StorableBool SoftPhysicsOnJsb { get; set; }
        internal StorableBool LdpEnabledJsb { get; set; }
        internal JSONStorableBool ShowColliderPreviewsJsb { get; set; }

        internal bool SoftPhysicsSynced { get; set; }
        internal bool SoftPhysicsOn { get; set; }

        internal abstract bool IsLdpUpdateOn();

        internal void CalibrateSoftJointsDistortion(string direction)
        {
            foreach(var softJoint in AllSoftJointsList)
            {
                softJoint.CalibrateDistortion(direction);
            }
        }

        internal void ResetSoftJointsDistortion()
        {
            foreach(var softJoint in AllSoftJointsList)
            {
                softJoint.ResetDistortion();
            }
        }

        internal void EnableMultiplyFriction()
        {
            foreach(var softJoint in AllSoftJointsList)
            {
                softJoint.EnableMultiplyFriction();
            }
        }

        internal void Execute()
        {
            if(!SoftPhysicsOn)
            {
                return;
            }

            softJointsFixedUpdate.Run();
        }

        internal bool AnySoftJointMoved()
        {
            foreach(var x in AllSoftJointsList)
            {
                if(x.Moved(0.008f))
                {
                    return true;
                }
            }

            return false;
        }

        internal JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass();
            if(!IsLdpUpdateOn())
            {
                return jc;
            }

            foreach(var kvp in AllSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    jc[softJoint.Id] = softJoint.SerializeCalibrationData();
                }
            }

            return jc;
        }

        internal void CalibrateFromJSON(JSONClass jc)
        {
            if(!IsLdpUpdateOn())
            {
                return;
            }

            foreach(var kvp in AllSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.CalibrateFromJSON(jc[softJoint.Id].AsObject);
                }
            }
        }

        internal void ResetSoftJointsClothingAdjustments()
        {
            foreach(var kvp in AllSoftJoints)
            {
                foreach(var softJoint in kvp.Value)
                {
                    softJoint.ResetClothingAdjustments();
                }
            }
        }
    }
}
