using SimpleJSON;
using System.Diagnostics.CodeAnalysis;
using UnityEngine;

namespace NAT.Common.Models
{
    abstract class CalibratedJoint
    {
        protected readonly Rigidbody originRb;
        internal ConfigurableJoint Joint { get; private set; }
        internal Rigidbody Rigidbody { get; }
        protected readonly DAZBone bone;
        protected readonly Person person;
        protected bool softPhysicsOn;

        protected Rigidbody[] softJointOuterRbs;
        protected Rigidbody[] softJointCenterRbs;
        protected Rigidbody[] combinedSoftRbs;
        protected int[] centerVertexIndices;

        internal float AngleY { get; private set; }
        internal float AngleX { get; private set; }
        internal float DepthDiff { get; private set; }

        Vector3 _calibratedOffsetFromOrigin;
        Quaternion _calibratedOriginRotation;
        protected Vector3 simulatedJointPosition;
        Vector3 _positionRelativeToOrigin;
        Vector2 _angleXBaseVector;
        Vector2 _angleYBaseVector;
        float _depthBase;

        protected CalibratedJoint(ConfigurableJoint joint, Rigidbody originRb, Person person)
        {
            Joint = joint;
            Rigidbody = joint.GetComponent<Rigidbody>();
            bone = joint.GetComponent<DAZBone>();
            this.originRb = originRb;
            this.person = person;
        }

        internal virtual void SetFunctionsWhenSoftPhysicsOff(Person person, int[] centerPositionalVerts)
        {
        }

        [SuppressMessage("ReSharper", "ParameterTypeCanBeEnumerable.Global")]
        internal virtual void SetFunctionsWhenSoftPhysicsOn(
            Rigidbody[] softJointOuterRbs,
            Rigidbody[] softJointMainRbs,
            Rigidbody[] softJointCenterRbs
        )
        {
        }

        internal abstract void SetRotation(float x, float y, float z);

        float _prevXAngle;
        float _prevYAngle;

        internal bool Moved(float tolerance)
        {
            float xAngle = RelativeXAngle();
            float yAngle = RelativeYAngle();
            bool moved = Mathf.Abs(xAngle - _prevXAngle) > tolerance || Mathf.Abs(yAngle - _prevYAngle) > tolerance;
            _prevXAngle = xAngle;
            _prevYAngle = yAngle;
            return moved;
        }

        float RelativeXAngle() => Vector2.SignedAngle(
            _angleXBaseVector,
            new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.x)
        );

        float RelativeYAngle() => Vector2.SignedAngle(
            _angleYBaseVector,
            new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.y)
        );

        internal void Calibrate()
        {
            _calibratedOffsetFromOrigin = Rigidbody.position - originRb.position;
            _calibratedOriginRotation = originRb.rotation;
            simulatedJointPosition = SimulatedJointPosition();
            _positionRelativeToOrigin = RelativePosition();
            _angleXBaseVector = new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.x);
            _angleYBaseVector = new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.y);
            _depthBase = RelativeDepth();
        }

        internal void Measure(Vector3 offset)
        {
            simulatedJointPosition = SimulatedJointPosition();
            _positionRelativeToOrigin = RelativePosition();
            AngleX = RelativeXAngle() + offset.x;
            AngleY = RelativeYAngle() + offset.y;
            DepthDiff = _depthBase - RelativeDepth() + offset.z;
        }

        Vector3 SimulatedJointPosition()
        {
            var originRotationDiff = originRb.rotation * Quaternion.Inverse(_calibratedOriginRotation);
            return originRb.position + originRotationDiff * _calibratedOffsetFromOrigin;
        }

        protected abstract Vector3 RelativePosition();
        protected abstract float RelativeDepth();

        internal void ResetMeasurements()
        {
            AngleY = 0;
            AngleX = 0;
            DepthDiff = 0;
        }

        internal JSONClass SerializeCalibrationData() => new JSONClass
        {
            ["offset_from_origin"] = JSONUtils.SerializeVector3(_calibratedOffsetFromOrigin),
            ["origin_rotation"] = JSONUtils.SerializeQuaternion(_calibratedOriginRotation),
            ["angle_x_base_vector"] = JSONUtils.SerializeVector2(_angleXBaseVector),
            ["angle_y_base_vector"] = JSONUtils.SerializeVector2(_angleYBaseVector),
            ["depth_base"] = { AsFloat = _depthBase },
        };

        internal void CalibrateFromJSON(JSONClass jc)
        {
            _calibratedOffsetFromOrigin = JSONUtils.DeserializeVector3(jc["offset_from_origin"]);
            _calibratedOriginRotation = JSONUtils.DeserializeQuaternion(jc["origin_rotation"]);
            _angleXBaseVector = JSONUtils.DeserializeVector2(jc["angle_x_base_vector"]);
            _angleYBaseVector = JSONUtils.DeserializeVector2(jc["angle_y_base_vector"]);
            _depthBase = jc["depth_base"].AsFloat;
        }
    }
}
