﻿using UnityEngine;

namespace NAT.Common.Models
{
    class HardCollider
    {
        const float DEFAULT_FRICTION = 0.6f;

        internal AutoCollider AutoCollider { get; }
        internal Collider Collider { get; }
        readonly PhysicMaterial _colliderMaterial;
        readonly Transform _colliderTransform;
        internal JSONStorableString MassDisplayValueJss { get; }

        internal float MaxFrictionalDistance { private get; set; }
        float _sqrDistanceDiff;
        internal float SqrNeutralDistance { get; set; }

        float _clothingMassMultiplier;

        internal HardCollider(AutoCollider autoCollider)
        {
            AutoCollider = autoCollider;
            AutoCollider.resizeTrigger = AutoCollider.ResizeTrigger.None;

            Collider = AutoCollider.jointCollider;
            Collider.enabled = true;
            _colliderMaterial = Collider.material;
            _colliderTransform = Collider.transform;

            MassDisplayValueJss = new JSONStorableString("massDisplayValue", "");
        }

        internal void UpdateRigidbodyMass(float mass)
        {
            if(!Collider || !Collider.attachedRigidbody)
            {
                return;
            }

            Collider.attachedRigidbody.mass = _clothingMassMultiplier * mass;
            string displayValue = $"{mass:0.000} kg".Bold();
            MassDisplayValueJss.val = $"Collider Mass: {displayValue}";
        }

        internal float GetRigidbodyMass() => Collider.attachedRigidbody.mass;

        internal void UpdateDimensions(float radiusOffset)
        {
            AutoCollider.autoRadiusBuffer = radiusOffset;
            /* Scale length buffer with radius to ensure that the capsule colliders stay spherical.
             * 2x radius is exact scaling, preserving current center cylinder length.
             * Adding 0.1f ensures cylinder length decreases enough to prevent it
             * having any length. (+buffer => -length)
             */
            AutoCollider.autoLengthBuffer = 2f * radiusOffset + 0.1f;
        }

        internal void UpdatePosition(float rightOffset, float upOffset, float lookOffset)
        {
            AutoCollider.colliderRightOffset = rightOffset;
            AutoCollider.colliderUpOffset = upOffset;
            AutoCollider.colliderLookOffset = lookOffset;
        }

        internal void UpdateFriction(float multiplier, float max)
        {
            float normalizedDistance = Mathf.InverseLerp(0, MaxFrictionalDistance, _sqrDistanceDiff);
            float dynamicFriction = max - Mathf.SmoothStep(0, max, normalizedDistance);
            _colliderMaterial.dynamicFriction = multiplier * dynamicFriction;

            float staticMax = 0.1f + max;
            float staticFriction = staticMax - Mathf.SmoothStep(0, staticMax, normalizedDistance);
            _colliderMaterial.staticFriction = multiplier * staticFriction;
        }

        internal void SetClothingMassMultiplier(float massMultiplier) => _clothingMassMultiplier = massMultiplier;

        internal void AutoColliderSizeSet()
        {
            AutoCollider.resizeTrigger = AutoCollider.ResizeTrigger.Always;
            AutoCollider.AutoColliderSizeSet();
            AutoCollider.resizeTrigger = AutoCollider.ResizeTrigger.None;
        }

        internal void Calibrate(Vector3 breastCenter, Rigidbody chestRb) =>
            SqrNeutralDistance = SqrDistanceFromBreastCenter(breastCenter, chestRb);

        internal void UpdateDistanceDiff(Vector3 breastCenter, Rigidbody chestRb) =>
            _sqrDistanceDiff = Mathf.Abs(SqrNeutralDistance - SqrDistanceFromBreastCenter(breastCenter, chestRb));

        float SqrDistanceFromBreastCenter(Vector3 relativeBreastCenter, Rigidbody chestRb)
        {
            var relativeTransformPosition = Calc.RelativePosition(chestRb, _colliderTransform.position);
            /* Scale up by 1000 (millimeters) */
            return (relativeTransformPosition - relativeBreastCenter).sqrMagnitude * 1000;
        }

        internal void ResetDistanceDiff() => _sqrDistanceDiff = 0;

        internal void EnableMultiplyFriction() => _colliderMaterial.frictionCombine = PhysicMaterialCombine.Multiply;

        internal void RestoreDefaults(float defaultMass)
        {
            UpdateRigidbodyMass(defaultMass);

            AutoCollider.autoRadiusBuffer = 0;
            AutoCollider.autoLengthBuffer = 0;
            AutoCollider.colliderRightOffset = 0;
            AutoCollider.colliderUpOffset = 0;
            AutoCollider.colliderLookOffset = 0;
            AutoCollider.resizeTrigger = AutoCollider.ResizeTrigger.Always;
            AutoCollider.AutoColliderSizeSet(true);
            AutoCollider.resizeTrigger = AutoCollider.ResizeTrigger.MorphChangeOnly;

            /* Restore default friction. */
            var material = Collider.material;
            material.frictionCombine = PhysicMaterialCombine.Average;
            material.dynamicFriction = DEFAULT_FRICTION;
            material.staticFriction = DEFAULT_FRICTION;
        }
    }
}
