﻿using NAT.Common.Components;
using NAT.Common.Storables;
using UnityEngine;

namespace NAT.Common.Models
{
    sealed class HardColliderGroup
    {
        const float DEFAULT_MASS = 0.04f; // Seems to be a hard coded value in VaM.

        internal string Id { get; }
        internal string VisualizerId { get; }

        readonly Scaler _baseRbMassScaler;
        readonly Scaler _radiusScaler;
        readonly Scaler _rightOffsetScaler;
        readonly Scaler _upOffsetScaler;
        readonly Scaler _lookOffsetScaler;

        readonly float _frictionMultiplier;

        internal HardCollider Left { get; }
        internal HardCollider Right { get; }

        internal StorableFloat MassMultiplierJsf { get; set; }
        internal StorableFloat RadiusJsf { get; set; }
        internal StorableFloat RightJsf { get; set; }
        internal StorableFloat UpJsf { get; set; }
        internal StorableFloat LookJsf { get; set; }

        internal HardColliderGroup(
            string id,
            string visualizerId,
            HardCollider left,
            HardCollider right,
            Scaler radiusScaler,
            Scaler rightOffsetScaler,
            Scaler upOffsetScaler,
            Scaler lookOffsetScaler,
            float frictionMultiplier
        )
        {
            Id = id;
            VisualizerId = visualizerId;
            Left = left;
            Right = right;
            _radiusScaler = radiusScaler;
            _rightOffsetScaler = rightOffsetScaler;
            _upOffsetScaler = upOffsetScaler;
            _lookOffsetScaler = lookOffsetScaler;
            _frictionMultiplier = frictionMultiplier;
        }

        internal HardColliderGroup(
            string id,
            string visualizerId,
            HardCollider left,
            HardCollider right,
            Scaler baseRbMassScaler,
            Scaler radiusScaler,
            Scaler rightOffsetScaler,
            Scaler upOffsetScaler,
            Scaler lookOffsetScaler,
            float frictionMultiplier
        )
        {
            Id = id;
            VisualizerId = visualizerId;
            Left = left;
            Right = right;
            _baseRbMassScaler = baseRbMassScaler;
            _radiusScaler = radiusScaler;
            _rightOffsetScaler = rightOffsetScaler;
            _upOffsetScaler = upOffsetScaler;
            _lookOffsetScaler = lookOffsetScaler;
            _frictionMultiplier = frictionMultiplier;
        }

        internal void UpdateRigidbodyMass(float value)
        {
            Left.UpdateRigidbodyMass(value);
            Right.UpdateRigidbodyMass(value);
        }

        internal void UpdateRigidbodyMass(float combinedMultiplier, float mass, float softness, bool softPhysicsOn)
        {
            float rbMass = softPhysicsOn
                ? combinedMultiplier * _baseRbMassScaler.Scale(DEFAULT_MASS, mass, softness)
                : combinedMultiplier * 0.125f * _baseRbMassScaler.Scale(DEFAULT_MASS, mass);
            UpdateRigidbodyMass(rbMass);
        }

        internal void UpdateDimensions(float mass, float softness)
        {
            float radius = -_radiusScaler.Scale(RadiusJsf.val, mass, softness);
            Left.UpdateDimensions(radius);
            Right.UpdateDimensions(radius);
        }

        internal void UpdatePosition(float mass, float softness)
        {
            float rightOffset = _rightOffsetScaler.Scale(RightJsf.val, mass, softness);
            float upOffset = -_upOffsetScaler.Scale(UpJsf.val, mass, softness);
            float lookOffset = -_lookOffsetScaler.Scale(LookJsf.val, mass, softness);
            Left.UpdatePosition(-rightOffset, upOffset, lookOffset);
            Right.UpdatePosition(rightOffset, upOffset, lookOffset);
        }

        internal void UpdateMaxFrictionalDistance(float sizeMultiplierLeft, float sizeMultiplierRight, float smallestRadius)
        {
            float halfSqrSmallestRadiusMillimeters = 0.5f * smallestRadius * smallestRadius * 1000;
            Left.MaxFrictionalDistance = sizeMultiplierLeft * halfSqrSmallestRadiusMillimeters;
            Right.MaxFrictionalDistance = sizeMultiplierRight * halfSqrSmallestRadiusMillimeters;
        }

        internal void UpdateFriction(float max)
        {
            Left.UpdateFriction(_frictionMultiplier, max);
            Right.UpdateFriction(_frictionMultiplier, max);
        }

        internal void SetClothingMassMultipliers(float massMultiplierLeft, float massMultiplierRight)
        {
            Left.SetClothingMassMultiplier(massMultiplierLeft);
            Right.SetClothingMassMultiplier(massMultiplierRight);
        }

        internal void AutoColliderSizeSet()
        {
            Left.AutoColliderSizeSet();
            Right.AutoColliderSizeSet();
        }

        internal void Calibrate(Vector3 leftCenter, Vector3 rightCenter, Rigidbody centerRb)
        {
            Left.Calibrate(leftCenter, centerRb);
            Right.Calibrate(rightCenter, centerRb);
        }

        internal void UpdateDistanceDiffs(Vector3 leftCenter, Vector3 rightCenter, Rigidbody centerRb)
        {
            Left.UpdateDistanceDiff(leftCenter, centerRb);
            Right.UpdateDistanceDiff(rightCenter, centerRb);
        }

        internal void ResetDistanceDiffs()
        {
            Left.ResetDistanceDiff();
            Right.ResetDistanceDiff();
        }

        internal void EnableMultiplyFriction()
        {
            Left.EnableMultiplyFriction();
            Right.EnableMultiplyFriction();
        }

        internal void RestoreDefaults()
        {
            Left.RestoreDefaults(DEFAULT_MASS);
            Right.RestoreDefaults(DEFAULT_MASS);
        }

        internal bool HasRigidbodies() => Left.Collider.attachedRigidbody && Right.Collider.attachedRigidbody;

        internal string GetJSONId() => Id.Replace(" ", "_").ToLower();
    }
}
