using NAT.Common.Handlers.Configs;
using System;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Models
{
    sealed class JointPhysicsParam : PhysicsParam
    {
        internal float Value { get; private set; }

        internal Dictionary<string, PhysicsCalculator> GravityCalculators { get; set; } = new Dictionary<string, PhysicsCalculator>();
        readonly Dictionary<string, float> _additiveGravityAdjustments = new Dictionary<string, float>();

        internal Dictionary<string, PhysicsCalculator> ForceCalculators { get; set; } = new Dictionary<string, PhysicsCalculator>();
        readonly Dictionary<string, float> _additiveForceAdjustments = new Dictionary<string, float>();

        float _clothingAdjustmentAmount;
        // internal float ClothingAdjustmentValue { get; private set; }

        internal Action<float> Sync { private get; set; }

        internal JointPhysicsParam(string side, string name, float min, float max) : base(side, name, min, max)
        {
        }

        internal bool HasAdjustableGravityPhysics()
        {
            if(GravityCalculators == null || GravityCalculators.Count == 0)
            {
                return false;
            }

            foreach(var pair in GravityCalculators)
            {
                if(pair.Value.MultiplierJsf == null)
                {
                    return false;
                }
            }

            return true;
        }

        internal override void DoSync()
        {
            float gravitySum = 0f;
            foreach(var pair in _additiveGravityAdjustments)
            {
                gravitySum += pair.Value;
            }

            float forceSum = 0f;
            foreach(var pair in _additiveForceAdjustments)
            {
                forceSum += pair.Value;
            }

            float value = OffsetJsf.val + BaseValueJsf.val + gravitySum + forceSum;
            float clamped = Mathf.Clamp(value, BaseValueJsf.min, BaseValueJsf.max);
            Value = Mathf.Lerp(clamped, BaseValueJsf.max, _clothingAdjustmentAmount);
            // ClothingAdjustmentValue = Value - clamped; // just for info
            Sync?.Invoke(Value); // sync is null for target rotation parameters
        }

        internal void UpdateMassValue(float mass)
        {
            BaseValueJsf.val = mass;
            DoSync();
        }

        // For temp use, call UpdateValue after
        internal void OverrideValue(float value)
        {
            Value = value;
            Sync?.Invoke(value);
        }

        internal void UpdateGravityValue(string direction, float effect, float mass, float softness)
        {
            if(!GravityCalculators.ContainsKey(direction))
            {
                return;
            }

            float gravityValue = GravityCalculators[direction].CalculateDynamic(effect, mass, softness, true);
            _additiveGravityAdjustments[direction] = gravityValue;
        }

        internal void ResetGravityValue(string direction)
        {
            if(!GravityCalculators.ContainsKey(direction))
            {
                return;
            }

            _additiveGravityAdjustments[direction] = 0;
        }

        internal void UpdateForceValue(string direction, float effect, float mass, float softness)
        {
            if(!ForceCalculators.ContainsKey(direction))
            {
                return;
            }

            float forceValue = ForceCalculators[direction].CalculateDynamic(effect, mass, softness, true);
            _additiveForceAdjustments[direction] = forceValue;
        }

        internal void ResetForceValue(string direction)
        {
            if(!ForceCalculators.ContainsKey(direction))
            {
                return;
            }

            _additiveForceAdjustments[direction] = 0;
        }

        internal void UpdateClothingAdjustmentAmount(float value) => _clothingAdjustmentAmount = value;
        internal void ResetClothingAdjustmentAmount() => _clothingAdjustmentAmount = 0f;

        internal void SaveCalibrated()
        {
            OffsetJsf.SaveCalibrated();
            if(HasAdjustableGravityPhysics())
            {
                foreach(var pair in GravityCalculators)
                {
                    pair.Value.MultiplierJsf.SaveCalibrated();
                }
            }
        }
    }
}
