using NAT.BM.Handlers;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;
using NAT.TM.Handlers;
using System.Collections.Generic;

namespace NAT.Common.Models
{
    class JointPhysicsParamGroup : PhysicsParamGroup
    {
        internal bool RequiresCalibration { get; set; }
        internal bool RightInverted { get; set; }
        internal string GravityInfoText { get; set; }
        internal JSONStorableBool EnabledJsb { get; set; }

        readonly ScriptBase _script;
        JointPhysicsParam _left, _right;

        internal void DoSync()
        {
            _left.DoSync();
            _right.DoSync();
        }

        internal JointPhysicsParam Left
        {
            get { return _left; }
            set
            {
                _left = value;
                leftParam = value;
            }
        }

        internal JointPhysicsParam Right
        {
            get { return _right; }
            set
            {
                _right = value;
                rightParam = value;
            }
        }

        internal JointPhysicsParamGroup(ScriptBase script)
        {
            _script = script;
        }

        internal void UpdateMassValue(float mass)
        {
            Left.UpdateMassValue(mass);
            Right.UpdateMassValue(mass);
        }

        internal void OverrideValue(float value)
        {
            Left.OverrideValue(value);
            Right.OverrideValue(value);
        }

        internal void SetGravityCalculators(
            Dictionary<string, PhysicsCalculator> leftGravityCalculators,
            Dictionary<string, PhysicsCalculator> rightGravityCalculators
        )
        {
            Left.GravityCalculators = leftGravityCalculators;
            Right.GravityCalculators = rightGravityCalculators;
        }

        internal void SetForceCalculators(Dictionary<string, Dictionary<string, PhysicsCalculator>> configs)
        {
            Left.ForceCalculators = configs[Side.LEFT];
            Right.ForceCalculators = configs[Side.RIGHT];
        }

        internal override void SetCallbackFunctions()
        {
            Left.OffsetJsf.setJSONCallbackFunction = jsf =>
            {
                var leftStorable = (StorableFloat) jsf;
                var rightStorable = Right.OffsetJsf;
                Left.UpdateOffsetValue(leftStorable.val);
                float rightValue = RightInverted ? -leftStorable.val : leftStorable.val;
                Right.UpdateOffsetValue(ApplyChangesOnlyToLeftParamJsb.val ? rightStorable.defaultVal : rightValue);
                if(RequiresCalibration)
                {
                    _script.SetCalibrationPending(leftStorable);
                }

                _script.Tabs.PhysicsWindow.OnMainParamAdjusted(this);
            };
            ApplyChangesOnlyToLeftParamJsb.setCallbackFunction = value =>
            {
                /* Offset */
                {
                    var leftStorable = Left.OffsetJsf;
                    var rightStorable = Right.OffsetJsf;
                    float rightValue = RightInverted ? -leftStorable.val : leftStorable.val;
                    Right.UpdateOffsetValue(value ? rightStorable.defaultVal : rightValue);
                    if(RequiresCalibration)
                    {
                        _script.SetCalibrationPending(rightStorable);
                    }
                }

                /* Custom Gravity */
                if(Name == BREAST_JOINT_MASS.NAME || Name == GLUTE_JOINT_MASS.NAME)
                {
                    var gravityPhysicsHandler = _script.IsBootyMagic
                        ? (GravityPhysicsHandler) _script.Provider.GetInstance<GluteGravityPhysicsHandler>()
                        : _script.Provider.GetInstance<BreastGravityPhysicsHandler>();
                    gravityPhysicsHandler.SetGravityValues();
                }

                /* Gravity Physics */
                {
                    foreach(var pair in Left.GravityCalculators)
                    {
                        var config = pair.Value;
                        var leftStorable = config.MultiplierJsf;
                        var rightStorable = Right.GravityCalculators[pair.Key].MultiplierJsf;
                        rightStorable.val = ApplyChangesOnlyToLeftParamJsb.val ? rightStorable.defaultVal : leftStorable.val;
                        if(RequiresCalibration)
                        {
                            _script.SetCalibrationPending(rightStorable);
                        }
                    }
                }
            };
        }

        internal void SetGravityMultiplierCallbacks()
        {
            foreach(var pair in Left.GravityCalculators)
            {
                var config = pair.Value;
                config.MultiplierJsf.setJSONCallbackFunction = jsf =>
                {
                    var leftStorable = (StorableFloat) jsf;
                    var rightStorable = Right.GravityCalculators[pair.Key].MultiplierJsf;
                    rightStorable.val = ApplyChangesOnlyToLeftParamJsb.val ? rightStorable.defaultVal : leftStorable.val;
                    if(RequiresCalibration)
                    {
                        _script.SetCalibrationPending(leftStorable);
                    }

                    _script.Tabs.PhysicsWindow.OnMainParamAdjusted(this);
                };
            }
        }

        internal void ResetClothingAdjustmentAmount()
        {
            Left.ResetClothingAdjustmentAmount();
            Right.ResetClothingAdjustmentAmount();
        }

        internal override bool IsModified => base.IsModified || IsGravityPhysicsModified();

        bool IsGravityPhysicsModified()
        {
            if(!Left.HasAdjustableGravityPhysics())
            {
                return false;
            }

            foreach(var pair in Left.GravityCalculators)
            {
                if(pair.Value.MultiplierJsf.IsModified)
                {
                    return true;
                }
            }

            return false;
        }

        internal void SaveCalibrated()
        {
            if(!RequiresCalibration)
            {
                return;
            }

            Left.SaveCalibrated();
            Right.SaveCalibrated();
        }
    }
}
