using System;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common
{
    sealed class MorphListener
    {
        const float FILTER_STRENGTH = 0.005f;

        readonly HashSet<int> _vertexIndices;
        readonly HashSet<string> _excludeMorphsNames;
        List<DAZMorph> _listenedMorphsList;
        readonly Dictionary<DAZMorph, float> _listenedMorphs = new Dictionary<DAZMorph, float>();
        internal bool ProcessingDone { get; private set; }

        internal MorphListener(HashSet<int> vertexIndices, HashSet<string> excludeMorphsNames)
        {
            _vertexIndices = vertexIndices;
            _excludeMorphsNames = excludeMorphsNames;
        }

        internal void ProcessMorphs(DAZMorphBank morphBank)
        {
            if(morphBank.morphs == null)
            {
                return;
            }

            foreach(var morph in morphBank.morphs)
            {
                try
                {
                    if(
                        morph.visible &&
                        !morph.isPoseControl &&
                        morph.group != null &&
                        !morph.group.Contains("Pose/") &&
                        !_listenedMorphs.ContainsKey(morph) &&
                        !_excludeMorphsNames.Contains(morph.morphName)
                    )
                    {
                        if(InSet(morph, FILTER_STRENGTH))
                        {
                            _listenedMorphs[morph] = morph.morphValue;
                        }
                    }
                }
                catch(Exception e)
                {
                    Debug.LogError($"Unable to add morph '{morph.morphName}'. Error: {e}");
                }
            }

            _listenedMorphsList = _listenedMorphs.Keys.ToList();
            ProcessingDone = true;
        }

        internal bool IsChanged { get; private set; }
        bool _prevIsChanged;

        internal bool ChangeDetected()
        {
            if(Time.frameCount % 20 == 0)
            {
                IsChanged = AnyMorphIsChanged();
                if(IsChanged != _prevIsChanged)
                {
                    _prevIsChanged = IsChanged;
                    return true;
                }
            }

            return false;
        }

        bool AnyMorphIsChanged()
        {
            bool result = false;

            foreach(var pair in _listenedMorphs)
            {
                if(Mathf.Abs(pair.Key.morphValue - pair.Value) > 0.001f)
                {
                    result = true;
                    break;
                }
            }

            return result;
        }

        internal void SaveCalibrated()
        {
            if(!ProcessingDone)
            {
                return;
            }

            foreach(var morph in _listenedMorphsList)
            {
                _listenedMorphs[morph] = morph.morphValue;
            }
        }

        bool InSet(DAZMorph morph, float filterStrength)
        {
            if(morph.deltas == null)
            {
                morph.LoadDeltas();
            }

            if(morph.deltas == null)
            {
                return false;
            }

            float hitDelta = 0.0f;
            float deltaMagnitudesSum = 0f;
            foreach(var vertex in morph.deltas)
            {
                deltaMagnitudesSum += vertex.delta.magnitude;
            }

            float hitDeltaMax = deltaMagnitudesSum * filterStrength;
            foreach(var delta in morph.deltas)
            {
                if(_vertexIndices.Contains(delta.vertex))
                {
                    hitDelta += delta.delta.magnitude;
                    if(hitDelta >= hitDeltaMax)
                    {
                        return true;
                    }
                }
            }

            return false;
        }
    }
}
