//
using NAT.Common.Components;
using NAT.Common.Handlers;
using NAT.Common.Storables;
using NAT.Common.UI;
using System;
using System.Collections;
using SimpleJSON;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.Common
{
    class ScriptBase : MVRScript
    {
        internal string DomainName { get; private set; }
        internal Color DomainColor { get; private set; }
        internal Loggr Loggr { get; private set; }
        internal Bootstrap Bootstrap { get; set; }
        internal bool? Initialized { get; set; }
        protected bool InitializedFromDefaultPreset { get; set; }
        internal bool RestoringFromJSON { get; set; }
        internal Person Person { get; set; }
        internal PluginPresetHandler PresetHandler { get; set; }
        internal string LoadedPresetName { get; private set; }

        internal StorableFloat SoftnessJsf { get; set; }
        internal StorableFloat QuicknessJsf { get; set; }
        internal float Softness { get; set; }
        internal float Quickness { get; set; }
        internal StorableAction CalibrationAction { get; set; }
        internal StorableFloat LeftRightAngleOffsetJsf { get; set; }
        internal StorableFloat UpDownAngleOffsetJsf { get; set; }
        internal StorableFloat DepthOffsetJsf { get; set; }

        protected Vector3 leftMeasurementOffset, rightMeasurementOffset;

        protected UnityEventsListener pluginUIEventsListener;
        internal Bindings Bindings { get; set; }
        internal CalibrationStateHandler CalibrationStateHandler { get; set; }
        internal UIModManagerBase UIModManager { get; set; }
        internal Tabs Tabs { get; set; }
        internal ProviderBase Provider { get; set; }

        protected bool isUpdatePaused;
        internal bool IsDevelopmentEnv => Bootstrap.IsDevelopmentEnv;
        internal bool IsTittyMagic { get; private set; }
        internal bool IsBootyMagic { get; private set; }
        internal bool IsInsideRestore => insideRestore;

        List<StorableDomain> _storableDomains;
        Dictionary<string, StorableDomain> _storableDomainsByName;

        internal JSONStorable GetOtherPlugin()
        {
            if(!Bootstrap)
            {
                Debug.Log($"{DomainName}: Bootstrap was null (Memory leak)");
                return null;
            }

            return IsBootyMagic ? Bootstrap.TittyMagicStorable : Bootstrap.BootyMagicStorable;
        }

        internal bool IsGlobalUISideLeft { get; } = UISideAlign.globalSide == UISideAlign.Side.Left;
        readonly Dictionary<string, UIDynamic> _customElements = new Dictionary<string, UIDynamic>();
        bool _uiOpened;
        bool _gravityControlsAdded;

        protected void Init(string domainName, Color domainColor)
        {
            DomainName = domainName;
            DomainColor = domainColor;
            IsBootyMagic = domainName == Constant.BOOTY_MAGIC;
            IsTittyMagic = domainName == Constant.TITTY_MAGIC;
            Loggr = new Loggr(DomainName);
            _storableDomains = new List<StorableDomain>
            {
                new StorableDomain(StorableDomainName.GENERAL,
                    "General plugin parameters."
                ),
                new StorableDomain(StorableDomainName.JOINT_PHYSICS,
                    "Joint physics parameters."
                ),
                new StorableDomain(StorableDomainName.SOFT_PHYSICS,
                    "Soft physics parameters."
                ),
                new StorableDomain(StorableDomainName.FORCE_PHYSICS,
                    "Force physics parameters. These are not configurable in the UI."
                ),
                new StorableDomain(StorableDomainName.MORPHS,
                    "Directional force morphing and other morphing related parameters. Clothing morph settings are at the end of the list!"
                ),
                new StorableDomain(StorableDomainName.HARD_COLLIDERS,
                    "Hard colliders parameters."
                ),
                new StorableDomain(StorableDomainName.FRICTION,
                    "Friction parameters."
                ),
                new StorableDomain(StorableDomainName.UI_ACTIONS,
                    "Actions for opening specific UI windows. These actions are also available via the Keybindings plugin."
                ),
                new StorableDomain(StorableDomainName.INTERNAL,
                    "Internal parameters the used for inter-plugin communication. Do not touch."
                ),
                new StorableDomain(StorableDomainName.CLOTHING,
                    "Clothing parameters. Item specific settings are not stored - use clothing profile presets to save your settings."
                ),
            };
            for(int i = 0; i < _storableDomains.Count; i++)
            {
                _storableDomains[i].PreRegisterHeader(i);
            }

            _storableDomainsByName = new Dictionary<string, StorableDomain>();
            foreach(var domain in _storableDomains)
            {
                _storableDomainsByName[domain.Name] = domain;
            }
        }

        // Prevent ScriptBase from showing up as a plugin in Plugins tab.
        public override bool ShouldIgnore() => true;

        public override void InitUI()
        {
            if(ShouldIgnore() || DomainName == null)
            {
                return;
            }

            base.InitUI();
            if(!UITransform || pluginUIEventsListener)
            {
                return;
            }

            StartCoroutine(InitUICo());
        }

        protected virtual void OnUIEnabled()
        {
        }

        protected virtual void OnUIDisabled()
        {
        }

        IEnumerator InitUICo()
        {
            pluginUIEventsListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
            pluginUIEventsListener.EnabledHandlers += () => StartCoroutine(OnUIEnabledCo());

            while(Initialized == null)
            {
                yield return null;
            }

            if(Initialized == false)
            {
                enabledJSON.val = false;
                yield break;
            }

            pluginUIEventsListener.DisabledHandlers += () => StartCoroutine(OnUIDisabledCo());
        }

        bool _inEnabledCo;

        IEnumerator OnUIEnabledCo()
        {
            if(_inEnabledCo)
            {
                /* When VAM UI is toggled back on with the plugin UI already active, onEnable gets called twice and onDisable once.
                 * This ensures onEnable logic executes just once.
                 */
                yield break;
            }

            _inEnabledCo = true;

            if(!_uiOpened)
            {
                SetGrayBackground();
            }

            while(Initialized == null)
            {
                yield return null;
            }

            if(Initialized == false)
            {
                yield break;
            }

            if(!_uiOpened)
            {
                Tabs.CreateNavigationButtons();
                Tabs.GoToMainWindow();
                _uiOpened = true;
            }

            yield return null;
            yield return null;
            yield return null;
            OnUIEnabled();

            #if ENV_DEVELOPMENT
            {
                if(!_devUICreated)
                {
                    CreateDevUI();
                }
            }
            #endif

            _inEnabledCo = false;
        }

        IEnumerator OnUIDisabledCo()
        {
            if(_inEnabledCo)
            {
                /* When VAM UI is toggled back on with the plugin UI already active, onEnable gets called twice and onDisable once.
                 * This ensures only onEnable logic executes.
                 */
                yield break;
            }

            OnUIDisabled();
        }

        void SetGrayBackground()
        {
            var background = rightUIContent.parent.parent.parent.transform.GetComponent<Image>();
            background.color = Colors.backgroundGray;
        }

        protected GravityPhysicsHandler gravityPhysicsHandler;
        UIDynamicButton _toggleGravityControlsButton;
        Button _gravityControlsHeader;
        Button _stopButton;

        protected void AddGravityControls()
        {
            if(_gravityControlsAdded)
            {
                return;
            }

            float posX;

            {
                posX = IsGlobalUISideLeft ? 1088 : -63;
                string label = IsGlobalUISideLeft ? "<".Bold() : ">".Bold();
                var button = CreateCustomButton(label, new Vector2(posX, -1121), new Vector2(-1025, 50));
                button.textColor = Color.white;
                button.buttonColor = Colors.transparentGray;
                button.AddListener(ToggleGravityControls);
                _toggleGravityControlsButton = button;
            }
            {
                posX = IsGlobalUISideLeft ? 1141 : -307;
                var button = CreateCustomButton("Simulate Gravity", new Vector2(posX, -1121), new Vector2(-834, 50));
                button.textColor = Color.white;
                button.buttonColor = Colors.darkTransparentGray;
                var colors = button.button.colors;
                colors.highlightedColor = Color.white;
                colors.pressedColor = Color.white;
                button.button.colors = colors;
                _gravityControlsHeader = button.button;
            }

            float controlsPosX = IsGlobalUISideLeft ? 1113 : -125;
            float controlsSpacing = IsGlobalUISideLeft ? 91 : -91;

            CreateGravityControlButton(Direction.UP, "↑", new Vector2(controlsPosX + controlsSpacing, -1180));
            CreateGravityControlButton(Direction.DOWN, "↓", new Vector2(controlsPosX + controlsSpacing, -1239));

            posX = IsGlobalUISideLeft ? controlsPosX : controlsPosX + 2 * controlsSpacing;
            CreateGravityControlButton(Direction.BACK, "BK", new Vector2(posX, -1180));

            posX = IsGlobalUISideLeft ? controlsPosX + 2 * controlsSpacing : controlsPosX;
            CreateGravityControlButton(Direction.FORWARD, "FW", new Vector2(posX, -1180));

            posX = IsGlobalUISideLeft ? controlsPosX : controlsPosX + 2 * controlsSpacing;
            CreateGravityControlButton(Direction.RIGHT, "←", new Vector2(posX, -1239));

            posX = IsGlobalUISideLeft ? controlsPosX + 2 * controlsSpacing : controlsPosX;
            CreateGravityControlButton(Direction.LEFT, "→", new Vector2(posX, -1239));

            {
                posX = IsGlobalUISideLeft ? controlsPosX : controlsPosX + 2 * controlsSpacing;
                var button = CreateCustomButton("Stop".Bold(), new Vector2(posX, -1289), new Vector2(-806, 50));
                button.buttonText.fontSize = 28;
                button.buttonColor = Colors.transparentPink;
                gravityPhysicsHandler.StopGravitySimulationAction.button = button.button;
                _stopButton = button.button;
            }

            _gravityControlsAdded = true;
        }

        void CreateGravityControlButton(string direction, string label, Vector2 position)
        {
            var size = new Vector2(-988, 60);
            var button = CreateCustomButton(label, position, size);
            button.textColor = gravityPhysicsHandler.GravitySimulationDirection == direction && !CalibrationStateHandler.CalibratingJsb.val
                ? Colors.funkyCyan
                : Color.white;
            button.buttonColor = Colors.transparentGray;
            button.buttonText.fontSize = direction == Direction.BACK || direction == Direction.FORWARD ? 28 : 36;
            gravityPhysicsHandler.SimulateGravityActions[direction].button = button.button;
        }

        void ToggleGravityControls()
        {
            bool active = !_gravityControlsHeader.gameObject.activeSelf;
            if(active)
            {
                _toggleGravityControlsButton.label = IsGlobalUISideLeft ? "<" : ">";
            }
            else
            {
                _toggleGravityControlsButton.label = IsGlobalUISideLeft ? ">" : "<";
            }

            _gravityControlsHeader.gameObject.SetActive(active);
            foreach(var pair in gravityPhysicsHandler.SimulateGravityActions)
            {
                var uiDynamicButton = pair.Value.button.GetComponent<UIDynamicButton>();
                uiDynamicButton.gameObject.SetActive(active);
            }

            if(!active)
            {
                _stopButton.onClick.Invoke();
            }

            _stopButton.gameObject.SetActive(active);
        }

        internal void SetButtonColors(string direction)
        {
            if(!_gravityControlsAdded)
            {
                return;
            }

            var triggeredAction = gravityPhysicsHandler.SimulateGravityActions[direction];
            foreach(var pair in gravityPhysicsHandler.SimulateGravityActions)
            {
                var action = pair.Value;
                var uiDynamicButton = action.button.GetComponent<UIDynamicButton>();
                uiDynamicButton.textColor = action == triggeredAction ? Colors.funkyCyan : Color.white;
            }
        }

        internal void ResetButtonColors()
        {
            if(!_gravityControlsAdded)
            {
                return;
            }

            foreach(var pair in gravityPhysicsHandler.SimulateGravityActions)
            {
                var uiDynamicButton = pair.Value.button.GetComponent<UIDynamicButton>();
                uiDynamicButton.textColor = Color.white;
            }
        }

        internal void ConfigureColliderFriction()
        {
            if(!enabled)
            {
                Loggr.LogMessage($"Enable the plugin to configure collider friction.");
                return;
            }

            Bindings.Actions[$"{Constant.OPEN_UI}_{Constant.FRICTION}"].actionCallback();
        }

        // ReSharper disable MemberCanBePrivate.Global
        // ReSharper disable UnusedMember.Global

        internal UIDynamicButton CreateCustomButton(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomButton(label, label, position, size, devConfigurable);

        internal UIDynamicButton CreateCustomButton(string id, string label, Vector2 position, Vector2 size, bool configurable = true)
        {
            var t = this.InstantiateButton(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicButton>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(configurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            _customElements[id] = uiDynamic;
            return uiDynamic;
        }

        internal UIDynamicTextField CreateCustomTextField(string text, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomTextField(text, text, position, size, devConfigurable);

        internal UIDynamicTextField CreateCustomTextField(string id, string text, Vector2 position, Vector2 size, bool configurable = true)
        {
            var t = this.InstantiateTextField(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicTextField>();
            uiDynamic.text = text;

            #if ENV_DEVELOPMENT
            {
                if(configurable)
                {
                    AddRect(id.Substring(0, Math.Min(id.Length, 15)), rectT);
                }
            }
            #endif

            _customElements[id] = uiDynamic;
            return uiDynamic;
        }

        internal UIDynamicSlider CreateCustomSlider(string label, Vector2 position, bool devConfigurable = true) =>
            CreateCustomSlider(label, label, position, devConfigurable);

        internal UIDynamicSlider CreateCustomSlider(string id, string label, Vector2 position, bool configurable = true)
        {
            var t = this.InstantiateSlider(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = new Vector2(-555, 120);
            var uiDynamic = t.GetComponent<UIDynamicSlider>();
            uiDynamic.label = label;
            uiDynamic.rangeAdjustEnabled = false;

            #if ENV_DEVELOPMENT
            {
                if(configurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            _customElements[id] = uiDynamic;
            return uiDynamic;
        }

        internal UIDynamicToggle CreateCustomToggle(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomToggle(label, label, position, size, devConfigurable);

        internal UIDynamicToggle CreateCustomToggle(string id, string label, Vector2 position, Vector2 size, bool configurable = true)
        {
            var t = this.InstantiateToggle(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicToggle>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(configurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            _customElements[id] = uiDynamic;
            return uiDynamic;
        }

        internal UIDynamicPopup CreateCustomPopup(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomFilterablePopup(label, label, position, size, devConfigurable);

        internal UIDynamicPopup CreateCustomPopup(string id, string label, Vector2 position, Vector2 size, bool devConfigurable = true)
        {
            var t = this.InstantiatePopup(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicPopup>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(devConfigurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            _customElements[id] = uiDynamic;
            return uiDynamic;
        }

        internal UIDynamicPopup CreateCustomScrollablePopup(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomFilterablePopup(label, label, position, size, devConfigurable);

        internal UIDynamicPopup CreateCustomScrollablePopup(string id, string label, Vector2 position, Vector2 size, bool devConfigurable = true)
        {
            var t = this.InstantiateScrollablePopup(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicPopup>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(devConfigurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            _customElements[id] = uiDynamic;
            return uiDynamic;
        }

        internal UIDynamicPopup CreateCustomFilterablePopup(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomFilterablePopup(label, label, position, size, devConfigurable);

        internal UIDynamicPopup CreateCustomFilterablePopup(string id, string label, Vector2 position, Vector2 size, bool devConfigurable = true)
        {
            var t = this.InstantiateFilterablePopup(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicPopup>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(devConfigurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            _customElements[id] = uiDynamic;
            return uiDynamic;
        }

        // ReSharper restore UnusedMember.Global
        // ReSharper restore MemberCanBePrivate.Global

        internal static void ConfigureCustomSlider(UIDynamicSlider uiDynamic)
        {
            var uiDynamicT = uiDynamic.transform;
            var rectT = uiDynamicT.GetComponent<RectTransform>();
            var position = rectT.anchoredPosition;
            rectT.anchoredPosition = new Vector2(position.x - 55, position.y);
            rectT.sizeDelta = new Vector2(-444, 118);

            // Debug.Log(Utils.ObjectHierarchyToString(slider.transform));
            // foreach(Transform t1 in uiDynamic.transform)
            // {
            //     Debug.Log($"{t1.name}:\n{Utils.EnumerableToString(Utils.GetChildComponentData(t1))}");
            // }

            var panelT = uiDynamicT.Find("Panel");
            var sliderT = uiDynamicT.Find("Slider");
            var textT = uiDynamicT.Find("Text");
            var valueInputFieldT = uiDynamicT.Find("ValueInputField");
            var defaultValueButtonT = uiDynamicT.Find("DefaultValueButton");
            var quickButtonsLeftT = uiDynamicT.Find("QuickButtonsGroup/QuickButtonsLeft");
            var quickButtonsRightT = uiDynamicT.Find("QuickButtonsGroup/QuickButtonsRight");
            var m1ButtonT = quickButtonsLeftT.Find("M1Button");
            var m2ButtonT = quickButtonsLeftT.Find("M2Button");
            var m3ButtonT = quickButtonsLeftT.Find("M3Button");
            var m4ButtonT = quickButtonsLeftT.Find("M4Button");
            var p1ButtonT = quickButtonsRightT.Find("P1Button");
            var p2ButtonT = quickButtonsRightT.Find("P2Button");
            var p3ButtonT = quickButtonsRightT.Find("P3Button");
            var p4ButtonT = quickButtonsRightT.Find("P4Button");

            textT.GetComponent<Text>().fontSize = 27;
            foreach(var childText in valueInputFieldT.GetComponentsInChildren<Text>())
            {
                childText.fontSize = 27;
            }

            m4ButtonT.gameObject.SetActive(false);
            p4ButtonT.gameObject.SetActive(false);

            var panelRectT = panelT.GetComponent<RectTransform>();
            var sliderRectT = sliderT.GetComponent<RectTransform>();
            var textRectT = textT.GetComponent<RectTransform>();
            var valueInputFieldRectT = valueInputFieldT.GetComponent<RectTransform>();
            var defaultValueButtonRectT = defaultValueButtonT.GetComponent<RectTransform>();
            var m1ButtonRectT = m1ButtonT.GetComponent<RectTransform>();
            var m2ButtonRectT = m2ButtonT.GetComponent<RectTransform>();
            var m3ButtonRectT = m3ButtonT.GetComponent<RectTransform>();
            var p1ButtonRectT = p1ButtonT.GetComponent<RectTransform>();
            var p2ButtonRectT = p2ButtonT.GetComponent<RectTransform>();
            var p3ButtonRectT = p3ButtonT.GetComponent<RectTransform>();

            panelRectT.sizeDelta = new Vector2(-114, 0);
            sliderRectT.sizeDelta = new Vector2(-125, 35);
            sliderRectT.anchoredPosition = new Vector2(0, 53);
            textRectT.anchoredPosition = new Vector2(60, -25);
            valueInputFieldRectT.anchoredPosition = new Vector2(-65, -25);

            defaultValueButtonRectT.anchoredPosition = new Vector2(1, 18);
            defaultValueButtonRectT.sizeDelta = new Vector2(100, 32);
            defaultValueButtonRectT.GetComponentInChildren<Text>().fontSize = 20;

            ConfigureQuickButton(m1ButtonRectT);
            ConfigureQuickButton(m2ButtonRectT);
            ConfigureQuickButton(m3ButtonRectT);
            ConfigureQuickButton(p1ButtonRectT, true);
            ConfigureQuickButton(p2ButtonRectT, true);
            ConfigureQuickButton(p3ButtonRectT, true);

            // AddRect(panelT.name, panelRectT);
            // AddRect(sliderT.name, sliderRectT);
            // AddRect(textT.name, textRectT);
            // AddRect(valueInputFieldT.name, valueInputFieldRectT);
            // AddRect(defaultValueButtonT.name, defaultValueButtonRectT);
            // AddRect(m1ButtonT.name, m1ButtonRectT);
            // AddRect(m2ButtonT.name, m2ButtonRectT);
            // AddRect(m3ButtonT.name, m3ButtonRectT);
            // AddRect(p1ButtonT.name, p1ButtonRectT);
            // AddRect(p2ButtonT.name, p2ButtonRectT);
            // AddRect(p3ButtonT.name, p3ButtonRectT);
        }

        static void ConfigureQuickButton(RectTransform rectT, bool right = false)
        {
            rectT.GetComponentInChildren<Text>().fontSize = 20;
            rectT.anchoredPosition = new Vector2(right ? 18 : -17, 18);
            rectT.sizeDelta = new Vector2(0, 32);
        }

        [SuppressMessage("ReSharper", "MemberCanBeMadeStatic.Global")]
        internal void ConfigureColliderPreviewsOpacitySlider(UIDynamicSlider uiDynamic)
        {
            var panelT = uiDynamic.transform.Find("Panel");
            var sliderT = uiDynamic.transform.Find("Slider");
            // var textT = uiDynamicSlider.transform.Find("Text");
            // var valueInputFieldT = uiDynamicSlider.transform.Find("ValueInputField");

            uiDynamic.quickButtonsEnabled = false;
            uiDynamic.rangeAdjustEnabled = false;
            uiDynamic.defaultButtonEnabled = false;

            var panelRect = panelT.GetComponent<RectTransform>();
            panelRect.sizeDelta = new Vector2(panelRect.sizeDelta.x, -19);
            panelRect.anchoredPosition = new Vector2(panelRect.anchoredPosition.x, 9);

            foreach(var text in uiDynamic.GetComponentsInChildren<Text>())
            {
                text.color = Color.white;
            }

            var panelBackground = panelT.GetComponent<Image>();
            panelBackground.color = Colors.transparentGray;
            var sliderBackground = sliderT.GetComponent<Image>();
            sliderBackground.color = Colors.darkTransparentGray;
        }

        // ReSharper disable once MemberCanBeMadeStatic.Global
        internal void ConfigureCustomPopup(UIDynamicPopup uiDynamic, float height, bool upwards)
        {
            uiDynamic.popup.labelText.color = Color.black;

            if(height > 0f)
            {
                uiDynamic.popupPanelHeight = height;
            }

            if(upwards)
            {
                uiDynamic.popup.popupPanel.offsetMin += new Vector2(0, uiDynamic.popupPanelHeight + 60);
                uiDynamic.popup.popupPanel.offsetMax += new Vector2(0, uiDynamic.popupPanelHeight + 60);
            }
        }

        internal void AddPrevNextButtons(UIDynamicPopup uiDynamic, bool cycle = false)
        {
            var popup = uiDynamic.popup;
            popup.labelText.alignment = TextAnchor.UpperCenter;
            popup.labelText.GetComponent<RectTransform>().anchorMax = new Vector2(0, 0.89f);

            /* Prev */
            {
                var t = this.InstantiateButton(uiDynamic.transform);
                var uiDynamicButton = t.GetComponent<UIDynamicButton>();
                uiDynamicButton.label = "<";
                if(cycle)
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetPreviousOrLastValue();
                    });
                }
                else
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetPreviousValue();
                    });
                }

                var rectT = t.GetComponent<RectTransform>();
                rectT.pivot = Vector2.zero;
                rectT.anchoredPosition = new Vector2(10, 0);
                rectT.sizeDelta = Vector2.zero;
                rectT.offsetMin = new Vector2(5, 5f);
                rectT.offsetMax = new Vector2(80, 70);
                rectT.anchorMin = Vector2.zero;
                rectT.anchorMax = Vector2.zero;
            }

            /* Next */
            {
                var t = this.InstantiateButton(uiDynamic.transform);
                var uiDynamicButton = t.GetComponent<UIDynamicButton>();
                uiDynamicButton.label = ">";
                if(cycle)
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetNextOrFirstValue();
                    });
                }
                else
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetNextValue();
                    });
                }

                var rectT = t.GetComponent<RectTransform>();
                rectT.pivot = Vector2.zero;
                rectT.anchoredPosition = Vector2.zero;
                rectT.sizeDelta = Vector2.zero;
                rectT.offsetMin = new Vector2(82, 5);
                rectT.offsetMax = new Vector2(157, 70);
                rectT.anchorMin = Vector2.zero;
                rectT.anchorMax = Vector2.zero;
            }
        }

        internal T GetCustomElementAs<T>(string key)
        {
            UIDynamic element;
            if(_customElements.TryGetValue(key, out element))
            {
                if(element is T)
                {
                    return (T) Convert.ChangeType(element, typeof(T));
                }
            }

            return default(T);
        }

        internal void SetCalibrationPending(IStorableParam storableParam)
        {
            if(RestoringFromJSON)
            {
                return;
            }

            CalibrationStateHandler.SetCalibrationPending(storableParam.Name, !storableParam.IsCalibrated);
        }

        protected bool inCalibrationCo;
        internal bool FirstCalibrationDone { get; set; }
        protected Coroutine calibrationCo;

        internal void StartCalibration(bool forceUpdateMass = false, JSONClass calibrationData = null)
        {
            if(RestoringFromJSON || inCalibrationCo)
            {
                return;
            }

            if(!enabled)
            {
                Loggr.LogMessage($"Enable {DomainName} to calibrate physics and morphs.");
                return;
            }

            if(forceUpdateMass)
            {
                CalibrationStateHandler.UpdateMassWhenCalibrating = true;
            }

            calibrationCo = StartCoroutine(CalibrationCo(calibrationData));
        }

        protected virtual IEnumerator CalibrationCo(JSONClass calibrationData = null)
        {
            yield break;
        }

        internal virtual void SyncMorphsAngleOffsets(float? timer = null)
        {
        }

        // adapted from Timeline v4.3.1 (c) acidbubbles
        IEnumerator SelectPluginUICo(Action postAction = null)
        {
            while(Initialized == null)
            {
                yield return null;
            }

            if(Initialized == false)
            {
                yield break;
            }

            if(UITransform && UITransform.gameObject.activeInHierarchy)
            {
                if(enabled)
                {
                    postAction?.Invoke();
                }

                yield break;
            }

            yield return SelectContainingAtomTabCo("Plugins");

            float timeout = Time.unscaledTime + 1;
            while(Time.unscaledTime < timeout)
            {
                yield return null;

                if(!UITransform)
                {
                    continue;
                }

                /* Close any currently open plugin UI before opening this plugin's UI */
                foreach(Transform scriptController in manager.pluginContainer)
                {
                    var mvrScript = scriptController.gameObject.GetComponent<MVRScript>();
                    if(mvrScript && mvrScript != this)
                    {
                        mvrScript.UITransform.gameObject.SetActive(false);
                    }
                }

                if(enabled)
                {
                    UITransform.gameObject.SetActive(true);
                    postAction?.Invoke();
                    yield break;
                }
            }
        }

        internal void DeferOpenUIMain() => StartCoroutine(SelectPluginUICo(Tabs.GoToMainWindow));
        internal void DeferOpenUIPhysics() => StartCoroutine(SelectPluginUICo(Tabs.GoToPhysicsWindow));
        internal void DeferOpenUIMorphs() => StartCoroutine(SelectPluginUICo(Tabs.GoToMorphsWindow));
        internal void DeferOpenUIHardColliders() => StartCoroutine(SelectPluginUICo(Tabs.GoToHardCollidersWindow));

        internal void DeferOpenUIFriction()
        {
            if(enabled)
            {
                StartCoroutine(SelectContainingAtomTabCo(
                    "Skin Materials 2",
                    () => UIModManager.SwitchToOtherPluginAction.actionCallback()
                ));
            }
            else
            {
                StartCoroutine(SelectContainingAtomTabCo("Plugins"));
            }
        }

        internal void DeferOpenUIClothing() => StartCoroutine(SelectPluginUICo(Tabs.GoToClothingWindow));
        internal void DeferOpenUIDev() => StartCoroutine(SelectPluginUICo(Tabs.GoToDevWindow));
        internal void DeferOpenUIDevMorph() => StartCoroutine(SelectPluginUICo(Tabs.GoToDevMorphWindow));
        internal void DeferOpenUI() => StartCoroutine(SelectPluginUICo());

        IEnumerator SelectContainingAtomTabCo(string tabName, Action postAction = null)
        {
            if(SuperController.singleton.gameMode != SuperController.GameMode.Edit)
            {
                SuperController.singleton.gameMode = SuperController.GameMode.Edit;
            }

            SuperController.singleton.SelectController(containingAtom.mainController, false, false);
            SuperController.singleton.ShowMainHUDAuto();

            float timeout = Time.unscaledTime + 1;
            while(Time.unscaledTime < timeout)
            {
                yield return null;
                var selector = containingAtom.gameObject.GetComponentInChildren<UITabSelector>();
                if(selector)
                {
                    selector.SetActiveTab(tabName);
                    postAction?.Invoke();
                    break;
                }
            }
        }

        internal void AddTextFieldToJss(UIDynamicTextField textField, JSONStorableString jss)
        {
            jss.dynamicText = textField;
            textFieldToJSONStorableString.Add(textField, jss);
        }

        internal void AddToggleToJsb(UIDynamicToggle toggle, JSONStorableBool jsb)
        {
            jsb.toggle = toggle.toggle;
            toggleToJSONStorableBool.Add(toggle, jsb);
        }

        internal void AddSliderToJsf(UIDynamicSlider slider, JSONStorableFloat jsf)
        {
            jsf.slider = slider.slider;
            sliderToJSONStorableFloat.Add(slider, jsf);
        }

        internal void DeferSetSaveFileBrowser(string fileExt) => StartCoroutine(SetSaveFileBrowserCo(fileExt));

        static IEnumerator SetSaveFileBrowserCo(string fileExt)
        {
            yield return null;
            var browser = SuperController.singleton.mediaFileBrowserUI;
            browser.browseVarFilesAsDirectories = false;
            browser.SetTextEntry(true);
            browser.fileEntryField.text = $"{((int) (DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString()}.{fileExt}";
            browser.ActivateFileNameField();
        }

        internal void PreRegisterGeneralParam(IStorableParam param) => PreRegisterParam(StorableDomainName.GENERAL, param);
        internal void PreRegisterJointPhysicsParam(IStorableParam param) => PreRegisterParam(StorableDomainName.JOINT_PHYSICS, param);
        internal void PreRegisterSoftPhysicsParam(IStorableParam param) => PreRegisterParam(StorableDomainName.SOFT_PHYSICS, param);
        internal void PreRegisterForcePhysicsParam(IStorableParam param) => PreRegisterParam(StorableDomainName.FORCE_PHYSICS, param);
        internal void PreRegisterHardCollidersParam(IStorableParam param) => PreRegisterParam(StorableDomainName.HARD_COLLIDERS, param);
        internal void PreRegisterMorphsParam(IStorableParam param) => PreRegisterParam(StorableDomainName.MORPHS, param);
        internal void PreRegisterFrictionParam(IStorableParam param) => PreRegisterParam(StorableDomainName.FRICTION, param);
        internal void PreRegisterUIActionsParam(IStorableParam param) => PreRegisterParam(StorableDomainName.UI_ACTIONS, param);

        internal void PreRegisterInternalParam(IStorableParam param, bool temporary = true)
        {
            param.Temporary = temporary;
            PreRegisterParam(StorableDomainName.INTERNAL, param);
        }

        internal void PreRegisterClothingParam(IStorableParam param) => PreRegisterParam(StorableDomainName.CLOTHING, param);

        void PreRegisterParam(string domainName, IStorableParam param)
        {
            if(param == null)
            {
                throw new ArgumentException($"Parameter null! Cannot pre-register to {domainName}");
            }

            _storableDomainsByName[domainName].PreRegister(param);
        }

        protected void RegisterParams()
        {
            foreach(Handler handler in Provider.GetHandlers())
            {
                handler.PreRegisterParams();
            }

            foreach(var domain in _storableDomains)
            {
                domain.RegisterAllTo(this);
            }
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            foreach(var domain in _storableDomains)
            {
                domain.RemoveTemporaryParamsFrom(jc);
            }

            return jc;
        }

        internal JSONClass GetFullBaseJSON() => base.GetJSON(true, true, true);

        internal void SavePresetJSON(JSONClass jc, string fullPath)
        {
            jc.Remove("enabled");
            SaveJSON(jc, fullPath);
            SetLoadedPresetName(fullPath);
        }

        internal virtual void RestoreFromPresetJSON(JSONClass jc, string fullPath)
        {
        }

        internal virtual void InitializeFromPresetJSON(JSONClass jc, string fullPath)
        {
        }

        ParsedPath ParsePresetPath(string fullPath) => new ParsedPath(fullPath, $"{FileUtils.savesDir}/{DomainName}Presets");

        protected void SetLoadedPresetName(string fullPath)
        {
            string loadedPresetName = "None";
            if(!string.IsNullOrEmpty(fullPath))
            {
                var parsedPath = ParsePresetPath(fullPath);
                loadedPresetName = parsedPath.fileName.Replace(".json", "");
                if(!string.IsNullOrEmpty(parsedPath.packageId))
                {
                    loadedPresetName += $" ({parsedPath.packageId}.var)";
                }
            }

            LoadedPresetName = loadedPresetName;
            Tabs.MainWindow?.Refresh();
        }

        protected void OnDestroy()
        {
            DestroyImmediate(pluginUIEventsListener);
            pluginUIEventsListener = null;

            Destroy(CalibrationStateHandler);
            CalibrationStateHandler = null;

            Destroy(UIModManager);
            UIModManager = null;
        }

        #if ENV_DEVELOPMENT

        bool _devUICreated;
        const float SCALE_MULT = 100f;
        RectTransform _selectedRect;
        readonly Dictionary<string, RectTransform> _rectTransforms = new Dictionary<string, RectTransform>();
        readonly JSONStorableStringChooser _rectJssc = new JSONStorableStringChooser("Rect", new List<string>(), Constant.SELECT, "Rect");
        readonly JSONStorableFloat _posX = new JSONStorableFloat("Pos X", 0.00f, -20.00f, 20.00f);
        readonly JSONStorableFloat _posY = new JSONStorableFloat("Pos Y", 0.00f, -20.00f, 20.00f);
        readonly JSONStorableFloat _sizeX = new JSONStorableFloat("Size X", 0.00f, -15.00f, 15.00f);
        readonly JSONStorableFloat _sizeY = new JSONStorableFloat("Size Y", 0.00f, -15.00f, 15.00f);

        void AddRect(string name, RectTransform rectT) => _rectTransforms[$"{name}_{Guid.NewGuid().ToString().Substring(0, 4)}"] = rectT;

        void UpdateRectOptions()
        {
            var options = new List<string> { Constant.SELECT };
            options.AddRange(_rectTransforms.Keys);
            _rectJssc.choices = options;
        }

        [SuppressMessage("ReSharper", "UnusedMember.Global")]
        void CreateDevUI()
        {
            var rectPopup = CreateCustomFilterablePopup(_rectJssc.name, _rectJssc.name, new Vector2(1475, -1050), new Vector2(-555, 120), false);
            ConfigureCustomPopup(rectPopup, 800f, true);
            AddPrevNextButtons(rectPopup);
            _rectJssc.setCallbackFunction = SelectRect;
            _rectJssc.popup = rectPopup.popup;

            var syncButton = CreateCustomButton("Sync Options", new Vector2(2000, -1045), new Vector2(-921, 110), false);
            syncButton.AddListener(UpdateRectOptions);

            var posX = CreateCustomSlider(_posX.name, new Vector2(1475, -1200), false);
            posX.valueFormat = "F2";
            _posX.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.anchoredPosition = SCALE_MULT * new Vector2(value, _posY.val);
                }
            };
            _posX.slider = posX.slider;

            var posY = CreateCustomSlider(_posY.name, new Vector2(1475, -1350), false);
            posY.valueFormat = "F2";
            _posY.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.anchoredPosition = SCALE_MULT * new Vector2(_posX.val, value);
                }
            };
            _posY.slider = posY.slider;

            var sizeX = CreateCustomSlider(_sizeX.name, new Vector2(1475, -1500), false);
            sizeX.valueFormat = "F2";
            _sizeX.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.sizeDelta = SCALE_MULT * new Vector2(value, _sizeY.val);
                }
            };
            _sizeX.slider = sizeX.slider;

            var sizeY = CreateCustomSlider(_sizeY.name, new Vector2(1475, -1650), false);
            sizeY.valueFormat = "F2";
            _sizeY.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.sizeDelta = SCALE_MULT * new Vector2(_sizeX.val, value);
                }
            };
            _sizeY.slider = sizeY.slider;

            _devUICreated = true;
        }

        void SelectRect(string value)
        {
            _selectedRect = _rectTransforms[value];
            var anchoredPosition = _selectedRect.anchoredPosition;
            _posX.defaultVal = anchoredPosition.x / SCALE_MULT;
            _posY.defaultVal = anchoredPosition.y / SCALE_MULT;
            _posX.val = _posX.defaultVal;
            _posY.val = _posY.defaultVal;
            var sizeDelta = _selectedRect.sizeDelta;
            _sizeX.defaultVal = sizeDelta.x / SCALE_MULT;
            _sizeY.defaultVal = sizeDelta.y / SCALE_MULT;
            _sizeX.val = _sizeX.defaultVal;
            _sizeY.val = _sizeY.defaultVal;
        }

        #endif
    }
}
