using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using System.Text;
using UnityEngine;

namespace NAT.Common.UI
{
    static class ClothingUIHelpers
    {
        internal static string CurrentPreset(StringBuilder sb, ClothingAdjustmentProfile selectedProfile) => sb
            .Clear()
            .Append("Current: ")
            .Append(selectedProfile?.LoadedName ?? "None")
            .AppendColor(selectedProfile?.IsSaved ?? true ? "" : " (unsaved)", Color.red)
            .ToString();

        internal static string TransitionInfo(StringBuilder sb, ClothingAdjustmentProfile selectedProfile)
        {
            if(selectedProfile == null || !selectedProfile.HasTarget)
            {
                return "";
            }

            return sb
                .Clear()
                .Append("Transition: ")
                .Append(Calc.RoundToDecimals(selectedProfile.Transition.StorableParam.val, 1))
                .Append(" %")
                .ToString();
        }

        internal static string BuildAdjustmentsInfo(StringBuilder sb, ClothingHandler clothingHandler, SoftPhysicsHandler softPhysicsHandler, bool isBootyMagic)
        {
            sb.Clear();
            sb.AppendSize("\n", 8).AppendBold("Active profiles").AppendLine();

            foreach(var profile in clothingHandler.ActiveProfiles)
            {
                string itemName = profile.GetItemName();
                if(profile.Symmetric.Value)
                {
                    sb.AppendFormat("  {0}", itemName).AppendLine();
                }
                else
                {
                    if(profile.LeftActive.Value)
                    {
                        sb.AppendFormat("  (L) {0}\n", itemName).AppendLine();
                    }

                    if(profile.RightActive.Value)
                    {
                        sb.AppendFormat("  (R) {0}\n", itemName).AppendLine();
                    }
                }
            }

            AppendForcesInfo(sb, clothingHandler);
            AppendHoldInfo(sb, clothingHandler);
            AppendDampeningInfo(sb, clothingHandler);
            AppendSoftPhysicsInfo(sb, softPhysicsHandler);
            if(isBootyMagic)
            {
                AppendGluteMorphsOffsetsInfo(sb, clothingHandler);
            }
            else
            {
                AppendBreastMorphsOffsetsInfo(sb, clothingHandler);
                AppendExtraMorphsInfo(sb, clothingHandler);
            }

            return sb.ToString();
        }

        static void AppendForcesInfo(StringBuilder sb, ClothingHandler clothingHandler)
        {
            sb.AppendSize("\n", 8).AppendBold("Forces").AppendLine(" (m/s^2)");
            var leftForce = clothingHandler.LeftForce;
            var rightForce = clothingHandler.RightForce;
            if(Mathf.Abs(leftForce.y) > 0.001f || Mathf.Abs(rightForce.y) > 0.001f)
            {
                sb.AppendFormat("  Y: {0:0.00} (L), {1:0.00} (R)", leftForce.y, rightForce.y).AppendLine();
            }

            if(Mathf.Abs(leftForce.x) > 0.001f || Mathf.Abs(rightForce.x) > 0.001f)
            {
                sb.AppendFormat("  X: {0:0.00} (L), {1:0.00} (R)", leftForce.x, -rightForce.x).AppendLine();
            }

            if(Mathf.Abs(leftForce.z) > 0.001f || Mathf.Abs(rightForce.z) > 0.001f)
            {
                sb.AppendFormat("  Z: {0:0.00} (L), {1:0.00} (R)", -leftForce.z, -rightForce.z).AppendLine();
            }
        }

        static void AppendHoldInfo(StringBuilder sb, ClothingHandler clothingHandler)
        {
            sb.AppendSize("\n", 8).AppendBold("Hold").AppendLine();
            if(Mathf.Abs(clothingHandler.LeftYXHoldPct) > 0.1f || Mathf.Abs(clothingHandler.RightYXHoldPct) > 0.1f)
            {
                sb.AppendFormat("  YX: {0:0}% (L), {1:0}% (R)", clothingHandler.LeftYXHoldPct, clothingHandler.RightYXHoldPct).AppendLine();
            }

            if(Mathf.Abs(clothingHandler.LeftZHoldPct) > 0.1f || Mathf.Abs(clothingHandler.RightZHoldPct) > 0.1f)
            {
                sb.AppendFormat("  Z: {0:0}% (L), {1:0}% (R)", clothingHandler.LeftZHoldPct, clothingHandler.RightZHoldPct).AppendLine();
            }
        }

        static void AppendDampeningInfo(StringBuilder sb, ClothingHandler clothingHandler)
        {
            sb.AppendSize("\n", 8).AppendBold("Dampening").AppendLine();
            if(Mathf.Abs(clothingHandler.LeftYXDampeningPct) > 0.1f || Mathf.Abs(clothingHandler.RightYXDampeningPct) > 0.1f)
            {
                sb.AppendFormat("  YX: {0:0}% (L), {1:0}% (R)", clothingHandler.LeftYXDampeningPct, clothingHandler.RightYXDampeningPct).AppendLine();
            }

            if(Mathf.Abs(clothingHandler.LeftZDampeningPct) > 0.1f || Mathf.Abs(clothingHandler.RightZDampeningPct) > 0.1f)
            {
                sb.AppendFormat("  Z: {0:0}% (L), {1:0}% (R)", clothingHandler.LeftZDampeningPct, clothingHandler.RightZDampeningPct).AppendLine();
            }
        }

        static void AppendSoftPhysicsInfo(StringBuilder sb, SoftPhysicsHandler softPhysicsHandler)
        {
            sb.AppendSize("\n", 8).AppendBold("Covered soft physics").AppendLine();
            foreach(var group in softPhysicsHandler.SoftParamGroups)
            {
                float left = group.LeftClothingAdjustmentAverage;
                float right = group.RightClothingAdjustmentAverage;
                float tolerance = Utils.ValueFormatToTolerance(group.ValueFormat);

                if(Mathf.Abs(left) > tolerance || Mathf.Abs(right) > tolerance)
                {
                    string leftSign = left >= 0f ? "+" : "";
                    string rightSign = right >= 0f ? "+" : "";
                    sb.AppendFormat("  {0}", group.Name).AppendLine();
                    sb.AppendFormat("    {0}{1} (L), {2}{3} (R)", leftSign, left.ToString(group.ValueFormat), rightSign, right.ToString(group.ValueFormat)).AppendLine();
                }
            }
        }

        static void AppendGluteMorphsOffsetsInfo(StringBuilder sb, ClothingHandler clothingHandler)
        {
            sb.AppendSize("\n", 8).AppendBold("Morphs offsets").AppendLine();
            var leftOffsets = clothingHandler.LeftMorphsOffsets;
            var rightOffsets = clothingHandler.RightMorphsOffsets;

            if(Mathf.Abs(leftOffsets.y) > 0.001f || Mathf.Abs(rightOffsets.y) > 0.001f)
            {
                sb.AppendFormat("  Y: {0:0.00}° (L), {1:0.00}° (R)", -leftOffsets.y, -rightOffsets.y).AppendLine();
            }

            if(Mathf.Abs(leftOffsets.x) > 0.001f || Mathf.Abs(rightOffsets.x) > 0.001f)
            {
                sb.AppendFormat("  X: {0:0.00}° (L), {1:0.00}° (R)", -leftOffsets.x, rightOffsets.x).AppendLine();
            }

            float leftZ = 1000f * -leftOffsets.z;
            float rightZ = 1000f * -rightOffsets.z;
            if(Mathf.Abs(leftZ) > 0.001f || Mathf.Abs(rightZ) > 0.01f)
            {
                sb.AppendFormat("  Z: {0:0.0}mm (L), {1:0.0}mm (R)", leftZ, rightZ).AppendLine();
            }
        }

        static void AppendBreastMorphsOffsetsInfo(StringBuilder sb, ClothingHandler clothingHandler)
        {
            sb.AppendSize("\n", 8).AppendBold("Morphs offsets").AppendLine();
            var leftOffsets = clothingHandler.LeftMorphsOffsets;
            var rightOffsets = clothingHandler.RightMorphsOffsets;

            if(Mathf.Abs(leftOffsets.y) > 0.001f || Mathf.Abs(rightOffsets.y) > 0.001f)
            {
                sb.AppendFormat("  Y: {0:0.00}° (L), {1:0.00}° (R)", leftOffsets.y, rightOffsets.y).AppendLine();
            }

            if(Mathf.Abs(leftOffsets.x) > 0.001f || Mathf.Abs(rightOffsets.x) > 0.001f)
            {
                sb.AppendFormat("  X: {0:0.00}° (L), {1:0.00}° (R)", leftOffsets.x, -rightOffsets.x).AppendLine();
            }

            float leftZ = 1000f * leftOffsets.z;
            float rightZ = 1000f * rightOffsets.z;
            if(Mathf.Abs(leftZ) > 0.001f || Mathf.Abs(rightZ) > 0.01f)
            {
                sb.AppendFormat("  Z: {0:0.0}mm (L), {1:0.0}mm (R)", leftZ, rightZ).AppendLine();
            }
        }

        static void AppendExtraMorphsInfo(StringBuilder sb, ClothingHandler clothingHandler)
        {
            sb.AppendSize("\n", 8).AppendBold("Extra morphs").AppendLine();
            foreach(var pair in clothingHandler.MorphConfigs)
            {
                var config = pair.Value;
                var leftConfig = config.Left;
                var rightConfig = config.Right;
                if(Mathf.Abs(leftConfig.morphValue) > 0.0001f || Mathf.Abs(rightConfig.morphValue) > 0.0001f)
                {
                    sb.AppendFormat("  {0}", pair.Key).AppendLine();
                    sb.AppendFormat("    {0:0.000} (L), {1:0.000} (R)", leftConfig.morphValue, rightConfig.morphValue).AppendLine();
                }
            }
        }

        internal static string TransitionStatus(StringBuilder sb, ClothingHandler clothingHandler) => sb
            .Clear()
            .AppendFormat("From: {0}", clothingHandler.SelectedProfile?.LoadedName ?? "None")
            .AppendLine()
            .AppendFormat("To: {0}", clothingHandler.SelectedProfile?.TargetName ?? "")
            .ToString();
    }
}
