using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;

namespace NAT.Common.UI
{
    sealed class Tabs
    {
        readonly ScriptBase _script;
        Window _activeTabWindow;
        internal Window MainWindow { get; set; }
        internal Window PhysicsWindow { get; set; }
        internal Window MorphsWindow { get; set; }
        internal Window HardCollidersWindow { get; set; }
        internal Window ClothingWindow { get; set; }

        internal Window GetActiveWindow() => _activeTabWindow?.ActiveNestedWindow ?? _activeTabWindow;
        internal Window GetActiveNestedWindow() => _activeTabWindow?.ActiveNestedWindow;

        // ReSharper disable UnusedMember.Global

        internal bool IsMainTabActive() => _activeTabWindow == MainWindow;
        internal bool IsPhysicsTabActive() => _activeTabWindow == PhysicsWindow;

        internal bool IsMorphsTabActive() => _activeTabWindow == MorphsWindow;
        internal bool IsHardCollidersTabActive() => _activeTabWindow == HardCollidersWindow;
        internal bool IsClothingTabActive() => _activeTabWindow == ClothingWindow;
        internal bool IsNestedWindowActive() => GetActiveNestedWindow() != null;

        // ReSharper restore UnusedMember.Global

        internal bool IsActive(Window window) =>
            window == _activeTabWindow ? !IsNestedWindowActive() : GetActiveNestedWindow() == window;

        readonly Dictionary<Window, NavigationButton> _tabButtons;

        internal Tabs(ScriptBase script)
        {
            _script = script;
            _tabButtons = new Dictionary<Window, NavigationButton>();
        }

        internal void CreateNavigationButtons()
        {
            const int spacing = 215;
            CreateNavigationButton(MainWindow, _script.DomainName, 2f, GoToMainWindow);
            CreateNavigationButton(PhysicsWindow, "Physics", 2f + spacing, GoToPhysicsWindow);
            CreateNavigationButton(MorphsWindow, "Morphs", 2f + 2 * spacing, GoToMorphsWindow);
            CreateNavigationButton(HardCollidersWindow, "Hard Colliders", 2f + 3 * spacing, GoToHardCollidersWindow);
            CreateNavigationButton(ClothingWindow, "Clothing", 2f + 4 * spacing, GoToClothingWindow);
            _script.CreateSpacer().height = 50f;
            _script.CreateSpacer(true).height = 50f;
        }

        void CreateNavigationButton(Window window, string label, float posX, UnityAction callback)
        {
            var element = new NavigationButton(
                _script.CreateCustomButton(label, new Vector2(posX, -72), new Vector2(-864, 60)),
                _script.DomainColor,
                Colors.darkerGray
            );
            element.AddListener(callback);
            element.SetInactive();
            _tabButtons[window] = element;
        }

        internal void GoToMainWindow() => NavigateToTab(MainWindow);
        internal void GoToPhysicsWindow() => NavigateToTab(PhysicsWindow);
        internal void GoToMorphsWindow() => NavigateToTab(MorphsWindow);

        internal void GoToHardCollidersWindow()
        {
            if(!_script.enabled)
            {
                _script.Loggr.LogMessage("Enable the plugin to configure hard colliders.");
                return;
            }

            NavigateToTab(HardCollidersWindow);
        }

        internal void GoToClothingWindow() => NavigateToTab(ClothingWindow);

        internal void GoToDevWindow()
        {
            GoToMainWindow();
            if(_script.IsTittyMagic)
            {
                MainWindow.OpenNestedWindow<NAT.TM.UI.DevWindow>();
            }
            else if(_script.IsBootyMagic)
            {
                MainWindow.OpenNestedWindow<NAT.BM.UI.DevWindow>();
            }
        }

        internal void GoToDevMorphWindow()
        {
            GoToMainWindow();
            if(_script.IsTittyMagic)
            {
                MainWindow.OpenNestedWindow<NAT.TM.UI.DevMorphWindow>();
            }
            else if(_script.IsBootyMagic)
            {
                MainWindow.OpenNestedWindow<NAT.BM.UI.DevMorphWindow>();
            }
        }

        void NavigateToTab(Window window)
        {
            _activeTabWindow?.Clear();
            _activeTabWindow = window;
            window.Rebuild();

            var navigationButton = _tabButtons[window];
            navigationButton.SetActive();
            foreach(var pair in _tabButtons)
            {
                if(pair.Key != window)
                {
                    pair.Value.SetInactive();
                }
            }
        }

        internal void ClosePopups() => _activeTabWindow.ClosePopups();
    }
}
