using NAT.Common.Models;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.Common.UI
{
    abstract class Window
    {
        protected readonly ScriptBase script;
        readonly Dictionary<string, UIDynamic> _elements;
        readonly Dictionary<string, Element> _customElements;
        readonly List<Window> _nestedWindows;
        protected Window parent;
        internal Window ActiveNestedWindow { get; set; }
        protected bool isBuilt;
        protected readonly StringBuilder diagnosticsSb = new StringBuilder();

        protected Window(ScriptBase script)
        {
            this.script = script;
            _elements = new Dictionary<string, UIDynamic>();
            _customElements = new Dictionary<string, Element>();
            _nestedWindows = new List<Window>();
        }

        protected void AddSpacer(float height, bool rightSide) => AddElement(() => script.NewSpacer(height, rightSide));
        protected void AddElement(Func<UIDynamic> createUIDynamic) => AddElement(Guid.NewGuid().ToString(), createUIDynamic);
        protected void AddElement(string key, Func<UIDynamic> createUIDynamic) => _elements[key] = createUIDynamic();

        protected void AddToggle(
            JSONStorableBool jsb,
            List<Element> listToAdd,
            Func<JSONStorableBool, UIDynamic> createUIDynamic
        )
        {
            var element = AddToggle(jsb, createUIDynamic);
            if(element != null)
            {
                listToAdd.Add(element);
            }
        }

        protected Element AddToggle(JSONStorableBool jsb, Func<JSONStorableBool, UIDynamic> createUIDynamic, bool activeInNested = false)
        {
            var element = new Element(createUIDynamic(jsb), activeInNested);
            _customElements[$"{jsb.name} [{typeof(UIDynamicToggle)}]"] = element;
            return element;
        }

        protected void AddSlider(
            JSONStorableFloat jsf,
            List<Element> listToAdd,
            Func<JSONStorableFloat, UIDynamic> createUIDynamic
        )
        {
            var element = AddSlider(jsf, createUIDynamic);
            if(element != null)
            {
                listToAdd.Add(element);
            }
        }

        protected Element AddSlider(JSONStorableFloat jsf, Func<JSONStorableFloat, UIDynamic> createUIDynamic, bool activeInNested = false)
        {
            var element = new Element(createUIDynamic(jsf), activeInNested);
            _customElements[$"{jsf.name} [{typeof(UIDynamicSlider)}]"] = element;
            return element;
        }

        protected void AddButton(
            string name,
            List<Element> listToAdd,
            Func<string, UIDynamic> createUIDynamic
        )
        {
            var element = AddButton(name, createUIDynamic);
            if(element != null)
            {
                listToAdd.Add(element);
            }
        }

        protected Element AddButton(string name, Func<string, UIDynamic> createUIDynamic, bool activeInNested = false)
        {
            var element = new Element(createUIDynamic(name), activeInNested);
            _customElements[$"{name} [{typeof(UIDynamicButton)}]"] = element;
            return element;
        }

        protected void AddTextField(
            string name,
            List<Element> listToAdd,
            Func<string, UIDynamic> createUIDynamic
        )
        {
            var element = AddTextField(name, createUIDynamic);
            if(element != null)
            {
                listToAdd.Add(element);
            }
        }

        protected Element AddTextField(string name, Func<string, UIDynamic> createUIDynamic, bool activeInNested = false)
        {
            var element = new Element(createUIDynamic(name), activeInNested);
            _customElements[$"{name} [{typeof(UIDynamicTextField)}]"] = element;
            return element;
        }

        protected void AddPopup(
            JSONStorableStringChooser jssc,
            List<Element> listToAdd,
            Func<JSONStorableStringChooser, UIDynamic> createUIDynamic
        )
        {
            var element = AddPopup(jssc, createUIDynamic);
            if(element != null)
            {
                listToAdd.Add(element);
            }
        }

        protected Element AddPopup(JSONStorableStringChooser jssc, Func<JSONStorableStringChooser, UIDynamic> createUIDynamic, bool activeInNested = false)
        {
            var element = new Element(createUIDynamic(jssc), activeInNested);
            _customElements[$"{jssc.name} [{typeof(UIDynamicPopup)}]"] = element;
            return element;
        }

        protected void AddNestedWindow(Window window)
        {
            window.parent = this;
            _nestedWindows.Add(window);
        }

        void AddBackButton(bool rightSide) => AddElement(() =>
        {
            var button = script.CreateButton("Return", rightSide);
            button.textColor = Color.white;
            var colors = button.button.colors;
            colors.normalColor = Colors.sliderGray;
            colors.highlightedColor = Color.grey;
            colors.pressedColor = Color.grey;
            button.button.colors = colors;
            button.AddListener(parent.OnReturn);
            return button;
        });

        UIDynamicTextField CreateBasicTextField(string text, bool rightSide) =>
            script.CreateTextField(new JSONStorableString("text", text), rightSide);

        protected UIDynamicTextField CreateHeaderTextField(
            string text,
            int fontSize,
            float height,
            bool rightSide
        )
        {
            var textField = CreateBasicTextField(text, rightSide);
            ModifyTextField(textField, fontSize, height);
            textField.UItext.alignment = TextAnchor.MiddleCenter;
            return textField;
        }

        protected UIDynamicTextField CreateValueTextField(
            JSONStorableString jss,
            int fontSize,
            float height,
            bool rightSide
        )
        {
            var textField = script.CreateTextField(jss, rightSide);
            ModifyTextField(textField, fontSize, height);
            textField.UItext.alignment = TextAnchor.MiddleLeft;
            return textField;
        }

        static void ModifyTextField(UIDynamicTextField textField, int fontSize, float height)
        {
            textField.UItext.fontSize = fontSize;
            textField.backgroundColor = Color.clear;
            var layout = textField.GetComponent<LayoutElement>();
            layout.preferredHeight = height;
            layout.minHeight = height;
        }

        protected void AddHeaderTextField(string text, bool rightSide) => AddElement(() =>
        {
            var textField = CreateHeaderTextField("\n".Size(20) + text.Bold(), 30, 60, rightSide);
            textField.UItext.alignment = TextAnchor.LowerCenter;
            return textField;
        });

        protected void AddLabelTextField(string text, bool rightSide) => AddElement(() =>
        {
            var textField = CreateBasicTextField(text.Bold(), rightSide);
            textField.UItext.fontSize = 26;
            textField.backgroundColor = Color.clear;
            var layout = textField.GetComponent<LayoutElement>();
            layout.preferredHeight = 35;
            layout.minHeight = 35;
            return textField;
        });

        protected void AddInfoTextField(string text, bool rightSide, float height = 100, int fontSize = 26) => AddElement(() =>
        {
            var textField = CreateBasicTextField(text, rightSide);
            textField.UItext.fontSize = fontSize;
            textField.backgroundColor = Color.clear;
            var layout = textField.GetComponent<LayoutElement>();
            layout.preferredHeight = height;
            layout.minHeight = height;
            return textField;
        });

        internal void Rebuild()
        {
            if(ActiveNestedWindow != null)
            {
                ActiveNestedWindow.Rebuild();
            }
            else
            {
                // TODO all elements as custom
                foreach(var pair in _customElements)
                {
                    pair.Value.ShowPrev();
                }

                if(parent != null)
                {
                    AddBackButton(false);
                    foreach(var pair in parent._customElements)
                    {
                        var element = pair.Value;
                        if(element.showInNested)
                        {
                            element.ShowPrev();
                        }
                    }
                }

                OnBuild();
                isBuilt = true;
                PostBuild();
            }
        }

        protected abstract void OnBuild();

        protected virtual void PostBuild()
        {
        }

        protected virtual void OnClose()
        {
        }

        internal void OnReturn()
        {
            ActiveNestedWindow.Clear();
            ActiveNestedWindow = null;
            Rebuild();
        }

        protected UIDynamic GetUIDynamic(string key) => _elements.ContainsKey(key) ? _elements[key] : null;

        protected T GetUIDynamicAs<T>(string key)
        {
            UIDynamic element;
            if(_elements.TryGetValue(key, out element))
            {
                if(element is T)
                {
                    return (T) Convert.ChangeType(element, typeof(T));
                }
            }

            return default(T);
        }

        protected Element GetElement<T>(string name) where T : UIDynamic
        {
            Element element;
            _customElements.TryGetValue($"{name} [{typeof(T)}]", out element);
            return element;
        }

        internal void OpenNestedWindow<T>() where T : Window
        {
            ActiveNestedWindow?.Clear();
            ActiveNestedWindow = null;
            Clear();

            var nestedWindow = _nestedWindows.FirstOrDefault(window => window is T);
            if(nestedWindow == null)
            {
                script.Loggr.LogError($"Window '{typeof(T)}' not found");
                return;
            }

            ActiveNestedWindow = nestedWindow;
            ActiveNestedWindow.Rebuild();
        }

        internal virtual void OnMainParamAdjusted(PhysicsParamGroup group)
        {
        }

        internal virtual void OnSoftParamAdjusted(PhysicsParamGroup group)
        {
        }

        internal virtual void UpdateDiagnostics()
        {
        }

        // TODO execute on UI opened, prevent if UI not open
        internal virtual void Refresh()
        {
        }

        internal void ClosePopups()
        {
            if(ActiveNestedWindow != null)
            {
                ActiveNestedWindow.ClosePopups();
            }
            else
            {
                ClosePopupsSelf();
            }
        }

        internal virtual void Clear()
        {
            if(ActiveNestedWindow != null)
            {
                ActiveNestedWindow.Clear();
                // ActiveNestedWindow = null; // Do not set null - it gets rebuilt on return
            }
            else
            {
                ClearSelf();
                if(parent != null)
                {
                    ClearParent();
                }
            }

            OnClose();
        }

        protected void ClearSelf()
        {
            foreach(var pair in _elements)
            {
                script.RemoveElement(pair.Value);
            }

            foreach(var pair in _customElements)
            {
                pair.Value.Hide();
            }
        }

        void ClearParent()
        {
            foreach(var pair in parent._customElements)
            {
                var element = pair.Value;
                if(element.showInNested)
                {
                    element.Hide();
                }
            }
        }

        void ClosePopupsSelf()
        {
            foreach(var pair in _elements)
            {
                (pair.Value as UIDynamicPopup)?.SetVisible(false);
            }

            foreach(var pair in _customElements)
            {
                pair.Value.ClosePopup();
            }
        }
    }
}
