using ColliderEditor;
using NAT.Common;
using NAT.Common.Components;
using NAT.Common.Handlers;
using NAT.Common.Models;
using NAT.Common.Storables;
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.TM.Handlers
{
    sealed class BreastColliderHandler : HardColliderHandler
    {
        BreastJointPhysicsHandler _jointPhysicsHandler;
        BreastSoftPhysicsHandler _softPhysicsHandler;
        List<AutoColliderPair> _autoColliderPairs;

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            relativeToRigidbody = Person.ChestRb;
            _jointPhysicsHandler = provider.GetInstance<BreastJointPhysicsHandler>();
            _softPhysicsHandler = provider.GetInstance<BreastSoftPhysicsHandler>();
            InitColliderVisualizer();
            LoadSettings();

            var options = HardColliderGroups.Select(c => c.VisualizerId).ToList();
            var displayOptions = HardColliderGroups.Select(c => c.Id).ToList();
            ColliderGroupsJssc = new JSONStorableStringChooser(
                "colliderGroup",
                options,
                displayOptions,
                "",
                "Collider"
            )
            {
                setCallbackFunction = value => ColliderVisualizer.SelectablesJSON.val = value,
                val = options[0],
            };

            ColliderVisualizer.GroupsJSON.setJSONCallbackFunction = jssc =>
            {
                ColliderVisualizer.SelectedPreviewOpacityJSON.val = jssc.val == "Off" ? 0 : 1;
                if(jssc.val != "Off")
                {
                    ColliderVisualizer.SelectablesJSON.val = ColliderGroupsJssc.val;
                }
            };

            BaseMassMultiplierJsf = new StorableFloat($"Hard Colliders Base {Constant.MASS_MULTIPLIER}", Person.IsFemale ? 1.00f : 0.50f, 0.01f, 2.00f)
            {
                setCallbackFunction = _ => SyncCollidersMass(),
            };
            HighlightAllJsb = new JSONStorableBool(
                "Highlight All Colliders",
                false,
                value => ColliderVisualizer.PreviewOpacityJSON.val = value ? 1.00f : 0.67f
            );
        }

        void InitColliderVisualizer()
        {
            ColliderVisualizer = Script.gameObject.AddComponent<ColliderVisualizer>();
            var groups = new List<Group>
            {
                new Group("Off", @"$off"), //match nothing
                new Group("Both breasts", @"[lr](Pectoral\d)"),
                new Group("Left breast", @"lPectoral\d"),
            };
            ColliderVisualizer.Init(Script, groups);
            ColliderVisualizer.PreviewOpacityJSON.val = 0.67f;
            ColliderVisualizer.PreviewOpacityJSON.defaultVal = 0.67f;
            ColliderVisualizer.SelectedPreviewOpacityJSON.val = 1;
            ColliderVisualizer.SelectedPreviewOpacityJSON.defaultVal = 1;
            ColliderVisualizer.GroupsJSON.val = "Left breast";
            ColliderVisualizer.GroupsJSON.defaultVal = "Left breast";
            ColliderVisualizer.HighlightMirrorJSON.val = true;

            foreach(string option in new[] { Constant.SELECT, "Other", "All" })
            {
                ColliderVisualizer.GroupsJSON.choices.Remove(option);
            }
        }

        protected override void LoadSettings()
        {
            var leftAutoColliders = Person.LeftPectoralAutoColliderGroup.GetAutoColliders();
            var rightAutoColliders = Person.RightPectoralAutoColliderGroup.GetAutoColliders();

            _autoColliderPairs = new List<AutoColliderPair>();
            for(int i = 0; i < leftAutoColliders.Length; i++)
            {
                _autoColliderPairs.Add(new AutoColliderPair { Left = leftAutoColliders[i], Right = rightAutoColliders[i] });
            }

            HardColliderGroups = _autoColliderPairs.Select(pair => NewHardColliderGroup(pair.Left, pair.Right)).ToList();
            EnableMultiplyFriction();
        }

        static float ColliderRadiusAndPositionSizeCurve(float x) => Calc.InverseSmoothStep(x, 0.35f, 0.06f);
        static float ColliderRadiusAndPositionSizeCurve2(float x) => Calc.InverseSmoothStep(x, 0.07f, 0.04f);

        static Dictionary<string, Scaler> GetScalingConfig(string id)
        {
            switch(id)
            {
                case "Pectoral1":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = -0.26f,
                                Max = 40,
                                MassCurve = x => 0.27f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = -0.09f,
                                Max = 40,
                                MassCurve = x => -0.32f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = 0.35f,
                                Max = 40,
                                MassCurve = x => 0.38f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.12f,
                                Max = 40,
                            }
                        },
                    };
                case "Pectoral2":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = -0.15f,
                                Max = 40,
                                MassCurve = x => -0.45f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = 0.80f,
                                Max = 40,
                                MassCurve = x => 0.10f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = -0.12f,
                                Max = 40,
                                MassCurve = x => -0.70f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.43f,
                                Max = 40,
                                MassCurve = x => -0.07f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                    };
                case "Pectoral3":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = -0.19f,
                                Max = 40,
                                MassCurve = x => -0.94f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = -0.32f,
                                Max = 40,
                                MassCurve = x => -0.12f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = 0.21f,
                                Max = 40,
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.19f,
                                Max = 40,
                                MassCurve = x => -0.42f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                    };
                case "Pectoral4":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = -0.28f,
                                Max = 40,
                                MassCurve = x => -0.62f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = -0.32f,
                                Max = 40,
                                MassCurve = x => -0.42f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = -0.05f,
                                Max = 40,
                                MassCurve = x => 0.35f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.13f,
                                Max = 40,
                                MassCurve = x => -0.12f * ColliderRadiusAndPositionSizeCurve(x),
                            }
                        },
                    };
                case "Pectoral5":
                    return new Dictionary<string, Scaler>
                    {
                        {
                            Constant.RADIUS_OFFSET, new Scaler
                            {
                                Offset = -0.41f,
                                Max = 40,
                                MassCurve = x => -1.24f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.X_OFFSET, new Scaler
                            {
                                Offset = -0.37f,
                                Max = 40,
                                MassCurve = x => -0.84f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Y_OFFSET, new Scaler
                            {
                                Offset = 0.76f,
                                Max = 40,
                                MassCurve = x => 0.65f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                        {
                            Constant.Z_OFFSET, new Scaler
                            {
                                Offset = -0.29f,
                                Max = 40,
                                MassCurve = x => -0.29f * ColliderRadiusAndPositionSizeCurve2(x),
                            }
                        },
                    };
                default: throw new ArgumentException($"Invalid id {id}");
            }
        }

        HardColliderGroup NewHardColliderGroup(AutoCollider left, AutoCollider right)
        {
            string id = left.name.Replace("AutoColliderFemaleAutoCollidersl", ""); // -> Pectoral1, Pectoral2, etc.
            var leftCollider = new HardCollider(left);
            var rightCollider = new HardCollider(right);
            var scalingConfig = GetScalingConfig(id);

            var frictionMultipliers = new Dictionary<string, float>
            {
                { "Pectoral1", 0.5f },
                { "Pectoral2", 1.0f },
                { "Pectoral3", 1.0f },
                { "Pectoral4", 1.0f },
                { "Pectoral5", 1.0f },
            };

            string visualizerId = ColliderVisualizer.SelectablesJSON.choices.Find(option => option.EndsWith($"l{id}"));
            var group = new HardColliderGroup(
                id,
                visualizerId,
                leftCollider,
                rightCollider,
                new Scaler
                {
                    Offset = 0.050f,
                    Max = 1,
                    MassCurve = x => 0.022f * Calc.InverseSmoothStepWithMax(x, 8.93f, 1.25f, 1.14f),
                    SoftnessCurve = x =>
                        Mathf.Lerp(-0.052f, -0.069f, Calc.InverseSmoothStep(1 - _jointPhysicsHandler.NormalizedMass, 0.24f, 0.70f)) *
                        Calc.InverseSmoothStep(x, 0.10f, 0.11f),
                },
                scalingConfig[Constant.RADIUS_OFFSET],
                scalingConfig[Constant.X_OFFSET],
                scalingConfig[Constant.Y_OFFSET],
                scalingConfig[Constant.Z_OFFSET],
                frictionMultipliers[id]
            )
            {
                MassMultiplierJsf = new StorableFloat($"{id} {Constant.MASS_MULTIPLIER}", 0.50f, 0.01f, 1.00f),
                RadiusJsf = new StorableFloat($"{id} {Constant.RADIUS_OFFSET}", 0, -1f, 1f),
                RightJsf = new StorableFloat($"{id} {Constant.X_OFFSET}", 0, -1f, 1f),
                UpJsf = new StorableFloat($"{id} {Constant.Y_OFFSET}", 0, -1f, 1f),
                LookJsf = new StorableFloat($"{id} {Constant.Z_OFFSET}", 0, -1f, 1f),
            };

            group.MassMultiplierJsf.setCallbackFunction = _ => SyncColliderMass(group);
            group.RadiusJsf.setCallbackFunction = _ => SyncRadius(group);
            group.RightJsf.setCallbackFunction = _ => SyncPosition(group);
            group.UpJsf.setCallbackFunction = _ => SyncPosition(group);
            group.LookJsf.setCallbackFunction = _ => SyncPosition(group);

            return group;
        }

        void SyncRadius(HardColliderGroup group)
        {
            group.UpdateDimensions(_jointPhysicsHandler.NormalizedMass, Script.Softness);
            SyncSizeAuto();
        }

        void SyncPosition(HardColliderGroup group)
        {
            group.UpdatePosition(_jointPhysicsHandler.NormalizedMass, Script.Softness);
            SyncSizeAuto();
        }

        internal override void SyncAllOffsets()
        {
            float mass = _jointPhysicsHandler.NormalizedRealMass;
            float softness = Script.Softness;
            foreach(var group in HardColliderGroups)
            {
                group.UpdateDimensions(mass, softness);
                group.UpdatePosition(mass, softness);
            }

            SyncSizeAuto();
        }

        internal override void UpdateFrictionSizeMultipliers()
        {
            /* int[]: left and right skin width markers */
            frictionSizeMultiplierLeft = FrictionSizeMultiplier(new[] { 8956, 8967 });
            frictionSizeMultiplierRight = FrictionSizeMultiplier(new[] { 19622, 19633 });
        }

        float FrictionSizeMultiplier(int[] indices)
        {
            /* experimentally determined with 3kg breasts, is slightly different for different shapes */
            const float maxWidth = 0.17f;
            var rawSkinnedVerts = Person.Skin.rawSkinnedVerts;
            float width = (rawSkinnedVerts[indices[0]] - rawSkinnedVerts[indices[1]]).magnitude;
            float multiplier = Mathf.InverseLerp(0, maxWidth, width);
            return Calc.InverseSmoothStep(multiplier, 0.55f, 0.42f);
        }

        internal override void UpdateClothingAdjustmentAmounts(float massMultiplierLeft, float massMultiplierRight)
        {
            foreach(var group in HardColliderGroups)
            {
                group.SetClothingMassMultipliers(massMultiplierLeft, massMultiplierRight);
            }
        }

        internal override void SyncCollidersMass()
        {
            foreach(var group in HardColliderGroups)
            {
                SyncColliderMass(group);
            }
        }

        void SyncColliderMass(HardColliderGroup group)
        {
            float combinedMultiplier = BaseMassMultiplierJsf.val * group.MassMultiplierJsf.val;
            if(_softPhysicsHandler.IsLdpUpdateOn())
            {
                combinedMultiplier *= Mathf.Lerp(1.00f, 1.44f, Script.Softness);
            }

            group.UpdateRigidbodyMass(
                combinedMultiplier,
                _jointPhysicsHandler.NormalizedMass,
                Script.Softness,
                _softPhysicsHandler.SoftPhysicsOn
            );
        }

        internal override void PreRegisterParams()
        {
            Script.PreRegisterHardCollidersParam(BaseMassMultiplierJsf);
            foreach(var group in HardColliderGroups)
            {
                Script.PreRegisterHardCollidersParam(group.MassMultiplierJsf);
                Script.PreRegisterHardCollidersParam(group.RadiusJsf);
                Script.PreRegisterHardCollidersParam(group.RightJsf);
                Script.PreRegisterHardCollidersParam(group.UpJsf);
                Script.PreRegisterHardCollidersParam(group.LookJsf);
            }
        }

        internal override void CalibrateColliders()
        {
            var breastCenterLeft = CollidersCenter(_jointPhysicsHandler.leftVolumeCalculationVerts);
            var breastCenterRight = CollidersCenter(_jointPhysicsHandler.rightVolumeCalculationVerts);
            foreach(var group in HardColliderGroups)
            {
                group.Calibrate(breastCenterLeft, breastCenterRight, Person.ChestRb);
            }
        }

        internal override void UpdateDistanceDiffs()
        {
            var breastCenterLeft = CollidersCenter(_jointPhysicsHandler.leftVolumeCalculationVerts);
            var breastCenterRight = CollidersCenter(_jointPhysicsHandler.rightVolumeCalculationVerts);
            foreach(var group in HardColliderGroups)
            {
                group.UpdateDistanceDiffs(breastCenterLeft, breastCenterRight, Person.ChestRb);
            }
        }
    }
}
