using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Models;
using NAT.Common.Storables;
using System.Collections.Generic;
using UnityEngine;
using static NAT.Direction;

namespace NAT.TM.Handlers
{
    sealed class BreastForceMorphHandler : ForceHandler
    {
        BreastJointPhysicsHandler _jointPhysicsHandler;
        BreastSoftPhysicsHandler _softPhysicsHandler;

        internal Dictionary<string, List<ConfigurableMorph>> MorphSets { get; private set; }
        Dictionary<string, List<ConfigurableMorph>> _forwardMorphSets;

        internal JSONStorableStringChooser DirectionJssc { get; private set; }

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _jointPhysicsHandler = provider.GetInstance<BreastJointPhysicsHandler>();
            _softPhysicsHandler = provider.GetInstance<BreastSoftPhysicsHandler>();
            BaseJsf = NewMultiplierJsf("Force Morphing Base Multiplier");
            UpJsf = NewMultiplierJsf("Force Morphing Up Multiplier");
            DepthOutJsf = NewMultiplierJsf("Force Morphing Depth Out Multiplier");
            DepthInJsf = NewMultiplierJsf("Force Morphing Depth In Multiplier");
            SidewaysInJsf = NewMultiplierJsf("Force Morphing Sideways In Multiplier");
            SidewaysOutJsf = NewMultiplierJsf("Force Morphing Sideways Out Multiplier");
            LoadSettings();
        }

        protected override void OnMultiplierValueChanged(JSONStorableFloat jsf)
        {
            var storable = (StorableFloat) jsf;
            if(_softPhysicsHandler.IsLdpUpdateOn())
            {
                Script.SetCalibrationPending(storable);
            }
        }

        protected override void LoadSettings()
        {
            MorphSets = new Dictionary<string, List<ConfigurableMorph>>
            {
                { UP_L, UpForceMorphs("L") },
                { UP_R, UpForceMorphs("R") },
                { UP_C, UpForceCenterMorphs() },
                { BACK_L, DepthInForceMorphs("L") },
                { BACK_R, DepthInForceMorphs("R") },
                { BACK_C, DepthInForceCenterMorphs() },
                { FORWARD_L, DepthOutForceMorphs("L") },
                { FORWARD_R, DepthOutForceMorphs("R") },
                { FORWARD_C, DepthOutForceCenterMorphs() },
                { LEFT_L, LeftForceLeftMorphs() },
                { LEFT_R, LeftForceRightMorphs() },
                { RIGHT_L, RightForceLeftMorphs() },
                { RIGHT_R, RightForceRightMorphs() },
            };

            _forwardMorphSets = new Dictionary<string, List<ConfigurableMorph>>
            {
                { FORWARD_L, DepthOutForceMorphs2("L") },
                { FORWARD_R, DepthOutForceMorphs2("R") },
            };

            var directions = new List<string>
            {
                UP,
                UP_C,
                BACK,
                BACK_C,
                FORWARD,
                FORWARD_C,
                LEFT_L,
                RIGHT_L,
                LEFT_R,
                RIGHT_R,
            };
            DirectionJssc = new JSONStorableStringChooser("direction", directions, directions[0], "Direction");
        }

        ConfigurableMorph NewConfigurableMorph(string morphName, bool isNegative, float softMultiplier, float massMultiplier) =>
            new ConfigurableMorph(
                Person.GetBreastMorph(morphName),
                isNegative,
                new JSONStorableFloat("softMultiplier", softMultiplier, -3.00f, 3.00f),
                new JSONStorableFloat("massMultiplier", massMultiplier, -3.00f, 3.00f)
            );

        List<ConfigurableMorph> UpForceMorphs(string side)
        {
            string prefix = $"{MorphPrefix.UP}/{MorphPrefix.UP}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breast Height {side}", false, 1.00f, 0.50f),
                NewConfigurableMorph($"{prefix} Breast Sag1 {side}", true, -0.75f, -0.38f),
                NewConfigurableMorph($"{prefix} Breast Sag2 {side}", true, -1.00f, -0.50f),
                NewConfigurableMorph($"{prefix} Breasts Natural {side}", true, -0.75f, -0.38f),
                NewConfigurableMorph($"{prefix} Breast Rotate Up {side}", false, 0.80f, 0.80f),
                NewConfigurableMorph($"{prefix} Breast Top Curve2 {side}", true, -0.90f, -0.60f),
                NewConfigurableMorph($"{prefix} Breasts Implants {side}", false, 0.36f, 0.18f),
                NewConfigurableMorph($"{prefix} Breast Diameter {side}", false, 1.70f, 0.50f),
                NewConfigurableMorph($"{prefix} Breast Diameter(Pose) {side}", false, 1.70f, 0.50f),
                NewConfigurableMorph($"{prefix} Breast Width {side}", false, 0.21f, 0.14f),
                NewConfigurableMorph($"{prefix} Breast flat(Fixed) {side}", false, 0.25f, 0.06f),
                NewConfigurableMorph($"{prefix} BreastsShape2 {side}", false, 1.80f, 0.53f),
                NewConfigurableMorph($"{prefix} Breast Height Lower {side}", true, -1.50f, -0.60f),
                NewConfigurableMorph($"{prefix} Breast Move Up {side}", false, 1.60f, 0.60f),
                NewConfigurableMorph($"{prefix} Breasts TogetherApart {side}", false, 0.90f, 0.90f),
                NewConfigurableMorph($"{prefix} Breast Top Curve1 {side}", true, -0.40f, -0.20f),
                NewConfigurableMorph($"{prefix} Breasts Under Curve {side}", true, -1.50f, -0.75f),
                NewConfigurableMorph($"{prefix} ChestUnderBreast {side}", false, 0.80f, 0.38f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother1 {side}", false, 0.25f, 0.00f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 {side}", false, 0.40f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 {side}", false, 0.15f, 0.23f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother4 {side}", false, 0.30f, 0.15f),
                NewConfigurableMorph($"{prefix} Breast Height Upper {side}", false, 0.40f, 0.60f),
                NewConfigurableMorph($"{prefix} Breasts Upward Slope {side}", false, 1.30f, 3.00f),
                NewConfigurableMorph($"{prefix} Breast upper down {side}", false, 0.15f, 0.08f),
                NewConfigurableMorph($"{prefix} Breasts Small Top Slope {side}", false, 0.13f, 0.03f),
                NewConfigurableMorph($"{prefix} Areolae Depth {side}", false, 0.30f, 0.15f),
            };
        }

        List<ConfigurableMorph> UpForceCenterMorphs()
        {
            string prefix = $"{MorphPrefix.UP}/{MorphPrefix.UP}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Center Gap Smooth", false, 1.70f, 0.30f),
                NewConfigurableMorph($"{prefix} Center Gap Depth", true, -0.50f, -1.20f),
                NewConfigurableMorph($"{prefix} Center Gap Height", false, 1.20f, 0.75f),
                NewConfigurableMorph($"{prefix} Centre Gap Wide", false, 0.50f, 0.38f),
                NewConfigurableMorph($"{prefix} Center Gap UpDown", false, 2.00f, -0.45f),
                NewConfigurableMorph($"{prefix} Chest Smoother", false, 3.00f, 1.13f),
                NewConfigurableMorph($"{prefix} ChestSmoothCenter", false, 3.00f, 1.26f),
            };
        }

        List<ConfigurableMorph> DepthInForceMorphs(string side)
        {
            string prefix = $"{MorphPrefix.BK}/{MorphPrefix.BK}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breast Diameter {side}", false, 0.84f, 0.28f),
                NewConfigurableMorph($"{prefix} Breast flat(Fixed) {side}", false, 0.54f, 0.27f),
                NewConfigurableMorph($"{prefix} Breast Pointed {side}", true, -0.50f, -0.17f),
                NewConfigurableMorph($"{prefix} Breast Top Curve1 {side}", true, -0.66f, -0.33f),
                NewConfigurableMorph($"{prefix} Breast Top Curve2 {side}", false, 0.30f, 0.15f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother1 {side}", false, 0.70f, 0.53f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 {side}", false, 0.07f, 0.32f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 {side}", false, 0.60f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Zero {side}", false, 0.22f, 0.12f),
                NewConfigurableMorph($"{prefix} Breasts Flatten {side}", false, 1.00f, 0.80f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward {side}", true, -0.40f, -0.15f),
                NewConfigurableMorph($"{prefix} Breasts Implants {side}", false, 0.12f, 0.00f),
                NewConfigurableMorph($"{prefix} Breasts TogetherApart {side}", false, 0.15f, 0.23f),
                NewConfigurableMorph($"{prefix} Depth Squash {side}", false, 1.60f, 0.75f),
            };
        }

        List<ConfigurableMorph> DepthInForceCenterMorphs()
        {
            string prefix = $"{MorphPrefix.BK}/{MorphPrefix.BK}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Sternum Depth", true, 0.10f, -0.60f),
                NewConfigurableMorph($"{prefix} Sternum Width", true, -0.30f, -0.45f),
                NewConfigurableMorph($"{prefix} Center Gap Depth", true, 0.00f, -0.60f),
            };
        }

        List<ConfigurableMorph> DepthOutForceMorphs(string side)
        {
            string prefix = $"{MorphPrefix.FW}/{MorphPrefix.FW}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breast Diameter(Pose) {side}", false, 0.40f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Height2 {side}", true, -0.40f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Rotate Down {side}", true, -0.40f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Width {side}", true, -0.24f, -0.12f),
                NewConfigurableMorph($"{prefix} Breast Depth {side}", false, 1.60f, -0.12f),
                NewConfigurableMorph($"{prefix} Breast Small {side}", false, 0.64f, 0.27f),
                NewConfigurableMorph($"{prefix} Breast Round {side}", true, -1.80f, -0.90f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward {side}", true, -0.12f, -0.06f),
                NewConfigurableMorph($"{prefix} Lg Breast Hang {side}", false, 0.00f, 0.00f),
            };
        }

        List<ConfigurableMorph> DepthOutForceMorphs2(string side)
        {
            string prefix = $"{MorphPrefix.FW}/{MorphPrefix.FW}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breast Diameter(Pose) {side}", false, 0.60f, 0.40f),
                NewConfigurableMorph($"{prefix} Breast Height2 {side}", true, -0.15f, -0.45f),
                NewConfigurableMorph($"{prefix} Breast Rotate Down {side}", false, 0.14f, 0.22f),
                NewConfigurableMorph($"{prefix} Breast Width {side}", true, -0.15f, -0.75f),
                NewConfigurableMorph($"{prefix} Breast Depth {side}", false, 0.88f, -0.62f),
                NewConfigurableMorph($"{prefix} Breast Small {side}", false, 0.75f, -0.25f),
                NewConfigurableMorph($"{prefix} Breast Round {side}", true, -1.40f, -0.70f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward {side}", false, 0.45f, 0.30f),
                NewConfigurableMorph($"{prefix} Lg Breast Hang {side}", false, 0.55f, 0.75f),
            };
        }

        List<ConfigurableMorph> DepthOutForceCenterMorphs()
        {
            string prefix = $"{MorphPrefix.FW}/{MorphPrefix.FW}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Sternum Width", false, 0.41f, 0.21f),
                NewConfigurableMorph($"{prefix} ChestSeparateBreasts", false, 0.00f, 0.45f),
            };
        }

        List<ConfigurableMorph> LeftForceLeftMorphs()
        {
            string prefix = $"{MorphPrefix.LT}/{MorphPrefix.LT}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breasts Shift S2S Left L", false, 1.50f, 0.75f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward L", true, -1.00f, -1.20f),
                NewConfigurableMorph($"{prefix} Breast Width L", true, -0.40f, -0.20f),
                NewConfigurableMorph($"{prefix} Breast Rotate X Out L", true, -0.60f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Rotate X In L", true, -0.40f, -0.20f),
                NewConfigurableMorph($"{prefix} Breast Diameter L", true, -0.40f, -0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 L", false, 0.10f, 0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 L", false, 0.40f, 0.60f),
            };
        }

        List<ConfigurableMorph> LeftForceRightMorphs()
        {
            string prefix = $"{MorphPrefix.LT}/{MorphPrefix.LT}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breasts Shift S2S Left R", false, 1.50f, 0.45f),
                NewConfigurableMorph($"{prefix} Breast Rotate X Out R", true, -0.84f, -0.42f),
                NewConfigurableMorph($"{prefix} Breasts Flatten R", true, -0.40f, -0.30f),
                NewConfigurableMorph($"{prefix} Breast Width R", false, 0.90f, -0.75f),
                NewConfigurableMorph($"{prefix} Breast Rotate X In R", false, 0.20f, 0.00f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward R", false, 0.20f, 0.11f),
                NewConfigurableMorph($"{prefix} Areola S2S R", false, 0.60f, 2.70f),
                NewConfigurableMorph($"{prefix} Breast Diameter R", true, -0.04f, -0.06f),
                NewConfigurableMorph($"{prefix} Breast Depth Squash R", false, 0.20f, 3.00f),
                NewConfigurableMorph($"{prefix} Breast Move S2S In R", false, 0.00f, 0.45f),
                NewConfigurableMorph($"{prefix} Breasts Implants R", true, 0.10f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 R", false, 0.10f, 0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 R", false, 0.40f, 0.60f),
            };
        }

        List<ConfigurableMorph> RightForceLeftMorphs()
        {
            string prefix = $"{MorphPrefix.RT}/{MorphPrefix.RT}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breasts Shift S2S Right L", false, 1.50f, 0.45f),
                NewConfigurableMorph($"{prefix} Breast Rotate X Out L", true, -0.84f, -0.42f),
                NewConfigurableMorph($"{prefix} Breasts Flatten L", true, -0.40f, -0.30f),
                NewConfigurableMorph($"{prefix} Breast Width L", false, 0.90f, -0.75f),
                NewConfigurableMorph($"{prefix} Breast Rotate X In L", false, 0.20f, 0.00f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward L", false, 0.20f, 0.11f),
                NewConfigurableMorph($"{prefix} Areola S2S L", false, 0.60f, 2.70f),
                NewConfigurableMorph($"{prefix} Breast Diameter L", true, -0.04f, -0.06f),
                NewConfigurableMorph($"{prefix} Breast Depth Squash L", false, 0.20f, 3.00f),
                NewConfigurableMorph($"{prefix} Breast Move S2S In L", false, 0.00f, 0.45f),
                NewConfigurableMorph($"{prefix} Breasts Implants L", true, 0.10f, -0.15f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 L", false, 0.10f, 0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 L", false, 0.40f, 0.60f),
            };
        }

        List<ConfigurableMorph> RightForceRightMorphs()
        {
            string prefix = $"{MorphPrefix.RT}/{MorphPrefix.RT}";
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Breasts Shift S2S Right R", false, 1.50f, 0.75f),
                NewConfigurableMorph($"{prefix} Breasts Hang Forward R", true, -1.00f, -1.20f),
                NewConfigurableMorph($"{prefix} Breast Width R", true, -0.40f, -0.20f),
                NewConfigurableMorph($"{prefix} Breast Rotate X Out R", true, -0.60f, 0.30f),
                NewConfigurableMorph($"{prefix} Breast Rotate X In R", true, -0.40f, -0.20f),
                NewConfigurableMorph($"{prefix} Breast Diameter R", true, -0.40f, -0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother2 R", false, 0.10f, 0.60f),
                NewConfigurableMorph($"{prefix} Breast Under Smoother3 R", false, 0.40f, 0.60f),
            };
        }

        internal override void PreRegisterParams()
        {
            Script.PreRegisterMorphsParam(BaseJsf);
            Script.PreRegisterMorphsParam(UpJsf);
            Script.PreRegisterMorphsParam(DepthOutJsf);
            Script.PreRegisterMorphsParam(DepthInJsf);
            Script.PreRegisterMorphsParam(SidewaysInJsf);
            Script.PreRegisterMorphsParam(SidewaysOutJsf);
        }

        float _mass;
        float _softness;

        internal float UpLimit { get; private set; }
        internal float DepthOutLimit { get; private set; }
        internal float SidewaysOutLimit { get; private set; }
        internal float SidewaysInLimit { get; private set; }
        internal float UpNormalizer { get; private set; }
        internal float DepthOutNormalizer { get; private set; }
        internal float DepthInNormalizer { get; private set; }
        internal float SidewaysNormalizer { get; private set; }

        float _upMultiplier;
        float _depthOutMultiplier;
        float _depthInMultiplier;
        float _sidewaysInMultiplier;
        float _sidewaysOutMultiplier;

        static float SoftnessMultiplier(float x) => Calc.InverseSmoothStep(x, 0.80f, 0.03f);

        // https://www.desmos.com/calculator/ychmmr0gqu
        static float MassMultiplier(float x) => Calc.Exponential3(x, 2.17f, 1.56f, 1.00f, 1.30f, 1.70f);

        internal void SetMultipliers(float mass, float softness)
        {
            _mass = mass;
            _softness = SoftnessMultiplier(softness);
            SetUpMultiplier();
            SetDepthOutMultiplier();
            SetDepthInMultiplier();
            SetSidewaysMultipliers();
        }

        void SetUpMultiplier()
        {
            // https://www.desmos.com/calculator/zke8k2x7cu
            UpLimit = 50f * Calc.Exponential1(_mass, 0.28f, 0.13f, 0.95f);
            float softnessMultiplier = Mathf.Lerp(0.65f, 1.46f, _softness);
            UpNormalizer = softnessMultiplier * UpNormalizerCurve();
            _upMultiplier = 0.0060f * softnessMultiplier * UpNormalizer * MassMultiplier(_mass);
        }

        float UpNormalizerCurve()
        {
            const float cutoff1 = 0.12f;
            const float cutoff2 = 0.43f;

            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(37.0f, 50.0f, ratioToCutoff1);
            }

            if(_mass < cutoff2)
            {
                float ratioFromCutoff1To2 = (_mass - cutoff1) / (cutoff2 - cutoff1);
                return Mathf.Lerp(50.0f, 55.4f, Calc.Exponential1(ratioFromCutoff1To2, 1.76f, 1.43f, 0.67f));
            }

            float ratioFromCutoff2 = (_mass - cutoff2) / cutoff2;
            return Mathf.Lerp(47.10f, 55.40f, 1 - ratioFromCutoff2);
        }

        void SetDepthOutMultiplier()
        {
            // https://www.desmos.com/calculator/bhfmr1kcn5
            DepthOutLimit = 0.00888f + 0.03774f * Calc.InverseSmoothStepWithMax(_mass, 0.42f, 0.2f, 0);
            float softnessMin = Mathf.Lerp(0.21f, 0.56f, Calc.InverseSmoothStep(1 - _mass, 0.25f, 0.20f));
            float softnessMax = Mathf.Lerp(1.06f, 1.16f, Calc.InverseSmoothStep(1 - _mass, 0.25f, 0.20f));
            float softnessMultiplier = Mathf.Lerp(softnessMin, softnessMax, _softness);
            DepthOutNormalizer = 1 / (softnessMultiplier * DepthOutNormalizerCurve());
            _depthOutMultiplier = 12.50f * softnessMultiplier * DepthOutNormalizer * Mathf.Lerp(3.50f, 3.90f, _mass);
        }

        float DepthOutNormalizerCurve()
        {
            const float cutoff1 = 0.12f;
            const float cutoff2 = 0.25f;

            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(16f, 31.0f, 1 - ratioToCutoff1);
            }

            if(_mass > cutoff2)
            {
                float ratioFromCutoff2 = (_mass - cutoff2) / cutoff2;
                return Mathf.Lerp(16f, 24f, Mathf.Pow(ratioFromCutoff2, 1.75f));
            }

            return 16f;
        }

        // https://www.desmos.com/calculator/ntxyujsadd
        float DepthInMassMultiplier()
        {
            float value = 9.50f / Mathf.Sqrt(3 * _mass + 3.40f) - 2.40f;
            return 2.40f * value;
        }

        void SetDepthInMultiplier()
        {
            // https://www.desmos.com/calculator/o7ezjrco9z
            // Depth in limit is not practical, as depth morphing needs to occur continuously for collision to look good
            // DepthInLimit = 0.0105f + 0.0375f * Calc.Exponential1(_mass, -3.57f, 0.92f, 1.17f);
            float softnessMin = Mathf.Lerp(0.58f, 0.900f, Calc.InverseSmoothStep(_mass, 0.25f, 0.20f));
            float softnessMax = Mathf.Lerp(1.16f, 0.900f, Calc.InverseSmoothStep(_mass, 0.25f, 0.20f));
            float softnessMultiplier = Mathf.Lerp(softnessMin, softnessMax, _softness);
            DepthInNormalizer = 1 / (softnessMultiplier * DepthInNormalizerCurve());
            _depthInMultiplier = 12.50f * softnessMultiplier * DepthInNormalizer * DepthInMassMultiplier();
        }

        float DepthInNormalizerCurve()
        {
            const float cutoff1 = 0.12f;
            const float cutoff2 = 0.25f;

            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(30f, 39f, 1 - ratioToCutoff1);
            }

            if(_mass > cutoff2)
            {
                float ratioFromCutoff2 = (_mass - cutoff2) / cutoff2;
                return Mathf.Lerp(30f, 10.0f, Mathf.Pow(ratioFromCutoff2, 1.75f));
            }

            return 30f;
        }

        void SetSidewaysMultipliers()
        {
            // https://www.desmos.com/calculator/brldbo3jf3
            SidewaysInLimit = 8.436f + 17.982f * Calc.InverseSmoothStepWithMax(_mass, 0.341f, -1.10f, 0.57f);
            // https://www.desmos.com/calculator/zgpn36qszk
            SidewaysOutLimit = 6.882f + 27.861f * Calc.Exponential1(_mass, 0.81f, 0.12f, 1.1f);
            float softnessMultiplier = Mathf.Lerp(1.00f, 1.33f, _softness);
            SidewaysNormalizer = softnessMultiplier * SidewaysNormalizerCurve();
            _sidewaysInMultiplier = softnessMultiplier * 0.0145f * SidewaysNormalizer * MassMultiplier(_mass);
            _sidewaysOutMultiplier = softnessMultiplier * 0.0170f * SidewaysNormalizer * MassMultiplier(_mass);
        }

        // https://www.desmos.com/calculator/fusmae0li0
        float SidewaysNormalizerCurve()
        {
            const float cutoff1 = 0.25f;
            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(33.4f, 40.0f, Calc.InverseSmoothStep(ratioToCutoff1, -0.16f, 0.00f));
            }

            float ratioFromCutoff1 = (_mass - cutoff1) / cutoff1;
            return Mathf.Lerp(27.8f, 40.0f, 1 - Calc.InverseSmoothStep(ratioFromCutoff1, -0.33f, 0.00f));
        }

        float _leftRoll, _rightRoll;
        float _leftPitch, _rightPitch;
        internal float leftLeanBackFixer, rightLeanBackFixer;
        internal float leftUpsideDownFixer, rightUpsideDownFixer;
        internal float leftRollFixer, rightRollFixer;

        internal void Execute()
        {
            _leftRoll = AngleCalculator.LeftRoll;
            _rightRoll = -AngleCalculator.RightRoll;
            _leftPitch = AngleCalculator.LeftPitch;
            _rightPitch = AngleCalculator.RightPitch;
            float mass = _jointPhysicsHandler.NormalizedMass;
            // TODO world gravity
            leftLeanBackFixer = HandlerUtils.CalculateLeanBackFixerMultiplier(mass, _leftPitch, _leftRoll);
            rightLeanBackFixer = HandlerUtils.CalculateLeanBackFixerMultiplier(mass, _rightPitch, _rightRoll);
            leftUpsideDownFixer = HandlerUtils.CalculateUpsideDownFixerMultiplier(_leftPitch, _leftRoll);
            rightUpsideDownFixer = HandlerUtils.CalculateUpsideDownFixerMultiplier(_rightPitch, _rightRoll);
            leftRollFixer = HandlerUtils.CalculateRollFixerMultiplier(mass, _leftRoll);
            rightRollFixer = HandlerUtils.CalculateRollFixerMultiplier(mass, _rightRoll);

            AdjustUpMorphs();
            AdjustDepthOutMorphs();
            AdjustDepthInMorphs();
            AdjustSidewaysMorphs();
        }

        static float UpForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.44f, 1.28f, 0.96f);

        float UpEffect(float angle, float fixerMultiplier) =>
            fixerMultiplier *
            _upMultiplier *
            UserUpMultiplier() *
            UpForceEffectCurve(Mathf.Min(Mathf.Abs(angle), UpLimit) / UpNormalizer);

        void AdjustUpMorphs()
        {
            float leftFixer = 0;
            float rightFixer = 0;

            if(Person.LeftBreast.AngleY >= 0)
            {
                leftFixer = Mathf.Lerp(0.63f, 1f, leftLeanBackFixer * AngleCalculator.LeftDiffFromHorizontal);
                UpdateMorphs(UP_L, UpEffect(Person.LeftBreast.AngleY, leftFixer));
            }
            else
            {
                ResetMorphs(UP_L);
            }

            if(Person.RightBreast.AngleY >= 0)
            {
                rightFixer = Mathf.Lerp(0.63f, 1f, rightLeanBackFixer * AngleCalculator.RightDiffFromHorizontal);
                UpdateMorphs(UP_R, UpEffect(Person.RightBreast.AngleY, rightFixer));
            }
            else
            {
                ResetMorphs(UP_R);
            }

            float angleYCenter = (Person.RightBreast.AngleY + Person.LeftBreast.AngleY) / 2;
            if(angleYCenter >= 0)
            {
                UpdateMorphs(UP_C, UpEffect(angleYCenter, (leftFixer + rightFixer) / 2));
            }
            else
            {
                ResetMorphs(UP_C);
            }
        }

        static float DepthInForceEffectCurve(float x) => Calc.Exponential1(x, 1.66f, 1.70f, 1.06f);

        float DepthInEffect(float depthDiff, float fixerMultiplier) =>
            fixerMultiplier *
            _depthInMultiplier *
            UserDepthInMultiplier() *
            DepthInForceEffectCurve(Mathf.Abs(depthDiff) / DepthInNormalizer);

        void AdjustDepthInMorphs()
        {
            if(Person.LeftBreast.DepthDiff < 0)
            {
                ResetMorphs(BACK_L);
            }
            else
            {
                UpdateMorphs(BACK_L, DepthInEffect(Person.LeftBreast.DepthDiff, leftLeanBackFixer));
            }

            if(Person.RightBreast.DepthDiff < 0)
            {
                ResetMorphs(BACK_R);
            }
            else
            {
                UpdateMorphs(BACK_R, DepthInEffect(Person.RightBreast.DepthDiff, rightLeanBackFixer));
            }

            float centerDepthDiff = (Person.LeftBreast.DepthDiff + Person.RightBreast.DepthDiff) / 2;
            if(centerDepthDiff < 0)
            {
                ResetMorphs(BACK_C);
            }
            else
            {
                UpdateMorphs(BACK_C, DepthInEffect(centerDepthDiff, (leftLeanBackFixer + rightLeanBackFixer) / 2));
            }
        }

        float DepthOutEffect(float depthDiff, float fixerMultiplier) =>
            fixerMultiplier *
            _depthOutMultiplier *
            UserDepthOutMultiplier() *
            Calc.Exponential1(Mathf.Min(Mathf.Abs(depthDiff), DepthOutLimit) / DepthOutNormalizer, 1.42f, 3.20f, 1.15f);

        void AdjustDepthOutMorphs()
        {
            float leftFixer = 0;
            float rightFixer = 0;

            if(Person.LeftBreast.DepthDiff < 0)
            {
                leftFixer = leftRollFixer;
                // UpdateMorphs(FORWARD_L, DepthOutEffect(Person.LeftBreast.DepthDiff, leftFixer));
                UpdateDepthOutMorphs(FORWARD_L, DepthOutEffect(Person.LeftBreast.DepthDiff, leftFixer), AngleCalculator.LeftForwardLeanAmount);
            }
            else
            {
                ResetMorphs(FORWARD_L);
            }

            if(Person.RightBreast.DepthDiff < 0)
            {
                rightFixer = rightRollFixer;
                // UpdateMorphs(FORWARD_R, DepthOutEffect(Person.RightBreast.DepthDiff, rightFixer));
                UpdateDepthOutMorphs(FORWARD_R, DepthOutEffect(Person.RightBreast.DepthDiff, rightFixer), AngleCalculator.RightForwardLeanAmount);
            }
            else
            {
                ResetMorphs(FORWARD_R);
            }

            float centerDepthDiff = (Person.LeftBreast.DepthDiff + Person.RightBreast.DepthDiff) / 2;
            if(centerDepthDiff < 0)
            {
                UpdateMorphs(FORWARD_C, DepthOutEffect(centerDepthDiff, (leftFixer + rightFixer) / 2));
            }
            else
            {
                ResetMorphs(FORWARD_C);
            }
        }

        static float SidewaysForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.44f, 1.28f, 0.96f);

        float SidewaysInEffect(float angle, float userMultiplier) =>
            _sidewaysInMultiplier *
            userMultiplier *
            SidewaysForceEffectCurve(Mathf.Min(Mathf.Abs(angle), SidewaysInLimit) / SidewaysNormalizer);

        float SidewaysOutEffect(float angle, float userMultiplier) =>
            _sidewaysOutMultiplier *
            userMultiplier *
            SidewaysForceEffectCurve(Mathf.Min(Mathf.Abs(angle), SidewaysOutLimit) / SidewaysNormalizer);

        void AdjustSidewaysMorphs()
        {
            float sidewaysInMultiplier = UserSidewaysInMultiplier();
            float sidewaysOutMultiplier = UserSidewaysOutMultiplier();

            if(Person.LeftBreast.AngleX >= 0)
            {
                ResetMorphs(LEFT_L);
                UpdateMorphs(RIGHT_L, SidewaysInEffect(Person.LeftBreast.AngleX, sidewaysInMultiplier));
            }
            else
            {
                ResetMorphs(RIGHT_L);
                UpdateMorphs(LEFT_L, SidewaysOutEffect(Person.LeftBreast.AngleX, sidewaysOutMultiplier));
            }

            if(Person.RightBreast.AngleX >= 0)
            {
                ResetMorphs(LEFT_R);
                UpdateMorphs(RIGHT_R, SidewaysOutEffect(Person.RightBreast.AngleX, sidewaysOutMultiplier));
            }
            else
            {
                ResetMorphs(RIGHT_R);
                UpdateMorphs(LEFT_R, SidewaysInEffect(Person.RightBreast.AngleX, sidewaysInMultiplier));
            }
        }

        void UpdateMorphs(string direction, float effect)
        {
            float halfEffect = 0.5f * effect;
            foreach(var configurableMorph in MorphSets[direction])
            {
                // softness fixed to 0.62f due to historical reasons
                configurableMorph.SetValue(halfEffect, _mass, 0.62f);
            }
        }

        void UpdateDepthOutMorphs(string direction, float effect, float leanForwardAmount)
        {
            float halfEffect = 0.5f * effect;
            var morphs = MorphSets[direction];
            var forwardMorphs = _forwardMorphSets[direction];
            for(int i = 0; i < morphs.Count; i++)
            {
                var configurableMorph = morphs[i];
                // softness fixed to 0.62f due to historical reasons
                float normalValue = configurableMorph.Calculate(halfEffect, _mass, 0.62f);
                float forwardValue = forwardMorphs[i].Calculate(halfEffect, _mass, 0.62f);
                float value = Mathf.Lerp(normalValue, forwardValue, leanForwardAmount);
                configurableMorph.SetValue(value);
            }
        }

        internal override void ResetAll()
        {
            if(MorphSets == null)
            {
                return;
            }

            foreach(var pair in MorphSets)
            {
                foreach(var configurableMorph in pair.Value)
                {
                    configurableMorph.Reset();
                }
            }
        }

        void ResetMorphs(string direction)
        {
            foreach(var configurableMorph in MorphSets[direction])
            {
                configurableMorph.Reset();
            }
        }
    }
}
