using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;
using System.Collections.Generic;
using UnityEngine;
using static NAT.Direction;

namespace NAT.TM.Handlers
{
    sealed class BreastForcePhysicsHandler : ForceHandler
    {
        BreastJointPhysicsHandler _jointPhysicsHandler;
        BreastSoftPhysicsHandler _softPhysicsHandler;

        List<JointPhysicsParam> _leftParams, _rightParams;

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _jointPhysicsHandler = provider.GetInstance<BreastJointPhysicsHandler>();
            _softPhysicsHandler = provider.GetInstance<BreastSoftPhysicsHandler>();
            BaseJsf = NewMultiplierJsf("Force Physics Base Multiplier");
            UpJsf = NewMultiplierJsf("Force Physics Up Multiplier");
            DepthOutJsf = NewMultiplierJsf("Force Physics Depth Out Multiplier");
            DepthInJsf = NewMultiplierJsf("Force Physics Depth In Multiplier");
            SidewaysInJsf = NewMultiplierJsf("Force Physics Sideways In Multiplier");
            SidewaysOutJsf = NewMultiplierJsf("Force Physics Sideways Out Multiplier");
            LoadSettings();
        }

        protected override void OnMultiplierValueChanged(JSONStorableFloat jsf)
        {
            var storable = (StorableFloat) jsf;
            if(_softPhysicsHandler.IsLdpUpdateOn())
            {
                Script.SetCalibrationPending(storable);
            }
        }

        protected override void LoadSettings()
        {
            var centerOfGravity = _jointPhysicsHandler.CenterOfGravity;
            var inOutDamper = _jointPhysicsHandler.InOutDamper;

            centerOfGravity.SetForceCalculators(CenterOfGravityCalculators());
            inOutDamper.SetForceCalculators(InOutDamperCalculators());

            _leftParams = new List<JointPhysicsParam>
            {
                centerOfGravity.Left,
                inOutDamper.Left,
            };

            _rightParams = new List<JointPhysicsParam>
            {
                centerOfGravity.Right,
                inOutDamper.Right,
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> CenterOfGravityCalculators()
        {
            var backCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = -0.060f,
                    MultiplierB = -0.060f,
                    CurveA = x => 1 - x,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var forwardCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = 0.100f,
                    MultiplierB = 0.100f,
                    CurveA = x => 1 - x,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var left = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, backCalculator },
                { FORWARD, forwardCalculator },
            };
            var right = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, backCalculator },
                { FORWARD, forwardCalculator },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> InOutDamperCalculators()
        {
            var depthInCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = 6f,
                    MultiplierB = -5f,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var depthOutCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0,
                    MultiplierA = 6f,
                    MultiplierB = -5f,
                    CurveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var left = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, depthInCalculator },
                { FORWARD, depthOutCalculator },
            };
            var right = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, depthInCalculator },
                { FORWARD, depthOutCalculator },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        internal override void PreRegisterParams()
        {
            Script.PreRegisterForcePhysicsParam(BaseJsf);
            // Script.PreRegisterForcePhysicsParam(UpJsf);
            Script.PreRegisterForcePhysicsParam(DepthOutJsf);
            Script.PreRegisterForcePhysicsParam(DepthInJsf);
            // Script.PreRegisterForcePhysicsParam(SidewaysInJsf);
            // Script.PreRegisterForcePhysicsParam(SidewaysOutJsf);
        }

        /* TODO duplicates ForceMorphHandler */

        float _mass;
        float _softness;

        float _upLimit;
        float _depthOutLimit;
        float _depthInLimit;
        float _sidewaysLimit;

        float _upMultiplier;
        float _depthOutMultiplier;
        float _depthInMultiplier;
        float _sidewaysMultiplier;

        static float SoftnessMultiplier(float x) => Calc.InverseSmoothStep(x, 0.80f, 0.03f);

        // https://www.desmos.com/calculator/ychmmr0gqu
        static float MassMultiplier(float x) => Calc.Exponential3(x, 2.17f, 1.56f, 1.00f, 1.30f, 1.70f);

        internal void SetMultipliers(float mass, float softness)
        {
            _mass = mass;
            _softness = SoftnessMultiplier(softness);
            SetUpMultiplier();
            SetDepthOutMultiplier();
            SetDepthInMultiplier();
            SetSidewaysMultiplier();
        }

        void SetUpMultiplier()
        {
            float softnessMultiplier = Mathf.Lerp(0.65f, 1.33f, _softness);
            _upLimit = softnessMultiplier * UpLimit();
            _upMultiplier = 0.0070f * softnessMultiplier * _upLimit * MassMultiplier(_mass);
        }

        float UpLimit()
        {
            const float cutoff1 = 0.12f;
            const float cutoff2 = 0.43f;

            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(37.0f, 50.0f, ratioToCutoff1);
            }

            if(_mass < cutoff2)
            {
                float ratioFromCutoff1To2 = (_mass - cutoff1) / (cutoff2 - cutoff1);
                return Mathf.Lerp(50.0f, 55.4f, Calc.Exponential1(ratioFromCutoff1To2, 1.76f, 1.43f, 0.67f));
            }

            float ratioFromCutoff2 = (_mass - cutoff2) / cutoff2;
            return Mathf.Lerp(47.10f, 55.40f, 1 - ratioFromCutoff2);
        }

        float DepthOutMassMultiplier() => Mathf.Lerp(4.00f, 2.80f, 1 - _mass);

        void SetDepthOutMultiplier()
        {
            float softnessMultiplier = Mathf.Lerp(0.60f, 1.20f, _softness);
            _depthOutLimit = softnessMultiplier * DepthOutLimit();
            _depthOutMultiplier = softnessMultiplier * (12.50f / _depthOutLimit) * DepthOutMassMultiplier();
        }

        float DepthOutLimit()
        {
            const float cutoff1 = 0.12f;
            const float cutoff2 = 0.25f;

            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(16f, 31.0f, 1 - ratioToCutoff1);
            }

            if(_mass > cutoff2)
            {
                float ratioFromCutoff2 = (_mass - cutoff2) / cutoff2;
                return Mathf.Lerp(16f, 24f, Mathf.Pow(ratioFromCutoff2, 1.75f));
            }

            return 16f;
        }

        // https://www.desmos.com/calculator/ntxyujsadd
        float DepthInMassMultiplier()
        {
            float value = Mathf.Sqrt(3 * _mass + 3.40f) - 2.40f;
            return 2.00f * value;
        }

        void SetDepthInMultiplier()
        {
            float softnessMultiplier = Mathf.Lerp(
                Mathf.Lerp(0.55f, 1.10f, Calc.InverseSmoothStep(_mass, 0.50f, 0.05f)),
                1.100f,
                _softness
            );
            _depthInLimit = softnessMultiplier * DepthInLimit();
            _depthInMultiplier = softnessMultiplier * 12.50f / _depthInLimit * DepthInMassMultiplier();
        }

        float DepthInLimit()
        {
            const float cutoff1 = 0.12f;
            const float cutoff2 = 0.25f;

            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(30f, 39f, 1 - ratioToCutoff1);
            }

            if(_mass > cutoff2)
            {
                float ratioFromCutoff2 = (_mass - cutoff2) / cutoff2;
                return Mathf.Lerp(30f, 10.0f, Mathf.Pow(ratioFromCutoff2, 1.75f));
            }

            return 30f;
        }

        void SetSidewaysMultiplier()
        {
            float softnessMultiplier = Mathf.Lerp(0.90f, 1.20f, _softness);
            _sidewaysLimit = softnessMultiplier * SidewaysLimit();
            _sidewaysMultiplier = softnessMultiplier * 0.0109f * _sidewaysLimit * MassMultiplier(_mass);
        }

        // https://www.desmos.com/calculator/fusmae0li0
        float SidewaysLimit()
        {
            const float cutoff1 = 0.25f;
            if(_mass < cutoff1)
            {
                float ratioToCutoff1 = _mass / cutoff1;
                return Mathf.Lerp(33.4f, 40.0f, Calc.InverseSmoothStep(ratioToCutoff1, -0.16f, 0.00f));
            }

            float ratioFromCutoff1 = (_mass - cutoff1) / cutoff1;
            return Mathf.Lerp(27.8f, 40.0f, 1 - Calc.InverseSmoothStep(ratioFromCutoff1, -0.33f, 0.00f));
        }

        float _leftRoll, _rightRoll;
        float _leftPitch, _rightPitch;
        float _leftRollAngleMultiplier, _rightRollAngleMultiplier;
        float _leftLeanBackFixerMultiplier, _rightLeanBackFixerMultiplier;

        internal void Execute()
        {
            _leftRoll = AngleCalculator.LeftRoll;
            _rightRoll = -AngleCalculator.RightRoll;
            _leftPitch = AngleCalculator.LeftPitch;
            _rightPitch = AngleCalculator.RightPitch;
            _leftRollAngleMultiplier = Mathf.Lerp(1.25f, 1f, Mathf.Abs(_leftRoll));
            _rightRollAngleMultiplier = Mathf.Lerp(1.25f, 1f, Mathf.Abs(_rightRoll));
            float mass = _jointPhysicsHandler.NormalizedMass;
            _leftLeanBackFixerMultiplier = HandlerUtils.CalculateLeanBackFixerMultiplier(mass, _leftPitch, _leftRoll);
            _rightLeanBackFixerMultiplier = HandlerUtils.CalculateLeanBackFixerMultiplier(mass, _rightPitch, _rightRoll);

            AdjustUpPhysics();
            AdjustDepthOutPhysics();
            AdjustDepthInPhysics();
            AdjustSidewaysPhysics();
        }

        static float UpForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.44f, 1.28f, 0.96f);

        float UpEffect(float angle, float pitchMultiplier) =>
            pitchMultiplier *
            _upMultiplier *
            UserUpMultiplier() *
            UpForceEffectCurve(Mathf.Abs(angle) / _upLimit);

        void AdjustUpPhysics()
        {
            if(Person.LeftBreast.AngleY >= 0)
            {
                float leftPitchMultiplier = Mathf.Lerp(0.63f, 1f, _leftLeanBackFixerMultiplier * AngleCalculator.LeftDiffFromHorizontal);
                UpdatePhysicsLeft(UP, UpEffect(Person.LeftBreast.AngleY, leftPitchMultiplier));
            }
            else
            {
                ResetPhysicsLeft(UP);
            }

            if(Person.RightBreast.AngleY >= 0)
            {
                float rightPitchMultiplier = Mathf.Lerp(0.63f, 1f, _rightLeanBackFixerMultiplier * AngleCalculator.RightDiffFromHorizontal);
                UpdatePhysicsRight(UP, UpEffect(Person.RightBreast.AngleY, rightPitchMultiplier));
            }
            else
            {
                ResetPhysicsRight(UP);
            }
        }

        static float DepthInForceEffectCurve(float x) => Calc.Exponential1(x, 1.66f, 1.70f, 1.06f);

        float DepthInEffect(float depthDiff, float leanBackFixerMultiplier) =>
            leanBackFixerMultiplier *
            _depthInMultiplier *
            UserDepthInMultiplier() *
            DepthInForceEffectCurve(Mathf.Abs(depthDiff) * _depthInLimit);

        void AdjustDepthInPhysics()
        {
            if(Person.LeftBreast.DepthDiff < 0)
            {
                ResetPhysicsLeft(BACK);
            }
            else
            {
                UpdatePhysicsLeft(BACK, DepthInEffect(Person.LeftBreast.DepthDiff, _leftLeanBackFixerMultiplier));
            }

            if(Person.RightBreast.DepthDiff < 0)
            {
                ResetPhysicsRight(BACK);
            }
            else
            {
                UpdatePhysicsRight(BACK, DepthInEffect(Person.RightBreast.DepthDiff, _rightLeanBackFixerMultiplier));
            }
        }

        float DepthOutEffect(float depthDiff, float upsideDownFixerMultiplier) =>
            upsideDownFixerMultiplier *
            _depthOutMultiplier *
            UserDepthOutMultiplier() *
            Calc.Exponential1(Mathf.Abs(depthDiff) * _depthOutLimit, 1.42f, 3.20f, 1.15f);

        void AdjustDepthOutPhysics()
        {
            float leftUpsideDownFixerMultiplier = HandlerUtils.CalculateUpsideDownFixerMultiplier(_leftPitch, _leftRoll);
            float rightUpsideDownFixerMultiplier = HandlerUtils.CalculateUpsideDownFixerMultiplier(_rightPitch, _rightRoll);

            if(Person.LeftBreast.DepthDiff < 0)
            {
                UpdatePhysicsLeft(FORWARD, DepthOutEffect(Person.LeftBreast.DepthDiff, leftUpsideDownFixerMultiplier));
            }
            else
            {
                ResetPhysicsLeft(FORWARD);
            }

            if(Person.RightBreast.DepthDiff <= 0)
            {
                UpdatePhysicsRight(FORWARD, DepthOutEffect(Person.RightBreast.DepthDiff, rightUpsideDownFixerMultiplier));
            }
            else
            {
                ResetPhysicsRight(FORWARD);
            }
        }

        static float SidewaysForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.44f, 1.28f, 0.96f);

        float SidewaysEffect(float angle, float userMultiplier, float rollAngleMultiplier) =>
            _sidewaysMultiplier *
            userMultiplier *
            SidewaysForceEffectCurve(rollAngleMultiplier * Mathf.Abs(angle) / _sidewaysLimit);

        void AdjustSidewaysPhysics()
        {
            if(Person.LeftBreast.AngleX >= 0)
            {
                // right force on left breast
                ResetPhysicsLeft(SIDEWAYS_OUT);
                UpdatePhysicsLeft(SIDEWAYS_IN, SidewaysEffect(Person.LeftBreast.AngleX, UserSidewaysInMultiplier(), _leftRollAngleMultiplier));
            }
            else
            {
                // left force on left breast
                ResetPhysicsLeft(SIDEWAYS_IN);
                UpdatePhysicsLeft(SIDEWAYS_OUT, SidewaysEffect(Person.LeftBreast.AngleX, UserSidewaysOutMultiplier(), _leftRollAngleMultiplier));
            }

            if(Person.RightBreast.AngleX >= 0)
            {
                // right force on right breast
                ResetPhysicsRight(SIDEWAYS_IN);
                UpdatePhysicsRight(SIDEWAYS_OUT, SidewaysEffect(Person.RightBreast.AngleX, UserSidewaysOutMultiplier(), _rightRollAngleMultiplier));
            }
            else
            {
                // left force on right breast
                ResetPhysicsRight(SIDEWAYS_OUT);
                UpdatePhysicsRight(SIDEWAYS_IN, SidewaysEffect(Person.RightBreast.AngleX, UserSidewaysInMultiplier(), _rightRollAngleMultiplier));
            }
        }

        void UpdatePhysicsLeft(string direction, float effect)
        {
            foreach(var param in _leftParams)
            {
                param.UpdateForceValue(direction, effect, _mass, _softness);
            }
        }

        void UpdatePhysicsRight(string direction, float effect)
        {
            foreach(var param in _rightParams)
            {
                param.UpdateForceValue(direction, effect, _mass, _softness);
            }
        }

        internal override void ResetAll()
        {
            foreach(string direction in new[]
            {
                LEFT,
                RIGHT,
                UP,
                DOWN,
                FORWARD,
                BACK,
            })
            {
                ResetPhysicsLeft(direction);
                ResetPhysicsRight(direction);
            }
        }

        void ResetPhysicsLeft(string direction)
        {
            foreach(var param in _leftParams)
            {
                param.ResetForceValue(direction);
            }
        }

        void ResetPhysicsRight(string direction)
        {
            foreach(var param in _rightParams)
            {
                param.ResetForceValue(direction);
            }
        }
    }
}
