using NAT.Common;
using NAT.Common.Handlers;
using UnityEngine;

namespace NAT.TM.Handlers
{
    sealed class BreastFrictionHandler : FrictionHandler
    {
        JointPhysicsHandler _jointPhysicsHandler;
        SoftPhysicsHandler _softPhysicsHandler;

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _jointPhysicsHandler = provider.GetInstance<BreastJointPhysicsHandler>();
            _softPhysicsHandler = provider.GetInstance<BreastSoftPhysicsHandler>();
            SetupStorables();
            LoadSettings();
        }

        protected override void LoadSettings()
        {
            if(!IsEnabled)
            {
                return;
            }

            _softPhysicsHandler.FatColliderRadius.SetFrictionCalculators(
                FatColliderRadiusCalculator(),
                FatColliderRadiusCalculator()
            );
        }

        protected override void UpdateSoftColliders(float friction)
        {
            if(!_softPhysicsHandler.SoftPhysicsOn)
            {
                return;
            }

            if(AdaptiveFrictionJsb.val)
            {
                _softPhysicsHandler.FatColliderRadius.UpdateInverseFrictionValue(
                    1 - Mathf.InverseLerp(0, DrySkinFrictionJsf.val, friction),
                    _softPhysicsHandler.FatColliderRadius.Left.UsesRealMass
                        ? _jointPhysicsHandler.NormalizedRealMass
                        : _jointPhysicsHandler.NormalizedMass,
                    Script.Softness
                );
            }
            else
            {
                _softPhysicsHandler.FatColliderRadius.ResetInverseFrictionValue();
            }

            _softPhysicsHandler.FatColliderFriction.Left.OffsetJsf.val = friction;
            _softPhysicsHandler.FatColliderFriction.Right.OffsetJsf.val = friction;
        }

        /* Maximum friction that a collider can have, drops off dynamically with distance from collider's normal position */
        internal override void CalculateFriction()
        {
            if(!IsEnabled || !Script.IsEnabledNullSafe() || glossJsf == null || specularBumpinessJsf == null)
            {
                return;
            }

            if(!AdaptiveFrictionJsb.val)
            {
                MaxHardColliderFriction = Mathf.Clamp(0.60f + FrictionOffsetJsf.val, 0.00f, 1.00f);
                FrictionValueJsf.valNoCallback = Mathf.Clamp(0.80f + FrictionOffsetJsf.val, 0.00f, 1.00f);
            }
            else
            {
                float normalizedLinearGloss = Mathf.InverseLerp(3.000f, 7.500f, glossJsf.val);
                float normalizedSpecBump = Mathf.InverseLerp(0.200f, 3.000f, specularBumpinessJsf.val);

                /* Gloss based multiplier for how much effect bumpiness has on friction */
                float bumpinessEffectMultiplier = Calc.InverseSmoothStep(normalizedLinearGloss, 0.50f, 0.90f);

                /* Inverse of gloss determines the minimum value for friction along a curve. */
                float minFriction = DrySkinFrictionJsf.val * (1 - Calc.InverseSmoothStep(normalizedLinearGloss, 0.50f, 0.90f));

                /* Hard colliders */
                {
                    float specBumpComponent = 0.350f * bumpinessEffectMultiplier * normalizedSpecBump;
                    float friction = 0.80f * Mathf.Lerp(minFriction, 1.000f, specBumpComponent);
                    MaxHardColliderFriction = Mathf.Clamp(friction + FrictionOffsetJsf.val, 0.00f, 1.00f);
                }

                /* Soft colliders */
                {
                    float specBumpComponent = 0.350f * bumpinessEffectMultiplier * normalizedSpecBump;
                    float friction = Mathf.Lerp(minFriction, 1.000f, specBumpComponent);
                    FrictionValueJsf.valNoCallback = Mathf.Clamp(friction + FrictionOffsetJsf.val, 0.00f, 1.00f);
                }
            }

            UpdateSoftColliders(FrictionValueJsf.val);
        }
    }
}
