using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Models;
using NAT.Common.Storables;
using System.Collections.Generic;
using static NAT.Direction;

namespace NAT.TM.Handlers
{
    sealed class BreastGravityOffsetMorphHandler : Handler
    {
        BreastJointPhysicsHandler _jointPhysicsHandler;
        BreastGravityPhysicsHandler _gravityPhysicsHandler;
        AngleCalculator _angleCalculator;
        Dictionary<string, List<ConfigurableMorph>> _morphSets;

        internal StorableFloat OffsetMorphingJsf { get; private set; }

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _jointPhysicsHandler = provider.GetInstance<BreastJointPhysicsHandler>();
            _gravityPhysicsHandler = provider.GetInstance<BreastGravityPhysicsHandler>();
            _angleCalculator = provider.GetInstance<AngleCalculator>();
            OffsetMorphingJsf = new StorableFloat("Gravity Offset Morphing", 1.00f, 0.00f, 2.00f)
            {
                setCallbackFunction = _ => Script.SetCalibrationPending(OffsetMorphingJsf),
            };
            LoadSettings();
        }

        protected override void LoadSettings() => _morphSets = new Dictionary<string, List<ConfigurableMorph>>
        {
            {
                DOWN_L, new List<ConfigurableMorph>
                {
                    new ConfigurableMorph(
                        Person.GetBreastMorph("DN/DN Breast Rotate Up L"),
                        false,
                        new JSONStorableFloat("softnessMultiplier", 0.76f, 0, 3),
                        new JSONStorableFloat("massMultiplier", 0.855f, 0, 3)
                    ),
                    new ConfigurableMorph(
                        Person.GetBreastMorph("DN/DN Breasts Natural Reverse L"),
                        false,
                        new JSONStorableFloat("softnessMultiplier", 0.36f, 0, 3),
                        new JSONStorableFloat("massMultiplier", 0.405f, 0, 3)
                    ),
                }
            },
            {
                DOWN_R, new List<ConfigurableMorph>
                {
                    new ConfigurableMorph(
                        Person.GetBreastMorph("DN/DN Breast Rotate Up R"),
                        false,
                        new JSONStorableFloat("softnessMultiplier", 0.76f, 0, 3),
                        new JSONStorableFloat("massMultiplier", 0.855f, 0, 3)
                    ),
                    new ConfigurableMorph(
                        Person.GetBreastMorph("DN/DN Breasts Natural Reverse R"),
                        false,
                        new JSONStorableFloat("softnessMultiplier", 0.37f, 0, 3),
                        new JSONStorableFloat("massMultiplier", 0.405f, 0, 3)
                    ),
                }
            },
        };

        internal override void PreRegisterParams()
        {
            Script.PreRegisterMorphsParam(OffsetMorphingJsf);
        }

        float _upDownExtraMultiplier;

        internal void SetMultipliers()
        {
            float curvedMass = Calc.Exponential1(1 - _jointPhysicsHandler.NormalizedRealMass, 0.26f, 2.60f, 5.38f, 0.84f, 0.33f);
            _upDownExtraMultiplier = 0.25f * curvedMass;
        }

        internal void SaveCalibrated() => OffsetMorphingJsf.SaveCalibrated();

        float _mass;
        float _softness;

        internal void Execute(float mass, float softness)
        {
            _mass = mass;
            _softness = softness;
            AdjustUpDownMorphs(
                DOWN_L,
                _gravityPhysicsHandler.ExtraDownGravityLeft,
                _angleCalculator.SmoothLeftRoll,
                _angleCalculator.Smooth2XLeftPitch
            );
            AdjustUpDownMorphs(
                DOWN_R,
                _gravityPhysicsHandler.ExtraDownGravityRight,
                _angleCalculator.SmoothRightRoll,
                _angleCalculator.Smooth2XRightPitch
            );
        }

        void AdjustUpDownMorphs(string direction, float extraGravity, float roll, float pitch)
        {
            float effect =
                OffsetMorphingJsf.val *
                _upDownExtraMultiplier *
                extraGravity *
                AngleCalculator.RollMultiplier(roll) *
                AngleCalculator.UpDownAdjustByAngle(pitch);

            if(pitch > 0)
            {
                // leaning forward
                if(pitch < 1)
                {
                    // upright
                    UpdateMorphs(direction, effect);
                }
                else
                {
                    // upside down
                    ResetMorphs(direction);
                }
            }
            else
            {
                // leaning back
                if(pitch > -1)
                {
                    // upright
                    UpdateMorphs(direction, effect);
                }
                else
                {
                    // upside down
                    ResetMorphs(direction);
                }
            }
        }

        void UpdateMorphs(string direction, float effect)
        {
            foreach(var configurableMorph in _morphSets[direction])
            {
                configurableMorph.SetValue(effect, _mass, _softness);
            }
        }

        internal void ResetAll()
        {
            if(_morphSets == null)
            {
                return;
            }

            foreach(var pair in _morphSets)
            {
                foreach(var configurableMorph in pair.Value)
                {
                    configurableMorph.Reset();
                }
            }
        }

        void ResetMorphs(string direction)
        {
            foreach(var configurableMorph in _morphSets[direction])
            {
                configurableMorph.Reset();
            }
        }
    }
}
