using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;
using System.Collections.Generic;
using UnityEngine;
using static NAT.Direction;

namespace NAT.TM.Handlers
{
    sealed class BreastGravityPhysicsHandler : GravityPhysicsHandler
    {
        BreastJointPhysicsHandler _jointPhysicsHandler;
        internal StorableFloat TwistAngleTargetJsf { get; private set; }

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _jointPhysicsHandler = provider.GetInstance<BreastJointPhysicsHandler>();
            JointMass = _jointPhysicsHandler.JointMass;
            TwistAngleTargetJsf = new StorableFloat(TWIST_ANGLE_TARGET.NAME, 0.00f, -30.00f, 30.00f);
            LoadSettings();
            Person.SetBreastsUseGravity(false);
        }

        protected override void LoadSettings()
        {
            {
                var paramGroup = _jointPhysicsHandler.RotationSpring;
                var leftCalculators = SpringCalculators(Side.LEFT);
                var rightCalculators = SpringCalculators(Side.RIGHT);
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText = "Rotation Spring decreases when leaning forward or back.";
            }

            {
                var paramGroup = _jointPhysicsHandler.RotationDamper;
                var leftCalculators = DamperCalculators(Side.LEFT);
                var rightCalculators = DamperCalculators(Side.RIGHT);
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText =
                    "Rotation Damper increases when leaning back. When leaning to the side, it increases for the breast that rests against the chest.";
            }

            {
                var paramGroup = _jointPhysicsHandler.RightLeftSpring;
                var leftCalculators = RightLeftSpringCalculators(Side.LEFT);
                var rightCalculators = RightLeftSpringCalculators(Side.RIGHT);
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText = "Right/Left Spring increases when leaning to the side.";
            }

            {
                var paramGroup = _jointPhysicsHandler.InOutSpring;
                var leftCalculators = InOutSpringCalculators(Side.LEFT);
                var rightCalculators = InOutSpringCalculators(Side.RIGHT);
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText = "In/Out Spring decreases when leaning forward or back.";
            }

            {
                var paramGroup = _jointPhysicsHandler.UpDownAngleTarget;
                var leftCalculators = UpDownAngleTargetCalculators(Side.LEFT);
                var rightCalculators = UpDownAngleTargetCalculators(Side.RIGHT);
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText =
                    "Up/Down Angle Target decreases when upright and increases when upside down.";
            }

            {
                var paramGroup = _jointPhysicsHandler.RightLeftAngleTarget;
                var leftCalculators = RightLeftAngleTargetCalculatorsLeft();
                var rightCalculators = RightLeftAngleTargetCalculatorsRight();
                paramGroup.SetGravityCalculators(leftCalculators, rightCalculators);
                paramGroup.SetGravityMultiplierCallbacks();
                paramGroup.GravityInfoText =
                    "Right/Left Angle Target is positive when leaning to the right and negative when leaning to the left (the other way around for the right breast).";
            }

            {
                var paramGroup = _jointPhysicsHandler.TwistAngleTarget;
                var calculators = TwistAngleTargetCalculators();
                paramGroup.SetGravityCalculators(calculators[Side.LEFT], calculators[Side.RIGHT]);
            }
        }

        static PhysicsCalculator ForwardSpringCalculator(string side) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_SPRING.NAME, side), FORWARD),
            PrimaryConfig = new MultiplierConfig
            {
                MultiplierA = -18.0f,
                MultiplierB = -9.0f,
            },
        };

        static PhysicsCalculator BackSpringCalculator(string side) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_SPRING.NAME, side), BACK),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0,
                MultiplierA = -15.0f,
                MultiplierB = -7.5f,
            },
        };

        static Dictionary<string, PhysicsCalculator> SpringCalculators(string side) => new Dictionary<string, PhysicsCalculator>
        {
            { BACK, BackSpringCalculator(side) },
            { FORWARD, ForwardSpringCalculator(side) },
        };

        static PhysicsCalculator UpDamperCalculator(string side, string direction) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_DAMPER.NAME, side), direction),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0,
                MultiplierA = 0.25f,
                CurveA = UpGravityMassCurve,
            },
        };

        static PhysicsCalculator ForwardDamperCalculator(string side, string direction) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_DAMPER.NAME, side), direction),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0,
                MultiplierA = 0.60f,
                MultiplierB = 0.30f,
            },
        };

        static PhysicsCalculator BackDamperCalculator(string side, string direction) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_DAMPER.NAME, side), direction),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0,
                MultiplierA = 0.40f,
                MultiplierB = 0.20f,
            },
        };

        static PhysicsCalculator SidewaysInDamperCalculator(string side, string direction) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_DAMPER.NAME, side), direction),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0,
                MultiplierA = 0.24f,
                MultiplierB = 0.12f,
                CurveA = x => 1 - x,
            },
        };

        static Dictionary<string, PhysicsCalculator> DamperCalculators(string side) => new Dictionary<string, PhysicsCalculator>
        {
            { UP, UpDamperCalculator(side, UP) },
            { BACK, BackDamperCalculator(side, BACK) },
            { FORWARD, ForwardDamperCalculator(side, FORWARD) },
            { SIDEWAYS_IN, SidewaysInDamperCalculator(side, SIDEWAYS_IN) },
        };

        static PhysicsCalculator RightLeftSpringCalculator(string side, string direction) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_SPRING.NAME, side), direction),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 1500f,
            },
        };

        static Dictionary<string, PhysicsCalculator> RightLeftSpringCalculators(string side) => new Dictionary<string, PhysicsCalculator>
        {
            { SIDEWAYS_IN, RightLeftSpringCalculator(side, SIDEWAYS_IN) },
            { SIDEWAYS_OUT, RightLeftSpringCalculator(side, SIDEWAYS_OUT) },
        };

        static PhysicsCalculator InOutSpringCalculator(string side, string direction) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(IN_OUT_SPRING.NAME, side), direction),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = -50f,
                MultiplierA = -67f,
                MultiplierB = -83f,
                CurveA = x => Calc.Exponential1(1 - x, 0.3f, 10f, 0.84f, 1.01f),
                CurveB = Calc.InOutSpringSoftnessCurve,
            },
        };

        static Dictionary<string, PhysicsCalculator> InOutSpringCalculators(string side) => new Dictionary<string, PhysicsCalculator>
        {
            { BACK, InOutSpringCalculator(side, BACK) },
            { FORWARD, InOutSpringCalculator(side, FORWARD) },
        };

        static PhysicsCalculator UpAngleTargetCalculator(string side) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(BREAST_UP_DOWN_ANGLE_TARGET.NAME, side), UP),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 2.00f,
            },
        };

        static PhysicsCalculator DownAngleTargetCalculator(string side) => new PhysicsCalculator
        {
            MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(BREAST_UP_DOWN_ANGLE_TARGET.NAME, side), DOWN),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = -2.00f,
            },
        };

        static Dictionary<string, PhysicsCalculator> UpDownAngleTargetCalculators(string side) => new Dictionary<string, PhysicsCalculator>
        {
            { UP, UpAngleTargetCalculator(side) },
            { DOWN, DownAngleTargetCalculator(side) },
        };

        static Dictionary<string, PhysicsCalculator> RightLeftAngleTargetCalculatorsLeft() => new Dictionary<string, PhysicsCalculator>
        {
            {
                SIDEWAYS_IN, new PhysicsCalculator
                {
                    MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_IN),
                    PrimaryConfig = new MultiplierConfig
                    {
                        BaseValue = 2.50f,
                    },
                }
            },
            {
                SIDEWAYS_OUT, new PhysicsCalculator
                {
                    MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_OUT),
                    PrimaryConfig = new MultiplierConfig
                    {
                        BaseValue = -2.50f,
                    },
                }
            },
        };

        static Dictionary<string, PhysicsCalculator> RightLeftAngleTargetCalculatorsRight() => new Dictionary<string, PhysicsCalculator>
        {
            {
                SIDEWAYS_IN, new PhysicsCalculator
                {
                    MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_IN),
                    PrimaryConfig = new MultiplierConfig
                    {
                        BaseValue = -2.50f,
                    },
                }
            },
            {
                SIDEWAYS_OUT, new PhysicsCalculator
                {
                    MultiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_OUT),
                    PrimaryConfig = new MultiplierConfig
                    {
                        BaseValue = 2.50f,
                    },
                }
            },
        };

        static Dictionary<string, PhysicsCalculator> TwistAngleTargetCalculator() => new Dictionary<string, PhysicsCalculator>
        {
            { UP, DefaultPhysicsCalculator() },
            { DOWN, DefaultPhysicsCalculator() },
        };

        Dictionary<string, Dictionary<string, PhysicsCalculator>> TwistAngleTargetCalculators()
        {
            var left = TwistAngleTargetCalculator();
            var right = TwistAngleTargetCalculator();

            TwistAngleTargetJsf.setJSONCallbackFunction = jsf =>
            {
                var storable = (StorableFloat) jsf;
                float rounded = Calc.RoundToDecimals(storable.val);
                left[UP].PrimaryConfig.BaseValue = -rounded;
                left[UP].SecondaryConfig.BaseValue = -rounded;
                left[DOWN].PrimaryConfig.BaseValue = rounded;
                left[DOWN].SecondaryConfig.BaseValue = rounded;
                right[UP].PrimaryConfig.BaseValue = rounded;
                right[UP].SecondaryConfig.BaseValue = rounded;
                right[DOWN].PrimaryConfig.BaseValue = -rounded;
                right[DOWN].SecondaryConfig.BaseValue = -rounded;
                Script.SetCalibrationPending(storable);
                Script.Tabs.PhysicsWindow.OnMainParamAdjusted(_jointPhysicsHandler.TwistAngleTarget);
            };

            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        internal new void SaveCalibrated()
        {
            base.SaveCalibrated();
            TwistAngleTargetJsf.SaveCalibrated();
        }

        static float UpGravityMassCurve(float x) => Calc.InverseSmoothStep(1 - x, 0.67f, 0.98f);
        static float DownGravityMassCurve(float x) => Calc.Exponential1(x, 0.60f, 3.00f, 0.86f);
        static float BackForwardGravityMassCurve(float x) => 1 - x;
        static float SidewaysGravityMassCurve(float x) => Calc.InverseSmoothStep(1 - x, 0.40f, 0.71f);
        static float GravitySoftnessCurve(float x) => Calc.InverseSmoothStep(x, 0.35f, 0.11f);

        internal void SetMultipliers(float mass, float softness)
        {
            float baseMultiplier = Mathf.Lerp(Mathf.Lerp(0, 0.75f, mass), 1.00f, GravitySoftnessCurve(softness));
            upGravityMultiplier = baseMultiplier * Mathf.Lerp(1.00f, 3.20f, UpGravityMassCurve(mass));
            downGravityMultiplier = baseMultiplier * Mathf.Lerp(1.50f, 1.74f, DownGravityMassCurve(mass));
            backGravityMultiplier = baseMultiplier * Mathf.Lerp(1.00f, 1.73f, BackForwardGravityMassCurve(mass));
            forwardGravityMultiplier = baseMultiplier * Mathf.Lerp(0.52f, 2.28f, BackForwardGravityMassCurve(mass));
            sidewaysInGravityMultiplier = baseMultiplier * Mathf.Lerp(1.00f, 1.75f, SidewaysGravityMassCurve(mass));
            sidewaysOutGravityMultiplier = baseMultiplier * Mathf.Lerp(1.00f, 1.75f, SidewaysGravityMassCurve(mass));
            SetGravityValues();
        }

        bool _isCalibrating;
        float _mass;
        float _softness;
        float _leftRoll, _rightRoll;
        float _leftPitch, _rightPitch;

        internal void Execute(bool isCalibrating = false)
        {
            _isCalibrating = isCalibrating;
            _mass = _jointPhysicsHandler.NormalizedMass;
            _softness = Script.Softness;
            _leftRoll = AngleCalculator.SmoothLeftRoll;
            _rightRoll = AngleCalculator.SmoothRightRoll;
            _leftPitch = AngleCalculator.Smooth2XLeftPitch;
            _rightPitch = AngleCalculator.Smooth2XRightPitch;

            if(!IsSimulatingGravity)
            {
                ApplyRealGravity();
            }

            AdjustSidewaysPhysics();
            AdjustVerticalPhysics();
            AdjustHorizontalPhysics();
        }

        void AdjustSidewaysPhysics()
        {
            float leftEffect = Mathf.Abs(_leftRoll);
            if(_leftRoll > 0)
            {
                ResetLeftBreast(SIDEWAYS_IN);
                UpdateLeftBreast(SIDEWAYS_OUT, leftEffect);
            }
            else
            {
                ResetLeftBreast(SIDEWAYS_OUT);
                UpdateLeftBreast(SIDEWAYS_IN, leftEffect);
            }

            float rightEffect = Mathf.Abs(_rightRoll);
            if(_rightRoll > 0)
            {
                ResetRightBreast(SIDEWAYS_OUT);
                UpdateRightBreast(SIDEWAYS_IN, rightEffect);
            }
            else
            {
                ResetRightBreast(SIDEWAYS_IN);
                UpdateRightBreast(SIDEWAYS_OUT, rightEffect);
            }
        }

        void AdjustVerticalPhysics()
        {
            float leftEffect = AngleCalculator.UpDownAdjustByAngle(_leftPitch) * AngleCalculator.RollMultiplier(_leftRoll);
            if(_leftPitch > 1 || _leftPitch < -1)
            {
                // upside down, leaning forward or back
                ResetLeftBreast(DOWN);
                UpdateLeftBreast(UP, leftEffect);
            }
            else
            {
                ResetLeftBreast(UP);
                UpdateLeftBreast(DOWN, leftEffect);
            }

            float rightEffect = AngleCalculator.UpDownAdjustByAngle(_rightPitch) * AngleCalculator.RollMultiplier(_rightRoll);
            if(_rightPitch > 1 || _rightPitch < -1)
            {
                // upside down, leaning forward or back
                ResetRightBreast(DOWN);
                UpdateRightBreast(UP, rightEffect);
            }
            else
            {
                ResetRightBreast(UP);
                UpdateRightBreast(DOWN, rightEffect);
            }
        }

        void AdjustHorizontalPhysics()
        {
            float leftEffect = AngleCalculator.DepthAdjustByAngle(_leftPitch) * AngleCalculator.RollMultiplier(_leftRoll);
            if(_leftPitch > 0)
            {
                // leaning forward, upright or upside down
                ResetLeftBreast(BACK);
                UpdateLeftBreast(FORWARD, leftEffect);
            }
            else
            {
                // leaning back
                ResetLeftBreast(FORWARD);
                UpdateLeftBreast(BACK, leftEffect);
            }

            float rightEffect = AngleCalculator.DepthAdjustByAngle(_rightPitch) * AngleCalculator.RollMultiplier(_rightRoll);
            if(_rightPitch > 0)
            {
                // leaning forward, upright or upside down
                ResetRightBreast(BACK);
                UpdateRightBreast(FORWARD, rightEffect);
            }
            else
            {
                // leaning back
                UpdateRightBreast(BACK, rightEffect);
                ResetRightBreast(FORWARD);
            }
        }

        void UpdateLeftBreast(string direction, float effect)
        {
            _jointPhysicsHandler.RotationSpring.Left.UpdateGravityValue(direction, effect, _mass, _softness);
            if(!_isCalibrating)
            {
                _jointPhysicsHandler.RotationDamper.Left.UpdateGravityValue(direction, effect, _mass, _softness);
            }

            _jointPhysicsHandler.RightLeftSpring.Left.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.InOutSpring.Left.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.UpDownAngleTarget.Left.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.RightLeftAngleTarget.Left.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.TwistAngleTarget.Left.UpdateGravityValue(direction, effect, _mass, _softness);
        }

        void UpdateRightBreast(string direction, float effect)
        {
            _jointPhysicsHandler.RotationSpring.Right.UpdateGravityValue(direction, effect, _mass, _softness);
            if(!_isCalibrating)
            {
                _jointPhysicsHandler.RotationDamper.Right.UpdateGravityValue(direction, effect, _mass, _softness);
            }

            _jointPhysicsHandler.RightLeftSpring.Right.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.InOutSpring.Right.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.UpDownAngleTarget.Right.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.RightLeftAngleTarget.Right.UpdateGravityValue(direction, effect, _mass, _softness);
            _jointPhysicsHandler.TwistAngleTarget.Right.UpdateGravityValue(direction, effect, _mass, _softness);
        }

        internal void ResetAll()
        {
            foreach(string direction in new[]
            {
                SIDEWAYS_IN,
                SIDEWAYS_OUT,
                UP,
                DOWN,
                FORWARD,
                BACK,
            })
            {
                ResetLeftBreast(direction);
                ResetRightBreast(direction);
            }
        }

        void ResetLeftBreast(string direction)
        {
            _jointPhysicsHandler.RotationSpring.Left.ResetGravityValue(direction);
            if(!_isCalibrating)
            {
                _jointPhysicsHandler.RotationDamper.Left.ResetGravityValue(direction);
            }

            _jointPhysicsHandler.RightLeftSpring.Left.ResetGravityValue(direction);
            _jointPhysicsHandler.InOutSpring.Left.ResetGravityValue(direction);
            _jointPhysicsHandler.UpDownAngleTarget.Left.ResetGravityValue(direction);
            _jointPhysicsHandler.RightLeftAngleTarget.Left.ResetGravityValue(direction);
            _jointPhysicsHandler.TwistAngleTarget.Left.ResetGravityValue(direction);
        }

        void ResetRightBreast(string direction)
        {
            _jointPhysicsHandler.RotationSpring.Right.ResetGravityValue(direction);
            if(!_isCalibrating)
            {
                _jointPhysicsHandler.RotationDamper.Right.ResetGravityValue(direction);
            }

            _jointPhysicsHandler.RightLeftSpring.Right.ResetGravityValue(direction);
            _jointPhysicsHandler.InOutSpring.Right.ResetGravityValue(direction);
            _jointPhysicsHandler.UpDownAngleTarget.Right.ResetGravityValue(direction);
            _jointPhysicsHandler.RightLeftAngleTarget.Right.ResetGravityValue(direction);
            _jointPhysicsHandler.TwistAngleTarget.Right.ResetGravityValue(direction);
        }
    }
}
