using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using System.Collections.Generic;
using System.Text;
using UnityEngine;
using static NAT.Side;

namespace NAT.TM.Handlers
{
    sealed class BreastJointPhysicsHandler : JointPhysicsHandler
    {
        BreastGravityPhysicsHandler _gravityPhysicsHandler;

        internal JointPhysicsParamGroup RightLeftSpring { get; private set; }
        internal JointPhysicsParamGroup RightLeftDamper { get; private set; }
        internal JointPhysicsParamGroup TwistAngleTarget { get; private set; }

        JSONStorableBool XMotionEnabledJsb { get; set; }
        JSONStorableBool YMotionEnabledJsb { get; set; }
        JSONStorableBool ZMotionEnabledJsb { get; set; }

        /* Subsets of all breast indices, only ones that actually move */
        internal readonly int[] leftVolumeCalculationVerts =
        {
            2403, 2411, 2592, 7197, 7208, 8902, 8904, 8905, 8908, 8909, 8911, 8915, 8917, 8922, 8924, 8925, 8926, 8928,
            8930, 8933, 8934, 8935, 8939, 8941, 8942, 8943, 8945, 8946, 8947, 8948, 8949, 8951, 8952, 8954, 8959, 8969,
            8970,
        };

        internal readonly int[] rightVolumeCalculationVerts =
        {
            13233, 13241, 13410, 17904, 17915, 19568, 19570, 19571, 19574, 19575, 19577, 19581, 19583, 19588, 19590,
            19591, 19592, 19594, 19596, 19599, 19600, 19601, 19605, 19607, 19608, 19609, 19611, 19612, 19613, 19614,
            19615, 19617, 19618, 19620, 19625, 19635, 19636,
        };

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _gravityPhysicsHandler = provider.GetInstance<BreastGravityPhysicsHandler>();
            XMotionEnabledJsb = new JSONStorableBool("Right/Left Motion Enabled", true, SetXMotionEnabled);
            YMotionEnabledJsb = new JSONStorableBool("Up/Down Motion Enabled", false, SetYMotionEnabled);
            ZMotionEnabledJsb = new JSONStorableBool("In/Out Motion Enabled", true, SetZMotionEnabled);
            SetXMotionEnabled(XMotionEnabledJsb.val);
            SetYMotionEnabled(YMotionEnabledJsb.val);
            SetZMotionEnabled(ZMotionEnabledJsb.val);
            LoadSettings();
        }

        void SetXMotionEnabled(bool value)
        {
            var motion = value ? ConfigurableJointMotion.Limited : ConfigurableJointMotion.Locked;
            Person.LeftBreast.Joint.xMotion = motion;
            Person.RightBreast.Joint.xMotion = motion;
        }

        void SetYMotionEnabled(bool value)
        {
            var motion = value ? ConfigurableJointMotion.Limited : ConfigurableJointMotion.Locked;
            Person.LeftBreast.Joint.yMotion = motion;
            Person.RightBreast.Joint.yMotion = motion;
        }

        void SetZMotionEnabled(bool value)
        {
            var motion = value ? ConfigurableJointMotion.Limited : ConfigurableJointMotion.Locked;
            Person.LeftBreast.Joint.zMotion = motion;
            Person.RightBreast.Joint.zMotion = motion;
        }

        internal void CalculateMass()
        {
            float leftVolume = CalculateVolume(leftVolumeCalculationVerts);
            float rightVolume = CalculateVolume(rightVolumeCalculationVerts);
            float volume = (leftVolume + rightVolume) / 2;

            // https://www.desmos.com/calculator/aw68taob0a
            float value = Mathf.Pow(0.614f * volume, 1.290f);
            float massBaseValue = Mathf.Clamp(value, BREAST_JOINT_MASS.MIN, BREAST_JOINT_MASS.MAX);
            UpdateMassValueAndAmounts(massBaseValue);
        }

        internal void UpdateMassValueAndAmounts(float mass)
        {
            CalibratedMass = mass;
            JointMass.UpdateMassValue(mass);
            JointMass.UpdateOffsetMinMax();
            const float min = BREAST_JOINT_MASS.MIN;
            const float max = BREAST_JOINT_MASS.MAX;
            NormalizedMass = Mathf.InverseLerp(min, max, JointMass.Left.Value);
            NormalizedRealMass = Mathf.InverseLerp(min, max, JointMass.Left.BaseValueJsf.val);
        }

        float CalculateVolume(int[] vertexIndices)
        {
            var rawSkinnedVerts = Person.Skin.rawSkinnedVerts;
            var positions = new Vector3[vertexIndices.Length];
            for(int i = 0; i < vertexIndices.Length; i++)
            {
                positions[i] = Calc.RelativePosition(Person.ChestRb, rawSkinnedVerts[vertexIndices[i]]);
            }

            return 1000 * VolumeCalculator.CupVolume(positions, Person.ScaleJsf.val);
        }

        protected override void LoadSettings()
        {
            JointMass = new JointPhysicsParamGroup(Script)
            {
                Name = BREAST_JOINT_MASS.NAME,
                ValueFormat = "F3",
                Left = BreastMassParam(LEFT, Person.LeftBreast),
                Right = BreastMassParam(RIGHT, Person.RightBreast),
                RequiresCalibration = true,
                InfoText = new StringBuilder()
                    .AppendBold(BREAST_JOINT_MASS.NAME)
                    .AppendLine(" is the mass of the joint.")
                    .AppendLine()
                    .Append("Since mass represents breast size, other physics parameters are adjusted based on its value.")
                    .ToString(),
            };

            CenterOfGravity = new JointPhysicsParamGroup(Script)
            {
                Name = CENTER_OF_GRAVITY.NAME,
                ValueFormat = "F2",
                Left = CenterOfGravityParam(LEFT, Person.LeftBreast),
                Right = CenterOfGravityParam(RIGHT, Person.RightBreast),
                RequiresCalibration = true,
                InfoText = new StringBuilder()
                    .AppendBold(CENTER_OF_GRAVITY.NAME)
                    .AppendLine(" is the position of the joint's center of mass.")
                    .AppendLine()
                    .Append("At 0, the center of mass is inside the chest at the pectoral joint. At 1, it is at the nipple.")
                    .ToString(),
            };

            RotationSpring = new JointPhysicsParamGroup(Script)
            {
                Name = ROTATION_SPRING.NAME,
                ValueFormat = "0",
                Left = RotationSpringParam(LEFT, Person.LeftBreast),
                Right = RotationSpringParam(RIGHT, Person.RightBreast),
                RequiresCalibration = true,
                InfoText = new StringBuilder()
                    .AppendBold(ROTATION_SPRING.NAME)
                    .AppendLine(" pushes the joint towards its angle target.")
                    .AppendLine()
                    .Append("The angle target is defined by the Up/Down, Left/Right and Twist Angle Target parameters.")
                    .ToString(),
            };

            RotationDamper = new JointPhysicsParamGroup(Script)
            {
                Name = ROTATION_DAMPER.NAME,
                ValueFormat = "F2",
                Left = RotationDamperParam(LEFT, Person.LeftBreast),
                Right = RotationDamperParam(RIGHT, Person.RightBreast),
                DependsOnPhysicsRate = true,
                InfoText = new StringBuilder()
                    .AppendBold(ROTATION_DAMPER.NAME)
                    .AppendLine(" reduces oscillation around the joint's angle target.")
                    .AppendLine()
                    .Append("The higher the damper, the quicker breasts will stop swinging.")
                    .ToString(),
            };

            // AngularSpring = new JointPhysicsParamGroup(Script)
            // {
            //     Name = ANGULAR_SPRING.NAME,
            //     ValueFormat = "0",
            //     Left = AngularSpringParam(LEFT, Person.LeftBreast),
            //     Right = AngularSpringParam(RIGHT, Person.RightBreast),
            //     InfoText = new StringBuilder()
            //         .AppendBold(ANGULAR_SPRING.NAME)
            //         .ToString(),
            // };

            // AngularDamper = new JointPhysicsParamGroup(Script)
            // {
            //     Name = ANGULAR_DAMPER.NAME,
            //     valueFormat = "F2",
            //     Left = AngularDamperParam(LEFT, Person.LeftBreast),
            //     Right = AngularDamperParam(RIGHT, Person.RightBreast),
            //     DependsOnPhysicsRate = true,
            //     InfoText = new StringBuilder()
            //         .AppendBold(ANGULAR_DAMPER.NAME)
            //         .ToString(),
            // };

            // UpDownSpring = new JointPhysicsParamGroup(Script)
            // {
            //     Name = UP_DOWN_SPRING.NAME,
            //     ValueFormat = "0",
            //     Left = UpDownSpringParam(LEFT, Person.LeftBreast),
            //     Right = UpDownSpringParam(RIGHT, Person.RightBreast),
            //     RequiresCalibration = true,
            //     EnabledJsb = YMotionEnabledJsb,
            //     InfoText = new StringBuilder()
            //         .AppendBold(UP_DOWN_SPRING.NAME)
            //         .Append(" pushes the joint towards its position target along the Y axis.")
            //         .ToString(),
            // };

            // UpDownDamper = new JointPhysicsParamGroup(Script)
            // {
            //     Name = UP_DOWN_DAMPER.NAME,
            //     valueFormat = "F2",
            //     Left = UpDownDamperParam(LEFT, Person.LeftBreast),
            //     Right = UpDownDamperParam(RIGHT, Person.RightBreast),
            //     DependsOnPhysicsRate = true,
            //     EnabledJsb = YMotionEnabledJsb,
            //     InfoText = new StringBuilder()
            //         .AppendBold(UP_DOWN_DAMPER.NAME)
            //         .Append(" reduces oscillation around the joint's position target along the Y axis.")
            //         .ToString(),
            // };

            RightLeftSpring = new JointPhysicsParamGroup(Script)
            {
                Name = RIGHT_LEFT_SPRING.NAME,
                ValueFormat = "0",
                Left = RightLeftSpringParam(LEFT, Person.LeftBreast),
                Right = RightLeftSpringParam(RIGHT, Person.RightBreast),
                RequiresCalibration = true,
                EnabledJsb = XMotionEnabledJsb,
                InfoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_SPRING.NAME)
                    .Append(" pushes the joint towards its position target along the X axis.")
                    .ToString(),
            };

            RightLeftDamper = new JointPhysicsParamGroup(Script)
            {
                Name = RIGHT_LEFT_DAMPER.NAME,
                ValueFormat = "F2",
                Left = RightLeftDamperParam(LEFT, Person.LeftBreast),
                Right = RightLeftDamperParam(RIGHT, Person.RightBreast),
                DependsOnPhysicsRate = true,
                EnabledJsb = XMotionEnabledJsb,
                InfoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_DAMPER.NAME)
                    .Append(" reduces oscillation around the joint's position target along the X axis.")
                    .ToString(),
            };

            InOutSpring = new JointPhysicsParamGroup(Script)
            {
                Name = IN_OUT_SPRING.NAME,
                ValueFormat = "0",
                Left = InOutSpringParam(LEFT, Person.LeftBreast),
                Right = InOutSpringParam(RIGHT, Person.RightBreast),
                RequiresCalibration = true,
                EnabledJsb = ZMotionEnabledJsb,
                InfoText = new StringBuilder()
                    .AppendBold(IN_OUT_SPRING.NAME)
                    .AppendLine(" pushes the joint towards its position target along the Z axis.")
                    .AppendLine()
                    .Append("Directional force morphing along the forward-back axis depends on In/Out Spring being suitably low for the given breast mass.")
                    .ToString(),
            };

            InOutDamper = new JointPhysicsParamGroup(Script)
            {
                Name = IN_OUT_DAMPER.NAME,
                ValueFormat = "F2",
                Left = InOutDamperParam(LEFT, Person.LeftBreast),
                Right = InOutDamperParam(RIGHT, Person.RightBreast),
                DependsOnPhysicsRate = true,
                EnabledJsb = ZMotionEnabledJsb,
                InfoText = new StringBuilder()
                    .AppendBold(IN_OUT_DAMPER.NAME)
                    .Append(" reduces oscillation around the joint's position target along the Z axis.")
                    .ToString(),
            };

            var leftUpDownAngleTarget = UpDownAngleTargetParam(LEFT);
            var leftRightLeftAngleTarget = RightLeftAngleTargetParam(LEFT);
            var leftTwistAngleTarget = TwistAngleTargetParam(LEFT);

            var rightUpDownAngleTarget = UpDownAngleTargetParam(RIGHT);
            var rightRightLeftAngleTarget = RightLeftAngleTargetParam(RIGHT);
            var rightTwistAngleTarget = TwistAngleTargetParam(RIGHT);

            leftTwistAngleTarget.Sync = value => SyncAngleTarget(
                LEFT,
                Person.LeftBreast,
                leftUpDownAngleTarget.Value,
                leftRightLeftAngleTarget.Value,
                value
            );
            rightTwistAngleTarget.Sync = value => SyncAngleTarget(
                RIGHT,
                Person.RightBreast,
                rightUpDownAngleTarget.Value,
                rightRightLeftAngleTarget.Value,
                value
            );

            UpDownAngleTarget = new JointPhysicsParamGroup(Script)
            {
                Name = BREAST_UP_DOWN_ANGLE_TARGET.NAME,
                ValueFormat = "F2",
                Left = leftUpDownAngleTarget,
                Right = rightUpDownAngleTarget,
                RequiresCalibration = true,
                InfoText = new StringBuilder()
                    .AppendBold(BREAST_UP_DOWN_ANGLE_TARGET.NAME)
                    .Append(" is the joint's vertical target angle. Negative values pull breasts down, positive values push them up.")
                    .ToString(),
            };

            RightLeftAngleTarget = new JointPhysicsParamGroup(Script)
            {
                Name = RIGHT_LEFT_ANGLE_TARGET.NAME,
                ValueFormat = "F2",
                Left = leftRightLeftAngleTarget,
                Right = rightRightLeftAngleTarget,
                RequiresCalibration = true,
                RightInverted = true,
                InfoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_ANGLE_TARGET.NAME)
                    .Append(" is the joint's horizontal target angle. A negative offset pulls breasts apart, while a positive offset pushes them together.")
                    .ToString(),
            };

            TwistAngleTarget = new JointPhysicsParamGroup(Script)
            {
                Name = TWIST_ANGLE_TARGET.NAME,
                ValueFormat = "F2",
                Left = leftTwistAngleTarget,
                Right = rightTwistAngleTarget,
                RequiresCalibration = true,
                RightInverted = true, // correct but has no effect because offset not adjusted directly
                InfoText = new StringBuilder()
                    .AppendBold(TWIST_ANGLE_TARGET.NAME)
                    .AppendLine(" is the joint's forward axis target angle.")
                    .AppendLine()
                    .Append("The final value depends on the person's pose. The offset determines the max angle when the person is upright.")
                    .Append(" The angle is inverted when upside down and zero when horizontal.")
                    .ToString(),
            };

            ParamGroups = new List<JointPhysicsParamGroup>
            {
                CenterOfGravity,
                RotationSpring,
                RotationDamper,
                // AngularSpring,
                // AngularDamper,
                // UpDownSpring,
                // UpDownDamper,
                RightLeftSpring,
                RightLeftDamper,
                InOutSpring,
                InOutDamper,
                UpDownAngleTarget,
                RightLeftAngleTarget,
                TwistAngleTarget,
            };

            JointMass.SetCallbackFunctions();
            foreach(var group in ParamGroups)
            {
                group.SetCallbackFunctions();
            }
        }

        JointPhysicsParam BreastMassParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            BREAST_JOINT_MASS.NAME,
            BREAST_JOINT_MASS.MIN,
            BREAST_JOINT_MASS.MAX
        )
        {
            Sync = value => SyncBreastMass(target.Rigidbody, value),
        };

        JointPhysicsParam CenterOfGravityParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            CENTER_OF_GRAVITY.NAME,
            CENTER_OF_GRAVITY.MIN,
            CENTER_OF_GRAVITY.MAX
        )
        {
            Sync = value => SyncCenterOfGravity(target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0.51f,
                MultiplierA = 0.30f,
                MultiplierB = 0.00f,
                // TODO curveA
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = -0.06f,
                MultiplierA = 0.24f,
                MultiplierB = 0.00f,
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = 0.06f,
                MultiplierA = 0.24f,
                MultiplierB = 0.00f,
            },
        };

        JointPhysicsParam RotationSpringParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            ROTATION_SPRING.NAME,
            ROTATION_SPRING.MIN,
            ROTATION_SPRING.MAX
        )
        {
            Sync = value => SyncRotationSpring(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 88f,
                MultiplierA = 0.38f,
                MultiplierB = -0.60f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.46f, 0.10f),
                CurveB = x => Calc.InverseSmoothStep(x, 0.10f, 0.13f),
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = 6f,
                MultiplierA = 0.67f,
                MultiplierB = 0.00f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = -9f,
                MultiplierA = 0.67f,
                MultiplierB = 0.00f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
            },
        };

        JointPhysicsParam RotationDamperParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            ROTATION_DAMPER.NAME,
            ROTATION_DAMPER.MIN,
            ROTATION_DAMPER.MAX
        )
        {
            Sync = value => SyncRotationDamper(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 1.10f,
                MultiplierA = 0.00f,
                MultiplierB = -0.75f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.46f, 0.10f),
                CurveB = x => Calc.InverseSmoothStep(x, 0.24f, 0.61f),
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = -0.08f,
                MultiplierA = 8.375f,
                MultiplierB = 0.00f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = 0.04f,
                MultiplierA = 8.375f,
                MultiplierB = 0.00f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
            },
        };

        // JointPhysicsParam AngularSpringParam(string side, CalibratedJoint target) => new JointPhysicsParam(
        //     side,
        //     ANGULAR_SPRING.NAME,
        //     ANGULAR_SPRING.MIN,
        //     ANGULAR_SPRING.MAX
        // )
        // {
        //     Sync = value => SyncAngularSpring(target.Joint, target.Rigidbody, value),
        //     PrimaryConfig = new MultiplierConfig(),
        // };

        // JointPhysicsParam AngularDamperParam(string side, CalibratedJoint target) => new JointPhysicsParam(
        //     side,
        //     ANGULAR_DAMPER.NAME,
        //     ANGULAR_DAMPER.MIN,
        //     ANGULAR_DAMPER.MAX
        // )
        // {
        //     Sync = value => SyncAngularDamper(target.Joint, target.Rigidbody, value),
        //     PrimaryConfig = new MultiplierConfig(),
        // };

        // JointPhysicsParam UpDownSpringParam(string side, CalibratedJoint target) => new JointPhysicsParam(
        //     side,
        //     UP_DOWN_SPRING.NAME,
        //     UP_DOWN_SPRING.MIN,
        //     UP_DOWN_SPRING.MAX
        // )
        // {
        //     Sync = value => SyncUpDownSpring(target.Joint, target.Rigidbody, value),
        //     PrimaryConfig = new MultiplierConfig
        //     {
        //         BaseValue = 2000f,
        //     },
        // };

        // JointPhysicsParam UpDownDamperParam(string side, CalibratedJoint target) => new JointPhysicsParam(
        //     side,
        //     UP_DOWN_DAMPER.NAME,
        //     UP_DOWN_DAMPER.MIN,
        //     UP_DOWN_DAMPER.MAX
        // )
        // {
        //     Sync = value => SyncUpDownDamper(target.Joint, target.Rigidbody, value),
        //     PrimaryConfig = new MultiplierConfig
        //     {
        //         BaseValue = 10f,
        //     },
        // };

        JointPhysicsParam RightLeftSpringParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            RIGHT_LEFT_SPRING.NAME,
            RIGHT_LEFT_SPRING.MIN,
            RIGHT_LEFT_SPRING.MAX
        )
        {
            Sync = value => SyncRightLeftSpring(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 1500f,
                MultiplierA = 0.50f,
                MultiplierB = -0.77f,
                CurveA = x => Calc.InverseSmoothStep(x, 0.53f, 0.70f),
                CurveB = x => Calc.InverseSmoothStep(x, 0.60f, 0.62f),
            },
        };

        JointPhysicsParam RightLeftDamperParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            RIGHT_LEFT_DAMPER.NAME,
            RIGHT_LEFT_DAMPER.MIN,
            RIGHT_LEFT_DAMPER.MAX
        )
        {
            Sync = value => SyncRightLeftDamper(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 25f,
                MultiplierB = -0.69f,
                CurveB = x => Calc.InverseSmoothStep(x, 0.60f, 0.62f),
            },
        };

        JointPhysicsParam InOutSpringParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            IN_OUT_SPRING.NAME,
            IN_OUT_SPRING.MIN,
            IN_OUT_SPRING.MAX
        )
        {
            Sync = value => SyncInOutSpring(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 1000f,
                MultiplierA = -0.40f,
                CurveA = x => 1 - x, // TODO smoothstep
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = 42f,
                MultiplierA = -0.28f,
                MultiplierB = 0.00f,
                CurveA = x => 1 - x,
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = -63f,
                MultiplierA = -0.28f,
                MultiplierB = 0.00f,
                CurveA = x => 1 - x,
            },
        };

        JointPhysicsParam InOutDamperParam(string side, CalibratedJoint target) => new JointPhysicsParam(
            side,
            IN_OUT_DAMPER.NAME,
            IN_OUT_DAMPER.MIN,
            IN_OUT_DAMPER.MAX
        )
        {
            Sync = value => SyncInOutDamper(target.Joint, target.Rigidbody, value),
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 11.5f,
                MultiplierA = 0f,
                MultiplierB = -0.37f,
            },
            QuicknessConfig = new MultiplierConfig
            {
                BaseValue = -1.33f,
                MultiplierA = 1.33f,
                MultiplierB = 0.00f,
            },
            SlownessConfig = new MultiplierConfig
            {
                BaseValue = 1.33f,
                MultiplierA = 1.33f,
                MultiplierB = 0.00f,
            },
        };

        static JointPhysicsParam UpDownAngleTargetParam(string side) => new JointPhysicsParam(
            side,
            BREAST_UP_DOWN_ANGLE_TARGET.NAME,
            BREAST_UP_DOWN_ANGLE_TARGET.MIN,
            BREAST_UP_DOWN_ANGLE_TARGET.MAX
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                BaseValue = -2.00f,
            },
        };

        static JointPhysicsParam RightLeftAngleTargetParam(string side) => new JointPhysicsParam(
            side,
            RIGHT_LEFT_ANGLE_TARGET.NAME,
            RIGHT_LEFT_ANGLE_TARGET.MIN,
            RIGHT_LEFT_ANGLE_TARGET.MAX
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        static JointPhysicsParam TwistAngleTargetParam(string side) => new JointPhysicsParam(
            side,
            TWIST_ANGLE_TARGET.NAME,
            TWIST_ANGLE_TARGET.MIN,
            TWIST_ANGLE_TARGET.MAX
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        // Reimplements AdjustJoints.cs method SyncMass
        void SyncBreastMass(Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            if(Mathf.Abs(rb.mass - value) <= 0.001f)
            {
                return;
            }

            rb.mass = value;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncCenterOfGravity
        void SyncCenterOfGravity(Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var newCenterOfMass = Vector3.Lerp(
                Person.BreastControl.lowCenterOfGravity,
                Person.BreastControl.highCenterOfGravity,
                value
            );
            if(Calc.IsVectorEqualWithin(1 / 100f, rb.centerOfMass, newCenterOfMass))
            {
                return;
            }

            rb.centerOfMass = newCenterOfMass;
            rb.WakeUp();
        }

        void SyncRotationSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            // TODO set multiplier
            float scalePow = Mathf.Pow(1.7f, Person.BreastControl.scale - 1f);

            float scaledSpring = value * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionSpring - scaledSpring) <= 1)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionSpring = scaledSpring;
            slerpDrive.maximumForce = DEFAULT_SLERP_MAX_FORCE * scalePow; // TODO set once
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionSpring = scaledSpring;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionSpring = 2 * scaledSpring;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.spring = 2 * scaledSpring * Person.BreastControl.limitSpringMultiplier; // 1.5f
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.spring = 2 * scaledSpring * Person.BreastControl.limitSpringMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        void SyncRotationDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            // TODO set multiplier
            float scalePow = Mathf.Pow(1.7f, Person.BreastControl.scale - 1f);

            float scaledDamper = value * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionDamper - scaledDamper) <= 0.01f)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionDamper = scaledDamper;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionDamper = 2 * scaledDamper;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionDamper = 2 * scaledDamper;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.damper = 2 * scaledDamper * Person.BreastControl.limitDamperMultiplier; // 0
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.damper = 2 * scaledDamper * Person.BreastControl.limitDamperMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        // void SyncAngularSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        // {
        //     if(!Script.enabled)
        //     {
        //         return;
        //     }
        //
        //     // see AdjustJoints.cs method ScaleChanged
        //     float scalePow = Mathf.Pow(1.7f, Person.BreastControl.scale - 1f);
        //
        //     float scaledSpring = value * scalePow;
        //     if(Mathf.Abs(joint.angularXDrive.positionSpring - scaledSpring) <= 1)
        //     {
        //         return;
        //     }
        //
        //     var angularXDrive = joint.angularXDrive;
        //     angularXDrive.positionSpring = scaledSpring;
        //     joint.angularXDrive = angularXDrive;
        //
        //     var angularYZDrive = joint.angularYZDrive;
        //     angularYZDrive.positionSpring = scaledSpring;
        //     joint.angularYZDrive = angularYZDrive;
        //
        //     var angularXLimitSpring = joint.angularXLimitSpring;
        //     angularXLimitSpring.spring = scaledSpring * Person.BreastControl.limitSpringMultiplier;
        //     joint.angularXLimitSpring = angularXLimitSpring;
        //
        //     var angularYZLimitSpring = joint.angularYZLimitSpring;
        //     angularYZLimitSpring.spring = scaledSpring * Person.BreastControl.limitSpringMultiplier;
        //     joint.angularYZLimitSpring = angularYZLimitSpring;
        //
        //     rb.WakeUp();
        // }

        // void SyncAngularDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        // {
        //     if(!Script.enabled)
        //     {
        //         return;
        //     }
        //
        //     // see AdjustJoints.cs method ScaleChanged
        //     float scalePow = Mathf.Pow(1.7f, Person.BreastControl.scale - 1f);
        //
        //     float scaledDamper = value * scalePow;
        //     if(Mathf.Abs(joint.angularXDrive.positionDamper - scaledDamper) <= 0.01f)
        //     {
        //         return;
        //     }
        //
        //     Debug.Log($"limitDamperMultiplier {Person.BreastControl.limitDamperMultiplier}" +
        //         $"\nlimitSpringMultiplier {Person.BreastControl.limitSpringMultiplier}");
        //
        //     var angularXDrive = joint.angularXDrive;
        //     angularXDrive.positionDamper = scaledDamper;
        //     joint.angularXDrive = angularXDrive;
        //
        //     var angularYZDrive = joint.angularYZDrive;
        //     angularYZDrive.positionDamper = scaledDamper;
        //     joint.angularYZDrive = angularYZDrive;
        //
        //     var angularXLimitSpring = joint.angularXLimitSpring;
        //     angularXLimitSpring.damper = scaledDamper * Person.BreastControl.limitDamperMultiplier;
        //     joint.angularXLimitSpring = angularXLimitSpring;
        //
        //     var angularYZLimitSpring = joint.angularYZLimitSpring;
        //     angularYZLimitSpring.damper = scaledDamper * Person.BreastControl.limitDamperMultiplier;
        //     joint.angularYZLimitSpring = angularYZLimitSpring;
        //
        //     rb.WakeUp();
        // }

        // void SyncUpDownSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        // {
        //     if(!Script.enabled)
        //     {
        //         return;
        //     }
        //
        //     var drive = joint.yDrive;
        //     if(Mathf.Abs(drive.positionSpring - value) <= 1)
        //     {
        //         return;
        //     }
        //
        //     drive.positionSpring = value;
        //     joint.yDrive = drive;
        //     rb.WakeUp();
        // }

        // void SyncUpDownDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        // {
        //     if(!Script.enabled)
        //     {
        //         return;
        //     }
        //
        //     var drive = joint.yDrive;
        //     if(Mathf.Abs(drive.positionDamper - value) <= 1)
        //     {
        //         return;
        //     }
        //
        //     drive.positionDamper = value;
        //     joint.yDrive = drive;
        //     rb.WakeUp();
        // }

        void SyncRightLeftSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.xDrive;
            if(Mathf.Abs(drive.positionSpring - value) <= 1)
            {
                return;
            }

            drive.positionSpring = value;
            joint.xDrive = drive;
            rb.WakeUp();
        }

        void SyncRightLeftDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.xDrive;
            if(Mathf.Abs(drive.positionDamper - value) <= 0.01f)
            {
                return;
            }

            drive.positionDamper = value;
            joint.xDrive = drive;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        void SyncInOutSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.zDrive;
            if(Mathf.Abs(drive.positionSpring - value) <= 1)
            {
                return;
            }

            drive.positionSpring = value;
            joint.zDrive = drive;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        void SyncInOutDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.zDrive;
            if(Mathf.Abs(drive.positionDamper - value) <= 0.01f)
            {
                return;
            }

            drive.positionDamper = value;
            joint.zDrive = drive;
            rb.WakeUp();
        }

        readonly Dictionary<string, int> _rotationZZeroCount = new Dictionary<string, int>
        {
            { LEFT, 0 },
            { RIGHT, 0 },
        };

        void SyncAngleTarget(string side, CalibratedJoint target, float x, float y, float z)
        {
            if(!Script.enabled)
            {
                return;
            }

            bool noZRotation = _gravityPhysicsHandler.TwistAngleTargetJsf.val == 0;
            if(_rotationZZeroCount[side] < 3)
            {
                if(noZRotation)
                {
                    _rotationZZeroCount[side]++;
                }
            }
            else if(!noZRotation)
            {
                _rotationZZeroCount[side] = 0;
            }

            target.SetRotation(x, y, z);
        }

        internal override void PreRegisterParams()
        {
            Script.PreRegisterJointPhysicsParam(JointMass.Left.OffsetJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.LinkDirectionsJsb);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.UpGravityJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.DownGravityJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.BackGravityJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.ForwardGravityJsf);
            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.SidewaysGravityJsf);

            foreach(var paramGroup in ParamGroups)
            {
                if(paramGroup == TwistAngleTarget)
                {
                    continue;
                }

                var leftParam = paramGroup.Left;
                Script.PreRegisterJointPhysicsParam(leftParam.OffsetJsf);

                foreach(var pair in leftParam.GravityCalculators)
                {
                    Script.PreRegisterJointPhysicsParam(pair.Value.MultiplierJsf);
                }
            }

            Script.PreRegisterJointPhysicsParam(_gravityPhysicsHandler.TwistAngleTargetJsf);
        }

        internal void SaveCalibrated()
        {
            JointMass.SaveCalibrated();
            foreach(var group in ParamGroups)
            {
                group.SaveCalibrated();
            }
        }

        internal void UpdateBaseValues(bool updateOffsetMinMax = true)
        {
            foreach(var group in ParamGroups)
            {
                group.UpdateBaseValue(
                    NormalizedMass,
                    Script.Softness,
                    Script.Quickness
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }
        }

        internal void SetGravitySimulationDamper()
        {
            float interpolant = Calc.InverseSmoothStep(NormalizedMass, 0.50f, 0.10f);
            RotationDamper.OverrideValue(Mathf.Lerp(2.5f, 10.00f, interpolant));
        }

        internal void UnsetGravitySimulationDamper() =>
            RotationDamper.UpdateBaseValue(
                NormalizedMass,
                Script.Softness,
                Script.Quickness
            );

        internal void RestoreOriginalPhysics()
        {
            if(Script.Initialized != true)
            {
                return;
            }

            XMotionEnabledJsb.val = false;
            YMotionEnabledJsb.val = false;
            ZMotionEnabledJsb.val = true;

            foreach(string name in new[]
            {
                "mass",
                "centerOfGravityPercent",
                "spring",
                "damper",
                "positionSpringZ",
                "positionDamperZ",
                "targetRotationX",
                "targetRotationY",
                "targetRotationZ",
            })
            {
                /* Set a value that is different from the original, then restore the original
                 * in order to trigger VaM's internal sync
                 */
                var paramJsf = Person.BreastControl.GetFloatJSONParam(name);
                float original = paramJsf.val;
                paramJsf.valNoCallback = Mathf.Abs(paramJsf.val - paramJsf.min) > 0.01f
                    ? paramJsf.min
                    : paramJsf.max;
                paramJsf.val = original;
            }
        }
    }
}
