using ColliderEditor;
using NAT.Common;
using NAT.Common.Diagnostics;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;
using NAT.TM.Models;
using NAT.TM.UI;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using static NAT.Side;
using static NAT.SoftVerticesGroupName;

namespace NAT.TM.Handlers
{
    sealed class BreastSoftPhysicsHandler : SoftPhysicsHandler
    {
        internal Dictionary<string, SoftJoint> NippleSoftJoints { get; private set; }

        Dictionary<string, SoftJoint[]> _nonGradientJoints;
        Dictionary<string, SoftJointGradientGroup[]> _gradientGroups;
        Dictionary<string, Dictionary<string, DAZPhysicsMeshSoftVerticesGroup>> _softVerticesGroups;

        internal StorableBool SoftPhysicsSelfCollideJsb { get; private set; }

        bool _pluginSoftPhysicsToggled;
        bool _ldpOn;

        internal override bool IsLdpUpdateOn() => _ldpOn && Script.Bootstrap.PersonMonitor.IsBreastSoftPhysicsOn();

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            angleCalculator = provider.GetInstance<AngleCalculator>();
            AllSoftJoints = new Dictionary<string, SoftJoint[]>();
            _nonGradientJoints = new Dictionary<string, SoftJoint[]>();
            _gradientGroups = new Dictionary<string, SoftJointGradientGroup[]>();
            _softVerticesGroups = new Dictionary<string, Dictionary<string, DAZPhysicsMeshSoftVerticesGroup>>();

            InitColliderVisualizer();
            SetupSoftGroups();
            EnableMultiplyFriction();

            softJointsFixedUpdate = new CustomFixedUpdate(Script.Bootstrap.LdpUpdateRateJssc);
            softJointsFixedUpdate.Actions += () => UpdateSoftJointsDistortion();
            SetupStorables();
            LoadSettings();

            SetupSoftJoints();
            SoftJointDiagnostics = new SoftJointDiagnostics(
                AllSoftJoints[LEFT],
                AllSoftJoints[RIGHT],
                Script.gameObject.AddComponent<DrawService>(),
                Script.Bootstrap.Camera
            );
            if(LdpEnabledJsb.val)
            {
                SoftJointDiagnostics.VisualizeLdpJsb.val = LdpEnabledJsb.val;
            }

            SoftPhysicsOn = SoftPhysicsOnJsb.val;
        }

        void InitColliderVisualizer()
        {
            ColliderVisualizer = Script.gameObject.AddComponent<ColliderVisualizer>();
            var groups = new List<Group>
            {
                new Group("Off", @"$off"), //match nothing
                new Group("Both breasts", @"PhysicsMeshJoint(left|right)(areola|nipple|outer)?\d+$"),
                new Group("Left breast", @"PhysicsMeshJointleft(areola|nipple|outer)?\d+$"),
                new Group("Right breast", @"PhysicsMeshJointright(areola|nipple|outer)?\d+$"),
            };
            ColliderVisualizer.Init(Script, groups);
            ColliderVisualizer.PreviewOpacityJSON.val = 0.500f;
            ColliderVisualizer.PreviewOpacityJSON.defaultVal = 0.500f;
            ColliderVisualizer.SelectedPreviewOpacityJSON.val = 0.500f;
            ColliderVisualizer.SelectedPreviewOpacityJSON.defaultVal = 0.500f;
            ColliderVisualizer.GroupsJSON.val = "Left breast";
            ColliderVisualizer.GroupsJSON.defaultVal = "Left breast";
            ColliderVisualizer.HighlightMirrorJSON.val = true;
            ColliderVisualizer.ShowPreviewsJSON.val = false;

            foreach(string option in new[] { Constant.SELECT, "Other", "All" })
            {
                ColliderVisualizer.GroupsJSON.choices.Remove(option);
            }

            ShowColliderPreviewsJsb = new JSONStorableBool("Show Collider Previews", true, (bool _) =>
            {
                if(Script.Tabs.IsPhysicsTabActive() && Script.Tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    ((SoftParameterWindow) Script.Tabs.GetActiveNestedWindow()).SetShowColliderPreviews();
                }
            });
        }

        void SetupSoftGroups()
        {
            SetupSoftVerticesGroups(LEFT);
            SetupSoftVerticesGroups(RIGHT);
        }

        void SetupSoftVerticesGroups(string side)
        {
            var groups = Person.BreastPhysicsMesh.softVerticesGroups;
            _softVerticesGroups[side] = new Dictionary<string, DAZPhysicsMeshSoftVerticesGroup>
            {
                { OUTER, groups.Find(group => group.name == $"{side}{OUTER}") },
                { MAIN, groups.Find(group => group.name == side) },
                { AREOLA, groups.Find(group => group.name == $"{side}{AREOLA}") },
                { NIPPLE, groups.Find(group => group.name == $"{side}{NIPPLE}") },
            };
        }

        void SetupSoftJoints()
        {
            OuterSoftJoints = new Dictionary<string, SoftJoint[]>();
            MainSoftJoints = new Dictionary<string, SoftJoint[]>();
            CenterSoftJoints = new Dictionary<string, SoftJoint[]>();
            NippleSoftJoints = new Dictionary<string, SoftJoint>();
            SetupSoftJoints(LEFT);
            SetupSoftJoints(RIGHT);
            AllSoftJointsList.AddRange(AllSoftJoints[LEFT]);
            AllSoftJointsList.AddRange(AllSoftJoints[RIGHT]);
        }

        void SetupSoftJoints(string side)
        {
            var allSJs = new List<SoftJoint>();
            var exceptNippleSJs = new List<SoftJoint>();
            var outerSJs = new List<SoftJoint>();
            var mainSJs = new List<SoftJoint>();
            var areolaSJs = new List<SoftJoint>();
            var nippleSJs = new List<SoftJoint>();

            foreach(var kvp in _softVerticesGroups[side])
            {
                string category = kvp.Key;
                var group = kvp.Value;
                foreach(var set in group.softVerticesSets)
                {
                    string id = $"{side}.{category}.{set.targetVertex.ToString()}";
                    var softJoint = new SoftJoint(id, set);
                    softJoint.SetParamGroups(
                        FatMass,
                        FatSpring,
                        FatDamper,
                        FatColliderRadius,
                        FatColliderDepth,
                        FatColliderFriction,
                        FatDistanceLimit
                    );
                    switch(category)
                    {
                        case OUTER:
                            outerSJs.Add(softJoint);
                            break;
                        case MAIN:
                            mainSJs.Add(softJoint);
                            break;
                        case AREOLA:
                            areolaSJs.Add(softJoint);
                            break;
                        case NIPPLE:
                            nippleSJs.Add(softJoint);
                            break;
                    }

                    allSJs.Add(softJoint);
                    if(category != NIPPLE)
                    {
                        exceptNippleSJs.Add(softJoint);
                    }
                }
            }

            AllSoftJoints[side] = allSJs.ToArray();
            OuterSoftJoints[side] = outerSJs.ToArray();
            MainSoftJoints[side] = mainSJs.ToArray();
            CenterSoftJoints[side] = areolaSJs.ToArray();
            NippleSoftJoints[side] = nippleSJs.First();

            var nearbyVerts = VertexIndices.NearbyVerts(side);
            var adjacentVerts = VertexIndices.AdjacentVerts(side);
            foreach(var softJoint in allSJs)
            {
                int target = softJoint.SoftVerticesSet.targetVertex;
                int[] vert;
                softJoint.NearbyJoints = nearbyVerts.TryGetValue(target, out vert)
                    ? allSJs.Where(x => vert.Contains(x.SoftVerticesSet.targetVertex)).ToArray()
                    : new SoftJoint[] { }; // nipple verts do not have nearby verts

                int[] adjacentVert;
                softJoint.AdjacentJoints = adjacentVerts.TryGetValue(target, out adjacentVert)
                    ? allSJs.Where(x => adjacentVert.Contains(x.SoftVerticesSet.targetVertex)).ToArray()
                    : new SoftJoint[] { }; // areola and nipple verts do not have adjacent verts
            }

            var gradientGroupsList = new List<SoftJointGradientGroup>();
            foreach(int[] indexGroup in VertexIndices.CenterToOuterVerts(side))
            {
                var gradientGroup = new SoftJointGradientGroup
                {
                    Areola = areolaSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[0])),
                    Main1 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[1])),
                    Main2 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2])),
                    Outer1 = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[3])),
                };

                if(indexGroup.Length > 4)
                {
                    gradientGroup.Outer2 = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[4]));
                }

                if(indexGroup.Length > 5)
                {
                    gradientGroup.Outer3 = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[5]));
                }

                gradientGroupsList.Add(gradientGroup);
            }

            _gradientGroups[side] = gradientGroupsList.ToArray();

            // TODO hard code
            _nonGradientJoints[side] = exceptNippleSJs
                .Where(softJoint => !_gradientGroups[side].Any(x => x.HasJoint(softJoint)))
                .ToArray();
        }

        static bool MatchByVertex(SoftJoint softJoint, int vertex) => softJoint.SoftVerticesSet.targetVertex == vertex;

        void SetupStorables()
        {
            bool ldpEnabled = Person.BreastPhysicsMesh.on && Script.Bootstrap.BreastsLdpDefaultEnabledJsb.val;
            SoftPhysicsOnJsb = new StorableBool(Constant.BREAST_SOFT_PHYSICS_ENABLED, Person.BreastPhysicsMesh.on);
            LdpEnabledJsb = new StorableBool("Local Distortion Physics Enabled", ldpEnabled);
            SoftPhysicsSelfCollideJsb = new StorableBool("Breast Soft Physics Self Collide", true);

            SoftPhysicsOnJsb.setCallbackFunction = OnToggleSoftPhysics;
            LdpEnabledJsb.setJSONCallbackFunction = OnToggleLdp;
            SoftPhysicsSelfCollideJsb.setCallbackFunction = SyncAllowSelfCollision;
        }

        void OnToggleSoftPhysics(bool value)
        {
            if(Script.CalibrationStateHandler.CalibratingJsb.val)
            {
                Script.Loggr.LogMessage("Cannot enable/disable soft physics while calibrating.");
                SoftPhysicsOnJsb.valNoCallback = !value;
                return;
            }

            LdpEnabledJsb.val = Script.Bootstrap.BreastsLdpDefaultEnabledJsb.val && value;

            if(!_reverseSyncingSoftPhysics)
            {
                _pluginSoftPhysicsToggled = true;
                if(Script.enabled)
                {
                    Person.BreastPhysicsMesh.on = value;
                }
                else
                {
                    Script.Loggr.LogMessage($"Enable the plugin to set {SoftPhysicsOnJsb.name}.");
                }

                if(Script.Tabs.IsMainTabActive())
                {
                    Script.Tabs.MainWindow.Refresh();
                }
            }

            _reverseSyncingSoftPhysics = false;
        }

        void OnToggleLdp(JSONStorableBool jsb)
        {
            if(!Script.enabled)
            {
                Script.Loggr.LogMessage($"Enable the plugin to set {LdpEnabledJsb.name}.");
                return;
            }

            var storable = (StorableBool) jsb;
            if(storable.val)
            {
                if(!SoftPhysicsOnJsb.val)
                {
                    Script.Loggr.LogMessage("Enable breast soft physics to use local distortion physics.");
                }
                else
                {
                    Script.SetCalibrationPending(storable);
                }
            }
            else
            {
                if(SoftPhysicsOn)
                {
                    Script.SetCalibrationPending(storable);
                }
            }

            SoftJointDiagnostics.VisualizeLdpJsb.val = jsb.val;

            if(Script.Tabs.IsMainTabActive())
            {
                Script.Tabs.MainWindow.Refresh();
            }
        }

        protected override void LoadSettings()
        {
            FatSpring = new BreastSoftPhysicsParamGroup(Script, this)
            {
                Name = FAT_SPRING.NAME,
                Left = FatSpringParam(LEFT),
                Right = FatSpringParam(RIGHT),
                ValueFormat = "F1",
                VisualizeRelativeChange = true,
                VisualizationLimit = 125f,
                ClothingAdjustmentVisualizationLimit = 1000f,
                DistortionLimit = 500f,
                InfoText = new StringBuilder()
                    .AppendBold(FAT_SPRING.NAME)
                    .AppendLine(" holds each soft joint in its target position.")
                    .AppendLine()
                    .Append("Low fat spring makes breast fat soft and slow. High fat spring makes it rigid and quick to return to its normal shape.")
                    .ToString(),
            };

            FatDamper = new BreastSoftPhysicsParamGroup(Script, this)
            {
                Name = FAT_DAMPER.NAME,
                Left = FatDamperParam(LEFT),
                Right = FatDamperParam(RIGHT),
                ValueFormat = "F2",
                VisualizeRelativeChange = true,
                VisualizationLimit = 2.50f,
                ClothingAdjustmentVisualizationLimit = 5.00f,
                DistortionLimit = 5.00f,
                InfoText = new StringBuilder()
                    .AppendBold(FAT_DAMPER.NAME)
                    .AppendLine(" reduces oscillation of each soft joint around its target position.")
                    .AppendLine()
                    .Append("Low fat damper makes breast fat jiggle more easily.")
                    .ToString(),
            };

            FatMass = new BreastSoftPhysicsParamGroup(Script, this)
            {
                Name = FAT_MASS.NAME,
                Left = FatMassParam(LEFT),
                Right = FatMassParam(RIGHT),
                ValueFormat = "F3",
                VisualizeRelativeChange = true,
                VisualizationLimit = 0.150f,
                ClothingAdjustmentVisualizationLimit = 0.150f,
                DistortionLimit = 0.300f,
                InfoText = new StringBuilder()
                    .AppendBold(FAT_MASS.NAME)
                    .Append(" determines the mass of each soft joint and the overall density of breast fat.")
                    .Append(" Increasing breast size without increasing fat mass reduces density.")
                    .ToString(),
            };

            FatColliderRadius = new BreastSoftPhysicsParamGroup(Script, this)
            {
                Name = FAT_COLLIDER_RADIUS.NAME,
                Left = FatColliderRadiusParam(LEFT),
                Right = FatColliderRadiusParam(RIGHT),
                ValueFormat = "F4",
                HasSoftColliderVisualization = true,
                VisualizeRelativeChange = false,
                VisualizationLimit = 0.0100f,
                ClothingAdjustmentVisualizationLimit = 0.0100f,
                DistortionLimit = 0.0200f,
                InfoText = new StringBuilder()
                    .AppendBold(FAT_COLLIDER_RADIUS.NAME)
                    .AppendLine("is a scaling factor for the size of the colliders connected to each soft joint.")
                    .AppendLine()
                    .Append("The actual size varies per collider and automatically scales with breast size.")
                    .ToString(),
            };

            FatColliderDepth = new BreastSoftPhysicsParamGroup(Script, this)
            {
                Name = FAT_COLLIDER_DEPTH.NAME,
                ValueFormat = "F4",
                Left = FatColliderDepthParam(LEFT),
                Right = FatColliderDepthParam(RIGHT),
                HasSoftColliderVisualization = true,
                VisualizeRelativeChange = false,
                VisualizationLimit = 0.0020f,
                ClothingAdjustmentVisualizationLimit = 0.0020f,
                DistortionLimit = 0.0020f,
                InfoText = new StringBuilder()
                    .AppendBold(FAT_COLLIDER_DEPTH.NAME)
                    .AppendLine(" offsets collider positions relative to skin surface")
                    .AppendLine()
                    .Append("Negative values pull colliders out from the breast, positive values push them into the breast.")
                    .ToString(),
            };

            FatColliderFriction = new BreastSoftPhysicsParamGroup(Script, this)
            {
                Name = FAT_COLLIDER_FRICTION.NAME,
                ValueFormat = "F3",
                Left = FatColliderFrictionParam(LEFT),
                Right = FatColliderFrictionParam(RIGHT),
                HasSoftColliderVisualization = true,
                VisualizeRelativeChange = true,
                VisualizationLimit = 0.500f,
                ClothingAdjustmentVisualizationLimit = 0.500f,
                DistortionLimit = 1.000f,
                InfoText = new StringBuilder()
                    .AppendBold(FAT_COLLIDER_FRICTION.NAME)
                    .Append(" sets the stickiness of individual soft colliders. Click below to configure adaptive friction and to offset the amount of friction.")
                    .ToString(),
            };

            FatDistanceLimit = new BreastSoftPhysicsParamGroup(Script, this)
            {
                Name = FAT_DISTANCE_LIMIT.NAME,
                ValueFormat = "F3",
                Left = FatDistanceLimitParam(LEFT),
                Right = FatDistanceLimitParam(RIGHT),
                VisualizeRelativeChange = true,
                VisualizationLimit = 0.020f,
                ClothingAdjustmentVisualizationLimit = 0.020f,
                DistortionLimit = 0.020f,
                InfoText = new StringBuilder()
                    .AppendBold(FAT_DISTANCE_LIMIT.NAME)
                    .Append(" is the maximum distance each soft joint can move away from its target position.")
                    .ToString(),
            };

            SoftParamGroups = new[]
            {
                FatSpring,
                FatDamper,
                FatMass,
                FatColliderRadius,
                FatColliderDepth,
                FatColliderFriction,
                FatDistanceLimit,
            };

            FatBackForce = new JointPhysicsParamGroup(Script)
            {
                Name = FAT_BACK_FORCE.NAME,
                ValueFormat = "F2",
                Left = FatBackForceParam(LEFT),
                Right = FatBackForceParam(RIGHT),
                InfoText = new StringBuilder()
                    .AppendBold(FAT_BACK_FORCE.NAME)
                    .Append(" applies a force to the pectoral joint based on the movement of soft joints. A high value can cause an out of control feedback loop.")
                    .ToString(),
            };

            FatBackForceMaxForce = new JointPhysicsParamGroup(Script)
            {
                Name = FAT_BACK_FORCE_MAX_FORCE.NAME,
                ValueFormat = "F2",
                Left = FatBackForceMaxForceParam(LEFT),
                Right = FatBackForceMaxForceParam(RIGHT),
                InfoText = new StringBuilder()
                    .AppendBold(FAT_BACK_FORCE_MAX_FORCE.NAME)
                    .AppendLine(" sets an upper limit on the magnitude of back force.")
                    .AppendLine()
                    .AppendFormat("Along with {0}, this can be used to prevent an out of control feedback loop.", FAT_BACK_FORCE_THRESHOLD.NAME)
                    .ToString(),
            };

            FatBackForceThreshold = new JointPhysicsParamGroup(Script)
            {
                Name = FAT_BACK_FORCE_THRESHOLD.NAME,
                ValueFormat = "F3",
                Left = FatBackForceThresholdParam(LEFT),
                Right = FatBackForceThresholdParam(RIGHT),
                InfoText = new StringBuilder()
                    .AppendBold(FAT_BACK_FORCE_THRESHOLD.NAME)
                    .AppendLine(" sets a minimum distance each soft joint needs to move for back force to be applied.")
                    .AppendLine()
                    .AppendFormat("Along with {0}, this can be used to prevent an out of control feedback loop.", FAT_BACK_FORCE_MAX_FORCE.NAME)
                    .ToString(),
            };

            ParamGroups = new[]
            {
                FatBackForce,
                FatBackForceMaxForce,
                FatBackForceThreshold,
            };

            foreach(var group in SoftParamGroups)
            {
                group.SetCallbackFunctions();
            }

            foreach(var group in ParamGroups)
            {
                group.SetCallbackFunctions();
            }
        }

        Rigidbody[] GetPositionalRigidbodies(string side, string group)
        {
            var rigidbodies = new List<Rigidbody>();
            foreach(var set in _softVerticesGroups[side][group].softVerticesSets)
            {
                var indices = VertexIndices.PositionalSoftVerts(side, group);
                if(indices.Contains(set.targetVertex))
                {
                    rigidbodies.Add(set.jointRB);
                }
            }

            return rigidbodies.ToArray();
        }

        Dictionary<string, Dictionary<int, int>> _softJointMirrorMaps;

        internal IEnumerable<SoftJoint> FindMirrorJoints(string fromSide, string toSide, List<SoftJoint> joints)
        {
            _softJointMirrorMaps = _softJointMirrorMaps ?? new Dictionary<string, Dictionary<int, int>>();
            if(!_softJointMirrorMaps.ContainsKey(fromSide))
            {
                _softJointMirrorMaps[fromSide] = VertexIndices.MirrorVerts(toSide);
            }

            var mirrorMap = _softJointMirrorMaps[fromSide];
            var allJoints = AllSoftJoints[toSide];

            foreach(var joint in joints)
            {
                yield return allJoints.First(x => x.SoftVerticesSet.targetVertex == mirrorMap[joint.SoftVerticesSet.targetVertex]);
            }
        }

        static float FatSpringSoftnessCurve(float x) => Calc.InverseSmoothStep(x, 0.35f, 0.11f);

        SoftPhysicsParam NewSoftPhysicsParameter(string side, string name, float min, float max)
        {
            var parameter = new SoftPhysicsParam(Constant.TITTY_MAGIC, side, name, min, max);
            parameter.Sync = (value, additionalOffset) => SyncGradientParameter(side, name, parameter.GradientConfig, value, additionalOffset);
            return parameter;
        }

        SoftPhysicsParam FatSpringParam(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_SPRING.NAME, FAT_SPRING.MIN, FAT_SPRING.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 1000f,
                MultiplierA = 0.00f,
                MultiplierB = -0.959f,
                CurveB = FatSpringSoftnessCurve,
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                BaseValue = 1000f,
                MultiplierA = 0.00f,
                MultiplierB = -0.954f,
                CurveB = FatSpringSoftnessCurve,
            };
            parameter.QuicknessConfig = new MultiplierConfig
            {
                BaseValue = 50f,
                MultiplierA = 0.00f,
                MultiplierB = -0.75f,
            };
            parameter.SlownessConfig = new MultiplierConfig
            {
                BaseValue = -125f,
                MultiplierA = 0.00f,
                MultiplierB = -0.96f,
            };

            var gradientConfig = parameter.GradientConfig;
            gradientConfig.MiddleShiftJsf.SetValueAndDefault(-0.40f);
            gradientConfig.OuterMultiplierJsf.SetValueAndDefault(2.00f);
            gradientConfig.CenterMultiplierJsf.SetValueAndDefault(1.50f);
            gradientConfig.NippleMultiplierJsf.SetValueAndDefault(2.25f);
            gradientConfig.OuterToMiddleCurvature = 0.30f;
            gradientConfig.OuterToMiddleMidPoint = 0.50f;
            gradientConfig.MiddleToCenterCurvature = 0.40f;
            gradientConfig.MiddleToCenterMidPoint = 0.70f;

            var ldpConfig = parameter.LdpConfig;
            ldpConfig.CompressionEffectJsf.SetValueAndDefault(0.30f);
            ldpConfig.CompressionSensitivityJsf.SetValueAndDefault(0.85f);
            ldpConfig.StretchingEffectJsf.SetValueAndDefault(0.30f);
            ldpConfig.StretchingSensitivityJsf.SetValueAndDefault(0.85f);

            return parameter;
        }

        SoftPhysicsParam FatDamperParam(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_DAMPER.NAME, FAT_DAMPER.MIN, FAT_DAMPER.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 1.60f,
                MultiplierA = 0.18f,
                MultiplierB = -0.55f,
                CurveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                CurveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                BaseValue = 1.60f,
                MultiplierA = 0.18f,
                MultiplierB = -0.55f,
                CurveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                CurveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.QuicknessConfig = new MultiplierConfig
            {
                BaseValue = -1.60f,
                MultiplierA = 0.18f,
                MultiplierB = -0.55f,
                CurveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                CurveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.SlownessConfig = new MultiplierConfig
            {
                BaseValue = 0.80f,
                MultiplierA = 0.18f,
                MultiplierB = -0.55f,
                CurveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                CurveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.GradientConfig;
            gradientConfig.MiddleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.OuterMultiplierJsf.SetValueAndDefault(2.30f);
            gradientConfig.CenterMultiplierJsf.SetValueAndDefault(1.50f);
            gradientConfig.NippleMultiplierJsf.SetValueAndDefault(1.80f);
            gradientConfig.OuterToMiddleCurvature = 0.30f;
            gradientConfig.OuterToMiddleMidPoint = 0.50f;
            gradientConfig.MiddleToCenterCurvature = 0.40f;
            gradientConfig.MiddleToCenterMidPoint = 0.70f;

            var ldpConfig = parameter.LdpConfig;
            ldpConfig.CompressionEffectJsf.SetValueAndDefault(-0.40f);
            ldpConfig.CompressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.StretchingEffectJsf.SetValueAndDefault(-0.40f);
            ldpConfig.StretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        SoftPhysicsParam FatMassParam(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_MASS.NAME, FAT_MASS.MIN, FAT_MASS.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0.045f,
                MultiplierA = 1.31f,
                MultiplierB = 0.60f,
                CurveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                CurveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                BaseValue = 0.045f,
                MultiplierA = 1.31f,
                MultiplierB = 0.60f,
                CurveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                CurveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.QuicknessConfig = new MultiplierConfig
            {
                BaseValue = -0.038f,
                MultiplierA = 1.31f,
                MultiplierB = 0.60f,
                CurveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.SlownessConfig = new MultiplierConfig
            {
                BaseValue = 0.038f,
                MultiplierA = 1.31f,
                MultiplierB = 0.60f,
                CurveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.GradientConfig;
            gradientConfig.MiddleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.OuterMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.CenterMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.NippleMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.OuterToMiddleCurvature = 0.30f;
            gradientConfig.OuterToMiddleMidPoint = 0.50f;
            gradientConfig.MiddleToCenterCurvature = 0.40f;
            gradientConfig.MiddleToCenterMidPoint = 0.70f;

            var ldpConfig = parameter.LdpConfig;
            ldpConfig.CompressionEffectJsf.SetValueAndDefault(0.50f);
            ldpConfig.CompressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.StretchingEffectJsf.SetValueAndDefault(-0.25f);
            ldpConfig.StretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        SoftPhysicsParam FatColliderRadiusParam(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_COLLIDER_RADIUS.NAME, FAT_COLLIDER_RADIUS.MIN, FAT_COLLIDER_RADIUS.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0.0142f,
                MultiplierA = 1.80f,
                MultiplierB = 0.20f,
                CurveA = x => Calc.Exponential1(x, 1.42f, 4.25f, 1.17f),
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                BaseValue = 0.0153f,
                MultiplierA = 1.80f,
                MultiplierB = 0.20f,
                CurveA = x => Calc.Exponential1(x, 1.42f, 4.25f, 1.17f),
            };

            var gradientConfig = parameter.GradientConfig;
            gradientConfig.MiddleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.OuterMultiplierJsf.SetValueAndDefault(1.10f);
            gradientConfig.CenterMultiplierJsf.SetValueAndDefault(1.25f);
            gradientConfig.NippleMultiplierJsf.SetValueAndDefault(0.33f);
            gradientConfig.OuterToMiddleCurvature = 0.30f;
            gradientConfig.OuterToMiddleMidPoint = 0.50f;
            gradientConfig.MiddleToCenterCurvature = 0.40f;
            gradientConfig.MiddleToCenterMidPoint = 0.70f;

            var ldpConfig = parameter.LdpConfig;
            ldpConfig.CompressionEffectJsf.SetValueAndDefault(0.10f);
            ldpConfig.CompressionSensitivityJsf.SetValueAndDefault(0.70f);
            ldpConfig.StretchingEffectJsf.SetValueAndDefault(0.40f);
            ldpConfig.StretchingSensitivityJsf.SetValueAndDefault(0.70f);

            parameter.UsesRealMass = true;
            return parameter;
        }

        SoftPhysicsParam FatColliderDepthParam(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_COLLIDER_DEPTH.NAME, FAT_COLLIDER_DEPTH.MIN, FAT_COLLIDER_DEPTH.MAX);
            parameter.PrimaryConfig = new MultiplierConfig();
            parameter.SecondaryConfig = new MultiplierConfig();

            var gradientConfig = parameter.GradientConfig;
            gradientConfig.MiddleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.OuterMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.CenterMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.NippleMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.OuterToMiddleCurvature = 0.30f;
            gradientConfig.OuterToMiddleMidPoint = 0.50f;
            gradientConfig.MiddleToCenterCurvature = 0.30f;
            gradientConfig.MiddleToCenterMidPoint = 0.50f;

            var ldpConfig = parameter.LdpConfig;
            ldpConfig.CompressionEffectJsf.SetValueAndDefault(0.00f);
            ldpConfig.CompressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.StretchingEffectJsf.SetValueAndDefault(1.00f);
            ldpConfig.StretchingSensitivityJsf.SetValueAndDefault(0.90f);

            return parameter;
        }

        SoftPhysicsParam FatColliderFrictionParam(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_COLLIDER_FRICTION.NAME, FAT_COLLIDER_FRICTION.MIN, FAT_COLLIDER_FRICTION.MAX);
            parameter.PrimaryConfig = new MultiplierConfig();
            parameter.SecondaryConfig = new MultiplierConfig();

            var gradientConfig = parameter.GradientConfig;
            gradientConfig.MiddleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.OuterMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.CenterMultiplierJsf.SetValueAndDefault(1.25f);
            gradientConfig.NippleMultiplierJsf.SetValueAndDefault(1.50f);
            gradientConfig.OuterToMiddleCurvature = 0.60f;
            gradientConfig.OuterToMiddleMidPoint = 0.60f;

            var ldpConfig = parameter.LdpConfig;
            ldpConfig.CompressionEffectJsf.SetValueAndDefault(0.00f);
            ldpConfig.CompressionSensitivityJsf.SetValueAndDefault(0.50f);
            ldpConfig.StretchingEffectJsf.SetValueAndDefault(0.30f);
            ldpConfig.StretchingSensitivityJsf.SetValueAndDefault(0.50f);

            return parameter;
        }

        SoftPhysicsParam FatDistanceLimitParam(string side)
        {
            var parameter = NewSoftPhysicsParameter(side, FAT_DISTANCE_LIMIT.NAME, FAT_DISTANCE_LIMIT.MIN, FAT_DISTANCE_LIMIT.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                BaseValue = 0.0155f,
                MultiplierA = 3.68f,
                MultiplierB = 0.52f,
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                BaseValue = 0.020f,
                MultiplierA = 3.68f,
                MultiplierB = 0.52f,
            };

            var gradientConfig = parameter.GradientConfig;
            gradientConfig.MiddleShiftJsf.SetValueAndDefault(-0.80f);
            gradientConfig.OuterMultiplierJsf.SetValueAndDefault(0.80f);
            gradientConfig.CenterMultiplierJsf.SetValueAndDefault(0.80f);
            gradientConfig.NippleMultiplierJsf.SetValueAndDefault(0.86f);
            gradientConfig.OuterToMiddleCurvature = 0.30f;
            gradientConfig.OuterToMiddleMidPoint = 0.50f;
            gradientConfig.MiddleToCenterCurvature = 0.30f;
            gradientConfig.MiddleToCenterMidPoint = 0.50f;

            var ldpConfig = parameter.LdpConfig;
            ldpConfig.CompressionEffectJsf.SetValueAndDefault(0.00f);
            ldpConfig.CompressionSensitivityJsf.SetValueAndDefault(0.80f);
            ldpConfig.StretchingEffectJsf.SetValueAndDefault(1.00f);
            ldpConfig.StretchingSensitivityJsf.SetValueAndDefault(0.80f);

            parameter.UsesRealMass = true;
            return parameter;
        }

        JointPhysicsParam FatBackForceParam(string side)
        {
            var parameter = new JointPhysicsParam(
                side,
                FAT_BACK_FORCE.NAME,
                FAT_BACK_FORCE.MIN,
                FAT_BACK_FORCE.MAX
            )
            {
                Sync = value => SyncBackForce(side, value),
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0.72f,
                    MultiplierA = 4.50f,
                    MultiplierB = 17.50f,
                    CurveA = x => Calc.InverseSmoothStep(x, 0.73f, 0.93f),
                    CurveB = x => 1 - Calc.InverseSmoothStep(x, 0.70f, 0.07f),
                },
                SecondaryConfig = new MultiplierConfig
                {
                    BaseValue = 0.36f,
                    MultiplierA = 18.00f,
                    MultiplierB = 35.00f,
                    CurveA = x => Calc.InverseSmoothStep(x, 0.73f, 0.93f),
                    CurveB = x => 1 - Calc.InverseSmoothStep(x, 0.70f, 0.07f),
                },
                QuicknessConfig = new MultiplierConfig
                {
                    BaseValue = -0.36f,
                    MultiplierA = 8.25f,
                    MultiplierB = 0.00f,
                    CurveA = x => Calc.InverseSmoothStep(x, 0.73f, 0.93f),
                },
                SlownessConfig = new MultiplierConfig
                {
                    BaseValue = 0.72f,
                    MultiplierA = 8.25f,
                    MultiplierB = 0.00f,
                },
            };

            return parameter;
        }

        JointPhysicsParam FatBackForceMaxForceParam(string side)
        {
            var parameter = new JointPhysicsParam(
                side,
                FAT_BACK_FORCE_MAX_FORCE.NAME,
                FAT_BACK_FORCE_MAX_FORCE.MIN,
                FAT_BACK_FORCE_MAX_FORCE.MAX
            )
            {
                Sync = value => SyncBackForceMaxForce(side, value),
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 50.00f,
                    MultiplierA = 0.00f,
                    MultiplierB = 0.00f,
                },
                SecondaryConfig = new MultiplierConfig
                {
                    BaseValue = 50.00f,
                    MultiplierA = 0.00f,
                    MultiplierB = 0.00f,
                },
            };

            return parameter;
        }

        JointPhysicsParam FatBackForceThresholdParam(string side)
        {
            var parameter = new JointPhysicsParam(
                side,
                FAT_BACK_FORCE_THRESHOLD.NAME,
                FAT_BACK_FORCE_THRESHOLD.MIN,
                FAT_BACK_FORCE_THRESHOLD.MAX
            )
            {
                Sync = value => SyncBackForceThresholdDistance(side, value),
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0.001f,
                    MultiplierA = 0.00f,
                    MultiplierB = 0.00f,
                },
                SecondaryConfig = new MultiplierConfig
                {
                    BaseValue = 0.001f,
                    MultiplierA = 0.00f,
                    MultiplierB = 0.00f,
                },
            };

            return parameter;
        }

        void SyncGradientParameter(string side, string paramName, GradientConfig gradientConfig, float value, float additionalOffset)
        {
            if(Script.Initialized != true || !Script.enabled || Script.RestoringFromJSON)
            {
                // Syncing parameter before first time calibration (e.g. due to offset storable callback) results in NaN errors
                return;
            }

            var gradientGroups = _gradientGroups[side];
            foreach(var group in gradientGroups)
            {
                group.SetGradientBaseValues(paramName, gradientConfig, value);
                group.SetAdditionalOffsets(paramName, additionalOffset);
            }

            /* Calculate remaining physics values for joints not included in gradients.
             * Values are based on neighboring (immediately adjacent) joints on the same "ring".
             */
            foreach(var joint in _nonGradientJoints[side])
            {
                float average = 0f;
                foreach(var x in joint.AdjacentJoints)
                {
                    average += x.GetBaseValue(paramName);
                }

                average /= joint.AdjacentJoints.Length;

                joint.SetBaseValue(paramName, average);
                joint.SetAdditionalOffset(paramName, additionalOffset);
            }

            NippleSoftJoints[side]
                .SetBaseValue(
                    paramName,
                    value * gradientConfig.NippleMultiplierJsf.val
                );
            NippleSoftJoints[side].SetAdditionalOffset(paramName, additionalOffset);
        }

        void SyncBackForce(string side, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            foreach(var kvp in _softVerticesGroups[side])
            {
                kvp.Value.jointBackForce = value;
            }
        }

        void SyncBackForceMaxForce(string side, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            foreach(var kvp in _softVerticesGroups[side])
            {
                kvp.Value.jointBackForceMaxForce = value;
            }
        }

        void SyncBackForceThresholdDistance(string side, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            foreach(var kvp in _softVerticesGroups[side])
            {
                kvp.Value.jointBackForceThresholdDistance = value;
            }
        }

        bool _reverseSyncingSoftPhysics;

        /* Update value if external value changed */
        internal void ReverseSyncSoftPhysicsOn()
        {
            if(!_pluginSoftPhysicsToggled)
            {
                _reverseSyncingSoftPhysics = true;
                SoftPhysicsOnJsb.valNoCallback = Person.BreastPhysicsMesh.on;
            }

            _pluginSoftPhysicsToggled = false;
        }

        /* Update value if external value changed */
        internal void ReverseSyncAllowSelfCollision()
        {
            SoftPhysicsSelfCollideJsb.valNoCallback = Person.BreastPhysicsMesh.allowSelfCollision;
        }

        internal void SyncSoftPhysicsOn()
        {
            SoftPhysicsOn = Script.Bootstrap.PersonMonitor.IsBreastSoftPhysicsOn();
            _ldpOn = SoftPhysicsOn && LdpEnabledJsb.val;
            if(!SoftPhysicsSynced)
            {
                if(SoftPhysicsOn)
                {
                    Person.LeftBreast.SetFunctionsWhenSoftPhysicsOn(
                        GetPositionalRigidbodies(LEFT, OUTER),
                        GetPositionalRigidbodies(LEFT, MAIN),
                        GetPositionalRigidbodies(LEFT, AREOLA)
                    );
                    Person.RightBreast.SetFunctionsWhenSoftPhysicsOn(
                        GetPositionalRigidbodies(RIGHT, OUTER),
                        GetPositionalRigidbodies(RIGHT, MAIN),
                        GetPositionalRigidbodies(RIGHT, AREOLA)
                    );
                }
                else
                {
                    Person.LeftBreast.SetFunctionsWhenSoftPhysicsOff(Person, VertexIndices.NippleSkinVerts(LEFT));
                    Person.RightBreast.SetFunctionsWhenSoftPhysicsOff(Person, VertexIndices.NippleSkinVerts(RIGHT));
                }
            }

            SoftPhysicsSynced = true;
        }

        internal void SyncAllowSelfCollision() => SyncAllowSelfCollision(SoftPhysicsSelfCollideJsb.val);

        void SyncAllowSelfCollision(bool value)
        {
            if(!Script.enabled)
            {
                Script.Loggr.LogMessage($"Enable the plugin to set {SoftPhysicsSelfCollideJsb.name}.");
                return;
            }

            Person.BreastPhysicsMesh.allowSelfCollision = value;
        }

        internal void SetShowColliderPreviews(bool hasSoftColliderVisualization) =>
            ColliderVisualizer.ShowPreviewsJSON.val = ShowColliderPreviewsJsb.val && hasSoftColliderVisualization;

        internal override void PreRegisterParams()
        {
            foreach(var paramGroup in SoftParamGroups)
            {
                var left = paramGroup.Left;
                if(paramGroup != FatColliderFriction)
                {
                    Script.PreRegisterSoftPhysicsParam(left.OffsetJsf);
                }

                Script.PreRegisterSoftPhysicsParam(left.GradientConfig.MiddleShiftJsf);
                Script.PreRegisterSoftPhysicsParam(left.GradientConfig.OuterMultiplierJsf);
                Script.PreRegisterSoftPhysicsParam(left.GradientConfig.CenterMultiplierJsf);
                Script.PreRegisterSoftPhysicsParam(left.GradientConfig.NippleMultiplierJsf);
                Script.PreRegisterSoftPhysicsParam(left.LdpConfig.CompressionEffectJsf);
                Script.PreRegisterSoftPhysicsParam(left.LdpConfig.CompressionSensitivityJsf);
                Script.PreRegisterSoftPhysicsParam(left.LdpConfig.StretchingEffectJsf);
                Script.PreRegisterSoftPhysicsParam(left.LdpConfig.StretchingSensitivityJsf);
            }

            foreach(var paramGroup in ParamGroups)
            {
                var left = paramGroup.Left;
                Script.PreRegisterSoftPhysicsParam(left.OffsetJsf);
            }

            Script.PreRegisterGeneralParam(SoftPhysicsOnJsb);
            Script.PreRegisterGeneralParam(LdpEnabledJsb);
            Script.PreRegisterGeneralParam(SoftPhysicsSelfCollideJsb);

        }

        internal void SaveCalibrated() => LdpEnabledJsb.SaveCalibrated();

        internal void UpdatePhysics(float realMass, float mass, bool updateOffsetMinMax = true)
        {
            if(!SoftPhysicsOn)
            {
                return;
            }

            foreach(var kvp in _gradientGroups)
            {
                foreach(var group in kvp.Value)
                {
                    group.CalculateDistances();
                }
            }

            foreach(var group in SoftParamGroups)
            {
                group.UpdateBaseValue(
                    group.Left.UsesRealMass ? realMass : mass,
                    Script.Softness,
                    Script.Quickness,
                    IsLdpUpdateOn()
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }

            foreach(var group in ParamGroups)
            {
                group.UpdateBaseValue(
                    mass,
                    Script.Softness,
                    Script.Quickness,
                    IsLdpUpdateOn()
                );
                if(updateOffsetMinMax)
                {
                    group.UpdateOffsetMinMax();
                }
            }
        }

        internal void UpdateSoftJointsDistortion(bool force = false)
        {
            UpdateSoftJointsDistortion(LEFT, force);
            UpdateSoftJointsDistortion(RIGHT, force);
        }

        internal void UpdateSoftJointsDistortion(string side, bool force = false)
        {
            if(!IsLdpUpdateOn())
            {
                return;
            }

            foreach(var softJoint in AllSoftJoints[side])
            {
                if(softJoint.UpdateDistortion(angleCalculator.CenterRoll, angleCalculator.CenterPitch2X, force))
                {
                    softJoint.SyncSpringAndDamperDistortion(side);
                    softJoint.SyncMassDistortion(side);
                    softJoint.SyncColliderRadiusAndDepthDistortion(side);
                    softJoint.SyncColliderFrictionDistortion(side);
                    softJoint.SyncDistanceLimitDistortion(side);
                }
            }
        }

        bool _originalAllowSelfCollision;

        internal void SaveOriginalBoolParamValues() => _originalAllowSelfCollision = Person.BreastPhysicsMesh.allowSelfCollision;

        internal void RestoreOriginalPhysics()
        {
            if(Script.Initialized != true)
            {
                return;
            }

            var breastPhysicsMesh = Person.BreastPhysicsMesh;
            breastPhysicsMesh.allowSelfCollision = _originalAllowSelfCollision;

            foreach(string name in breastPhysicsMesh.GetFloatParamNames())
            {
                /* Set a value that is different from the original, then restore the original
                 * in order to trigger VaM's internal sync
                 */
                var paramJsf = breastPhysicsMesh.GetFloatJSONParam(name);
                float original = paramJsf.val;
                paramJsf.valNoCallback = Mathf.Abs(paramJsf.val - paramJsf.min) > 0.01f
                    ? paramJsf.min
                    : paramJsf.max;
                paramJsf.val = original;
            }

            foreach(var softJoint in AllSoftJointsList)
            {
                softJoint.RestoreOriginalFriction();
            }
        }

        internal void Destroy()
        {
            Object.Destroy(ColliderVisualizer);
            ColliderVisualizer = null;
            SoftJointDiagnostics?.Destroy();
            SoftJointDiagnostics = null;
        }
    }
}
