using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;
using System.Linq;
using UnityEngine;

namespace NAT.TM.Handlers
{
    sealed class ChestClothingHandler : ClothingHandler
    {
        BreastJointPhysicsHandler _breastJointPhysicsHandler;
        BreastSoftPhysicsHandler _breastSoftPhysicsHandler;

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            leftRb = Person.LeftBreast.Rigidbody;
            rightRb = Person.RightBreast.Rigidbody;
            jointPhysicsHandler = provider.GetInstance<BreastJointPhysicsHandler>();
            softPhysicsHandler = provider.GetInstance<BreastSoftPhysicsHandler>();
            _breastJointPhysicsHandler = (BreastJointPhysicsHandler) jointPhysicsHandler;
            _breastSoftPhysicsHandler = (BreastSoftPhysicsHandler) softPhysicsHandler;
            gravityPhysicsHandler = provider.GetInstance<BreastGravityPhysicsHandler>();
            ClothingProfilesDir = "ChestClothingProfiles";
            validRegionTags = new[] { "full body", "torso" };
            invalidRegionTags = new[] { "hip", "arms", "feet", "hands", "head", "legs", "neck" };
            PresetHandler = new ClothingPresetHandler(Script, this, ClothingProfilesDir);
            CreateMorphConfigs();
            LoadSettings();
            initialized = true;
        }

        void CreateMorphConfigs()
        {
            const string id = "Reduce Size";
            var config = NewMorphConfig(id);
            config.OffsetsCompensation = new Vector3(0f, 0f, -0.0030f);
            MorphConfigs[id] = config;
            AddMorphConfigCallbacks();
        }

        ClothingMorphConfig NewMorphConfig(string id) => new ClothingMorphConfig(
            Person.GetBreastMorph($"Clothing/{id} L"),
            Person.GetBreastMorph($"Clothing/{id} R"),
            new StorableFloat(id, 0.000f, 0.000f, 2.000f)
        );

        internal void SetMultipliers(float mass)
        {
            float massCurve = Calc.InverseSmoothStep(1 - mass, 0.60f, 0.95f);
            upDownMassMultiplier = Mathf.Lerp(0.5f, 1.0f, massCurve);
            sidewaysMassMultiplier = Mathf.Lerp(0.17f, 1.0f, massCurve);
            inOutMassMultiplier = Mathf.Lerp(0.17f, 1.0f, massCurve);
        }

        protected override void SyncJointPhysicsAdjustments()
        {
            base.SyncJointPhysicsAdjustments();
            _breastJointPhysicsHandler.RightLeftSpring.Left.UpdateClothingAdjustmentAmount(CalculateHoldAmount(LeftYXHoldPct));
            _breastJointPhysicsHandler.RightLeftSpring.Right.UpdateClothingAdjustmentAmount(CalculateHoldAmount(RightYXHoldPct));
            _breastJointPhysicsHandler.RightLeftDamper.Left.UpdateClothingAdjustmentAmount(0.01f * LeftYXDampeningPct);
            _breastJointPhysicsHandler.RightLeftDamper.Right.UpdateClothingAdjustmentAmount(0.01f * RightYXDampeningPct);
        }

        protected override ClothingAdjustmentProfile GetProfile(ClothingItem item) => item.ChestProfile;
        protected override void SetProfile(ClothingItem item, ClothingAdjustmentProfile profile) => item.ChestProfile = profile;

        internal override void MirrorCoverageJoints(string fromSide, string toSide)
        {
            var joints = SelectedProfile.CoveredSoftJoints[fromSide];
            var mirrorJoints = _breastSoftPhysicsHandler.FindMirrorJoints(fromSide, toSide, joints).ToList();
            SelectedProfile.CoveredSoftJoints[toSide] = mirrorJoints;
            SelectedProfile.ExposedSoftJoints[toSide] = softPhysicsHandler.AllSoftJoints[toSide].Except(mirrorJoints).ToList();
        }
    }
}
