using NAT.Common;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using UnityEngine;

namespace NAT.TM.Models
{
    sealed class SoftJointGradientGroup
    {
        internal SoftJoint Areola { get; set; }
        internal SoftJoint Main1 { get; set; }
        internal SoftJoint Main2 { get; set; }
        internal SoftJoint Outer1 { get; set; }
        internal SoftJoint Outer2 { get; set; }
        internal SoftJoint Outer3 { get; set; }

        // ratios of distances to total distance
        float _main1Ratio;
        float _main2Ratio;
        float _outer1Ratio;
        float _outer2Ratio;

        internal void CalculateDistances()
        {
            float total;
            float areolaToMain1 = Vector3.Distance(Areola.JointRb.position, Main1.JointRb.position);
            float main1ToMain2 = Vector3.Distance(Main1.JointRb.position, Main2.JointRb.position);
            float main2ToOuter1 = Vector3.Distance(Main2.JointRb.position, Outer1.JointRb.position);
            if(Outer2 != null)
            {
                float outer1ToOuter2 = Vector3.Distance(Outer1.JointRb.position, Outer2.JointRb.position);
                if(Outer3 != null)
                {
                    float outer2ToOuter3 = Vector3.Distance(Outer2.JointRb.position, Outer3.JointRb.position);
                    total = areolaToMain1 + main1ToMain2 + main2ToOuter1 + outer1ToOuter2 + outer2ToOuter3;
                }
                else
                {
                    total = areolaToMain1 + main1ToMain2 + main2ToOuter1 + outer1ToOuter2;
                }

                _outer2Ratio = (areolaToMain1 + main1ToMain2 + main2ToOuter1 + outer1ToOuter2) / total;
            }
            else
            {
                total = areolaToMain1 + main1ToMain2 + main2ToOuter1;
            }

            _main1Ratio = areolaToMain1 / total;
            _main2Ratio = (areolaToMain1 + main1ToMain2) / total;
            _outer1Ratio = (areolaToMain1 + main1ToMain2 + main2ToOuter1) / total;
        }

        /* Value is parameter's overall base value based on softness, quickness and breast mass.
         * Individual joints get adjusted values based on the gradient.
         */
        internal void SetGradientBaseValues(string paramName, GradientConfig gradientConfig, float value)
        {
            Areola.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 0));
            Main1.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main1Ratio));
            Main2.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main2Ratio));

            if(Outer2 == null)
            {
                Outer1.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 1));
            }
            else
            {
                Outer1.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _outer1Ratio));
                if(Outer3 == null)
                {
                    Outer2.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 1));
                }
                else
                {
                    Outer2.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _outer2Ratio));
                    Outer3.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 1));
                }
            }
        }

        // Gradient starts from outer (ratio 1) and ends in areola (ratio 0)
        static float CalculateJointBaseValue(GradientConfig config, float value, float ratio)
        {
            float middle = 1 -
                Mathf.Lerp(
                    0.01f,
                    0.99f,
                    Mathf.InverseLerp(config.MiddleShiftJsf.min, config.MiddleShiftJsf.max, config.MiddleShiftJsf.val)
                );

            if(ratio > middle)
            {
                float smoothRatio = Calc.InverseSmoothStepWithMax(
                    ratio - middle,
                    1 - middle,
                    config.OuterToMiddleCurvature,
                    config.OuterToMiddleMidPoint
                );
                return Mathf.Lerp(value, value * config.OuterMultiplierJsf.val, smoothRatio);
            }
            else
            {
                float smoothRatio = 1 -
                    Calc.InverseSmoothStepWithMax(
                        ratio,
                        middle,
                        config.MiddleToCenterCurvature,
                        config.MiddleToCenterMidPoint
                    );
                return Mathf.Lerp(value, value * config.CenterMultiplierJsf.val, smoothRatio);
            }
        }

        internal void SetAdditionalOffsets(string paramName, float value)
        {
            Areola.SetAdditionalOffset(paramName, value);
            Main1.SetAdditionalOffset(paramName, value);
            Main2.SetAdditionalOffset(paramName, value);
            Outer1.SetAdditionalOffset(paramName, value);
            Outer2?.SetAdditionalOffset(paramName, value);
            Outer3?.SetAdditionalOffset(paramName, value);
        }

        internal bool HasJoint(SoftJoint joint) =>
            joint == Areola || joint == Main1 || joint == Main2 || joint == Outer1 || joint == Outer2 || joint == Outer3;
    }
}
