using System;
using System.Collections;
using System.Collections.Generic;
using NAT;
using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Storables;
using NAT.Common.UI;
using NAT.TM;
using NAT.TM.Handlers;
using NAT.TM.UI;
using SimpleJSON;
using UnityEngine;

namespace Naturalis
{
    sealed class TittyMagic : ScriptBase
    {
        public override bool ShouldIgnore() => false;

        const string CALIBRATION_DATA_VERSION = "1.1.0";

        protected override void OnUIEnabled()
        {
            _softPhysicsHandler.ReverseSyncAllowSelfCollision();
            AddGravityControls();

            if(Tabs.IsMainTabActive())
            {
                if(!Tabs.IsNestedWindowActive())
                {
                    Tabs.MainWindow.Refresh();
                }
                else if(Tabs.GetActiveNestedWindow() is DevWindow)
                {
                    _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(true);
                }
            }
            else if(Tabs.IsHardCollidersTabActive())
            {
                if(!enabled)
                {
                    Tabs.GoToMainWindow();
                }

                if(_hardColliderHandler.ColliderVisualizer)
                {
                    _hardColliderHandler.ColliderVisualizer.ShowPreviewsJSON.val = true;
                }
            }
            else if(Tabs.IsPhysicsTabActive())
            {
                if(!Tabs.IsNestedWindowActive())
                {
                    if(_softPhysicsHandler.SoftPhysicsSynced)
                    {
                        Tabs.PhysicsWindow.Refresh();
                    }
                }
                else if(Tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    _softPhysicsHandler.SoftJointDiagnostics?.DrawServiceSetEnabled(true);
                    var softParameterWindow = (SoftParameterWindow) Tabs.GetActiveNestedWindow();
                    softParameterWindow.Refresh();
                    softParameterWindow.SetShowColliderPreviews();
                }
            }
            else if(Tabs.IsClothingTabActive())
            {
                _softPhysicsHandler.SoftJointDiagnostics?.DrawServiceSetEnabled(true);
            }
        }

        protected override void OnUIDisabled()
        {
            _gravityPhysicsHandler.UnsetGravityDirection();

            if(Bootstrap.CalibrateOnUIClosedJsb.val && CalibrationStateHandler.Pending)
            {
                StartCalibration();
            }

            if(_hardColliderHandler.ColliderVisualizer)
            {
                _hardColliderHandler.ColliderVisualizer.ShowPreviewsJSON.val = false;
            }

            if(_softPhysicsHandler.ColliderVisualizer)
            {
                _softPhysicsHandler.ColliderVisualizer.ShowPreviewsJSON.val = false;
            }

            Tabs.ClosePopups();
            _softPhysicsHandler.SoftJointDiagnostics?.DrawServiceSetEnabled(false);

            if(Tabs.IsMainTabActive())
            {
                if(Tabs.GetActiveNestedWindow() is SettingsWindow)
                {
                    Tabs.MainWindow.OnReturn();
                }
            }
        }

        AngleCalculator _angleCalculator;
        BreastJointPhysicsHandler _jointPhysicsHandler;
        BreastSoftPhysicsHandler _softPhysicsHandler;
        BreastGravityPhysicsHandler _gravityPhysicsHandler;
        BreastForcePhysicsHandler _forcePhysicsHandler;
        FrictionHandler _frictionHandler;
        HardColliderHandler _hardColliderHandler;
        BreastGravityOffsetMorphHandler _gravityOffsetMorphHandler;
        BreastForceMorphHandler _forceMorphHandler;
        NippleErectionHandler _nippleErectionHandler;
        ChestClothingHandler _clothingHandler;
        MorphListener _breastMorphListener;

        public override void Init()
        {
            base.Init(Constant.TITTY_MAGIC, Colors.funkyCyan);
            StartCoroutine(SetupRoutine());
        }

        IEnumerator SetupRoutine()
        {
            yield return InitBootstrapCo();
            if(Bootstrap.Done == false)
            {
                Initialized = false;
                yield break;
            }

            Person = Bootstrap.Person;
            if(enabled)
            {
                if(!Person.IsFemale)
                {
                    Person.SetupFuta();
                }

                Person.SetPectoralRbCollisions(false);
            }

            Provider = new Provider(this);

            /* Used to store version in save JSON and communicate version to other plugin instances */
            var versionJss = new StorableString(Constant.NATURALIS_VERSION, "")
            {
                val = Bootstrap.NATURALIS_VERSION,
            };
            PreRegisterInternalParam(versionJss, false);

            _angleCalculator = Provider.GetInstance<AngleCalculator>();
            _angleCalculator.Init(
                Person.LeftBreast.Rigidbody,
                Person.RightBreast.Rigidbody,
                Person.ChestRb
            );
            SetupHandlers();
            SetupBreastMorphListener();
            SetupStorables();
            SetupCallbacks();
            SetupCalibrationHelper();

            Bindings = Bindings ?? new Bindings(this); // Might already be setup in OnBindingsListRequested.
            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);

            SetupUI();
            RegisterParams();

            if(!RestoringFromJSON)
            {
                PresetHandler.LoadDefaultPreset(Person.IsFemale, true);
                if(!InitializedFromDefaultPreset)
                {
                    _clothingHandler.AutoLoadDefaultProfiles();
                }

                _softPhysicsHandler.SaveOriginalBoolParamValues();
                StartCalibration(true);
            }

            Initialized = true;
        }

        IEnumerator InitBootstrapCo()
        {
            var bootyMagic = (BootyMagic) this.GetOwnScript("BootyMagic");
            Bootstrap = bootyMagic.Bootstrap;
            while(!Bootstrap || Bootstrap.Done == null)
            {
                yield return null;
            }
        }

        void SetupHandlers()
        {
            PresetHandler = new PluginPresetHandler(this, $"{DomainName}Presets");
            _jointPhysicsHandler = Provider.GetInstance<BreastJointPhysicsHandler>();
            _softPhysicsHandler = Provider.GetInstance<BreastSoftPhysicsHandler>();
            _gravityPhysicsHandler = Provider.GetInstance<BreastGravityPhysicsHandler>();
            gravityPhysicsHandler = _gravityPhysicsHandler;
            _forcePhysicsHandler = Provider.GetInstance<BreastForcePhysicsHandler>();
            _frictionHandler = Provider.GetInstance<FrictionHandler>();
            _hardColliderHandler = Provider.GetInstance<HardColliderHandler>();
            _forceMorphHandler = Provider.GetInstance<BreastForceMorphHandler>();
            _gravityOffsetMorphHandler = Provider.GetInstance<BreastGravityOffsetMorphHandler>();
            _nippleErectionHandler = Provider.GetInstance<NippleErectionHandler>();
            _clothingHandler = Provider.GetInstance<ChestClothingHandler>();

            foreach(Handler handler in Provider.GetHandlers())
            {
                handler.Init<TittyMagic>(Provider);
            }

            _nippleErectionHandler.NippleErectionJsf.setCallbackFunction = _ =>
                _nippleErectionHandler.UpdateNipple(_jointPhysicsHandler.NormalizedMass, Softness);

            Bootstrap.ClothingStore.OnItemsReloadedHandlers += _clothingHandler.OnItemsReloaded;
            Bootstrap.ClothingStore.OnItemToggledHandlers += _clothingHandler.OnItemToggled;
            Bootstrap.ClothingStore.OnItemAllowDetachToggledHandlers += _clothingHandler.OnItemAllowDetachToggled;
        }

        void SetupBreastMorphListener()
        {
            _breastMorphListener = new MorphListener(
                new HashSet<int>
                {
                    8, 16, 17, 93, 135, 136, 137, 138, 2401, 2402, 2403, 2404, 2405, 2406, 2407, 2408, 2409, 2410, 2411, 2412,
                    2413, 2414, 2415, 2416, 2417, 2418, 2419, 2590, 2591, 2592, 2594, 2596, 2597, 2598, 2871, 2872, 2873, 7216,
                    7218, 7219, 7220, 7230, 7231, 7232, 7233, 7234, 7235, 7236, 7237, 7238, 7239, 7240, 7241, 7242, 7243, 7244,
                    7245, 7246, 7247, 7248, 8833, 8834, 8836, 8837, 8838, 8839, 8840, 8848, 8849, 8850, 8851, 8852, 8853, 8854,
                    10936, 10944, 10945, 11021, 11063, 11064, 11065, 11066, 13231, 13232, 13233, 13234, 13235, 13236, 13237,
                    13238, 13239, 13240, 13241, 13242, 13243, 13244, 13245, 13246, 13247, 13248, 13249, 13408, 13409, 13410,
                    13411, 13412, 13413, 13414, 13672, 13673, 13674, 17923, 17924, 17925, 17926, 17936, 17937, 17938, 17939,
                    17940, 17941, 17942, 17943, 17944, 17945, 17946, 17947, 17948, 17949, 17950, 17951, 17952, 17953, 17954,
                    19504, 19505, 19507, 19508, 19509, 19510, 19511, 19519, 19520, 19521, 19522, 19523, 19524, 19525, 19639,
                },
                new HashSet<string>
                {
                    "FBMFitnessDetails",
                    "FBMHeight",
                    "PBMShouldersSize",
                    "Alessa",
                    "Eisa",
                    "Izarra",
                    "PBMLegsLength",
                    "Shin Length",
                    "Thigh Length",
                    "Upper Body Length",
                    "G3F_Hip Abdomen Shape",
                    "bellybulge",
                    "deepthroat",
                    "Athletic",
                    "BodyShort",
                    "Hollow back",
                    "FBMBodyTone",
                    "FBMBodyBuilderDetails",
                    "Julian Body",
                    "Shoulders Width",
                    "Nipples Large",
                    "WeebU.My_morphs.3.SideBulge",
                    "Ribs Width",
                    "Rectus Width",
                    "Rectus Outer Detail",
                    "Distort 1",
                    "Distort 2",
                    "Amy AiMei",
                    "succubus body",
                    "Impact - Breast",
                    "Breath1",
                    "Ribcage undefined",
                    "Sternocleidomastoid",
                }
            );

            if(Bootstrap.CheckForSizeChangesJsb.val)
            {
                ProcessMorphsForListener();
            }
        }

        void SetupStorables()
        {
            SoftnessJsf = new StorableFloat("Breast Softness", 70f, 0f, 100f)
            {
                setCallbackFunction = _ => SetCalibrationPending(SoftnessJsf),
            };
            QuicknessJsf = new StorableFloat("Breast Quickness Offset", 0.00f, -1.00f, 1.00f)
            {
                setCallbackFunction = _ => SetCalibrationPending(QuicknessJsf),
            };

            CalibrationAction = new StorableAction(Constant.CALIBRATE, () => StartCalibration(true));
            var calibrateOnNextPendingAction = new StorableAction(
                Constant.CALIBRATE_ON_NEXT_PENDING,
                () => CalibrationStateHandler.CalibrateOnNextPending = true
            );

            bool softPhysicsOn =
                UserPreferences.singleton.softPhysics &&
                containingAtom.GetStorableByID("SoftBodyPhysicsEnabler").GetBoolParamValue("enabled") &&
                _softPhysicsHandler.SoftPhysicsOnJsb.val;
            var syncSoftPhysicsJsb = new StorableBool(Constant.INTERNAL_SYNC_SOFT_PHYSICS, softPhysicsOn)
            {
                setCallbackFunction = OnSoftPhysicsToggled,
            };

            var onPhysicsRateChangedAction = new StorableAction(Constant.INTERNAL_ON_PHYSICS_RATE_CHANGED, () =>
            {
                if(!CalibrationStateHandler.Pending && !CalibrationStateHandler.CalibrateOnNextPending)
                {
                    Loggr.LogMessage("Physics rate changed - calibration set to pending.");
                }

                CalibrationStateHandler.SetCalibrationPending(Constant.PHYSICS_RATE_CHANGED);
            });

            var onCharacterChangedAction = new StorableAction(Constant.INTERNAL_ON_CHARACTER_CHANGED, () =>
            {
                _frictionHandler.Refresh();
                UIModManager.SetFrictionUIVisibility();
            });

            var unsetPendingJss = new StorableString(Constant.INTERNAL_UNSET_CALIBRATION_PENDING, "")
            {
                setJSONCallbackFunction = jss =>
                {
                    CalibrationStateHandler.SetCalibrationPending(jss.val, false);
                    jss.valNoCallback = "";
                },
            };

            var processMorphsAction = new StorableAction(Constant.INTERNAL_PROCESS_MORPHS, () =>
            {
                if(!_breastMorphListener.ProcessingDone)
                {
                    StartCoroutine(ProcessMorphsForListenerCo());
                }
            });

            PreRegisterGeneralParam(SoftnessJsf);
            PreRegisterGeneralParam(QuicknessJsf);
            PreRegisterGeneralParam(CalibrationAction);
            PreRegisterGeneralParam(calibrateOnNextPendingAction);
            PreRegisterInternalParam(syncSoftPhysicsJsb);
            PreRegisterInternalParam(onPhysicsRateChangedAction);
            PreRegisterInternalParam(onCharacterChangedAction);
            PreRegisterInternalParam(unsetPendingJss);
            PreRegisterInternalParam(processMorphsAction);
        }

        IEnumerator ProcessMorphsForListenerCo()
        {
            Loggr.LogMessage("Processing all morphs to be able to detect changes to breast morphs. This could take a while if you have lots of morphs.");
            yield return null;
            ProcessMorphsForListener();
        }

        void ProcessMorphsForListener()
        {
            _breastMorphListener.ProcessMorphs(Person.Geometry.morphBank1);
            if(!Person.IsFemale)
            {
                _breastMorphListener.ProcessMorphs(Person.Geometry.morphBank1OtherGender);
            }
        }

        void SetupCallbacks()
        {
            Person.ScaleJsf.setCallbackFunction += OnAtomScaleChanged;
            SuperController.singleton.onBeforeSceneSaveHandlers += OnBeforeSceneSave;
            SuperController.singleton.onSceneSavedHandlers += OnSceneSaved;
        }

        internal override void SyncMorphsAngleOffsets(float? timer = null)
        {
            var leftMorphsOffsets = _clothingHandler.LeftMorphsOffsets + _clothingHandler.LeftMorphsOffsetsCompensation;
            var rightMorphsOffsets = _clothingHandler.RightMorphsOffsets + _clothingHandler.RightMorphsOffsetsCompensation;

            if(timer != null && timer.Value >= Time.deltaTime)
            {
                var transition = CoroutineUtils.ExponentialVector3ArrayTransition(
                    timer.Value,
                    new Vector3TransitionParams(
                        leftMeasurementOffset,
                        leftMorphsOffsets,
                        value => leftMeasurementOffset = value
                    ),
                    new Vector3TransitionParams(
                        rightMeasurementOffset,
                        rightMorphsOffsets,
                        value => rightMeasurementOffset = value
                    )
                );
                StartCoroutine(transition);
            }
            else
            {
                leftMeasurementOffset = leftMorphsOffsets;
                rightMeasurementOffset = rightMorphsOffsets;
            }
        }

        void SetupCalibrationHelper()
        {
            CalibrationStateHandler = gameObject.AddComponent<CalibrationStateHandler>();
            CalibrationStateHandler.Init(this, () => StartCalibration());
            PreRegisterInternalParam(CalibrationStateHandler.CalibratingJsb);
        }

        // https://github.com/vam-community/vam-plugins-interop-specs/blob/main/keybindings.md
        public void OnBindingsListRequested(List<object> bindingsList)
        {
            Bindings = Bindings ?? new Bindings(this); // Might already be setup in Init.
            bindingsList.Add(Bindings.NamespaceDict);
            bindingsList.AddRange(Bindings.GetActionsList());
        }

        void SetupUI()
        {
            var uiModManager = gameObject.AddComponent<UIModManager>();
            uiModManager.Init(Provider);
            UIModManager = uiModManager;

            Tabs = new Tabs(this)
            {
                MainWindow = new MainWindow(this),
                PhysicsWindow = new PhysicsWindow(this),
                MorphsWindow = new MorphsWindow(this),
                HardCollidersWindow = new HardCollidersWindow(this),
                ClothingWindow = new ClothingWindow(this),
            };
        }

        void Update()
        {
            if(Initialized != true)
            {
                return;
            }

            try
            {
                if(!isUpdatePaused)
                {
                    if(Bootstrap.CheckForSizeChangesJsb.val)
                    {
                        if(_breastMorphListener.ChangeDetected())
                        {
                            CalibrationStateHandler.SetCalibrationPending(Constant.SIZE_CHANGED, _breastMorphListener.IsChanged);
                        }
                    }

                    if(_softPhysicsHandler.SoftJointDiagnostics?.IsDrawingClothingCoverage ?? false)
                    {
                        _clothingHandler.HandleToggleCoverageJoint();
                    }

                    if(pluginUIEventsListener.IsEnabled)
                    {
                        Tabs.GetActiveWindow()?.UpdateDiagnostics();
                    }
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"Update error: {e}");
            }
        }

        void FixedUpdate()
        {
            if(Initialized != true ||
                RestoringFromJSON ||
                !Person.Skin ||
                Person.IsFreezeGrabbing ||
                Bootstrap.IsGrabbingSubscene ||
                !SuperController.singleton.autoSimulation
            )
            {
                return;
            }

            if(isUpdatePaused)
            {
                if(inCalibrationCo && _gravityPhysicsHandler.IsSimulatingGravity)
                {
                    SimulateGravity();
                }

                return;
            }

            try
            {
                Person.LeftBreast.Measure(leftMeasurementOffset);
                Person.RightBreast.Measure(rightMeasurementOffset);

                if(_gravityPhysicsHandler.IsSimulatingGravity)
                {
                    SimulateGravity();
                }
                else
                {
                    _angleCalculator.CalculateAngles();
                    _gravityPhysicsHandler.Execute();

                    if(!Bootstrap.ForceUpdatesPaused())
                    {
                        _forcePhysicsHandler.Execute();
                        _forceMorphHandler.Execute();
                        _gravityOffsetMorphHandler.Execute(_jointPhysicsHandler.NormalizedRealMass, Softness);
                    }
                }

                _hardColliderHandler.UpdateDistanceDiffs();
                _hardColliderHandler.UpdateFriction(_frictionHandler.MaxHardColliderFriction);
                _clothingHandler.Execute();
                _jointPhysicsHandler.Execute();
                _softPhysicsHandler.Execute();
            }
            catch(Exception e)
            {
                Loggr.LogError($"FixedUpdate error: {e}");
                isUpdatePaused = true;
            }
        }

        bool _updatingMass;

        void SimulateGravity()
        {
            string direction = _gravityPhysicsHandler.GravitySimulationDirection;
            _angleCalculator.OverrideAngles(direction);
            _gravityPhysicsHandler.AddGravitySimulationForce(!inCalibrationCo);

            if(!_updatingMass)
            {
                if(inCalibrationCo && direction != Direction.DOWN)
                {
                    Person.LeftBreast.Measure(Vector3.zero);
                    Person.RightBreast.Measure(Vector3.zero);
                }

                _gravityPhysicsHandler.Execute(inCalibrationCo);
                _forcePhysicsHandler.Execute();
                _forceMorphHandler.Execute();
                _gravityOffsetMorphHandler.Execute(_jointPhysicsHandler.NormalizedRealMass, Softness);
            }
        }

        protected override IEnumerator CalibrationCo(JSONClass calibrationData = null)
        {
            inCalibrationCo = true;
            isUpdatePaused = true;
            _gravityPhysicsHandler.UnsetGravityDirection();
            /* Ensure everything is ready for calibration to proceed in case of e.g. appearance preset load */
            {
                /* Ensures breast soft physics enabled value is correct on first calibration when soft physics enabled is synced to SoftPhysicsHandler */
                while(!Bootstrap.PersonMonitor.Initialized)
                {
                    yield return null;
                }

                /* Can occur when loading a look and VaM pauses to load assets */
                while(!Person.IsSelectedCharacterReady())
                {
                    yield return null;
                }

                /* Must be unfrozen before checking for hard collider rigidbodies */
                yield return CalibrationStateHandler.WaitForAtomPhysicsUnfrozen();

                Bootstrap.PersonMonitor.Check(true);
                Bootstrap.PersonMonitor.enabled = false;

                /* Sanity check. Might occur (?) even after hard colliders are enabled by settings monitor */
                float timeout = Time.unscaledTime + 5;
                while(!_hardColliderHandler.RigidbodiesFound() && Time.unscaledTime < timeout)
                {
                    yield return null;
                }

                if(!_hardColliderHandler.RigidbodiesFound())
                {
                    Loggr.LogError("Calibration error: collider rigidbodies not found after 5 seconds of waiting");
                    yield break;
                }
            }

            bool doFullCalibration =
                calibrationData == null ||
                !calibrationData.HasKey("mass") ||
                !calibrationData.HasKey("left_breast") ||
                !calibrationData.HasKey("right_breast") ||
                !calibrationData.HasKey("soft_joints") ||
                !calibrationData.HasKey("hard_colliders");

            if(!doFullCalibration)
            {
                try
                {
                    CalibrateFromJSON(calibrationData);
                }
                catch(Exception e)
                {
                    doFullCalibration = true;
                    Loggr.LogError($"Error loading calibration data from JSON, continuing with full calibration. Error: {e}");
                }
            }

            if(doFullCalibration)
            {
                yield return FullCalibrationCo();
            }

            SaveCalibratedValues();
            CalibrationStateHandler.UpdateMassWhenCalibrating = false;
            isUpdatePaused = false;
            inCalibrationCo = false;
            FirstCalibrationDone = true;

            /* Update UI */
            if(Tabs.IsMainTabActive())
            {
                if(!Tabs.IsNestedWindowActive())
                {
                    Tabs.MainWindow.Refresh();
                }
            }
            else if(Tabs.IsPhysicsTabActive())
            {
                if(!Tabs.IsNestedWindowActive())
                {
                    Tabs.PhysicsWindow.Refresh();
                }
                else if(Tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    ((SoftParameterWindow) Tabs.GetActiveNestedWindow()).Refresh();
                }
            }
            else if(Tabs.IsClothingTabActive())
            {
                Tabs.ClothingWindow.Refresh();
            }
        }

        void CalibrateFromJSON(JSONClass jc)
        {
            _softPhysicsHandler.SyncSoftPhysicsOn();
            _softPhysicsHandler.SyncAllowSelfCollision();
            _jointPhysicsHandler.UpdateMassValueAndAmounts(jc["mass"].AsFloat);

            Softness = SoftnessJsf.val / 100f;
            Quickness = QuicknessJsf.val;
            SetMultipliersAndUpdatePhysics();

            Person.LeftBreast.CalibrateFromJSON(jc["left_breast"].AsObject);
            Person.RightBreast.CalibrateFromJSON(jc["right_breast"].AsObject);
            _softPhysicsHandler.CalibrateFromJSON(jc["soft_joints"].AsObject);

            _hardColliderHandler.SyncCollidersMass();
            _hardColliderHandler.CalibrateFromJSON(jc["hard_colliders"].AsObject);
            _hardColliderHandler.SyncAllOffsets();

            _clothingHandler.SetAllAdjustmentValues();
            _clothingHandler.SyncAdjustments();
            SyncMorphsAngleOffsets();

            CalibrationStateHandler.EnableSettingsMonitor();
        }

        IEnumerator FullCalibrationCo()
        {
            _softPhysicsHandler.SoftJointDiagnostics?.ClearLines();
            /* Must be called before calibration pose is applied because the current pose must be stored without duplicate morphs having >0 values. */
            ZeroDynamicAdjustments();
            Person.ZeroDuplicateTittyMagicMorphs();

            yield return CalibrationStateHandler.Begin();
            if(CalibrationStateHandler.Cancelling)
            {
                CalibrationStateHandler.Cancelling = false;
                yield break;
            }

            _softPhysicsHandler.SyncSoftPhysicsOn();
            _softPhysicsHandler.SyncAllowSelfCollision();

            {
                _gravityPhysicsHandler.EnableGravitySimulation(Direction.DOWN);
                yield return UpdateMassAndPhysicsCo();

                _jointPhysicsHandler.SetGravitySimulationDamper();
                yield return new WaitForSeconds(0.25f);

                float minTime = Time.time + 0.40f;
                float maxTime = Time.time + 1.00f;
                const float tolerance = 0.003f;
                while(
                    Time.time < minTime ||
                    (
                        (Person.LeftBreast.Moved(tolerance) || Person.RightBreast.Moved(tolerance)) &&
                        Time.time < maxTime
                    )
                )
                {
                    yield return null;
                }

                Person.LeftBreast.Calibrate();
                Person.RightBreast.Calibrate();
                if(_softPhysicsHandler.IsLdpUpdateOn())
                {
                    _softPhysicsHandler.CalibrateSoftJointsDistortion(Direction.DOWN);
                }

                _hardColliderHandler.SyncCollidersMass();
                _hardColliderHandler.CalibrateColliders();
                _hardColliderHandler.SyncAllOffsets();
            }

            if(_softPhysicsHandler.IsLdpUpdateOn())
            {
                foreach(string direction in new[] { Direction.FORWARD, Direction.UP, Direction.LEFT, Direction.RIGHT, Direction.BACK })
                {
                    _gravityPhysicsHandler.EnableGravitySimulation(direction);
                    yield return new WaitForSeconds(0.25f);

                    float timeout = Time.time + Mathf.Lerp(1.00f, 2.00f, _jointPhysicsHandler.NormalizedRealMass);
                    while(_softPhysicsHandler.AnySoftJointMoved() && Time.time < timeout)
                    {
                        yield return null;
                    }

                    _softPhysicsHandler.CalibrateSoftJointsDistortion(direction);
                }
            }

            _clothingHandler.SetAllAdjustmentValues();
            _clothingHandler.SyncAdjustments();
            SyncMorphsAngleOffsets();

            _gravityPhysicsHandler.StopGravitySimulation();
            _jointPhysicsHandler.UnsetGravitySimulationDamper();
            yield return CalibrationStateHandler.Finish();
        }

        void ZeroDynamicAdjustments()
        {
            Person.LeftBreast.ResetMeasurements();
            Person.RightBreast.ResetMeasurements();
            _softPhysicsHandler.ResetSoftJointsDistortion();
            _softPhysicsHandler.ResetSoftJointsClothingAdjustments();
            _hardColliderHandler.ResetDistanceDiffs();
            _gravityPhysicsHandler.ResetAll();
            _forcePhysicsHandler.ResetAll();
            _forceMorphHandler.ResetAll();
            _gravityOffsetMorphHandler.ResetAll();
            _clothingHandler.ResetAll();
        }

        IEnumerator UpdateMassAndPhysicsCo()
        {
            _updatingMass = true;
            if(CalibrationStateHandler.UpdateMassWhenCalibrating)
            {
                Softness = 0;
                Quickness = 0;
                _jointPhysicsHandler.UpdateBaseValues(false);
                _softPhysicsHandler.UpdatePhysics(_jointPhysicsHandler.NormalizedRealMass, _jointPhysicsHandler.NormalizedMass, false);

                yield return new WaitForSeconds(0.25f);
                /* Mass is calculated multiple times because each new mass value changes the exact breast
                 * shape and therefore the estimated volume.
                 */
                _jointPhysicsHandler.CalculateMass();
                yield return Utils.WaitAndRepeat(
                    () => _jointPhysicsHandler.CalculateMass(),
                    (int) Mathf.Lerp(4, 12, _jointPhysicsHandler.NormalizedRealMass),
                    0.13f
                );
            }

            Softness = SoftnessJsf.val / 100f;
            Quickness = QuicknessJsf.val;
            SetMultipliersAndUpdatePhysics();

            _updatingMass = false;
        }

        void SetMultipliersAndUpdatePhysics()
        {
            float normalizedRealMass = _jointPhysicsHandler.NormalizedRealMass;
            _gravityPhysicsHandler.SetMultipliers(normalizedRealMass, Softness);
            _forceMorphHandler.SetMultipliers(normalizedRealMass, Softness);
            _forcePhysicsHandler.SetMultipliers(normalizedRealMass, Softness);
            _gravityOffsetMorphHandler.SetMultipliers();
            _hardColliderHandler.UpdateFrictionSizeMultipliers();
            _clothingHandler.SetMultipliers(normalizedRealMass);

            _jointPhysicsHandler.UpdateBaseValues();
            _softPhysicsHandler.UpdatePhysics(normalizedRealMass, _jointPhysicsHandler.NormalizedMass);
            _nippleErectionHandler.UpdateNipple(_jointPhysicsHandler.NormalizedMass, Softness);
            _frictionHandler.CalculateFriction();
        }

        void SaveCalibratedValues()
        {
            SoftnessJsf.SaveCalibrated();
            QuicknessJsf.SaveCalibrated();
            _breastMorphListener.SaveCalibrated();
            _forceMorphHandler.SaveCalibrated();
            _gravityOffsetMorphHandler.SaveCalibrated();
            _jointPhysicsHandler.SaveCalibrated();
            _gravityPhysicsHandler.SaveCalibrated();
            _softPhysicsHandler.SaveCalibrated();
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            // Is set to "" for some reason, even though initialization sets it to the correct value
            SetStringParamValue(Constant.NATURALIS_VERSION, Bootstrap.NATURALIS_VERSION);
            var jc = base.GetJSON(includePhysical, includeAppearance, forceStore);

            if(_isSavingScene && FirstCalibrationDone && !CalibrationStateHandler.Pending)
            {
                jc["stored_in_scene"].AsBool = true; // Unused, might be useful later
                var calibrationData = new JSONClass
                {
                    ["version"] = CALIBRATION_DATA_VERSION,
                    ["mass"] = { AsFloat = _jointPhysicsHandler.CalibratedMass },
                    ["left_breast"] = Person.LeftBreast.SerializeCalibrationData(),
                    ["right_breast"] = Person.RightBreast.SerializeCalibrationData(),
                    ["soft_joints"] = _softPhysicsHandler.SerializeCalibrationData(),
                    ["hard_colliders"] = _hardColliderHandler.SerializeCalibrationData(),
                };
                jc["calibration_data"] = calibrationData;
            }

            var activeClothingProfiles = new JSONArray();
            foreach(var profile in _clothingHandler.ActiveProfiles)
            {
                if(!string.IsNullOrEmpty(profile.LoadedFullPath))
                {
                    activeClothingProfiles.Add(new JSONClass
                    {
                        // Hacky way to ensure that path is updated to SELF:/ by package manager when packaging the scene with the active preset files in the same var
                        ["presetName"] = profile.LoadedFullPath,
                    });
                }
            }

            jc["Active Clothing Profiles"] = activeClothingProfiles;

            needsStore = true;
            return jc;
        }

        public override void RestoreFromJSON(
            JSONClass jc,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            JSONArray presetAtoms = null,
            bool setMissingToDefault = true
        )
        {
            RestoringFromJSON = true;

            /* Disable early to allow correct enabled value to be used during Init */
            if(jc.HasKey("enabled") && !jc["enabled"].AsBool)
            {
                enabled = false;
            }

            /* Prevent overriding versionJss.val from JSON. Version stored in JSON just for information,
             * but could be intercepted here and used to save a "loadedFromVersion" value.
             */
            if(jc.HasKey(Constant.NATURALIS_VERSION))
            {
                jc[Constant.NATURALIS_VERSION] = Bootstrap.NATURALIS_VERSION;
            }

            /* Naturalis v1.0.0 -> v1.0.1+ fixes */
            JSONUtils.RenameKey(jc, "In/Out Damper Back Gravity", "In/Out Spring Back Gravity");
            JSONUtils.RenameKey(jc, "In/Out Damper Forward Gravity", "In/Out Spring Forward Gravity");

            StartCoroutine(
                RestoreFromJSONCo(
                    jc,
                    restorePhysical,
                    restoreAppearance,
                    presetAtoms,
                    setMissingToDefault
                )
            );
        }

        IEnumerator RestoreFromJSONCo(
            JSONClass jc,
            bool restorePhysical,
            bool restoreAppearance,
            JSONArray presetAtoms,
            bool setMissingToDefault
        )
        {
            while(Initialized == null)
            {
                yield return null;
            }

            if(Initialized == false)
            {
                yield break;
            }

            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            _softPhysicsHandler.SaveOriginalBoolParamValues();
            RestoringFromJSON = false;

            if(enabled)
            {
                JSONClass calibrationData = null;
                string scenePackageId = null;
                if(jc.HasKey("calibration_data"))
                {
                    #if VAM_GT_1_21
                    {
                        scenePackageId = FileUtils.ParsePackageIdFromPath(SuperController.singleton.LoadedSceneName);
                    }
                    #endif
                    var tmpData = jc["calibration_data"].AsObject;
                    if(
                        tmpData.HasKey("version") &&
                        new SemVer(tmpData["version"].Value).IsMajorMinorEqual(CALIBRATION_DATA_VERSION)
                    )
                    {
                        calibrationData = tmpData;
                    }
                }

                // ReSharper disable once ExpressionIsAlwaysNull
                _clothingHandler.AutoLoadProfiles(JSONUtils.GetJSONArrayEnumerable(jc, "Active Clothing Profiles"), scenePackageId);
                StartCalibration(true, calibrationData);
            }
        }

        internal override void RestoreFromPresetJSON(JSONClass jc, string fullPath)
        {
            if(Initialized != true || RestoringFromJSON)
            {
                Loggr.LogMessage("Preset file must be loaded after initialization.");
                return;
            }

            jc["enabled"].AsBool = enabled;
            base.RestoreFromJSON(jc);
            string packageId = FileUtils.ParsePackageIdFromPath(fullPath);
            _clothingHandler.AutoLoadProfiles(JSONUtils.GetJSONArrayEnumerable(jc, "Active Clothing Profiles"), packageId);
            _softPhysicsHandler.SaveOriginalBoolParamValues();
            SetLoadedPresetName(fullPath);

            if(Bootstrap.CalibrateOnPresetLoadedJsb.val)
            {
                StartCalibration(true);
            }
        }

        internal override void InitializeFromPresetJSON(JSONClass jc, string fullPath)
        {
            jc["enabled"].AsBool = enabled;
            base.RestoreFromJSON(jc);
            string packageId = FileUtils.ParsePackageIdFromPath(fullPath);
            _clothingHandler.AutoLoadProfiles(JSONUtils.GetJSONArrayEnumerable(jc, "Active Clothing Profiles"), packageId);
            SetLoadedPresetName(fullPath);
            InitializedFromDefaultPreset = true;
        }

        void OnAtomScaleChanged(float _)
        {
            if(Initialized != true || !enabled)
            {
                return;
            }

            if(!CalibrationStateHandler.Pending && !CalibrationStateHandler.CalibrateOnNextPending)
            {
                Loggr.LogMessage("Person atom scale changed - calibration set to pending.");
            }

            CalibrationStateHandler.SetCalibrationPending(Constant.SIZE_CHANGED);
        }

        bool _isSavingScene;

        void OnBeforeSceneSave()
        {
            if(Initialized != true || !enabled)
            {
                return;
            }

            try
            {
                _isSavingScene = true;
                isUpdatePaused = true;
                _gravityOffsetMorphHandler.ResetAll();
                _forceMorphHandler.ResetAll();
                _nippleErectionHandler.Reset();
                _clothingHandler.ResetMorphs();
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnBeforeSceneSave error: {e}");
            }
        }

        void OnSceneSaved()
        {
            if(Initialized != true || !enabled)
            {
                return;
            }

            try
            {
                _isSavingScene = false;
                if(!CalibrationStateHandler.CalibratingJsb.val)
                {
                    isUpdatePaused = false;
                    _nippleErectionHandler.UpdateNipple(_jointPhysicsHandler.NormalizedMass, Softness);
                    _clothingHandler.SyncMorphsOffsetsCompensation();
                    SyncMorphsAngleOffsets();
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnSceneSaved error: {e}");
            }
        }

        void OnSoftPhysicsToggled(bool value)
        {
            _softPhysicsHandler.ReverseSyncSoftPhysicsOn();
            if(!CalibrationStateHandler.CalibrateOnNextPending)
            {
                Loggr.LogMessage($"Soft physics {(value ? "enabled" : "disabled")} - updates paused, calibration set to pending.");
            }

            isUpdatePaused = true;
            CalibrationStateHandler.SetCalibrationPending(Constant.SOFT_PHYSICS_CHANGED);
            _softPhysicsHandler.SoftPhysicsSynced = false;

            if(Tabs.IsPhysicsTabActive())
            {
                if(Tabs.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    if(!value)
                    {
                        Tabs.PhysicsWindow.OnReturn();
                    }
                }
                else if(!value)
                {
                    Tabs.PhysicsWindow.Refresh();
                }
            }

            if(!value)
            {
                _softPhysicsHandler.SoftJointDiagnostics.VisualizeLdpJsb.val = false;
            }
        }

        void OnEnable()
        {
            if(Initialized != true)
            {
                return;
            }

            try
            {
                Bootstrap.PersonMonitor.Check(true);
                Person.EnableUseColliders();
                if(!Person.IsFemale)
                {
                    if(!Person.FutaSetupDone)
                    {
                        Person.SetupFuta();
                    }
                    else
                    {
                        Person.ParentBreastPhysicsMesh("MaleMorphers");
                    }
                }

                _hardColliderHandler.EnableMultiplyFriction();
                if(_hardColliderHandler.ColliderVisualizer)
                {
                    _hardColliderHandler.ColliderVisualizer.enabled = true;
                }

                _softPhysicsHandler.SaveOriginalBoolParamValues();
                _softPhysicsHandler.EnableMultiplyFriction();
                if(_softPhysicsHandler.ColliderVisualizer)
                {
                    _softPhysicsHandler.ColliderVisualizer.enabled = true;
                }

                _softPhysicsHandler.SoftJointDiagnostics?.DrawServiceSetEnabled(true);
                StartCalibration(true);
                UIModManager.enabled = true;
                Person.SetBreastsUseGravity(false);
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnEnable error: {e}");
            }
        }

        void OnDisable()
        {
            if(Initialized != true)
            {
                return;
            }

            try
            {
                if(calibrationCo != null)
                {
                    StopCoroutine(calibrationCo);
                }

                _hardColliderHandler.RestoreOriginalPhysics();
                if(_hardColliderHandler.ColliderVisualizer)
                {
                    _hardColliderHandler.ColliderVisualizer.enabled = false;
                }

                _jointPhysicsHandler.RestoreOriginalPhysics();
                _softPhysicsHandler.RestoreOriginalPhysics();
                if(_softPhysicsHandler.ColliderVisualizer)
                {
                    _softPhysicsHandler.ColliderVisualizer.enabled = false;
                }

                _softPhysicsHandler.SoftJointDiagnostics?.DrawServiceSetEnabled(false);
                UIModManager.enabled = false;

                CalibrationStateHandler.Finish(Bootstrap.BootyMagicStorable.IsEnabledNullSafe());
                if(!Person.IsFemale)
                {
                    Person.ParentBreastPhysicsMesh("FemaleMorphers");
                }

                Person.SetPectoralRbCollisions(true);
                _gravityOffsetMorphHandler.ResetAll();
                _forceMorphHandler.ResetAll();
                _nippleErectionHandler.Reset();
                _clothingHandler.ResetMorphs();
                Person.SetBreastsUseGravity(true);
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnDisable error: {e}");
            }
        }

        new void OnDestroy()
        {
            try
            {
                base.OnDestroy();
                if(Person != null)
                {
                    Person.ScaleJsf.setCallbackFunction -= OnAtomScaleChanged;
                }

                _frictionHandler?.Destroy();
                _hardColliderHandler?.Destroy();
                _softPhysicsHandler?.Destroy();

                Destroy(Bootstrap);
                Bootstrap = null;

                SuperController.singleton.onSceneSavedHandlers -= OnSceneSaved;
                SuperController.singleton.onBeforeSceneSaveHandlers -= OnBeforeSceneSave;
                SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
            }
            catch(Exception e)
            {
                if(Initialized == true)
                {
                    SuperController.LogError($"OnDestroy error: {e}");
                }
                else
                {
                    Debug.LogError($"OnDestroy error: {e}");
                }
            }
        }
    }
}
