﻿using NAT.Common;
using NAT.Common.Diagnostics;
using NAT.Common.Handlers;
using NAT.Common.UI;
using NAT.TM.Handlers;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.TM.UI
{
    sealed class ClothingWindow : Window
    {
        readonly ClothingHandler _clothingHandler;
        readonly SoftPhysicsHandler _softPhysicsHandler;
        readonly SoftJointDiagnostics _softJointDiagnostics;
        readonly Dictionary<string, Section> _sections;
        UIDynamicTextField _currentPresetTextField;
        UIDynamicTextField _transitionInfoTextField;
        UIDynamicTextField _adjustmentsInfoTextField;

        internal ClothingWindow(ScriptBase script) : base(script)
        {
            _clothingHandler = script.Provider.GetInstance<ChestClothingHandler>();
            _softPhysicsHandler = script.Provider.GetInstance<BreastSoftPhysicsHandler>();
            _softJointDiagnostics = _softPhysicsHandler.SoftJointDiagnostics;
            AddNestedWindow(new AdditionalClothingSettingsWindow(script));
            _sections = new Dictionary<string, Section>
            {
                [FORCES_HOLD_AND_DAMPENING] = new Section(BuildForcesHoldAndDampeningTab),
                [SOFT_PHYSICS] = new Section(BuildSoftPhysicsTab),
                [MORPH_ADJUSTMENTS] = new Section(BuildMorphAdjustmentsTab),
            };
        }

        const string FORCES_HOLD_AND_DAMPENING = "Forces, Hold & Dampening";
        const string SOFT_PHYSICS = "Soft Physics";
        const string MORPH_ADJUSTMENTS = "Morph Adjustments";
        const string LEFT_PROFILE = "Left Profile";
        const string RIGHT_PROFILE = "Right Profile";
        const string TOGGLE_ADJUSTMENTS_INFO = "Toggle Adjustments Info";
        const string ADJUSTMENTS_INFO_HEADER = "Adjustments Info Header";
        const string ADJUSTMENTS_INFO = "Adjustments Info";

        protected override void OnBuild()
        {
            if(isBuilt)
            {
                return;
            }

            BuildClothingPopupRow();
            BuildPresetArea();
            BuildSettingsArea();
            BuildAdjustmentsInfoPanel();
            NavigateToSection(FORCES_HOLD_AND_DAMPENING);
        }

        void BuildClothingPopupRow()
        {
            AddPopup(_clothingHandler.ClothingItemJssc, jssc =>
            {
                var uiDynamic = script.CreateCustomPopup(jssc.name, jssc.label, new Vector2(0, -190), new Vector2(-310, 110));
                script.ConfigureCustomPopup(uiDynamic, 510, false);
                jssc.popup = uiDynamic.popup;
                return uiDynamic;
            });
            AddToggle(_clothingHandler.ItemEnabledJsb, jsb =>
            {
                var uiDynamic = script.CreateCustomToggle(jsb.name, new Vector2(763, -136), new Vector2(-890, 50));
                jsb.toggle = uiDynamic.toggle;
                return uiDynamic;
            });
            AddButton("Customize", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(763, -185), new Vector2(-890, 50));
                uiDynamic.AddListener(() =>
                {
                    var selectedItem = _clothingHandler.SelectedItem;
                    if(selectedItem == null || !selectedItem.EnabledJsb.val)
                    {
                        return;
                    }

                    selectedItem.DazItem.OpenUI();
                });
                return uiDynamic;
            });
            AddButton("Reload Items", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, "Reload\nItems", new Vector2(960, -185), new Vector2(-970, 100));
                uiDynamic.buttonText.fontSize = 26;
                uiDynamic.AddListener(() => script.Bootstrap.ClothingStore.ReloadItems());
                return uiDynamic;
            });
        }

        void BuildPresetArea()
        {
            AddTextField("Profile Preset Header", name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Profile\nPreset", new Vector2(0, -275), new Vector2(-930, 70));
                uiDynamic.UItext.fontSize = 28;
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            });
            AddButton("Save", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(160, -253), new Vector2(-950, 50));
                uiDynamic.AddListener(() => _clothingHandler.SaveProfile());
                return uiDynamic;
            });
            AddButton("Load", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(288, -253), new Vector2(-950, 50));
                uiDynamic.AddListener(() => _clothingHandler.LoadProfile());
                return uiDynamic;
            });
            AddButton("Reset", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(415, -253), new Vector2(-950, 50));
                uiDynamic.AddListener(() => _clothingHandler.ResetProfile());
                return uiDynamic;
            });
            AddButton("Set As Item Default", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(551, -253), new Vector2(-820, 50));
                uiDynamic.AddListener(() => _clothingHandler.SaveDefault());
                uiDynamic.buttonText.fontSize = 26;
                return uiDynamic;
            });
            AddButton("Load Item Default", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(810, -253), new Vector2(-820, 50));
                uiDynamic.AddListener(() => _clothingHandler.LoadDefault());
                uiDynamic.buttonText.fontSize = 26;
                return uiDynamic;
            });
            AddTextField("Current Preset Info", name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Current: None", new Vector2(160, -289), new Vector2(-170, 32));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.MiddleLeft;
                uiDynamic.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
                var scrollView = uiDynamic.transform.Find("Scroll View");
                var scrollRect = scrollView.GetComponent<ScrollRect>();
                scrollRect.horizontal = true;
                scrollRect.horizontalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
                _currentPresetTextField = uiDynamic;
                return uiDynamic;
            });
        }

        void BuildSettingsArea()
        {
            AddTextField("Profile Settings Header", name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Profile\nSettings", new Vector2(0, -386), new Vector2(-930, 70));
                uiDynamic.UItext.fontSize = 28;
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            });

            AddToggle(_clothingHandler.SymmetricJsb, jsb =>
            {
                var uiDynamic = script.CreateCustomToggle(jsb.name, new Vector2(160, -373), new Vector2(-856, 50));
                jsb.toggle = uiDynamic.toggle;
                uiDynamic.SetActiveStyle(_clothingHandler.SelectedProfile != null, true);
                return uiDynamic;
            });
            AddButton(LEFT_PROFILE, name =>
            {
                var uiDynamic = script.CreateCustomButton(name, "Left", new Vector2(391, -347), new Vector2(-931, 50));
                uiDynamic.AddListener(() => _clothingHandler.SelectLeft());
                uiDynamic.SetActiveStyle(!(_clothingHandler.SelectedProfile?.Symmetric.Value ?? true), true);
                return uiDynamic;
            });
            AddButton(RIGHT_PROFILE, name =>
            {
                var uiDynamic = script.CreateCustomButton(name, "Right", new Vector2(539, -347), new Vector2(-931, 50));
                uiDynamic.AddListener(() => _clothingHandler.SelectRight());
                uiDynamic.SetActiveStyle(!(_clothingHandler.SelectedProfile?.Symmetric.Value ?? true), true);
                return uiDynamic;
            });
            AddToggle(_clothingHandler.LeftActiveJsb, jsb =>
            {
                var uiDynamic = script.CreateCustomToggle(jsb.name, "Active", new Vector2(391, -395), new Vector2(-931, 50));
                jsb.toggle = uiDynamic.toggle;
                uiDynamic.SetActiveStyle(_clothingHandler.SelectedProfile != null, true);
                return uiDynamic;
            });
            AddToggle(_clothingHandler.RightActiveJsb, jsb =>
            {
                var uiDynamic = script.CreateCustomToggle(jsb.name, "Active", new Vector2(539, -395), new Vector2(-931, 50));
                jsb.toggle = uiDynamic.toggle;
                uiDynamic.SetActiveStyle(_clothingHandler.SelectedProfile != null, true);
                return uiDynamic;
            });

            AddTextField("Additional Settings Header", name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Additional\nSettings", new Vector2(750, -356), new Vector2(-906, 70));
                uiDynamic.UItext.fontSize = 28;
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            });
            AddButton("Show Additional Settings", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, "Show...", new Vector2(920, -343), new Vector2(-931, 50));
                uiDynamic.AddListener(OpenNestedWindow<AdditionalClothingSettingsWindow>);
                return uiDynamic;
            });
            AddTextField("Transition Info", name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "", new Vector2(750, -395), new Vector2(-760, 32));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                _transitionInfoTextField = uiDynamic;
                return uiDynamic;
            });

            AddSectionNavigationButton(FORCES_HOLD_AND_DAMPENING, 60, -710);
            AddSectionNavigationButton(SOFT_PHYSICS, 430, -870);
            AddSectionNavigationButton(MORPH_ADJUSTMENTS, 640, -710);
        }

        void AddSectionNavigationButton(string id, int posX, int sizeX) => AddButton($"{id} Tab", name =>
        {
            var uiDynamic = script.CreateCustomButton(name, id, new Vector2(posX, -468), new Vector2(sizeX, 55));
            uiDynamic.buttonText.fontSize = 26;
            var navigationButton = new NavigationButton(uiDynamic, script.DomainColor, Color.gray);
            navigationButton.AddListener(() => NavigateToSection(id));
            if(id == SOFT_PHYSICS && !_softPhysicsHandler.SoftPhysicsOn)
            {
                navigationButton.SetEnabled(false);
            }

            _sections[id].NavigationButton = navigationButton;
            return uiDynamic;
        });

        void BuildAdjustmentsInfoPanel()
        {
            AddButton(TOGGLE_ADJUSTMENTS_INFO, name =>
            {
                int posX = script.IsGlobalUISideLeft ? 1088 : -63;
                string label = script.IsGlobalUISideLeft ? "<".Bold() : ">".Bold();
                var uiDynamic = script.CreateCustomButton(name, label, new Vector2(posX, -214), new Vector2(-1025, 50));
                uiDynamic.textColor = Color.white;
                uiDynamic.buttonColor = Colors.transparentGray;
                uiDynamic.AddListener(ToggleAdjustmentsInfoPanel);
                return uiDynamic;
            }, true);
            AddButton(ADJUSTMENTS_INFO_HEADER, name =>
            {
                int posX = script.IsGlobalUISideLeft ? 1141 : -382;
                var uiDynamic = script.CreateCustomButton(name, "Adjustments Info", new Vector2(posX, -214), new Vector2(-759, 50));
                uiDynamic.textColor = Color.white;
                uiDynamic.buttonColor = Colors.darkTransparentGray;
                var colors = uiDynamic.button.colors;
                colors.highlightedColor = Color.white;
                colors.pressedColor = Color.white;
                uiDynamic.button.colors = colors;
                return uiDynamic;
            }, true);
            AddTextField(ADJUSTMENTS_INFO, name =>
            {
                int posX = script.IsGlobalUISideLeft ? 1113 : -382;
                var uiDynamic = script.CreateCustomTextField(name, "", new Vector2(posX, -1058), new Vector2(-731, 845));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.textColor = Color.white;
                uiDynamic.backgroundColor = Colors.transparentGray;
                uiDynamic.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
                var scrollView = uiDynamic.transform.Find("Scroll View");
                var scrollRect = scrollView.GetComponent<ScrollRect>();
                scrollRect.horizontal = true;
                scrollRect.horizontalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
                _adjustmentsInfoTextField = uiDynamic;
                return uiDynamic;
            }, true);
        }

        void ToggleAdjustmentsInfoPanel()
        {
            var toggleStatusButton = GetElement<UIDynamicButton>(TOGGLE_ADJUSTMENTS_INFO);
            var statusHeader = GetElement<UIDynamicButton>(ADJUSTMENTS_INFO_HEADER);
            var statusTextField = GetElement<UIDynamicTextField>(ADJUSTMENTS_INFO);

            if(toggleStatusButton == null || statusHeader == null || statusTextField == null)
            {
                return;
            }

            if(statusHeader.IsVisible())
            {
                toggleStatusButton.SetLabel(script.IsGlobalUISideLeft ? ">".Bold() : "<".Bold());
                statusHeader.Hide();
                statusTextField.Hide();
            }
            else
            {
                toggleStatusButton.SetLabel(script.IsGlobalUISideLeft ? "<".Bold() : ">".Bold());
                statusHeader.Show();
                statusTextField.Show();
            }
        }

        void BuildForcesHoldAndDampeningTab(List<Element> elements)
        {
            AddTextField($"{FORCES_HOLD_AND_DAMPENING} Header", elements, name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Forces", new Vector2(12, -530), new Vector2(-558, 40));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            });

            AddTextField($"{FORCES_HOLD_AND_DAMPENING} Info", elements, name =>
            {
                string text = new StringBuilder()
                    .Append("Forces simulate the way clothing pushes breasts.")
                    .Append(" Forces automatically change the physical behavior and apply directional morphing.")
                    .AppendLine(" Gravity in the same direction can cancel out the force.")
                    .AppendLine()
                    .AppendLine("Hold limits the amount of movement by increasing joint springs. It also reduces hard collider mass.")
                    .AppendLine()
                    .Append("Dampening prevents jiggling by increasing joint damper parameters.")
                    .ToString();
                var uiDynamic = script.CreateCustomTextField(name, text, new Vector2(547, -883), new Vector2(-558, 371));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            });

            const float startYPos = -650;
            const float spacing = 127;

            AddSlider(_clothingHandler.YForceJsf, elements, jsf =>
            {
                var uiDynamic = script.CreateCustomSlider(jsf.name, "Y Force (Up/Down)", new Vector2(10, startYPos));
                uiDynamic.valueFormat = "F2";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });
            AddSlider(_clothingHandler.XForceJsf, elements, jsf =>
            {
                var uiDynamic = script.CreateCustomSlider(jsf.name, "X Force (Sideways)", new Vector2(10, startYPos - spacing));
                uiDynamic.valueFormat = "F2";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });
            AddSlider(_clothingHandler.ZForceJsf, elements, jsf =>
            {
                var uiDynamic = script.CreateCustomSlider(jsf.name, "Z Force (In/Out)", new Vector2(10, startYPos - 2 * spacing));
                uiDynamic.valueFormat = "F2";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });

            AddTextField("Hold Header", elements, name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Hold", new Vector2(12, -955), new Vector2(-558, 40));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.UpperCenter;
                return uiDynamic;
            });

            AddSlider(_clothingHandler.YXHoldPctJsf, elements, jsf =>
            {
                var uiDynamic = script.CreateCustomSlider(jsf.name, new Vector2(10, startYPos - 45 - 3 * spacing));
                uiDynamic.valueFormat = "F1";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });
            AddSlider(_clothingHandler.ZHoldPctJsf, elements, jsf =>
            {
                var uiDynamic = script.CreateCustomSlider(jsf.name, new Vector2(10, startYPos - 45 - 4 * spacing));
                uiDynamic.valueFormat = "F1";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });

            AddTextField("Dampening Header", elements, name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Dampening", new Vector2(547, -955), new Vector2(-558, 40));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.UpperCenter;
                return uiDynamic;
            });

            AddSlider(_clothingHandler.YXDampeningPctJsf, elements, jsf =>
            {
                var uiDynamic = script.CreateCustomSlider(jsf.name, new Vector2(545, startYPos - 45 - 3 * spacing));
                uiDynamic.valueFormat = "F1";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });
            AddSlider(_clothingHandler.ZDampeningPctJsf, elements, jsf =>
            {
                var uiDynamic = script.CreateCustomSlider(jsf.name, new Vector2(545, startYPos - 45 - 4 * spacing));
                uiDynamic.valueFormat = "F1";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });
        }

        void BuildSoftPhysicsTab(List<Element> elements)
        {
            AddTextField("Soft Physics Info", elements, name =>
            {
                var sb = new StringBuilder();
                if(script.Person.IsFemale)
                {
                    sb.Append("Find the area covered by the clothing item. To ensure full coverage, adjust the radius or");
                    sb.Append(" hover over a joint in desktop mode and press CTRL to toggle it manually.");
                }
                else
                {
                    sb.Append("Select covered joints with the quick buttons, and by hovering over a joint in desktop mode");
                    sb.Append("and pressing CTRL to toggle. (Auto-detection is only supported on a female character.");
                }

                var uiDynamic = script.CreateCustomTextField(name, sb.ToString(), new Vector2(10, -855), new Vector2(-10, 371));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            });

            if(script.Person.IsFemale)
            {
                AddSlider(_clothingHandler.SoftJointSearchRadiusJsf, elements, jsf =>
                {
                    var uiDynamic = script.CreateCustomSlider(jsf.name, new Vector2(10, -670));
                    uiDynamic.valueFormat = "F4";
                    ScriptBase.ConfigureCustomSlider(uiDynamic);
                    jsf.slider = uiDynamic.slider;
                    return uiDynamic;
                });
                AddButton("Find Covered Soft Joints", elements, name =>
                {
                    var uiDynamic = script.CreateCustomButton(name, new Vector2(12, -730), new Vector2(-557, 50));
                    uiDynamic.AddListener(_clothingHandler.FindCoveredSoftJoints);
                    return uiDynamic;
                });
            }
            else
            {
                const float sizeY = 55;
                var size = new Vector2(-817, sizeY);
                const float startYPos = -620;
                AddButton("Select All (Left)", elements, name =>
                {
                    var uiDynamic = script.CreateCustomButton(name, new Vector2(12, startYPos), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.SetAllJointsCovered(Side.LEFT));
                    return uiDynamic;
                });
                AddButton("Select All (Right)", elements, name =>
                {
                    var uiDynamic = script.CreateCustomButton(name, new Vector2(274, startYPos), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.SetAllJointsCovered(Side.RIGHT));
                    return uiDynamic;
                });
                AddButton("Select None (Left)", elements, name =>
                {
                    var uiDynamic = script.CreateCustomButton(name, new Vector2(12, startYPos - sizeY), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.SetAllJointsExposed(Side.LEFT));
                    return uiDynamic;
                });
                AddButton("Select None (Right)", elements, name =>
                {
                    var uiDynamic = script.CreateCustomButton(name, new Vector2(274, startYPos - sizeY), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.SetAllJointsExposed(Side.RIGHT));
                    return uiDynamic;
                });
                AddButton("Mirror Right to Left", elements, name =>
                {
                    var uiDynamic = script.CreateCustomButton(name, new Vector2(12, startYPos - 2 * sizeY), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.MirrorCoverageJoints(Side.RIGHT, Side.LEFT));
                    return uiDynamic;
                });
                AddButton("Mirror Left to Right", elements, name =>
                {
                    var uiDynamic = script.CreateCustomButton(name, new Vector2(274, startYPos - 2 * sizeY), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.MirrorCoverageJoints(Side.LEFT, Side.RIGHT));
                    return uiDynamic;
                });
            }

            AddPopup(_clothingHandler.VisualizeClothingCoverageJssc, elements, jssc =>
            {
                var uiDynamic = script.CreateCustomScrollablePopup(jssc.name, jssc.name, new Vector2(545, -670), new Vector2(-552, 118));
                script.ConfigureCustomPopup(uiDynamic, 400, false);
                script.AddPrevNextButtons(uiDynamic);
                uiDynamic.SetActiveStyle(_clothingHandler.SelectedProfile?.IsCoverageMapped ?? false, true);
                jssc.popup = uiDynamic.popup;
                return uiDynamic;
            });
            AddToggle(_clothingHandler.AutoSelectVisualizedParamJsb, elements, jsb =>
            {
                var uiDynamic = script.CreateCustomToggle(jsb.name, new Vector2(545, -730), new Vector2(-557, 50));
                jsb.toggle = uiDynamic.toggle;
                return uiDynamic;
            });

            AddTextField("Soft Physics Parameters Info", elements, name =>
            {
                string text = new StringBuilder()
                    .Append("Shift a parameter from its current value towards its maximum value (if positive %) or minimum value (if negative %).")
                    .Append(" Increase Fat Spring and Fat Damper to make the clothed area more rigid. The other parameters are provided for experimentation.")
                    .ToString();
                var uiDynamic = script.CreateCustomTextField(name, text, new Vector2(10, -1120), new Vector2(-10, 371));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            });

            const float sliderSpacing = 127;
            var list = _clothingHandler.SoftPhysicsJsfs.Values.ToList();
            for(int i = 0; i < list.Count; i++)
            {
                var parameterJsf = list[i];
                float posX = i < 3 ? 10 : 545;
                float currentSpacing = i < 3 ? i * sliderSpacing : (i - 3) * sliderSpacing;
                AddSlider(parameterJsf, elements, jsf =>
                {
                    var uiDynamic = script.CreateCustomSlider(jsf.name, new Vector2(posX, -970 - currentSpacing));
                    uiDynamic.valueFormat = "F1";
                    ScriptBase.ConfigureCustomSlider(uiDynamic);
                    jsf.slider = uiDynamic.slider;
                    return uiDynamic;
                });
            }
        }

        void BuildMorphAdjustmentsTab(List<Element> elements)
        {
            const float startYPos = -730;
            const float spacing = 127;

            AddTextField("Morphing Offsets Header", elements, name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Morphing Offsets", new Vector2(272, -530), new Vector2(-558, 40));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.UpperCenter;
                return uiDynamic;
            });

            AddTextField("Morphing Offsets Info", elements, name =>
            {
                const string text = "Apply directional morphs independently of forces.";
                var uiDynamic = script.CreateCustomTextField(name, text, new Vector2(272, -610), new Vector2(-558, 75));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            });

            AddSlider(_clothingHandler.YMorphsOffsetJsf, elements, jsf =>
            {
                string label = new StringBuilder()
                    .AppendFormat("{0} (°)", _clothingHandler.YMorphsOffsetJsf.name)
                    .ToString();
                var uiDynamic = script.CreateCustomSlider(jsf.name, label, new Vector2(270, startYPos));
                uiDynamic.valueFormat = "F2";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });

            AddSlider(_clothingHandler.XMorphsOffsetJsf, elements, jsf =>
            {
                string label = new StringBuilder()
                    .AppendFormat("{0} (°)", _clothingHandler.XMorphsOffsetJsf.name)
                    .ToString();
                var uiDynamic = script.CreateCustomSlider(jsf.name, label, new Vector2(270, startYPos - spacing));
                uiDynamic.valueFormat = "F2";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });
            AddSlider(_clothingHandler.ZMorphsOffsetJsf, elements, jsf =>
            {
                string label = new StringBuilder()
                    .AppendFormat("{0} (mm)", _clothingHandler.ZMorphsOffsetJsf.name)
                    .ToString();
                var uiDynamic = script.CreateCustomSlider(jsf.name, label, new Vector2(270, startYPos - 2 * spacing));
                uiDynamic.valueFormat = "F1";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });

            const float startYPos2 = -1050;
            AddTextField("Extra Morphs Header", elements, name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Extra Morphs", new Vector2(272, startYPos2), new Vector2(-558, 40));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.UpperCenter;
                return uiDynamic;
            });

            var configs = _clothingHandler.MorphConfigs.Values.ToArray();
            for(int i = 0; i < configs.Length; i++)
            {
                var config = configs[i];
                float currentSpacing = (i + 1) * spacing;
                AddSlider(config.StorableFloat, elements, jsf =>
                {
                    var uiDynamic = script.CreateCustomSlider(jsf.name, new Vector2(270, startYPos2 - currentSpacing));
                    uiDynamic.valueFormat = "F2";
                    ScriptBase.ConfigureCustomSlider(uiDynamic);
                    jsf.slider = uiDynamic.slider;
                    return uiDynamic;
                });
            }
        }

        protected override void PostBuild()
        {
            _clothingHandler.AutoSelectItem();
            _softJointDiagnostics?.DrawServiceSetEnabled(true);
            Refresh();
        }

        protected override void OnClose() =>
            _softJointDiagnostics?.DrawServiceSetEnabled(false);

        void NavigateToSection(string id)
        {
            foreach(var kvp in _sections)
            {
                if(kvp.Key == id)
                {
                    kvp.Value.Show();
                }
                else
                {
                    kvp.Value.Hide();
                }
            }
        }

        internal override void Refresh()
        {
            if(!isBuilt)
            {
                return;
            }

            var selectedProfile = _clothingHandler.SelectedProfile;
            bool profileExists = selectedProfile != null;
            bool isSymmetric = !profileExists || selectedProfile.Symmetric.Value;
            bool isCoverageMapped = profileExists && selectedProfile.IsCoverageMapped;

            var leftButton = GetElement<UIDynamicButton>(LEFT_PROFILE);
            var rightButton = GetElement<UIDynamicButton>(RIGHT_PROFILE);
            var symmetricToggle = GetElement<UIDynamicToggle>(_clothingHandler.SymmetricJsb.name);
            var leftActiveToggle = GetElement<UIDynamicToggle>(_clothingHandler.LeftActiveJsb.name);
            var rightActiveToggle = GetElement<UIDynamicToggle>(_clothingHandler.RightActiveJsb.name);

            leftButton.SetActiveStyle(!isSymmetric, true);
            rightButton.SetActiveStyle(!isSymmetric, true);
            symmetricToggle.SetActiveStyle(profileExists, true);
            leftActiveToggle.SetActiveStyle(profileExists, true);
            rightActiveToggle.SetActiveStyle(profileExists, true);

            if(isSymmetric)
            {
                leftButton.SetLabel("Left");
                rightButton.SetLabel("Right");
            }
            else
            {
                if(selectedProfile.IsRightSelected)
                {
                    leftButton.SetLabel("Left");
                    rightButton.SetLabel("Right".Bold());
                }
                else
                {
                    leftButton.SetLabel("Left".Bold());
                    rightButton.SetLabel("Right");
                }
            }

            var visualizeCoveragePopup = GetElement<UIDynamicPopup>(_clothingHandler.VisualizeClothingCoverageJssc.name);
            visualizeCoveragePopup?.SetActiveStyle(isCoverageMapped, true);
            _sections[SOFT_PHYSICS].NavigationButton.SetEnabled(_softPhysicsHandler.SoftPhysicsOn);
            if(!_softPhysicsHandler.SoftPhysicsOn && _sections[SOFT_PHYSICS].IsShown)
            {
                NavigateToSection(FORCES_HOLD_AND_DAMPENING);
            }
        }

        readonly StringBuilder _currentPresetSb = new StringBuilder();
        readonly StringBuilder _transitionInfoSb = new StringBuilder();

        // TODO base class jss?
        internal override void UpdateDiagnostics()
        {
            if(!isBuilt)
            {
                return;
            }

            if(ActiveNestedWindow != null)
            {
                ActiveNestedWindow.UpdateDiagnostics();
            }
            else
            {
                if(Time.frameCount % 6 == 0)
                {
                    _currentPresetTextField.text = ClothingUIHelpers.CurrentPreset(_currentPresetSb, _clothingHandler.SelectedProfile);
                    _transitionInfoTextField.text = ClothingUIHelpers.TransitionInfo(_transitionInfoSb, _clothingHandler.SelectedProfile);
                }
            }

            _softJointDiagnostics.ClearLines();
            UpdateSoftPhysicsVisualization();
            UpdateAdjustmentsInfo();
        }

        void UpdateSoftPhysicsVisualization()
        {
            if(!_sections[SOFT_PHYSICS].IsShown || _softJointDiagnostics == null)
            {
                return;
            }

            if(_clothingHandler.VisualizeClothingCoverageJssc.val == Constant.COVERAGE)
            {
                var profile = _clothingHandler.SelectedProfile;
                if(_clothingHandler.SelectedItem.EnabledJsb.val &&
                    profile != null &&
                    profile.IsActive &&
                    profile.IsCoverageMapped
                )
                {
                    _softJointDiagnostics.DrawClothingCoverage(profile);
                }
                else
                {
                    _softJointDiagnostics.IsDrawingClothingCoverage = false;
                }
            }
            else
            {
                _softJointDiagnostics.DrawClothingAdjustments(_clothingHandler.VisualizedParamGroup);
                _softJointDiagnostics.IsDrawingClothingCoverage = false;
            }
        }

        internal void UpdateAdjustmentsInfo()
        {
            if(Time.frameCount % 6 != 0)
            {
                return;
            }

            _adjustmentsInfoTextField.text = ClothingUIHelpers.BuildAdjustmentsInfo(diagnosticsSb, _clothingHandler, _softPhysicsHandler, false);
        }
    }
}
