// ReSharper disable UnusedMember.Local
using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Models;
using NAT.Common.UI;
using NAT.TM.Handlers;
using System;
using System.Linq;
using System.Text;
using UnityEngine;

namespace NAT.TM.UI
{
    sealed class DevWindow : Window
    {
        readonly AngleCalculator _angleCalculator;
        readonly BreastJointPhysicsHandler _jointPhysicsHandler;
        readonly BreastSoftPhysicsHandler _softPhysicsHandler;
        readonly GravityPhysicsHandler _gravityPhysicsHandler;
        readonly BreastForceMorphHandler _forceMorphHandler;
        readonly ChestClothingHandler _clothingHandler;

        readonly JSONStorableString _leftDebugArea;
        readonly JSONStorableString _rightDebugArea;
        readonly JSONStorableString _jointDiagnosticsJss;
        readonly JSONStorableStringChooser _softParamJssc;
        SoftPhysicsParamGroup _selectedParamGroup;

        readonly CalibratedJoint _leftJoint, _rightJoint;

        internal DevWindow(ScriptBase script) : base(script)
        {
            _angleCalculator = script.Provider.GetInstance<AngleCalculator>();
            _jointPhysicsHandler = script.Provider.GetInstance<BreastJointPhysicsHandler>();
            _softPhysicsHandler = script.Provider.GetInstance<BreastSoftPhysicsHandler>();
            _gravityPhysicsHandler = script.Provider.GetInstance<BreastGravityPhysicsHandler>();
            _forceMorphHandler = script.Provider.GetInstance<BreastForceMorphHandler>();
            _clothingHandler = script.Provider.GetInstance<ChestClothingHandler>();
            _leftDebugArea = new JSONStorableString("leftDebugArea", "");
            _rightDebugArea = new JSONStorableString("rightDebugArea", "");
            _jointDiagnosticsJss = new JSONStorableString("jointDiagnostics", "");
            var choices = _softPhysicsHandler.SoftParamGroups.Select(group => group.Name).ToList();
            _softParamJssc = new JSONStorableStringChooser("softPhysicsParams", choices, choices.First(), "Params")
            {
                setCallbackFunction = OnParamOptionSelected,
            };
            OnParamOptionSelected(_softParamJssc.val);

            _leftJoint = script.Person.LeftBreast;
            _rightJoint = script.Person.RightBreast;
        }

        void OnParamOptionSelected(string value)
        {
            _selectedParamGroup = _softPhysicsHandler.SoftParamGroups.First(group => group.Name == value);
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddElement(() =>
            {
                var textField = script.CreateTextField(_leftDebugArea, rightSide);
                textField.UItext.fontSize = 28;
                textField.height = Constant.UI_MAX_HEIGHT - 130;
                return textField;
            });
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(50, rightSide);
            // AddElement(() =>
            // {
            //     var textField = script.CreateTextField(_rightDebugArea, rightSide);
            //     textField.UItext.fontSize = 28;
            //     textField.height = Constant.UI_MAX_HEIGHT - 130;
            //     return textField;
            // });
        }

        void BuildRightSide2(bool rightSide = true)
        {
            AddElement(() =>
            {
                var popup = script.CreateFilterablePopup(_softPhysicsHandler.SoftJointDiagnostics.SoftJointRbDiagnosticJssc, rightSide);
                script.ConfigureCustomPopup(popup, 480f, false);
                script.AddPrevNextButtons(popup);
                return popup;
            });

            AddElement(() =>
            {
                var popup = script.CreateFilterablePopup(_softParamJssc, rightSide);
                script.ConfigureCustomPopup(popup, 0f, false);
                script.AddPrevNextButtons(popup);
                return popup;
            });

            AddElement(() =>
            {
                var textField = script.CreateTextField(_jointDiagnosticsJss, rightSide);
                textField.UItext.fontSize = 28;
                textField.height = 250;
                return textField;
            });
        }

        protected override void PostBuild() =>
            _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(true);

        protected override void OnClose() =>
            _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(false);

        internal override void UpdateDiagnostics()
        {
            if(Time.frameCount % 6 != 0)
            {
                return;
            }

            // DebugMorphAngleLimits();
            // DebugGravityAndAngles();
            // DebugClothingMorphs();
        }

        void DebugMorphAngleLimits()
        {
            diagnosticsSb
                .Clear()
                .AppendFormat("NRealM: {0:0.000}", _jointPhysicsHandler.NormalizedRealMass)
                .AppendLine()
                .AppendLine()
                .AppendFormat("UpLimit: {0:0.000}", _forceMorphHandler.UpLimit)
                .AppendLine()
                .AppendFormat("DepthOutLimit: {0:0.000}", _forceMorphHandler.DepthOutLimit)
                .AppendLine()
                .AppendFormat("SidewaysOutLimit: {0:0.000}", _forceMorphHandler.SidewaysOutLimit)
                .AppendLine()
                .AppendFormat("SidewaysInLimit: {0:0.000}", _forceMorphHandler.SidewaysInLimit)
                .AppendLine()
                .AppendLine()
                .AppendFormat("UpNormalizer: {0:0.000}", _forceMorphHandler.UpNormalizer)
                .AppendLine()
                .AppendFormat("DepthInNormalizer: {0:0.000}", _forceMorphHandler.DepthInNormalizer)
                .AppendLine()
                .AppendFormat("DepthOutNormalizer: {0:0.000}", _forceMorphHandler.DepthOutNormalizer)
                .AppendLine()
                .AppendFormat("SidewaysNormalizer: {0:0.000}", _forceMorphHandler.SidewaysNormalizer)
                .AppendLine()
                .AppendLine()
                .AppendFormat("AngleY L: {0:0.000}", _leftJoint.AngleY)
                .AppendLine()
                .AppendFormat("AngleY R: {0:0.000}", _rightJoint.AngleY)
                .AppendLine()
                .AppendFormat("AngleX L: {0:0.000}", _leftJoint.AngleX)
                .AppendLine()
                .AppendFormat("AngleX R: {0:0.000}", _rightJoint.AngleX)
                .AppendLine()
                .AppendFormat("DepthDiff L: {0:0.000}", _leftJoint.DepthDiff)
                .AppendLine()
                .AppendFormat("DepthDiff R: {0:0.000}", _rightJoint.DepthDiff)
                .AppendLine();
            _leftDebugArea.val = diagnosticsSb.ToString();
        }

        void DebugGravityAndAngles()
        {
            var leftT = _leftJoint.Rigidbody.transform;
            var rightT = _rightJoint.Rigidbody.transform;

            var gravity = _gravityPhysicsHandler.CustomGravityLeft;
            float relGLeftRight = gravity.RelativeTo(leftT, Vector3.right);
            float relGLeftUp = gravity.RelativeTo(leftT, Vector3.up);
            float relGLeftForward = gravity.RelativeTo(leftT, Vector3.forward);
            float relGRightRight = gravity.RelativeTo(rightT, Vector3.right);
            float relGRightUp = gravity.RelativeTo(rightT, Vector3.up);
            float relGRightForward = gravity.RelativeTo(rightT, Vector3.forward);

            diagnosticsSb
                .Clear()
                .AppendFormat("LeftPitch: {0:0.00}", _angleCalculator.LeftPitch)
                .AppendLine()
                .AppendFormat("LeftRoll: {0:0.00}", _angleCalculator.LeftRoll)
                .AppendLine()
                .AppendFormat("RightPitch: {0:0.00}", _angleCalculator.RightPitch)
                .AppendLine()
                .AppendFormat("RightRoll: {0:0.00}", _angleCalculator.RightRoll)
                .AppendLine()
                .AppendLine()
                .AppendFormat("leanback fix L: {0:0.00}", _forceMorphHandler.leftLeanBackFixer)
                .AppendLine()
                .AppendFormat("leanback fix R: {0:0.00}", _forceMorphHandler.rightLeanBackFixer)
                .AppendLine()
                .AppendFormat("upsidedown fix L {0:0.00}", _forceMorphHandler.leftUpsideDownFixer)
                .AppendLine()
                .AppendFormat("upsidedown fix R {0:0.00}", _forceMorphHandler.rightUpsideDownFixer)
                .AppendLine()
                .AppendFormat("roll fix L {0:0.00}", _forceMorphHandler.leftRollFixer)
                .AppendLine()
                .AppendFormat("roll fix R {0:0.00}", _forceMorphHandler.rightRollFixer)
                .AppendLine()
                .AppendLine()
                .AppendItalic("Breast joint gravity (m/s^2):")
                .AppendLine()
                .AppendFormat(" x: {0:0.00}", gravity.x)
                .AppendLine()
                .AppendFormat(" y: {0:0.00}", gravity.y)
                .AppendLine()
                .AppendFormat(" z: {0:0.00}", gravity.z)
                .AppendLine()
                .AppendLine()
                .AppendItalic("Rel G Left (m/s^2):")
                .AppendLine()
                .AppendFormat(" x: {0:0.00}", relGLeftRight)
                .AppendLine()
                .AppendFormat(" y: {0:0.00}", relGLeftUp)
                .AppendLine()
                .AppendFormat(" z: {0:0.00}", relGLeftForward)
                .AppendLine();
            _leftDebugArea.val = diagnosticsSb.ToString();

            var refGLeftRight = _gravityPhysicsHandler.RefRightGravityLeft;
            var refGLeftUp = _gravityPhysicsHandler.RefUpGravityLeft;
            var refGLeftDown = _gravityPhysicsHandler.RefDownGravityLeft;
            var refGLeftForward = _gravityPhysicsHandler.RefForwardGravityLeft;
            var refGLeftBack = _gravityPhysicsHandler.RefBackGravityLeft;
            diagnosticsSb
                .Clear()
                .AppendItalic("Ref G Left Right (m/s^2):")
                .AppendLine()
                .AppendFormat(" x: {0:0.00}", refGLeftRight.x)
                .AppendLine()
                .AppendFormat(" y: {0:0.00}", refGLeftRight.y)
                .AppendLine()
                .AppendFormat(" z: {0:0.00}", refGLeftRight.z)
                .AppendLine()
                .AppendLine()
                .AppendItalic("Ref G Left Up (m/s^2):")
                .AppendLine()
                .AppendFormat(" x: {0:0.00}", refGLeftUp.x)
                .AppendLine()
                .AppendFormat(" y: {0:0.00}", refGLeftUp.y)
                .AppendLine()
                .AppendFormat(" z: {0:0.00}", refGLeftUp.z)
                .AppendLine()
                .AppendLine()
                .AppendItalic("Ref G Left Down (m/s^2):")
                .AppendLine()
                .AppendFormat(" x: {0:0.00}", refGLeftDown.x)
                .AppendLine()
                .AppendFormat(" y: {0:0.00}", refGLeftDown.y)
                .AppendLine()
                .AppendFormat(" z: {0:0.00}", refGLeftDown.z)
                .AppendLine()
                .AppendLine()
                .AppendItalic("Ref G Left Forward (m/s^2):")
                .AppendLine()
                .AppendFormat(" x: {0:0.00}", refGLeftForward.x)
                .AppendLine()
                .AppendFormat(" y: {0:0.00}", refGLeftForward.y)
                .AppendLine()
                .AppendFormat(" z: {0:0.00}", refGLeftForward.z)
                .AppendLine()
                .AppendLine()
                .AppendItalic("Ref G Left Back (m/s^2):")
                .AppendLine()
                .AppendFormat(" x: {0:0.00}", refGLeftBack.x)
                .AppendLine()
                .AppendFormat(" y: {0:0.00}", refGLeftBack.y)
                .AppendLine()
                .AppendFormat(" z: {0:0.00}", refGLeftBack.z)
                .AppendLine();
            _rightDebugArea.val = diagnosticsSb.ToString();
        }

        void DebugClothingMorphs()
        {
            diagnosticsSb
                .Clear()
                .AppendLine(_clothingHandler.LeftMorphsOffsets.ToPrettyString())
                .AppendLine(_clothingHandler.RightMorphsOffsets.ToPrettyString())
                .AppendLine(_clothingHandler.LeftMorphsOffsetsCompensation.ToPrettyString())
                .AppendLine(_clothingHandler.RightMorphsOffsetsCompensation.ToPrettyString())
                .AppendLine();
            foreach(var pair in _clothingHandler.MorphConfigs)
            {
                var config = pair.Value;
                diagnosticsSb
                    .AppendFormat("{0} Left: {1}", pair.Key, config.Left.morphValue)
                    .AppendLine()
                    .AppendFormat("{0} Right: {1}", pair.Key, config.Right.morphValue)
                    .AppendLine();
            }

            _leftDebugArea.val = diagnosticsSb.ToString();
        }

        void DebugSoftJoints()
        {
            try
            {
                if(_selectedParamGroup != null)
                {
                    if(Time.frameCount % 6 == 0)
                    {
                        _jointDiagnosticsJss.val = _softPhysicsHandler.SoftJointDiagnostics.GetSelectedJointValues(_selectedParamGroup);
                    }
                }

                _softPhysicsHandler.SoftJointDiagnostics.ClearLines();
                _softPhysicsHandler.SoftJointDiagnostics.DrawSelectedJoint();
                if(_selectedParamGroup != null)
                {
                    _softPhysicsHandler.SoftJointDiagnostics.DrawDiagnostics(_selectedParamGroup);
                }
            }
            catch(Exception e)
            {
                Debug.Log($"DebugSoftJoints: {e}");
            }
        }
    }
}
