using ColliderEditor;
using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.UI;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.TM.UI
{
    sealed class HardCollidersWindow : Window
    {
        readonly Dictionary<string, UIDynamic> _sectionElements = new Dictionary<string, UIDynamic>();
        readonly HardColliderHandler _hardColliderHandler;
        readonly ColliderVisualizer _visualizer;
        readonly JSONStorableString _allCollidersMassDisplayValueJss;

        internal HardCollidersWindow(ScriptBase script) : base(script)
        {
            _hardColliderHandler = script.Provider.GetInstance<HardColliderHandler>();
            _visualizer = _hardColliderHandler.ColliderVisualizer;
            _allCollidersMassDisplayValueJss = new JSONStorableString("allCollidersMassDisplayValue", "");
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            AddShowPreviewsPopupChangeHandler();
            AddColliderPopupChangeHandler();
            _visualizer.enabled = true;
            _visualizer.ShowPreviewsJSON.val = true;
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddHeaderTextField("Configure Hard Colliders", rightSide);

            AddInfoTextField(
                "Hard colliders apply a force to the back to the pectoral joint, making the breast move in response to collision.",
                rightSide
            );

            AddElement(_hardColliderHandler.BaseMassMultiplierJsf.name, () =>
            {
                var slider = script.CreateSlider(_hardColliderHandler.BaseMassMultiplierJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Base Mass Multiplier";
                return slider;
            });

            AddElement(() => CreateValueTextField(_allCollidersMassDisplayValueJss, 28, 45, rightSide));

            AddInfoTextField(
                new StringBuilder()
                    .AppendBold("Collider mass")
                    .AppendLine(" affects the pectoral joint's inertia as well as the amount of force applied back to the pectoral joint.")
                    .AppendLine()
                    .AppendBold("Radius")
                    .AppendLine(" adjusts the size of the selected collider.")
                    .AppendLine()
                    .AppendBold("X, Y and Z offsets")
                    .AppendLine(" can be used to modify the position of the selected collider.")
                    .AppendLine()
                    .Append("The closer the collider matches the volume of the breast, the more accurately it responds to touch.")
                    .ToString(),
                rightSide,
                525f
            );

            AddElement(_visualizer.GroupsJSON.name, () =>
            {
                var popup = script.CreatePopup(_visualizer.GroupsJSON, rightSide);
                popup.label = "Show Previews";
                popup.popup.labelText.color = Color.black;
                popup.popup.popupPanel.offsetMin += new Vector2(5, popup.popup.popupPanelHeight + 55);
                popup.popup.popupPanel.offsetMax += new Vector2(5, popup.popup.popupPanelHeight + 55);
                return popup;
            });

            AddElement(_visualizer.XRayPreviewsJSON.name, () =>
            {
                var toggle = script.CreateToggle(_visualizer.XRayPreviewsJSON, rightSide);
                toggle.label = "Use XRay Previews";
                return toggle;
            });

            AddElement(_hardColliderHandler.HighlightAllJsb.name, () =>
                script.CreateToggle(_hardColliderHandler.HighlightAllJsb, rightSide)
            );
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(40, rightSide);
            AddElement(_hardColliderHandler.ColliderGroupsJssc.name, () =>
            {
                var popup = script.CreateFilterablePopup(_hardColliderHandler.ColliderGroupsJssc, rightSide);
                script.ConfigureCustomPopup(popup, 340f, false);
                script.AddPrevNextButtons(popup, true);
                return popup;
            });
        }

        protected override void OnClose()
        {
            _visualizer.ShowPreviewsJSON.val = false;
            _visualizer.enabled = false;
        }

        void RebuildSection(string colliderId, bool rightSide = true)
        {
            var hardColliderGroup = _hardColliderHandler.HardColliderGroups
                .Find(group => group.VisualizerId == colliderId);

            AddColliderSectionElement(hardColliderGroup.MassMultiplierJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.MassMultiplierJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Mass Multiplier";
                slider.AddListener((float value) => UpdateSliderColor(hardColliderGroup.MassMultiplierJsf));
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.Left.MassDisplayValueJss.name, () =>
                CreateValueTextField(hardColliderGroup.Left.MassDisplayValueJss, 28, 45, rightSide)
            );

            AddColliderSectionElement(hardColliderGroup.RadiusJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.RadiusJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Radius Offset";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.RightJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.RightJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "X Offset";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.UpJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.UpJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Y Offset";
                return slider;
            });

            AddColliderSectionElement(hardColliderGroup.LookJsf.name, () =>
            {
                var slider = script.CreateSlider(hardColliderGroup.LookJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Z Offset";
                return slider;
            });

            GetUIDynamic(_hardColliderHandler.BaseMassMultiplierJsf.name).AddListener(UpdateAllSliderColors);
            UpdateAllSliderColors(0);
        }

        void AddColliderSectionElement(string key, Func<UIDynamic> createElement)
        {
            _sectionElements[key] = createElement();
        }

        void UpdateAllSliderColors(float value)
        {
            foreach(var group in _hardColliderHandler.HardColliderGroups)
            {
                if(_sectionElements.ContainsKey(group.MassMultiplierJsf.name))
                {
                    UpdateSliderColor(group.MassMultiplierJsf);
                }
            }
        }

        void UpdateSliderColor(JSONStorableFloat storable)
        {
            var slider = (UIDynamicSlider) _sectionElements[storable.name];
            var images = slider.slider.gameObject.transform.GetComponentsInChildren<Image>();
            var fillImage = images.First(image => image.name == "Fill");
            var handleImage = images.First(image => image.name == "Handle");
            var color = MultiplierSliderColor(_hardColliderHandler.BaseMassMultiplierJsf.val * storable.val / 1.5f);
            fillImage.color = color;
            handleImage.color = color;
        }

        static Color MultiplierSliderColor(float value)
        {
            if(value <= 1 / 4f)
            {
                return Color.Lerp(new Color(1, 1, 1, 0.25f), Color.white, 4 * value);
            }

            if(value <= 1 / 3f)
            {
                return Color.white;
            }

            return value <= 2 / 3f
                ? Color.Lerp(Color.white, new Color(1.0f, 0.6f, 0.2f), Mathf.InverseLerp(1 / 3f, 2 / 3f, value))
                : Color.Lerp(new Color(1.0f, 0.6f, 0.2f), new Color(1.0f, 0.2f, 0.2f), Mathf.InverseLerp(2 / 3f, 1, value));

        }

        internal override void Clear()
        {
            base.Clear();
            ClearSection();
            RemoveShowPreviewsPopupChangeHandler();
            RemoveColliderPopupChangeHandler();
        }

        void ClearSection()
        {
            foreach(var pair in _sectionElements)
            {
                script.RemoveElement(pair.Value);
            }

            _sectionElements.Clear();
        }

        void AddShowPreviewsPopupChangeHandler()
        {
            var uiDynamicPopup = GetUIDynamicAs<UIDynamicPopup>(_visualizer.GroupsJSON.name);
            if(uiDynamicPopup)
            {
                uiDynamicPopup.popup.onValueChangeHandlers += OnShowPreviewsPopupValueChanged;
                GetUIDynamic(_visualizer.XRayPreviewsJSON.name).SetActiveStyle(uiDynamicPopup.popup.currentValue != "Off");
            }
        }

        void RemoveShowPreviewsPopupChangeHandler()
        {
            var popup = GetUIDynamicAs<UIDynamicPopup>(_visualizer.GroupsJSON.name);
            if(popup)
            {
                popup.popup.visible = false;
                popup.popup.onValueChangeHandlers -= OnShowPreviewsPopupValueChanged;
            }
        }

        void OnShowPreviewsPopupValueChanged(string value)
        {
            GetUIDynamic(_visualizer.XRayPreviewsJSON.name).SetActiveStyle(value != "Off");
            GetUIDynamic(_hardColliderHandler.HighlightAllJsb.name).SetActiveStyle(value != "Off");
        }

        void AddColliderPopupChangeHandler()
        {
            var popup = GetUIDynamicAs<UIDynamicPopup>(_hardColliderHandler.ColliderGroupsJssc.name);
            if(popup)
            {
                popup.popup.onValueChangeHandlers += OnColliderPopupValueChanged;
                RebuildSection(_hardColliderHandler.ColliderGroupsJssc.val);
            }
        }

        void RemoveColliderPopupChangeHandler()
        {
            var popup = GetUIDynamicAs<UIDynamicPopup>(_hardColliderHandler.ColliderGroupsJssc.name);
            if(popup)
            {
                popup.popup.visible = false;
                popup.popup.onValueChangeHandlers -= OnColliderPopupValueChanged;
            }
        }

        void OnColliderPopupValueChanged(string value)
        {
            ClearSection();
            RebuildSection(value);
        }

        internal override void UpdateDiagnostics()
        {
            if(!isBuilt || Time.frameCount % 6 != 0)
            {
                return;
            }

            float mass = _hardColliderHandler.HardColliderGroups.Sum(group => group.Left.GetRigidbodyMass());
            _allCollidersMassDisplayValueJss.val = diagnosticsSb
                .Clear()
                .AppendFormat("All Colliders Mass: <b>{0:0.000} kg</b>", mass)
                .ToString();
        }
    }
}
