using NAT.Common;
using NAT.Common.UI;
using NAT.TM.Handlers;
using SimpleJSON;
using System.Text;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.TM.UI
{
    sealed class MainWindow : Window
    {
        readonly BreastJointPhysicsHandler _jointPhysicsHandler;
        readonly BreastSoftPhysicsHandler _softPhysicsHandler;
        readonly JSONStorableString _massValueJss = new JSONStorableString("Mass Value", "");
        UIDynamicTextField _currentPresetTextField;

        internal MainWindow(ScriptBase script) : base(script)
        {
            var provider = script.Provider;
            _jointPhysicsHandler = provider.GetInstance<BreastJointPhysicsHandler>();
            _softPhysicsHandler = provider.GetInstance<BreastSoftPhysicsHandler>();

            AddNestedWindow(new SettingsWindow(script));
            if(script.IsDevelopmentEnv)
            {
                AddNestedWindow(new DevWindow(script));
                AddNestedWindow(new DevMorphWindow(script));
            }
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            BuildPresetArea();
            BuildLowerArea();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddSpacer(5, rightSide);
            AddElement(() =>
            {
                var slider = script.CreateSlider(script.SoftnessJsf, rightSide);
                slider.valueFormat = "F0";
                return slider;
            });

            AddInfoTextField(
                new StringBuilder()
                    .Append("Softness, quickness and mass affect physics base values, the effect of gravity,")
                    .Append(" the amount of morphing, and hard collider properties. Use them to determine")
                    .Append(" the overall behavior, then fine tune physics, morphs and colliders as needed.")
                    .ToString(),
                rightSide,
                180f
            );

            AddElement(_softPhysicsHandler.SoftPhysicsOnJsb.name, () =>
                script.CreateToggle(_softPhysicsHandler.SoftPhysicsOnJsb, rightSide)
            );

            AddInfoTextField(
                new StringBuilder()
                    .AppendFormat("Simulate soft-body dynamics for breasts. Soft physics is required to get the most out of {0}.", Constant.TITTY_MAGIC)
                    .ToString(),
                rightSide
            );

            AddElement(_softPhysicsHandler.SoftPhysicsSelfCollideJsb.name, () =>
                script.CreateToggle(_softPhysicsHandler.SoftPhysicsSelfCollideJsb, rightSide)
            );
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(5, rightSide);
            AddElement(() =>
            {
                var slider = script.CreateSlider(script.QuicknessJsf, rightSide);
                slider.valueFormat = "F2";
                return slider;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(_jointPhysicsHandler.JointMass.Left.OffsetJsf, rightSide);
                slider.valueFormat = _jointPhysicsHandler.JointMass.ValueFormat;
                return slider;
            });

            AddElement(() => CreateValueTextField(_massValueJss, 28, 45, rightSide));

            AddElement(_softPhysicsHandler.LdpEnabledJsb.name, () =>
                script.CreateToggle(_softPhysicsHandler.LdpEnabledJsb, rightSide)
            );

            AddInfoTextField(
                "Adjust soft physics parameters of individual soft joints dynamically with stretching and compression.",
                rightSide
            );

            AddElement(() =>
            {
                var button = script.CreateButton("Configure Collider Friction", rightSide);
                button.AddListener(script.ConfigureColliderFriction);
                return button;
            });

            AddInfoTextField(
                "Collider friction controls how easily colliders stick to or slip away from hands or other colliding objects.",
                rightSide
            );
        }

        void BuildPresetArea()
        {
            if(isBuilt)
            {
                return;
            }

            AddTextField("Preset Header", name =>
            {
                var sb = new StringBuilder().AppendLine(script.DomainName).Append("Preset");
                var uiDynamic = script.CreateCustomTextField(name, sb.ToString(), new Vector2(5, -1140), new Vector2(-890, 70));
                uiDynamic.UItext.fontSize = 28;
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            });
            AddButton("Save", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(200, -1118), new Vector2(-950, 50));
                uiDynamic.AddListener(() => script.PresetHandler.SavePresetJsu.FileBrowse());
                return uiDynamic;
            });
            AddButton("Load", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(328, -1118), new Vector2(-950, 50));
                uiDynamic.AddListener(() => script.PresetHandler.LoadPresetJsu.FileBrowse());
                return uiDynamic;
            });
            AddButton("Reset", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(455, -1118), new Vector2(-950, 50));
                uiDynamic.AddListener(() => script.RestoreFromPresetJSON(new JSONClass(), null));
                return uiDynamic;
            });
            AddButton("Set As Default", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(591, -1118), new Vector2(-840, 50));
                uiDynamic.AddListener(() => script.PresetHandler.SaveDefaultPreset(script.Person.IsFemale));
                uiDynamic.buttonText.fontSize = 26;
                return uiDynamic;
            });
            AddButton("Load Default", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(830, -1118), new Vector2(-840, 50));
                uiDynamic.AddListener(() => script.PresetHandler.LoadDefaultPreset(script.Person.IsFemale));
                uiDynamic.buttonText.fontSize = 26;
                return uiDynamic;
            });
            AddTextField("Current Preset Info", name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "Current: None", new Vector2(200, -1154), new Vector2(-210, 32));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.MiddleLeft;
                uiDynamic.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
                var scrollView = uiDynamic.transform.Find("Scroll View");
                var scrollRect = scrollView.GetComponent<ScrollRect>();
                scrollRect.horizontal = true;
                scrollRect.horizontalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
                _currentPresetTextField = uiDynamic;
                return uiDynamic;
            });
        }

        void BuildLowerArea()
        {
            if(isBuilt)
            {
                return;
            }

            AddButton("User Preferences", name =>
            {
                var sb = new StringBuilder().AppendFormat("{0} User Preferences", Constant.NATURALIS);
                var uiDynamic = script.CreateCustomButton(name, sb.ToString(), new Vector2(360, -1224), new Vector2(-700, 50));
                uiDynamic.AddListener(OpenNestedWindow<SettingsWindow>);
                return uiDynamic;
            });

            AddTextField("Version", name =>
            {
                var sb = new StringBuilder().AppendFormat("v{0}", Bootstrap.NATURALIS_VERSION);
                var uiDynamic = script.CreateCustomTextField(name, sb.ToString(), new Vector2(841, -1230), new Vector2(-850, 42));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.alignment = TextAnchor.LowerRight;
                uiDynamic.UItext.fontSize = 26;
                return uiDynamic;
            });
        }

        protected override void PostBuild() => Refresh();

        internal override void UpdateDiagnostics()
        {
            if(Time.frameCount % 6 != 0)
            {
                return;
            }

            var paramGroup = _jointPhysicsHandler.JointMass;
            _massValueJss.val = diagnosticsSb
                .Clear()
                .AppendFormat("{0}: <b>{1} kg</b>", paramGroup.Name, paramGroup.Left.Value.ToString(paramGroup.ValueFormat))
                .ToString();
        }

        internal override void Refresh()
        {
            if(!isBuilt)
            {
                return;
            }

            var breastSoftPhysicsOnToggle = GetUIDynamic(_softPhysicsHandler.SoftPhysicsOnJsb.name);
            var ldpEnabledtoggle = GetUIDynamic(_softPhysicsHandler.LdpEnabledJsb.name);

            var personMonitor = script.Bootstrap.PersonMonitor;
            breastSoftPhysicsOnToggle.SetActiveStyle(
                personMonitor.IsBreastSoftPhysicsOn(),
                false,
                _softPhysicsHandler.SoftPhysicsOnJsb.val && script.enabled
            );
            ldpEnabledtoggle.SetActiveStyle(
                _softPhysicsHandler.SoftPhysicsOn && personMonitor.IsBreastSoftPhysicsOn(),
                false,
                script.enabled
            );

            _currentPresetTextField.text = new StringBuilder()
                .AppendFormat("Current: {0}", script.LoadedPresetName ?? "None")
                .ToString();
        }
    }
}
