using NAT.Common;
using NAT.Common.UI;
using NAT.TM.Handlers;
using System.Linq;
using System.Text;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.TM.UI
{
    sealed class MorphsWindow : Window
    {
        readonly BreastForceMorphHandler _forceMorphHandler;
        readonly NippleErectionHandler _nippleErectionHandler;
        readonly BreastGravityOffsetMorphHandler _gravityOffsetMorphHandler;

        internal MorphsWindow(ScriptBase script) : base(script)
        {
            var provider = script.Provider;
            _forceMorphHandler = provider.GetInstance<BreastForceMorphHandler>();
            _nippleErectionHandler = provider.GetInstance<NippleErectionHandler>();
            _gravityOffsetMorphHandler = provider.GetInstance<BreastGravityOffsetMorphHandler>();
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            GetUIDynamic(_forceMorphHandler.BaseJsf.name).AddListener(UpdateAllSliderColors);
            UpdateAllSliderColors(0);
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddHeaderTextField("Directional Force Morphing", rightSide);

            AddSpacer(5, rightSide);
            AddElement(_forceMorphHandler.BaseJsf.name, () =>
            {
                var slider = script.CreateSlider(_forceMorphHandler.BaseJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Base Multiplier";
                return slider;
            });

            AddInfoTextField(
                new StringBuilder()
                    .AppendLine("Control how much breast shape is adjusted dynamically with morphs.")
                    .AppendLine()
                    .Append("The amount of morphing is based on the breast's movement away from its neutral position relative to the chest.")
                    .AppendLine("The neutral position is estimated as the position when the person is in an upright pose and only the force of gravity is applied.")
                    .AppendLine()
                    .Append("Anything that causes the breast to move will cause morphing: collision, gravity or any kind of animation.")
                    .ToString(),
                rightSide,
                620f,
                28
            );

            AddHeaderTextField("Other Settings", rightSide);

            AddInfoTextField(
                "Expand nipple morphs and harden nipple physics.\n\nCompensate for the droop caused by negative Up/Down Angle Target value.",
                rightSide,
                180f,
                28
            );
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(80, rightSide);
            AddMultiplierSlider(_forceMorphHandler.UpJsf, "Up", rightSide);
            AddMultiplierSlider(_forceMorphHandler.DepthOutJsf, "Depth Out", rightSide);
            AddMultiplierSlider(_forceMorphHandler.DepthInJsf, "Depth In", rightSide);
            AddMultiplierSlider(_forceMorphHandler.SidewaysInJsf, "Sideways In", rightSide);
            AddMultiplierSlider(_forceMorphHandler.SidewaysOutJsf, "Sideways Out", rightSide);

            AddSpacer(90, rightSide);
            AddElement(() =>
            {
                var slider = script.CreateSlider(_nippleErectionHandler.NippleErectionJsf, rightSide);
                slider.valueFormat = "F2";
                return slider;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(_gravityOffsetMorphHandler.OffsetMorphingJsf, rightSide);
                slider.valueFormat = "F2";
                return slider;
            });
        }

        void AddMultiplierSlider(JSONStorableFloat storable, string label, bool rightSide) => AddElement(storable.name, () =>
        {
            var slider = script.CreateSlider(storable, rightSide);
            slider.valueFormat = "F2";
            slider.label = label;
            slider.AddListener((float _) => UpdateSliderColor(storable));
            return slider;
        });

        void UpdateAllSliderColors(float _)
        {
            UpdateSliderColor(_forceMorphHandler.UpJsf);
            UpdateSliderColor(_forceMorphHandler.DepthOutJsf);
            UpdateSliderColor(_forceMorphHandler.DepthInJsf);
            UpdateSliderColor(_forceMorphHandler.SidewaysInJsf);
            UpdateSliderColor(_forceMorphHandler.SidewaysOutJsf);
        }

        void UpdateSliderColor(JSONStorableFloat storable)
        {
            var slider = GetUIDynamicAs<UIDynamicSlider>(storable.name);
            if(slider)
            {
                var images = slider.slider.gameObject.transform.GetComponentsInChildren<Image>();
                var fillImage = images.First(image => image.name == "Fill");
                var handleImage = images.First(image => image.name == "Handle");
                var color = MultiplierSliderColor(_forceMorphHandler.BaseJsf.val * storable.val);
                fillImage.color = color;
                handleImage.color = color;
            }
        }

        static Color MultiplierSliderColor(float value) => value <= 1
            ? Color.Lerp(new Color(1, 1, 1, 0.25f), Color.white, value)
            : Color.Lerp(Color.white, new Color(1.0f, 0.2f, 0.2f), (value - 1) / 3);
    }
}
