using NAT.Common;
using NAT.Common.Models;
using NAT.Common.UI;
using NAT.TM.Handlers;
using System;
using System.Linq;
using System.Text;
using UnityEngine;

namespace NAT.TM.UI
{
    sealed class SoftParameterWindow : Window
    {
        readonly BreastSoftPhysicsHandler _softPhysicsHandler;
        readonly SoftPhysicsParamGroup _paramGroup;
        readonly SoftPhysicsParam _param;
        bool IsFrictionParam => _paramGroup.Name == FAT_COLLIDER_FRICTION.NAME;

        readonly JSONStorableString _outerValueJss = new JSONStorableString("outerValue", "");
        readonly JSONStorableString _mainValueJss = new JSONStorableString("mainValue", "");
        readonly JSONStorableString _areolaValueJss = new JSONStorableString("areolaValue", "");
        readonly JSONStorableString _nippleValueJss = new JSONStorableString("nippleValue", "");
        readonly StringBuilder _outerValueSb = new StringBuilder();
        readonly StringBuilder _mainValueSb = new StringBuilder();
        readonly StringBuilder _areolaValueSb = new StringBuilder();
        readonly StringBuilder _nippleValueSb = new StringBuilder();

        internal SoftParameterWindow(ScriptBase script, SoftPhysicsParamGroup paramGroup) : base(script)
        {
            _softPhysicsHandler = script.Provider.GetInstance<BreastSoftPhysicsHandler>();
            _paramGroup = paramGroup;
            _param = _paramGroup.Left;
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();

            if(_paramGroup.HasSoftColliderVisualization && !isBuilt)
            {
                AddToggle(_softPhysicsHandler.ShowColliderPreviewsJsb, jsb =>
                {
                    /* Get toggle created in different soft parameter's window
                     * TODO: single parameter window instance, change which parameter's elements shown
                     */
                    var uiDynamic = script.GetCustomElementAs<UIDynamicToggle>(jsb.name);
                    if(uiDynamic)
                    {
                        uiDynamic.gameObject.SetActive(true);
                        return uiDynamic;
                    }

                    float posX = script.IsGlobalUISideLeft ? 1113 : -408;
                    uiDynamic = script.CreateCustomToggle(jsb.name, jsb.name, new Vector2(posX, -882), new Vector2(-705, 50));
                    uiDynamic.textColor = Color.white;
                    uiDynamic.backgroundColor = Colors.darkTransparentGray;
                    jsb.toggle = uiDynamic.toggle;
                    return uiDynamic;
                });

                AddSlider(_softPhysicsHandler.ColliderVisualizer.PreviewOpacityJSON, jsf =>
                {
                    var uiDynamic = script.GetCustomElementAs<UIDynamicSlider>(jsf.name);
                    if(uiDynamic)
                    {
                        uiDynamic.gameObject.SetActive(true);
                        return uiDynamic;
                    }

                    float posX = script.IsGlobalUISideLeft ? 1113 : -553;
                    uiDynamic = script.CreateCustomSlider(jsf.name, "Collider Previews Opacity", new Vector2(posX, -1000));
                    script.ConfigureColliderPreviewsOpacitySlider(uiDynamic);
                    jsf.slider = uiDynamic.slider;
                    return uiDynamic;
                });
            }
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddInfoTextField(
                new StringBuilder().AppendSize("\n", 8).Append(_paramGroup.InfoText).ToString(),
                rightSide,
                IsFrictionParam ? 160f : 225f,
                28
            );

            if(IsFrictionParam)
            {
                AddElement(() =>
                {
                    var button = script.CreateButton("Configure Collider Friction", rightSide);
                    button.AddListener(script.ConfigureColliderFriction);
                    return button;
                });
            }

            AddElement(() =>
            {
                var sb = new StringBuilder().AppendSize("\n", 8).AppendBold("Local Distortion Physics");
                var textField = CreateHeaderTextField(sb.ToString(), 30, 50, rightSide);
                textField.UItext.alignment = TextAnchor.LowerCenter;
                return textField;
            });

            AddInfoTextField(
                new StringBuilder()
                    .AppendBold("Effect")
                    .Append(" causes the value to increase (if > 0) or decrease (if < 0)")
                    .Append(" as joints move closer to (compress) or further from (stretch) each other.")
                    .AppendBold(" Sensitivity")
                    .Append(" controls how easily the value changes due to small distortions.")
                    .ToString(),
                rightSide,
                160f
            );

            AddElement(_softPhysicsHandler.SoftJointDiagnostics.VisualizeLdpJsb.name, () =>
            {
                var toggle = script.CreateToggle(_softPhysicsHandler.SoftJointDiagnostics.VisualizeLdpJsb, rightSide);
                toggle.label = "Visualize";
                return toggle;
            });

            var ldpConfig = _param.LdpConfig;
            AddElement(ldpConfig.CompressionEffectJsf.name, () =>
            {
                var slider = script.CreateSlider(ldpConfig.CompressionEffectJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Compression Effect";
                return slider;
            });

            AddElement(ldpConfig.CompressionSensitivityJsf.name, () =>
            {
                var slider = script.CreateSlider(ldpConfig.CompressionSensitivityJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Compression Sensitivity";
                return slider;
            });

            AddElement(ldpConfig.StretchingEffectJsf.name, () =>
            {
                var slider = script.CreateSlider(ldpConfig.StretchingEffectJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Stretching Effect";
                return slider;
            });

            AddElement(ldpConfig.StretchingSensitivityJsf.name, () =>
            {
                var slider = script.CreateSlider(ldpConfig.StretchingSensitivityJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Stretching Sensitivity";
                return slider;
            });
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddElement(() =>
            {
                var toggle = script.CreateToggle(_paramGroup.ApplyChangesOnlyToLeftParamJsb, rightSide);
                toggle.label = "Apply Changes Only To Left Breast";
                return toggle;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(_param.OffsetJsf, rightSide);
                slider.valueFormat = _paramGroup.ValueFormat;
                slider.label = IsFrictionParam ? "Friction Value" : "Offset";
                slider.SetActiveStyle(!IsFrictionParam, true);
                return slider;
            });

            AddSpacer(90, rightSide);
            AddElement(() => GroupValueTextField(_outerValueJss, new Vector2(550, -320)));
            AddElement(() => GroupValueTextField(_mainValueJss, new Vector2(550, -365)));
            AddElement(() => GroupValueTextField(_areolaValueJss, new Vector2(819, -320)));
            AddElement(() => GroupValueTextField(_nippleValueJss, new Vector2(819, -365)));

            AddElement(() =>
            {
                var sb = new StringBuilder().AppendSize("\n", 8).AppendBold("Value Gradient");
                var textField = CreateHeaderTextField(sb.ToString(), 30, 50, rightSide);
                textField.UItext.alignment = TextAnchor.LowerCenter;
                return textField;
            });

            AddInfoTextField(
                new StringBuilder()
                    .AppendBold("Middle Shift")
                    .Append(" moves the gradient's mid point from the outer joints")
                    .Append(" at -1.00 to the areola joints at 1.00. The ")
                    .AppendBold("Multiplier")
                    .Append(" sliders set target values for the gradient's end points.")
                    .ToString(),
                rightSide,
                160f
            );

            var visualizeGradient = _softPhysicsHandler.SoftJointDiagnostics.VisualizeGradientJsb;
            AddElement(() =>
            {
                var toggle = script.CreateToggle(visualizeGradient, rightSide);
                toggle.label = "Visualize";
                return toggle;
            });

            var gradientConfig = _param.GradientConfig;
            AddElement(() =>
            {
                var slider = script.CreateSlider(gradientConfig.MiddleShiftJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Middle Shift";
                return slider;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(gradientConfig.OuterMultiplierJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Outer Multiplier";
                return slider;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(gradientConfig.CenterMultiplierJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Areola Multiplier";
                return slider;
            });

            AddElement(() =>
            {
                var slider = script.CreateSlider(gradientConfig.NippleMultiplierJsf, rightSide);
                slider.valueFormat = "F2";
                slider.label = "Nipple Multiplier";
                return slider;
            });
        }

        UIDynamicTextField GroupValueTextField(JSONStorableString storable, Vector2 position)
        {
            var parentT = script.UITransform.Find("Scroll View/Viewport/Content");
            var fieldTransform = script.InstantiateTextField(parentT);
            var rectTransform = fieldTransform.GetComponent<RectTransform>();
            rectTransform.pivot = Vector2.zero;
            rectTransform.anchoredPosition = position;
            rectTransform.sizeDelta = new Vector2(-815, 42);
            var textField = fieldTransform.GetComponent<UIDynamicTextField>();
            textField.text = storable.val;
            textField.UItext.fontSize = 26;
            script.AddTextFieldToJss(textField, storable);
            return textField;
        }

        protected override void PostBuild()
        {
            Refresh();
            _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(true);
            SetShowColliderPreviews();
        }

        internal override void Refresh()
        {
            var ldpConfig = _param.LdpConfig;
            bool active = _softPhysicsHandler.IsLdpUpdateOn();
            GetUIDynamic(_softPhysicsHandler.SoftJointDiagnostics.VisualizeLdpJsb.name)
                .SetActiveStyle(active, true);
            GetUIDynamic(ldpConfig.CompressionEffectJsf.name)
                .SetActiveStyle(active, true);
            GetUIDynamic(ldpConfig.CompressionSensitivityJsf.name)
                .SetActiveStyle(active, true);
            GetUIDynamic(ldpConfig.StretchingEffectJsf.name)
                .SetActiveStyle(active, true);
            GetUIDynamic(ldpConfig.StretchingSensitivityJsf.name)
                .SetActiveStyle(active, true);
        }

        internal void SetShowColliderPreviews() =>
            _softPhysicsHandler.SetShowColliderPreviews(_paramGroup.HasSoftColliderVisualization);

        internal override void UpdateDiagnostics()
        {
            if(Time.frameCount % 6 == 0)
            {
                _outerValueJss.val = ValueInfoFieldString(_outerValueSb, SoftVerticesGroupName.OUTER);
                _mainValueJss.val = ValueInfoFieldString(_mainValueSb, SoftVerticesGroupName.MAIN);
                _areolaValueJss.val = ValueInfoFieldString(_areolaValueSb, SoftVerticesGroupName.AREOLA);
                _nippleValueJss.val = ValueInfoFieldString(_nippleValueSb, SoftVerticesGroupName.NIPPLE);
            }

            _softPhysicsHandler.SoftJointDiagnostics.ClearLines();
            _softPhysicsHandler.SoftJointDiagnostics.DrawDiagnostics(_paramGroup);
            if(_paramGroup.HasSoftColliderVisualization && _softPhysicsHandler.ShowColliderPreviewsJsb.val)
            {
                _softPhysicsHandler.ColliderVisualizer.SyncPreviews();
            }
        }

        string ValueInfoFieldString(StringBuilder sb, string groupName)
        {
            sb.Clear();
            sb.AppendSize("\n", 5);

            switch(groupName)
            {
                case SoftVerticesGroupName.OUTER:
                {
                    sb.Append("Outer (avg): ");
                    break;
                }
                case SoftVerticesGroupName.CENTER:
                {
                    sb.Append("Center (avg): ");
                    break;
                }
                case SoftVerticesGroupName.AREOLA:
                {
                    sb.Append("Areola (avg): ");
                    break;
                }
                case SoftVerticesGroupName.NIPPLE:
                {
                    sb.Append("Nipple: ");
                    break;
                }
            }

            return sb
                .AppendBold(GetAverageGroupValue(groupName).ToString(_paramGroup.ValueFormat))
                .ToString();
        }

        float GetAverageGroupValue(string groupName)
        {
            string paramName = _paramGroup.Name;
            switch(groupName)
            {
                case SoftVerticesGroupName.OUTER:
                {
                    var softJoints = _softPhysicsHandler.OuterSoftJoints[Side.LEFT];
                    return softJoints.Average(softJoint => softJoint.GetRealValue(paramName));
                }
                case SoftVerticesGroupName.MAIN:
                {
                    var softJoints = _softPhysicsHandler.MainSoftJoints[Side.LEFT];
                    return softJoints.Average(softJoint => softJoint.GetRealValue(paramName));
                }
                case SoftVerticesGroupName.AREOLA:
                {
                    var softJoints = _softPhysicsHandler.CenterSoftJoints[Side.LEFT];
                    return softJoints.Average(softJoint => softJoint.GetRealValue(paramName));
                }
                case SoftVerticesGroupName.NIPPLE: return _softPhysicsHandler.NippleSoftJoints[Side.LEFT].GetRealValue(paramName);
                default: throw new Exception($"Invalid group name {groupName}");
            }
        }

        protected override void OnClose()
        {
            _softPhysicsHandler.ColliderVisualizer.ShowPreviewsJSON.val = false;
            _softPhysicsHandler.SoftJointDiagnostics.DrawServiceSetEnabled(false);
        }
    }
}
