using NAT.Common;
using NAT.Common.Components;
using NAT.Common.Handlers;
using NAT.Common.Storables;
using System;
using System.Collections;
using Naturalis;
using System.Text;
using UnityEngine;

namespace NAT.TM
{
    sealed class UIModManager : UIModManagerBase
    {
        ScriptBase _script;
        JSONStorable _bootyMagic;
        FrictionHandler _frictionHandler;

        bool CheckPluginEnabled() => _script.enabled;

        internal void Init(ProviderBase provider)
        {
            _script = provider.GetInstance<TittyMagic>();
            _bootyMagic = _script.Bootstrap.BootyMagicStorable;
            _frictionHandler = provider.GetInstance<FrictionHandler>();
            parentUI = _script.Person.AtomUIContent;
            var showFrictionUIAction = new StorableAction(Constant.INTERNAL_SHOW_FRICTION_UI, () => ShowFrictionUI(_frictionHandler.IsEnabled));
            var hideFrictionUIAction = new StorableAction(Constant.INTERNAL_HIDE_FRICTION_UI, () => HideFrictionUI(_frictionHandler.IsEnabled));
            var setTopElementPositionAction = new StorableAction(
                Constant.INTERNAL_SET_TOP_ELEMENTS_POSITION,
                () => StartCoroutine(SetTopElementsPositionCo())
            );
            skinMaterials2ModifiedJsb = new StorableBool(Constant.INTERNAL_SKIN_MATERIALS_2_MODIFIED, false);
            SwitchToOtherPluginAction = new JSONStorableAction(
                $"Switch To {Constant.TITTY_MAGIC}",
                () =>
                {
                    _bootyMagic.CallActionNullSafe(Constant.INTERNAL_HIDE_FRICTION_UI);
                    if(enabled)
                    {
                        ShowFrictionUI(_frictionHandler.IsEnabled);
                    }
                }
            );

            _script.PreRegisterInternalParam(showFrictionUIAction);
            _script.PreRegisterInternalParam(hideFrictionUIAction);
            _script.PreRegisterInternalParam(setTopElementPositionAction);
            _script.PreRegisterInternalParam(skinMaterials2ModifiedJsb);

            NewUIMod("calibrationButton", null, CreateCalibrationButtonCo, CheckPluginEnabled, true);
            NewUIMod(Constant.M_PECTORAL_PHYSICS, "M Pectoral Physics", ReplaceWithPluginUIButtonCo, CheckPluginEnabled);
            NewUIMod(Constant.F_BREAST_PHYSICS_1, "F Breast Physics 1", ReplaceWithPluginUIButtonCo, CheckPluginEnabled);
            NewUIMod(Constant.F_BREAST_PHYSICS_2, "F Breast Physics 2", ReplaceWithPluginUIButtonCo, CheckPluginEnabled);
            NewUIMod(Constant.F_BREAST_PRESETS, "F Breast Presets", ReplaceWithPluginUIButtonCo, CheckPluginEnabled);
            NewUIMod(Constant.SKIN_MATERIALS_2_ADD_BTN, "Skin Materials 2", AddButtonToSkinMaterialsUICo, CheckPluginEnabled);
            NewUIMod(Constant.SKIN_MATERIALS_2_FRICTION, "Skin Materials 2", CreateFrictionUICo, CheckPluginEnabled);

            foreach(var uiMod in uiMods)
            {
                uiMod.DeferApply();
            }

            initialized = true;

            if(!_script.enabled)
            {
                enabled = false;
            }
        }

        RectTransform _calibrationRect;

        IEnumerator SetTopElementsPositionCo()
        {
            while(!_calibrationRect) // keeps waiting for person UI to be opened which triggers the elements to be created
            {
                yield return null;
            }

            yield return new WaitForSeconds(0.1f); // In case BootyMagic calls this action OnDestroy but isn't yet gone
            _calibrationRect.anchoredPosition = _bootyMagic ? new Vector2(277, 155) : new Vector2(0, 155);
        }

        UIDynamicButton _calibrationButton;

        IEnumerator CreateCalibrationButtonCo(UIMod uiMod)
        {
            while(_script.Initialized != true)
            {
                yield return null;
            }

            try
            {
                var customTransform = _script.InstantiateButton(parentUI);
                _calibrationRect = customTransform.GetComponent<RectTransform>();
                _calibrationRect.pivot = Vector2.zero;
                _calibrationRect.anchoredPosition = _script.IsBootyMagic ? new Vector2(277, 155) : new Vector2(0, 155);
                _calibrationRect.sizeDelta = new Vector2(-837, 80);
                uiMod.AddCustomObject(customTransform.gameObject);

                _calibrationButton = customTransform.GetComponent<UIDynamicButton>();
                _calibrationButton.AddListener(() => _script.StartCalibration());
                _script.CalibrationStateHandler.CalibrationButton = _calibrationButton;
            }
            catch(Exception e)
            {
                _script.Loggr.LogError($"Error modifying {uiMod.Id} UI: {e}");
            }
        }

        IEnumerator ReplaceWithPluginUIButtonCo(UIMod uiMod)
        {
            while(_script.Bindings == null && _script.Initialized != true)
            {
                yield return null;
            }

            try
            {
                uiMod.InactivateChildren();
                uiMod.AddCustomObject(OpenPluginUIButton(uiMod.Target).gameObject);
            }
            catch(Exception e)
            {
                _script.Loggr.LogError($"Error modifying {uiMod.Id} UI: {e}");
            }
        }

        IEnumerator CreateFrictionUICo(UIMod uiMod)
        {
            while(_script.Initialized != true)
            {
                yield return null;
            }

            if(_script.Person.HasDummySkin())
            {
                while(!_frictionHandler.IsEnabled)
                {
                    yield return null;
                }
            }
            else if(!_frictionHandler.IsEnabled)
            {
                yield break;
            }

            try
            {
                var leftSide = uiMod.Target.Find("LeftSide");
                var rightSide = uiMod.Target.Find("RightSide");

                /* Collider friction title */
                {
                    var fieldTransform = _script.InstantiateTextField(leftSide);
                    var rectTransform = fieldTransform.GetComponent<RectTransform>();
                    rectTransform.pivot = Vector2.zero;
                    rectTransform.anchoredPosition = new Vector2(20, -930);
                    rectTransform.sizeDelta = new Vector2(rectTransform.sizeDelta.x + 10, 100);
                    uiMod.AddCustomObject(fieldTransform.gameObject);

                    var uiDynamic = fieldTransform.GetComponent<UIDynamicTextField>();
                    uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                    uiDynamic.text = $"{Constant.TITTY_MAGIC} Collider Friction".Size(32).Bold();
                    uiDynamic.backgroundColor = Color.clear;
                    uiDynamic.textColor = Color.white;
                }

                /* Collider friction info text area */
                {
                    var fieldTransform = _script.InstantiateTextField(leftSide);
                    var rectTransform = fieldTransform.GetComponent<RectTransform>();
                    rectTransform.pivot = Vector2.zero;
                    rectTransform.anchoredPosition = new Vector2(20, -1290);
                    rectTransform.sizeDelta = new Vector2(rectTransform.sizeDelta.x + 10, 400);
                    uiMod.AddCustomObject(fieldTransform.gameObject);

                    var uiDynamic = fieldTransform.GetComponent<UIDynamicTextField>();
                    uiDynamic.text = new StringBuilder()
                        .AppendLine("Combined friction for hard colliders and soft colliders.")
                        .AppendLine()
                        .Append("Adaptive friction reduces friction when ")
                        .AppendItalic("Gloss")
                        .Append(" is increased. The higher the gloss, the more ")
                        .AppendItalic("Specular Bumpiness")
                        .AppendLine(" adds friction. Other skin material sliders are ignored.")
                        .AppendLine()
                        .Append("Dry skin friction represents the friction value when gloss and specular bumpiness are low.")
                        .ToString();
                    uiDynamic.backgroundColor = Color.clear;
                    uiDynamic.textColor = Color.white;
                }

                /* Adaptive friction toggle */
                {
                    var customTransform = CustomToggle(rightSide, new Vector2(0, -880));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    var jsb = _frictionHandler.AdaptiveFrictionJsb;
                    var uiDynamic = customTransform.GetComponent<UIDynamicToggle>();
                    uiDynamic.label = jsb.name;
                    _script.AddToggleToJsb(uiDynamic, jsb);
                }

                /* Dry skin friction slider */
                {
                    var customTransform = CustomSlider(rightSide, new Vector2(0, -1020));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    DrySkinFrictionSlider = customTransform.GetComponent<UIDynamicSlider>();
                    var storable = _frictionHandler.DrySkinFrictionJsf;
                    DrySkinFrictionSlider.Configure(
                        storable.name,
                        storable.min,
                        storable.max,
                        storable.defaultVal,
                        storable.constrained,
                        "F3"
                    );
                    _script.AddSliderToJsf(DrySkinFrictionSlider, storable);
                    DrySkinFrictionSlider.SetActiveStyle(_frictionHandler.AdaptiveFrictionJsb.val, true);
                }

                /* Friction offset slider */
                {
                    var customTransform = CustomSlider(rightSide, new Vector2(0, -1160));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    var uiDynamic = customTransform.GetComponent<UIDynamicSlider>();
                    var storable = _frictionHandler.FrictionOffsetJsf;
                    uiDynamic.Configure(
                        storable.name,
                        storable.min,
                        storable.max,
                        storable.defaultVal,
                        storable.constrained,
                        "F3"
                    );
                    _script.AddSliderToJsf(uiDynamic, storable);
                }

                /* Soft collider friction value slider */
                {
                    var customTransform = CustomSlider(rightSide, new Vector2(0, -1300));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    var uiDynamic = customTransform.GetComponent<UIDynamicSlider>();
                    var storable = _frictionHandler.FrictionValueJsf;
                    uiDynamic.Configure(
                        storable.name,
                        storable.min,
                        storable.max,
                        storable.defaultVal,
                        storable.constrained,
                        "F3"
                    );
                    _script.AddSliderToJsf(uiDynamic, storable);
                    uiDynamic.SetActiveStyle(false, true);
                }

                if(_frictionHandler.IsEnabled)
                {
                    HideBootyMagicFrictionUI();
                    skinMaterials2ModifiedJsb.val = true;
                }
                else
                {
                    HideFrictionUI(false);
                }
            }
            catch(Exception e)
            {
                _script.Loggr.LogError($"Error modifying {uiMod.Id} UI: {e}");
            }
        }

        IEnumerator AddButtonToSkinMaterialsUICo(UIMod uiMod)
        {
            while(_script.Initialized != true)
            {
                yield return null;
            }

            if(_script.Person.HasDummySkin())
            {
                while(!_frictionHandler.IsEnabled)
                {
                    yield return null;
                }
            }
            else if(!_frictionHandler.IsEnabled)
            {
                yield break;
            }

            try
            {
                var customTransform = CustomButton(uiMod.Target.Find("LeftSide"), new Vector2(20, -810));
                uiMod.AddCustomObject(customTransform.gameObject);
                var uiDynamicButton = customTransform.GetComponent<UIDynamicButton>();
                SwitchToOtherPluginAction.RegisterButton(uiDynamicButton);
                uiDynamicButton.label = SwitchToOtherPluginAction.name;
                uiMod.Disable();
            }
            catch(Exception e)
            {
                _script.Loggr.LogError($"Error modifying {uiMod.Id}  UI: {e}");
            }
        }

        Transform OpenPluginUIButton(Transform parent)
        {
            var button = _script.InstantiateButton(parent);
            var uiDynamic = button.GetComponent<UIDynamicButton>();
            uiDynamic.label = new StringBuilder()
                .AppendFormat("<b>Open {0} UI</b>", Constant.TITTY_MAGIC)
                .ToString();
            uiDynamic.AddListener(() => _script.Bindings.Actions["OpenUI"].actionCallback());
            return button;
        }

        Transform CustomToggle(Transform parent, Vector2 anchoredPosition)
        {
            var customTransform = _script.InstantiateToggle(parent);
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = Vector2.zero;
            rectTransform.anchoredPosition = anchoredPosition;
            var sizeDelta = rectTransform.sizeDelta;
            rectTransform.sizeDelta = new Vector2(sizeDelta.x + 10, sizeDelta.y);
            return customTransform;
        }

        Transform CustomSlider(Transform parent, Vector2 anchoredPosition)
        {
            var customTransform = _script.InstantiateSlider(parent);
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = Vector2.zero;
            rectTransform.anchoredPosition = anchoredPosition;
            var sizeDelta = rectTransform.sizeDelta;
            rectTransform.sizeDelta = new Vector2(sizeDelta.x + 10, sizeDelta.y);
            return customTransform;
        }

        Transform CustomButton(Transform parent, Vector2 anchoredPosition)
        {
            var customTransform = _script.InstantiateButton(parent);
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = Vector2.zero;
            rectTransform.anchoredPosition = anchoredPosition;
            var sizeDelta = rectTransform.sizeDelta;
            rectTransform.sizeDelta = new Vector2(sizeDelta.x + 10, sizeDelta.y);
            return customTransform;
        }

        internal override void SetFrictionUIVisibility()
        {
            if(!enabled)
            {
                return;
            }

            if(_frictionHandler.IsEnabled)
            {
                ShowFrictionUI(true);
                HideBootyMagicFrictionUI();
                _script.Bootstrap.UIModManager.SetSkinMaterials2ModEnabled(true);
            }
            else
            {
                uiMods.Find(uiMod => uiMod.Id == Constant.SKIN_MATERIALS_2_FRICTION).Disable();
                uiMods.Find(uiMod => uiMod.Id == Constant.SKIN_MATERIALS_2_ADD_BTN).Disable();
                skinMaterials2ModifiedJsb.val = false;
                if(!_bootyMagic.GetBoolParamValueNullSafe(Constant.INTERNAL_SKIN_MATERIALS_2_MODIFIED))
                {
                    _script.Bootstrap.UIModManager.SetSkinMaterials2ModEnabled(false);
                }
            }
        }

        void HideBootyMagicFrictionUI()
        {
            if(!_bootyMagic.IsEnabledNullSafe())
            {
                return;
            }

            if(_bootyMagic.GetBoolParamValueNullSafe(Constant.INTERNAL_SKIN_MATERIALS_2_MODIFIED))
            {
                _bootyMagic.CallActionNullSafe(Constant.INTERNAL_HIDE_FRICTION_UI);
            }
        }

        void OnEnable()
        {
            if(!initialized)
            {
                return;
            }

            foreach(var uiMod in uiMods)
            {
                if(uiMod.Id == Constant.SKIN_MATERIALS_2_FRICTION)
                {
                    if(_frictionHandler.IsEnabled)
                    {
                        uiMod.Enable();
                        skinMaterials2ModifiedJsb.val = true;
                    }
                }
                else if(uiMod.Id != Constant.SKIN_MATERIALS_2_ADD_BTN)
                {
                    uiMod.Enable();
                }
            }

            _script.Bootstrap.UIModManager.SetEnabled(_frictionHandler.IsEnabled);
            if(_bootyMagic.IsEnabledNullSafe())
            {
                _bootyMagic.CallActionNullSafe(Constant.INTERNAL_HIDE_FRICTION_UI);
            }
        }

        new void OnDisable()
        {
            base.OnDisable();
            skinMaterials2ModifiedJsb.val = false;
            if(_bootyMagic.IsEnabledNullSafe())
            {
                _bootyMagic.CallActionNullSafe(Constant.INTERNAL_SHOW_FRICTION_UI);
            }
            else
            {
                _script.Bootstrap.UIModManager.enabled = false;
            }
        }
    }
}
